Type: Package
Package: anticlust
Title: Subset Partitioning via Anticlustering
Version: 0.8.10-1
Authors@R: c(
    person("Martin", "Papenberg", , "martin.papenberg@hhu.de", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-9900-4268")),
    person("Meik", "Michalke", role = "ctb",
           comment = "centroid based clustering algorithm"),
    person(c("Gunnar", "W."), "Klau", role = "ths"),
    person(c("Juliane", "V."), "Nagel", role = "ctb",
           comment = "package logo"),
    person("Martin", "Breuer", role = "ctb",
           comment = "Bicriterion algorithm by Brusco et al."),
    person("Marie L.", "Schaper", role = "ctb",
           comment = "Example data set"),
    person("Max", "Diekhoff", role = "ctb",
           comment = "Optimal maximum dispersion algorithm"),
    person("Hannah", "Hengelbrock", role = "ctb",
           comment = "TPSDP heuristic by Yang et al.")
    )
Author: Martin Papenberg [aut, cre] (<https://orcid.org/0000-0002-9900-4268>),
  Meik Michalke [ctb] (centroid based clustering algorithm),
  Gunnar W. Klau [ths],
  Juliane V. Nagel [ctb] (package logo),
  Martin Breuer [ctb] (Bicriterion algorithm by Brusco et al.),
  Marie L. Schaper [ctb] (Example data set),
  Max Diekhoff [ctb] (Optimal maximum dispersion algorithm),
  Hannah Hengelbrock [ctb] (TPSDP heuristic by Yang et al.)
Maintainer: Martin Papenberg <martin.papenberg@hhu.de>
Description: The method of anticlustering partitions a pool of elements into groups (i.e., anticlusters) with the goal of maximizing between-group similarity or within-group heterogeneity. The anticlustering approach thereby reverses the logic of cluster analysis that strives for high within-group homogeneity and clear separation between groups.  Computationally, anticlustering is accomplished by maximizing instead of minimizing a clustering objective function, such as the intra-cluster variance (used in k-means clustering) or the sum of pairwise distances within clusters. The main function anticlustering() gives access to optimal and heuristic anticlustering methods described in Papenberg and Klau (2021; <doi:10.1037/met0000301>), Brusco et al. (2020; <doi:10.1111/bmsp.12186>), Papenberg (2024; <doi:10.1111/bmsp.12315>), and Papenberg et al. (2025; <doi:10.1101/2025.03.03.641320>). The optimal algorithms require that an integer linear programming solver is installed. This package will install 'lpSolve' (<https://cran.r-project.org/package=lpSolve>) as a default solver, but it is also possible to use the package 'Rglpk' (<https://cran.r-project.org/package=Rglpk>), which requires the GNU linear programming kit (<https://www.gnu.org/software/glpk/glpk.html>), the package 'Rsymphony' (<https://cran.r-project.org/package=Rsymphony>), which requires the SYMPHONY ILP solver (<https://github.com/coin-or/SYMPHONY>), or the commercial solver Gurobi, which provides its own R package that is not available via CRAN (<https://www.gurobi.com/downloads/>). 'Rglpk', 'Rsymphony', 'gurobi' and their system dependencies have to be manually installed by the user because they are only suggested dependencies. Full access to the bicriterion anticlustering method proposed by Brusco et al. (2020) is given via the function bicriterion_anticlustering(), while kplus_anticlustering() implements the full functionality of the k-plus anticlustering approach proposed by Papenberg (2024). Some other functions are available to solve classical clustering problems. The function balanced_clustering() applies a cluster analysis under size constraints, i.e., creates equal-sized clusters. The function matching() can be used for (unrestricted, bipartite, or K-partite) matching. The function wce() can be used optimally solve the (weighted) cluster editing problem, also known as correlation clustering, clique partitioning problem or transitivity clustering.
License: MIT + file LICENSE
URL: https://github.com/m-Py/anticlust,
        https://m-py.github.io/anticlust/
BugReports: https://github.com/m-Py/anticlust/issues
Depends: R (>= 3.6.0)
Imports: Matrix, RANN (>= 2.6.0), lpSolve
Suggests: knitr, palmerpenguins, Rglpk, rmarkdown, Rsymphony, tinytest,
        gurobi
VignetteBuilder: knitr, rmarkdown
Encoding: UTF-8
LazyData: true
NeedsCompilation: yes
RoxygenNote: 7.3.2
SystemRequirements: Rendering the vignette requires pandoc
        (<https://pandoc.org/>).
Packaged: 2025-03-26 08:55:13 UTC; martinp
Repository: CRAN
Date/Publication: 2025-03-26 17:40:10 UTC
Built: R 4.3.3; aarch64-apple-darwin20; 2025-03-26 17:59:35 UTC; unix
Archs: anticlust.so.dSYM

Type: Package
Package: anticlust
Title: Subset Partitioning via Anticlustering
Version: 0.8.14
Authors@R: c(
    person("Martin", "Papenberg", , "martin.papenberg@hhu.de", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-9900-4268")),
    person("Meik", "Michalke", role = "ctb",
           comment = "centroid based clustering algorithm"),
    person(c("Gunnar", "W."), "Klau", role = "ths"),
    person(c("Juliane", "V."), "Nagel", role = "ctb",
           comment = "package logo"),
    person("Martin", "Breuer", role = "ctb",
           comment = "Bicriterion algorithm by Brusco et al."),
    person("Marie L.", "Schaper", role = "ctb",
           comment = "Example data set"),
    person("Max", "Diekhoff", role = "ctb",
           comment = "Optimal maximum dispersion algorithm"),
    person("Hannah", "Hengelbrock", role = "ctb",
           comment = "TPSDP heuristic by Yang et al.")
    )
Author: Martin Papenberg [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9900-4268>),
  Meik Michalke [ctb] (centroid based clustering algorithm),
  Gunnar W. Klau [ths],
  Juliane V. Nagel [ctb] (package logo),
  Martin Breuer [ctb] (Bicriterion algorithm by Brusco et al.),
  Marie L. Schaper [ctb] (Example data set),
  Max Diekhoff [ctb] (Optimal maximum dispersion algorithm),
  Hannah Hengelbrock [ctb] (TPSDP heuristic by Yang et al.)
Maintainer: Martin Papenberg <martin.papenberg@hhu.de>
Description: The method of anticlustering partitions a pool of elements into groups (i.e., anticlusters) with the goal of maximizing between-group similarity or within-group heterogeneity. The anticlustering approach thereby reverses the logic of cluster analysis that strives for high within-group homogeneity and clear separation between groups.  Computationally, anticlustering is accomplished by maximizing instead of minimizing a clustering objective function, such as the intra-cluster variance (used in k-means clustering) or the sum of pairwise distances within clusters. The main function anticlustering() gives access to optimal and heuristic anticlustering methods described in Papenberg and Klau (2021; <doi:10.1037/met0000301>), Brusco et al. (2020; <doi:10.1111/bmsp.12186>), Papenberg (2024;  <doi:10.1111/bmsp.12315>), Papenberg, Wang, et al. (2025; <doi:10.1016/j.crmeth.2025.101137>), Papenberg, Breuer, et al. (2025; <doi:10.1017/psy.2025.10052>), and Yang et al. (2022; <doi:10.1016/j.ejor.2022.02.003>). The optimal algorithms require that an integer linear programming solver is installed. This package will install 'lpSolve' (<https://cran.r-project.org/package=lpSolve>) as a default solver, but it is also possible to use the package 'Rglpk' (<https://cran.r-project.org/package=Rglpk>), which requires the GNU linear programming kit (<https://www.gnu.org/software/glpk/glpk.html>), the package 'Rsymphony' (<https://cran.r-project.org/package=Rsymphony>), which requires the SYMPHONY ILP solver (<https://github.com/coin-or/SYMPHONY>), or the commercial solver Gurobi, which provides its own R package that is not available via CRAN (<https://www.gurobi.com/downloads/>). 'Rglpk', 'Rsymphony', 'gurobi' and their system dependencies have to be manually installed by the user because they are only suggested dependencies. Full access to the bicriterion anticlustering method proposed by Brusco et al. (2020) is given via the function bicriterion_anticlustering(), while kplus_anticlustering() implements the full functionality of the k-plus anticlustering approach proposed by Papenberg (2024). Some other functions are available to solve classical clustering problems. The function balanced_clustering() applies a cluster analysis under size constraints, i.e., creates equal-sized clusters. The function matching() can be used for (unrestricted, bipartite, or K-partite) matching. The function wce() can be used optimally solve the (weighted) cluster editing problem, also known as correlation clustering, clique partitioning problem or transitivity clustering.
License: MIT + file LICENSE
URL: https://github.com/m-Py/anticlust,
        https://m-py.github.io/anticlust/
BugReports: https://github.com/m-Py/anticlust/issues
Depends: R (>= 3.6.0)
Imports: Matrix, RANN (>= 2.6.0), lpSolve
Suggests: knitr, MASS, Rglpk, rmarkdown, Rsymphony, tinytest, gurobi,
        tableone, palmerpenguins
VignetteBuilder: knitr, rmarkdown
Encoding: UTF-8
LazyData: true
NeedsCompilation: yes
RoxygenNote: 7.3.3
SystemRequirements: Rendering the vignette requires pandoc
        (<https://pandoc.org/>).
Packaged: 2026-04-15 12:48:45 UTC; martinp
Repository: CRAN
Date/Publication: 2026-04-15 14:40:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-20 23:24:54 UTC; unix
Archs: anticlust.so.dSYM

# Generated by roxygen2: do not edit by hand

export(anticlustering)
export(balanced_clustering)
export(bicriterion_anticlustering)
export(categorical_sampling)
export(categories_to_binary)
export(dispersion_objective)
export(diversity_objective)
export(fast_anticlustering)
export(generate_exchange_partners)
export(generate_partitions)
export(kplus_anticlustering)
export(kplus_moment_variables)
export(matching)
export(mean_sd_tab)
export(n_partitions)
export(optimal_anticlustering)
export(optimal_dispersion)
export(plot_clusters)
export(plot_similarity)
export(three_phase_search_anticlustering)
export(variance_objective)
export(wce)
importFrom(Matrix,sparseMatrix)
importFrom(RANN,nn2)
importFrom(grDevices,palette)
importFrom(grDevices,rainbow)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(stats,as.dist)
importFrom(stats,as.formula)
importFrom(stats,contrasts)
importFrom(stats,dist)
importFrom(stats,model.matrix)
importFrom(stats,rbinom)
importFrom(stats,sd)
importFrom(utils,combn)
useDynLib(anticlust, .registration = TRUE)

# See COPYRIGHT file for source of the data, and LICENSE file.

library(anticlust)
library(readxl)
library(usethis)

tt <- read_excel("./inst/data_raw/Indices_PhraseEM.xlsx")

tt$target_word_emotionality <- factor(tt$`Word Emotionality`)
tt$sentence_emotionality <- factor(tt$`Sentence Emotionality`)
tt$percentage_target_word <- tt$`% of occurrence target word`
tt$valence_target_word <- tt$`Valence Target Word`
tt$arousal_target_word <- tt$`Arousal Target Word`
tt$target_word <- tt$`Target Word`
tt$sentence <- tt$Sentence

brunel2025 <- subset(as.data.frame(tt), select = c(target_word, sentence, target_word_emotionality:arousal_target_word))

usethis::use_data(brunel2025, overwrite = TRUE)

#' Bicriterion iterated local search heuristic
#'
#' This function implements the bicriterion algorithm BILS for
#' anticlustering by Brusco et al. (2020;
#' <doi:10.1111/bmsp.12186>). The description of their algorithm is
#' given in Section 3 of their paper (in particular, see the
#' Pseudocode in their Figure 2). It also implements some extensions
#' to the bicriterion approach for anticlustering described in 
#' Papenberg, Breuer, et al. (2025).
#' 
#' 
#' @param x The data input. Can be one of two structures: (1) A
#'     feature matrix where rows correspond to elements and columns
#'     correspond to variables (a single numeric variable can be
#'     passed as a vector). (2) An N x N matrix dissimilarity matrix;
#'     can be an object of class \code{dist} (e.g., returned by
#'     \code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
#'     where the entries of the upper and lower triangular matrix
#'     represent pairwise dissimilarities.
#' @param K How many anticlusters should be created. Alternatively:
#'     (a) A vector describing the size of each group, or (b) a vector
#'     of length \code{nrow(x)} describing how elements are assigned
#'     to anticlusters before the optimization starts.
#' @param R The desired number of restarts for the algorithm. By
#'     default, both phases (MBPI + ILS) of the algorithm are
#'     performed once.  See details.
#' @param W Optional argument, a vector of weights defining the
#'     relative importance of dispersion and diversity (0 <= W <=
#'     1). See details.
#' @param Xi Optional argument, specifies probability of swapping
#'     elements during the iterated local search. See examples.
#' @param dispersion_distances A distance matrix used to compute the
#'     dispersion if the dispersion should not be computed on the
#'     basis of argument \code{x}.
#' @param average_diversity Boolean. Compute the diversity not as a
#'     global sum across all pairwise within-group distances, but as
#'     the sum of the average of within-group distances.
#' @param init_partitions A matrix of initial partitions (rows =
#'     partitions; columns = elements) that serve as starting
#'     partitions during the iterations of the first phase of the BILS
#'     (i.e., the MBPI).  If not passed, a new random partition is
#'     generated at the start of each iteration (which is the default
#'     behaviour).
#' @param return Either "paretoset" (default), "best-diversity", 
#'     "best-average-diversity", "best-dispersion". See below.
#'     
#' @details
#'
#' The bicriterion algorithm by Brusco et al. (2020) aims to
#' simultaneously optimize two anticlustering criteria: the
#' \code{\link{diversity_objective}} and the
#' \code{\link{dispersion_objective}}. It returns a list of partitions
#' that approximate the pareto set of efficient solutions across both
#' criteria. By considering both the diversity and dispersion, this
#' algorithm is well-suited for maximizing overall within-group
#' heterogeneity. To select a partition among the approximated pareto
#' set, it is reasonable to plot the objectives for each partition
#' (see Examples).
#'
#' The arguments \code{R}, \code{W} and \code{Xi} are named for
#' consistency with Brusco et al. (2020). The argument \code{K} is
#' used for consistency with other functions in anticlust; Brusco et
#' al. used `G` to denote the number of groups. However, note that
#' \code{K} can not only be used to denote the number of equal-sized
#' groups, but also to specify group sizes, as in
#' \code{\link{anticlustering}}.
#' 
#' This function implements the combined bicriterion algorithm BILS,
#' which consists of two phases: The multistart bicriterion pairwise
#' interchange heuristic (MBPI, which is a local maximum search
#' heuristic similar to \code{method = "local-maximum"} in
#' \code{\link{anticlustering}}), and the iterated local search (ILS),
#' which is an improvement procedure that tries to overcome local
#' optima.  The argument \code{R} denotes the number of restarts of
#' the two phases of the algorithm. If \code{R} has length 1, half of
#' the repetitions perform the first phase MBPI), the other half
#' perform the ILS.  If \code{R} has length 2, the first entry
#' indicates the number of restarts of MBPI the second entry indicates
#' the number of restarts of ILS.  The argument \code{W} denotes the
#' relative weight given to the diversity and dispersion criterion in
#' a given run of the search heuristic. In each run, the a weight is
#' randomly selected from the vector \code{W}. As default values, we
#' use the weights that Brusco et al. used in their analyses. All
#' values in \code{W} have to be in [0, 1]; larger values indicate
#' that diversity is more important, whereas smaller values indicate
#' that dispersion is more important; \code{w = .5} implies the same
#' weight for both criteria. The argument \code{Xi} is the probability
#' that an element is swapped during the iterated local search
#' (specifically, Xi has to be a vector of length 2, denoting the
#' range of a uniform distribution from which the probability of
#' swapping is selected). For \code{Xi}, the default is selected
#' consistent with the analyses by Brusco et al.
#'
#' If the data input \code{x} is a feature matrix (that is: each row
#' is a "case" and each column is a "variable"), a matrix of the
#' Euclidean distances is computed as input to the algorithm. If a
#' different measure of dissimilarity is preferred, you may pass a
#' self-generated dissimilarity matrix via the argument \code{x}.  The
#' argument \code{dispersion_distances} can additionally be used if
#' the dispersion should be computed on the basis of a different
#' distance matrix.
#' 
#' As of anticlust version 0.8.6, this function includes some extensions 
#' to the original BILS by Brusco et al. (2020). These extensions are are 
#' implemented through the optional arguments \code{dispersion_distances}, 
#' \code{average_diversity}, \code{init_partitions}, and \code{return}, and
#' were described in Papenberg, Breuer, et al. (2025). 
#' If these arguments are not changed, the function performs the "vanilla" BILS 
#' as described in Brusco et al. If multiple \code{init_partitions} are given, 
#' ensure that each partition (i.e., each row of\code{init_partitions}) has the exact
#' same output of \code{\link{table}}.
#' 
#' @return By default, a \code{matrix} of anticlustering partitions
#'     (i.e., the approximated pareto set). Each row corresponds to a
#'     partition, each column corresponds to an input element. If the
#'     argument \code{return} is set to either "best-diversity", 
#'     "best-average-diversity", or "best-dispersion", it only returns one 
#'     partition (as a vector), that maximizes the respective objective.
#' 
#' @author Martin Breuer \email{M.Breuer@@hhu.de}, Martin Papenberg
#'     \email{martin.papenberg@@hhu.de}
#' 
#' @export
#' 
#' @examples 
#' # Generate some random data
#' M <- 3
#' N <- 80
#' K <- 4
#' data <- matrix(rnorm(N * M), ncol = M)
#'
#' # Perform bicriterion algorithm, use 200 repetitions
#' pareto_set <- bicriterion_anticlustering(data, K = K, R = 200)
#'
#' # Compute objectives for all solutions
#' diversities_pareto <- apply(pareto_set, 1, diversity_objective, x = data)
#' dispersions_pareto <- apply(pareto_set, 1, dispersion_objective, x = data)
#'
#' # Plot the pareto set
#' plot(
#'   diversities_pareto,
#'   dispersions_pareto,
#'   col = "blue",
#'   cex = 2,
#'   pch = as.character(1:NROW(pareto_set))
#' )
#' 
#' # Get some random solutions for comparison
#' rnd_solutions <- t(replicate(n = 200, sample(pareto_set[1, ])))
#' 
#' # Compute objectives for all random solutions
#' diversities_rnd <- apply(rnd_solutions, 1, diversity_objective, x = data)
#' dispersions_rnd <- apply(rnd_solutions, 1, dispersion_objective, x = data)
#' 
#' # Plot random solutions and pareto set. Random solutions are far away 
#' # from the good solutions in pareto set
#' plot(
#'   diversities_rnd, dispersions_rnd, 
#'   col = "red",
#'   xlab = "Diversity",
#'   ylab = "Dispersion",
#'   ylim = c(
#'     min(dispersions_rnd, dispersions_pareto), 
#'     max(dispersions_rnd, dispersions_pareto)
#'   ),
#'   xlim = c(
#'     min(diversities_rnd, diversities_pareto), 
#'     max(diversities_rnd, diversities_pareto)
#'   )
#' )
#' 
#' # Add approximated pareto set from bicriterion algorithm:
#' points(diversities_pareto, dispersions_pareto, col = "blue", cex = 2, pch = 19)
#'
#' @note
#'
#' For technical reasons, the pareto set returned by this function has
#' a limit of 500 partitions. Usually however, the
#' algorithm usually finds much fewer partitions. There is one following exception:
#' We do not recommend to use this method when the input data is
#' one-dimensional where the algorithm may identify too many
#' equivalent partitions causing it to run very slowly (see section 5.6 in 
#' Breuer, 2020).
#' 
#' @references
#' 
#' Brusco, M. J., Cradit, J. D., & Steinley, D. (2020). Combining
#' diversity and dispersion criteria for anticlustering: A bicriterion
#' approach. British Journal of Mathematical and Statistical
#' Psychology, 73, 275-396. https://doi.org/10.1111/bmsp.12186
#' 
#' Breuer (2020). Using anticlustering to maximize diversity and dispersion:
#' Comparing exact and heuristic approaches. Bachelor thesis.
#' 
#' Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. (2025). 
#' Extending the Bicriterion Approach for Anticlustering: Exact and Hybrid 
#' Approaches. Psychometrika. Advance online publication. https://doi.org/10.1017/psy.2025.10052
#' 

bicriterion_anticlustering <- function(
  x, K, R = NULL, 
  W = c(0.000001, 0.00001, 0.0001, 0.001, 0.01, 0.1, 0.5, 0.99, 0.999, 0.999999),
  Xi = c(0.05, 0.1),
  dispersion_distances = NULL, average_diversity = FALSE, init_partitions = NULL, return = "paretoset") {
  
  input_validation_bicriterion_anticlustering(x, K, R, W, Xi, dispersion_distances, average_diversity, init_partitions, return)

  distances <- convert_to_distances(x) 
  
  N <- NROW(distances)
  WL <- length(W)
  if (is.null(R)) {
    R <- c(1, 1)
  } else if (length(R) == 1) {
    R <- rep(ceiling(R / 2), 2)
  }
  
  if (is.null(dispersion_distances)) {
    dispersion_distances <- distances
  } else {
    dispersion_distances <- convert_to_distances(dispersion_distances)
  }

  clusters <- initialize_clusters(N, K, NULL) - 1
  
  if (argument_exists(init_partitions)) {
    use_init_partitions <- 1
    init_partitions <- t(apply(init_partitions, 1, to_numeric)) - 1 # ensure correct labeling of partitions
    clusters <- init_partitions[1, ]
  } else {
    init_partitions <- 0
    use_init_partitions <- 0
  }
  
  frequencies <- rep(1, length(unique(clusters)))
  if (average_diversity) {
    frequencies <- table(clusters)
  }
  upper_bound = 500 # limits number of resulting partitions 
  # create empty matrix for results to use in C
  result_matrix = matrix(data = -1, nrow = upper_bound , ncol = N) 
  
  # Call C function
  results <- .C(
    "bicriterion_iterated_local_search_call",
    as.double(distances),
    as.double(dispersion_distances),
    as.integer(N),
    as.integer(R),
    as.integer(upper_bound),
    as.integer(WL),
    as.double(W),
    as.double(Xi),
    as.integer(clusters),
    as.integer(frequencies),
    as.integer(use_init_partitions),
    as.integer(t(init_partitions)),
    result = integer(length(result_matrix)),
    mem_error = as.integer(0),
    PACKAGE = "anticlust" # important to call C
  )

  if (results[["mem_error"]] == 1) {
    stop("Could not allocate enough memory.")
  }
  
  results <- results$result
  
  # remove allocated space that was not needed
  results <- results[results != -1]

  # C returns the list of partitions as one vector, 
  # that we turn back into a matrix
  results <- matrix(results, ncol = N, byrow = TRUE)
  # Order each partition:
  results <- data.frame(t(apply(results, 1, order_cluster_vector)))
  # Remove duplicates
  results <- results[!duplicated(results), ]
  results <- unname(as.matrix(results))
  if (return == "paretoset") {
    return(results)
  }
  if (return == "best-dispersion") {
    best_obj <- which.max(apply(results, 1, dispersion_objective_, dispersion_distances))
    return(results[best_obj, ])
  } else if (return == "best-average-diversity" || average_diversity) {
    # manually compute objectives, apply() does not work here because table() has different results for the different clusterings
    average_diversities <- c()
    for (i in 1:nrow(results)) {
      average_diversities[i] <- weighted_diversity_objective_(distances, results[i, ], table(results[i, ]))
    }
    return(results[which.max(average_diversities), ])
  } else if (return == "best-diversity") {
    best_obj <- which.max(apply(results, 1, diversity_objective_, distances))
    return(results[best_obj, ])
  } 
}

input_validation_bicriterion_anticlustering <- function(
    x, K, R, W, Xi, dispersion_distances, 
    average_diversity, init_partitions, return) {

  input_validation_anticlustering(
    x, K, objective = "diversity", method = "brusco", 
    preclustering = FALSE, categories = NULL,
    repetitions = 1, standardize = FALSE
  )
  
  x <- convert_to_distances(x)
  N <- nrow(x)
  
  validate_input(
    return, "return", objmode = "character", len = 1,
    input_set = c("paretoset", "best-diversity", "best-average-diversity", "best-dispersion"), 
    not_na = TRUE, 
    not_function = TRUE
  )
  
  validate_input(
    average_diversity, "average_diversity", objmode = "logical", len = 1,
    input_set = c(TRUE, FALSE), 
    not_na = TRUE, 
    not_function = TRUE
  )
  
  if (average_diversity && return == "best-diversity") {
    stop("Cannot use 'return = best-diversity' with 'average_diversity = TRUE' (probably you want to use 'return = best-average-diversity').")
  }
  if (!average_diversity && return == "best-average-diversity") {
    stop("Cannot use 'return = best-average-diversity' with 'average_diversity = FALSE' (probably you want to use 'return = best-diversity').")
  }

  if (argument_exists(dispersion_distances)) {
    if (any(dim(dispersion_distances) != dim(x))) {
      stop("The dimensions of argument 'x' and 'dispersion_distances' do not match.")
    } 
  }

  checkweights(W)
  checkneighborhood(Xi)
  
  if (argument_exists(R)) {
    validate_input(R, "R", must_be_integer = TRUE, not_na = TRUE, not_function = TRUE, greater_than = -1)
    if (!length(R) %in% 1:2) {
      stop("Argument 'R' must have length 1 or 2.")
    }
  }
  
  if (argument_exists(init_partitions)) {
    if (length(R) != 2) {
      stop("If 'init_partitions' is used, specify argument R as vector of length 2.")
    }
    if (nrow(init_partitions) != R[1]) {
      stop("The number of repetitions (argument 'R') and the number of partitions (argument 'init_partitions') do not match.")
    } 
    if (ncol(init_partitions) != N) {
      stop("The number of columns in 'init_partitions' does not match the number of cases in 'x'.")
    }
  }
}


#verify input
#############
checkweights <- function(W){
  validate_input(W, "W", not_na = TRUE, not_function = TRUE, objmode = "numeric")
  for(i in W){
    if(i < 0 | i > 1){
      stop("All weights (argument 'W') must be percentages. Only values between 0 and 1 are allowed.")
    }
  }  
}

checkneighborhood <- function(Xi) {
  validate_input(Xi, "Xi", not_na = TRUE, not_function = TRUE, len = 2, objmode = "numeric")
  for (i in Xi) {
    if (i < 0 | i > 1) {
      stop("Both neighorhood indexes must be a percentage. Only values between 0 and 1 are allowed.")
    }
  } 
  if (Xi[1] > Xi[2]) {
    stop("First neighborhood percentage needs to be smaller than the second.")
  }
}

#' Completion norms for 403 French sentences
#'
#' @format A data frame with 403 rows and 7 variables
#' \describe{
#'   \item{target_word}{The word (in French)}
#'   \item{sentence}{The sentence (in French), which is completed by \code{target_word}}
#'   \item{target_word_emotionality}{Categorical: Is the target word classified as "positive", "negative" or "neutral"}
#'   \item{sentence_emotionality}{Categorical: Is the sentence classified as 
#'     "mixt", "negative", "neutral", "non specified", or "positive"}
#'   \item{percentage_target_word}{Numeric between 0 and 1; percentage of a test sample that identified the target word}
#'   \item{valence_target_word}{Numeric, average valence rating for the target word}
#'   \item{arousal_target_word}{Numeric, average arousal rating for the target word}
#' }
#'
#' @source
#' This data set was distributed as supplementary online materials to the paper by Brunel et al. (2025), 
#' and is available from the Open Science Repository via https://osf.io/7pc46. Retrieved on October 2, 2025.
#' The data was published under a CC-By Attribution 4.0 International license, which permits redistribution.
#' The anticlust packages includes the raw and unchanged data file as provided by Jeremy Brunel, and a 
#' script that converts the raw data into the package data.
#' 
#' @examples
#' 
#' groups <- anticlustering(
#'   brunel2025[, 3:7],
#'   K = 4, 
#'   objective = "kplus", 
#'   standardize = TRUE,
#'   method = "3phase"
#' )
#' # check out descriptive statistics for each variable for each of 4 groups:
#' # Numeric variables
#' mean_sd_tab(brunel2025[, 5:7], groups)
#' # Categorical variables:
#' table(groups, brunel2025$target_word_emotionality)
#' table(groups, brunel2025$sentence_emotionality)
#' 
#' @references
#'
#' Brunel, J., Dujardin, E., Delord, S. (2025) Emotional valence, cloze 
#' probability, and entropy: Completion norms for 403 French sentences. 
#' Behavior Research Methods 57(81). https://doi.org/10.3758/s13428-025-02604-7
#'
#'
"brunel2025"

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
set.seed(1)


## ----setup--------------------------------------------------------------------
library(anticlust)

## -----------------------------------------------------------------------------
library(MASS)
data(survey)       # load data set
nrow(survey)       # number of students
head(survey, n=10) # look at the data
sapply(survey, anyNA) # most variables contain missing values

## -----------------------------------------------------------------------------

features <- c("Sex", "W.Hnd", "Exer", "Smoke", "Pulse", "Height", "Age")

survey$House <- anticlustering(
  survey[, features], 
  K = 3
)


## -----------------------------------------------------------------------------
library(tableone)

CreateTableOne(features, strata = "House", data = survey)


## -----------------------------------------------------------------------------
survey$Rnd_House <- sample(survey$House)

CreateTableOne(features, strata = "Rnd_House", data = survey)


## -----------------------------------------------------------------------------
survey$House2 <- anticlustering(
  survey[, features], 
  K = 3,
  standardize = TRUE
)

CreateTableOne(features, strata = "House2", data = survey)


## -----------------------------------------------------------------------------
survey$House3 <- anticlustering(
  survey[, features], 
  K = 3,
  standardize = TRUE,
  method = "3phase"
)
CreateTableOne(features, strata = "House3", data = survey)

## -----------------------------------------------------------------------------
survey$House4 <- anticlustering(
  survey[, features], 
  K = 3,
  standardize = TRUE,
  method = "3phase",
  objective = "variance"
)
CreateTableOne(features, strata = "House4", data = survey)


## -----------------------------------------------------------------------------

colors <- c("#a9a9a9", "#df536b", "#61d04f")

ord <- order(survey$Pulse)
# Plot the data while visualizing the different clusters
plot(
  survey$Pulse[ord], 
  col = colors[survey$House4[ord]], 
  pch = 19, 
  ylab = "Pulse", 
  xlab = "Students (ordered by pulse)"
)
legend("bottomright", legend = paste("Group", 1:3), col = colors, pch = 19)


## -----------------------------------------------------------------------------
survey$House5 <- anticlustering(
  survey[, features], 
  K = 3,
  method = "3phase",
  objective = "kplus",
  standardize = TRUE
)
CreateTableOne(features, strata = "House5", data = survey)


## -----------------------------------------------------------------------------
survey$House6 <- anticlustering(
  survey[, features], 
  K = c(137, 50, 50),
  standardize = TRUE,
  method = "3phase"
)
CreateTableOne(features, strata = "House6", data = survey)

## -----------------------------------------------------------------------------
survey$House7 <- anticlustering(
  survey[, features], 
  K = c(137, 50, 50),
  standardize = TRUE,
  method = "local-maximum",
  repetitions = 10, # increasing repetitions may be helpful with method = "local-maximum"
  objective = "average-diversity"
)
CreateTableOne(features, strata = "House7", data = survey)

## -----------------------------------------------------------------------------
survey$House8 <- anticlustering(
  survey[, features], 
  K = c(137, 50, 50),
  standardize = TRUE,
  method = "local-maximum",
  repetitions = 10, 
  objective = "kplus"
)
CreateTableOne(features, strata = "House8", data = survey)

## -----------------------------------------------------------------------------
hist(survey$Age)
sort(survey$Age, decreasing = TRUE)[1:10]

## -----------------------------------------------------------------------------
survey$is_age_outlier <- factor(survey$Age > 70)
survey$House9 <- anticlustering(
  survey[, features], 
  K = c(137, 50, 50),
  standardize = TRUE,
  method = "local-maximum",
  repetitions = 10, 
  objective = "kplus",
  categories = survey$is_age_outlier
)
CreateTableOne(features, strata = "House9", data = survey)

## -----------------------------------------------------------------------------
table(survey$is_age_outlier, survey$House9) # new assignment using `categories` argument
table(survey$is_age_outlier, survey$House8) # old assignment not using `categories` argument

## The functions `get_init_assignments*()` get the assignments for the items
# that are restricted via cannot-link constraints. The other items still need 
# to be assigned (via `add_unassigned_elements()`, which is done in 
# init_must_link_groups()). 
# - get_init_assignments_optimal() implements an optimal
# bin packing algorithm. It is just a decision version because we are not interested
# in minimizing the number of bins (so the objective function is constant for 
# all possible assignments). 
# - get_init_assignments_heuristic() implements a randomized fit heuristic,
# where each must-link cluster is assigned to a random group where it fits.
# Only if this heuristic fails, the optimal algorithm will be called.

get_init_assignments <- function(N, ID, target_groups, method = "heuristic") {
  if (method == "optimal") {
    get_init_assignments_optimal(N, ID, target_groups)
  } 
  get_init_assignments_heuristic(N, ID, target_groups)
}

get_init_assignments_optimal <- function(N, ID, target_groups) {
  weights <- table(ID)[table(ID) > 1]
  multiple_IDs <- as.numeric(names(weights))
  opt_assignment <- optimal_binpacking_(target_groups, weights)
  # Optimal assignment is done for the "reduced" data set, we have to assign cluster
  # label to all elements that are part of must-link group:
  full_clusters <- rep(NA, N)
  for (i in seq_along(multiple_IDs)) {
    full_clusters[ID == multiple_IDs[i]] <- opt_assignment[i]
  }
  full_clusters
}

# Initialize must-link constraints by assigning all elements having the same ID to the same set
# all others remain free (i.e., as NA). This is a "randomized fit" algorithm for bin packing
get_init_assignments_heuristic <- function(N, ID, target_groups) {
  # Initialize all as NA
  init <- rep(NA, N)
  K <- length(target_groups)
  cluster_sizes_real <- rep(0, K)
  multiple_IDs <- as.numeric(names(table(ID)[table(ID) > 1]))
  
  for (current_id in multiple_IDs) {
    random_order_clusters <- sample(K)
    for (k in random_order_clusters) {
      # only fill into cluster if it fits
      if ((cluster_sizes_real[k] + sum(ID == current_id)) > target_groups[k]) {
        if (k == random_order_clusters[K]) {
          stop("I could not fulfil the `must_link` restrictions, sorry!")
        }
        next
      }
      init[ID == current_id] <- k
      cluster_sizes_real[k] <- cluster_sizes_real[k] + sum(ID == current_id)
      break
    }
  }
  stopifnot(sum(!is.na(init))  == sum(ID %in% multiple_IDs))
  init
}

# Initialize the groupings for the reduced sample, for all elements:
init_must_link_groups <- function(N, IDs_initial, IDs_reduced, target_groups) {
  init <- tryCatch(
    get_init_assignments(N, IDs_initial, target_groups, method = "heuristic"),
    error = function(e) e
  )
  if ("simpleError" %in% class(init)) {
    init <- tryCatch(
      get_init_assignments(N, IDs_initial, target_groups, method = "optimal"),
      error = function(e) e
    )
  }
  if ("simpleError" %in% class(init)) {
    stop("The must-link constraints cannot be fulfilled! I really tried.")
  }
  init <- add_unassigned_elements(target_groups, init, N, length(target_groups))
  # only return one index per must-link group:
  init[sapply(IDs_reduced, FUN = "[", 1)]
}

# Adjust distances for must-link anticlustering (which uses a "reduced" data set where
# each must-link group is treated as a single unit)
adjusted_distances_must_link <- function(distances, must_link) {
  distances <- as.matrix(distances)
  N <- nrow(distances)
  stopifnot(N == length(must_link))
  N_reduced <- length(unique(must_link))
  new_distances <- matrix(NA, ncol = N_reduced, nrow = N_reduced)
  list_must_link_indices <- get_must_link_indices(must_link)
  for (i in 1:N_reduced) {
    for (j in 1:N_reduced) {
      new_distances[i, j] <- sum(distances[list_must_link_indices[[i]], list_must_link_indices[[j]]])
    }
  }
  list(distances = new_distances, IDs = list_must_link_indices)
}

# This is the function that is called from anticlustering()
must_link_anticlustering <- function(x, K, must_link, method = "exchange", objective = "diversity", repetitions = NULL) {
  
  x <- to_matrix(x)
  N <- nrow(x)
  stopifnot(is_distance_matrix(x)) # caller must ensure that the distance matrix is here
  
  must_link <- to_numeric(must_link)
  must_link <- replace_na_by_index(must_link)
  
  dt <- adjusted_distances_must_link(x, must_link)
  
  target_groups <- table(initialize_clusters(N, K, NULL))
  
  # possibly use multiple initializing partitions:
  if (argument_exists(repetitions)) {
    init_partitions <- t(replicate(
      n = ifelse(method == "2PML", max(round(repetitions/2), 1), repetitions), # half repetitions for first phase
      init_must_link_groups(N, IDs_initial = must_link, IDs_reduced = dt$IDs, target_groups = target_groups)
    ))
    init_partitions <- init_partitions - 1 # yeah, this is not good code (where should this go?)
    init <- init_partitions[1, ]
  } else {
    init_partitions <- NULL
    init <- init_must_link_groups(N, IDs_initial = must_link, IDs_reduced = dt$IDs, target_groups = target_groups)
  }
  
  reduced_clusters <- c_anticlustering(
    dt$distances,
    K = init,
    categories = lengths(dt$IDs), # restrict exchanges to node with same number of elements
    objective = "diversity",
    local_maximum = ifelse(method == "exchange", FALSE, TRUE),
    init_partitions = init_partitions
  )
  
  full_clusters <- merged_cluster_to_original_cluster(reduced_clusters, must_link)
  
  ## This is the second phase of the 2PML algorithm:
  if (method == "2PML") {
    iterations <- ifelse(is.null(repetitions), 1, max(round(repetitions/2), 1)) # half repetitions for second phase
    BEST <- diversity_objective_(full_clusters, x)
    for (i in 1:iterations) {
      full_clusters_new <- improvement_search_2pml(x, full_clusters, must_link)
      # full clusters new is a perturbed partition, not locally optimal -> restore local optimality!
      reduced_clusters_new <- original_cluster_to_merged_cluster(full_clusters_new, must_link)
      # restore local optimality
      reduced_clusters_new <- c_anticlustering(
        dt$distances,
        K = reduced_clusters_new,
        categories = lengths(dt$IDs), # restrict exchanges to node with same number of elements
        objective = "diversity",
        local_maximum = TRUE
      )
      full_clusters <- merged_cluster_to_original_cluster(reduced_clusters_new, must_link)
    }
  }
  full_clusters
}

# Given an assignment of original elements to clusters, generate the
# corresponding clusters for the merged elements. 
original_cluster_to_merged_cluster <- function(clusters, must_link) {
  df <- data.frame(clusters, must_link)
  new_df <- df[!duplicated(df$must_link), ]
  # order by must_link grouping
  new_df[order(new_df$must_link), ]$clusters
}

# Given an assignment of "merged" elements to clusters, re-establish the
# corresponding clusters for the original elements. 
merged_cluster_to_original_cluster <- function(merged_clusters, must_link) {
  df <- data.frame(must_link, order = 1:length(must_link))
  new_order <- order(must_link)
  df <- df[new_order, ]
  df$clusters <- rep(merged_clusters, table(must_link))
  df[order(df$order), "clusters"]
}

## Do an improvement phase to overcome bad local optima
# For each must-link clique
improvement_search_2pml <- function(x, full_clusters, must_link) {
  N <- length(full_clusters)
  # get current objective for optimization
  OBJ_BY_CLUSTER <- diversity_objective_by_group(full_clusters, x)
  OBJ <- sum(OBJ_BY_CLUSTER)
  ## after this: set diagonal to zero, for more convenient updates of objective
  diag(x) <- 0

  # book keeping of indices (separately for must-linked elements and "singletons" that are not must-linked)
  list_must_link_indices <- get_must_link_indices(must_link) 
  cliques <- get_cliques(list_must_link_indices) 
  singletons <- get_singletons(list_must_link_indices)
  singleton_clusters <- full_clusters[singletons]
  
  # do swapping phase; iterate over cliques
  # for each clique, generate exchange partners twice
  for (i in seq_along(cliques)) {
    n_clique <- length(cliques[[i]])
    # randomly determine if the clique is exchanged with singletons only or with other cliques (that is, clique(s) + possibly singleton(s) in combination) 
    singleton_change <- sample(c(TRUE, FALSE), size = 1)
    if (singleton_change || n_clique <= 2) { # for cliques of size 2, we can only exchange with two singletons
      exchange_cluster <- get_exchange_partners_singletons( # generate exchange partner from singletons / not part of clique
        singletons, 
        singleton_clusters, 
        n_clique
      )
    } else { # only do the clique swapping for clique larger than 2 samples
      exchange_cluster <- get_exchange_partners_clique(cliques, i, full_clusters, must_link) # generate exchange partner from other cliques
    }
    if (!is.null(exchange_cluster$cluster_id)) {
      ## Do the swap
      tmp_clusters <- full_clusters
      tmp <- tmp_clusters[cliques[[i]]][1]
      tmp_clusters[cliques[[i]]] <- exchange_cluster$cluster_id
      tmp_clusters[exchange_cluster$sample_ids] <- tmp
      # perform swap if it improves objective
      ## Use local updating here instead of recomputing entirely, improves speed quite a bit
      OBJ_BY_CLUSTER_NEW <- update_diversity_must_link(
        x, OBJ_BY_CLUSTER, 
        full_clusters, 
        tmp_clusters, 
        indices_1 = cliques[[i]], 
        indices_2 = exchange_cluster$sample_ids, 
        cluster_1 = tmp, 
        cluster_2 = exchange_cluster$cluster_id
      )
      OBJ_NEW <- sum(OBJ_BY_CLUSTER_NEW)
      if (OBJ_NEW > OBJ) {
        OBJ <- OBJ_NEW
        OBJ_BY_CLUSTER <- OBJ_BY_CLUSTER_NEW
        full_clusters <- tmp_clusters
        singleton_clusters <- full_clusters[singletons]
      }
    }
  }
  full_clusters
}

update_diversity_must_link <- function(x, OBJ_BY_CLUSTER, clusters_old, clusters_new, indices_1, indices_2, cluster_1, cluster_2) {
  sum_distances_element_1_cluster_1 <- sum(x[indices_1, , drop = FALSE][, clusters_old == cluster_1, drop = FALSE])
  sum_distances_element_1_cluster_2 <- sum(x[indices_1, , drop = FALSE][, clusters_new == cluster_2, drop = FALSE])
  sum_distances_element_2_cluster_2 <- sum(x[indices_2, , drop = FALSE][, clusters_old == cluster_2, drop = FALSE])
  sum_distances_element_2_cluster_1 <- sum(x[indices_2, , drop = FALSE][, clusters_new == cluster_1, drop = FALSE])
  OBJ_BY_CLUSTER[cluster_1] <- OBJ_BY_CLUSTER[cluster_1] + sum_distances_element_2_cluster_1 - sum_distances_element_1_cluster_1
  OBJ_BY_CLUSTER[cluster_2] <- OBJ_BY_CLUSTER[cluster_2] + sum_distances_element_1_cluster_2 - sum_distances_element_2_cluster_2
  OBJ_BY_CLUSTER
}

# get list of indices of must-link cliques (and singletons)
get_must_link_indices <- function(must_link) {
  N <- length(must_link)
  tapply(1:N, must_link, c)
}

# using the list of must-link indices as input, get singleton indices (as vector)
get_singletons <- function(list_must_link_indices) {
  unname(unlist(list_must_link_indices[lengths(list_must_link_indices) == 1]))
}

#using the list  of must-link indices as input, get indices of cliques  (as list)
get_cliques <- function(list_must_link_indices) {
  list_must_link_indices[lengths(list_must_link_indices) > 1]
}

# determine a random cluster that fits a clique (and sample IDs for the exchange)
get_exchange_partners_singletons <- function(singletons, singleton_clusters, n_clique) {
  tab <- table(singleton_clusters)
  fitting_clusters <- as.numeric(names(tab[tab >= n_clique]))
  # If no swap attempt is possible, return list will NULL
  if (length(fitting_clusters) == 0) {
    return(list(
        cluster_id = NULL, 
        sample_ids = NULL
    ))
  }
  cluster_id <- sample_(fitting_clusters, size = 1)
  list(
    cluster_id = cluster_id, 
    sample_ids = sample(singletons[singleton_clusters == cluster_id], size = n_clique)
  )
}


## This function is quite horrible right now and I hope to make it better
# It generates multiple exchange partners for a must-link clique, using
# a combination of cliques / singletons. It randomly selects one combination that
# fits. First it has to generate all possible combinations that generates 
# the size of a must-link group (e.g., 6 = 1+2+3; 2+4; 5+1). This is quite hard
# in general I think.
get_exchange_partners_clique <- function(cliques, index, full_clusters, must_link) {
  
  n_clique <- length(cliques[[index]])
  stopifnot(n_clique > 2) # this must be here

  nl <- valid_sums_clique(n_clique) # returns a list of combinations that sum to n_clique
  
  # now check if any of these combinations exists in a cluster
  tab <- table(full_clusters, must_link)
  # iterate through combinations in random order and use the first exchange partners that work
  random_order <- sample(length(nl))
  for (i in random_order) {
    # determine where all samples are available
    cluster_ids_that_fit <- which(colSums(!apply(tab, 1, function(x) nl[[i]] %in% x)) == 0)
    # found fit!
    if (length(cluster_ids_that_fit) > 0) {
      one_cluster_id_that_fit <- sample_(cluster_ids_that_fit, size = 1)
      selected_cluster <- full_clusters == one_cluster_id_that_fit
      # select IDs of elements that serve as exchange ṕartners, via size of cliques
      tab2 <- table(full_clusters[selected_cluster], must_link[selected_cluster])
      must_link_ids <- as.numeric(dimnames(tab2)[[2]][random_match(nl[[i]], tab2)])
      # from IDs from selected must-link groups, get sample IDs
      sample_ids <- which(must_link %in% must_link_ids)
      if (any(is.na(must_link_ids)) || length(sample_ids) != n_clique) {
        next
      }
      return(list(
        cluster_id = one_cluster_id_that_fit, 
        sample_ids = sample_ids
      ))
    }

  }
  return(list(
    cluster_id = NULL, 
    sample_ids = NULL
  ))
}

## I generated these sums separately because recomputation is too expensive
valid_sums_clique <- function(n_clique) { # return a list of combinations that sum to n_clique
  if (n_clique <= 10) {
    return(
        list("3" = list(1:2), "4" = list(c(1L, 3L), c(2L, 2L), c(1L, 
1L, 2L)), "5" = list(c(1L, 4L), 2:3, c(1L, 1L, 3L), c(1L, 2L, 
2L), c(1L, 1L, 1L, 2L)), "6" = list(c(1L, 5L), c(2L, 4L), c(3L, 
3L), c(1L, 1L, 4L), 1:3, c(2L, 2L, 2L), c(1L, 1L, 1L, 3L), c(1L, 
1L, 2L, 2L), c(1L, 1L, 1L, 1L, 2L)), "7" = list(c(1L, 6L), c(2L, 
5L), 3:4, c(1L, 1L, 5L), c(1L, 2L, 4L), c(1L, 3L, 3L), c(2L, 
2L, 3L), c(1L, 1L, 1L, 4L), c(1L, 1L, 2L, 3L), c(1L, 2L, 2L, 
2L), c(1L, 1L, 1L, 1L, 3L), c(1L, 1L, 1L, 2L, 2L), c(1L, 1L, 
1L, 1L, 1L, 2L)), "8" = list(c(1L, 7L), c(2L, 6L), c(3L, 5L), 
    c(4L, 4L), c(1L, 1L, 6L), c(1L, 2L, 5L), c(1L, 3L, 4L), c(2L, 
    2L, 4L), c(2L, 3L, 3L), c(1L, 1L, 1L, 5L), c(1L, 1L, 2L, 
    4L), c(1L, 1L, 3L, 3L), c(1L, 2L, 2L, 3L), c(2L, 2L, 2L, 
    2L), c(1L, 1L, 1L, 1L, 4L), c(1L, 1L, 1L, 2L, 3L), c(1L, 
    1L, 2L, 2L, 2L), c(1L, 1L, 1L, 1L, 1L, 3L), c(1L, 1L, 1L, 
    1L, 2L, 2L), c(1L, 1L, 1L, 1L, 1L, 1L, 2L)), "9" = list(c(1L, 
8L), c(2L, 7L), c(3L, 6L), 4:5, c(1L, 1L, 7L), c(1L, 2L, 6L), 
    c(1L, 3L, 5L), c(1L, 4L, 4L), c(2L, 2L, 5L), 2:4, c(3L, 3L, 
    3L), c(1L, 1L, 1L, 6L), c(1L, 1L, 2L, 5L), c(1L, 1L, 3L, 
    4L), c(1L, 2L, 2L, 4L), c(1L, 2L, 3L, 3L), c(2L, 2L, 2L, 
    3L), c(1L, 1L, 1L, 1L, 5L), c(1L, 1L, 1L, 2L, 4L), c(1L, 
    1L, 1L, 3L, 3L), c(1L, 1L, 2L, 2L, 3L), c(1L, 2L, 2L, 2L, 
    2L), c(1L, 1L, 1L, 1L, 1L, 4L), c(1L, 1L, 1L, 1L, 2L, 3L), 
    c(1L, 1L, 1L, 2L, 2L, 2L), c(1L, 1L, 1L, 1L, 1L, 1L, 3L), 
    c(1L, 1L, 1L, 1L, 1L, 2L, 2L), c(1L, 1L, 1L, 1L, 1L, 1L, 
    1L, 2L)), "10" = list(c(1L, 9L), c(2L, 8L), c(3L, 7L), c(4L, 
6L), c(5L, 5L), c(1L, 1L, 8L), c(1L, 2L, 7L), c(1L, 3L, 6L), 
    c(1L, 4L, 5L), c(2L, 2L, 6L), c(2L, 3L, 5L), c(2L, 4L, 4L
    ), c(3L, 3L, 4L), c(1L, 1L, 1L, 7L), c(1L, 1L, 2L, 6L), c(1L, 
    1L, 3L, 5L), c(1L, 1L, 4L, 4L), c(1L, 2L, 2L, 5L), 1:4, c(1L, 
    3L, 3L, 3L), c(2L, 2L, 2L, 4L), c(2L, 2L, 3L, 3L), c(1L, 
    1L, 1L, 1L, 6L), c(1L, 1L, 1L, 2L, 5L), c(1L, 1L, 1L, 3L, 
    4L), c(1L, 1L, 2L, 2L, 4L), c(1L, 1L, 2L, 3L, 3L), c(1L, 
    2L, 2L, 2L, 3L), c(2L, 2L, 2L, 2L, 2L), c(1L, 1L, 1L, 1L, 
    1L, 5L), c(1L, 1L, 1L, 1L, 2L, 4L), c(1L, 1L, 1L, 1L, 3L, 
    3L), c(1L, 1L, 1L, 2L, 2L, 3L), c(1L, 1L, 2L, 2L, 2L, 2L), 
    c(1L, 1L, 1L, 1L, 1L, 1L, 4L), c(1L, 1L, 1L, 1L, 1L, 2L, 
    3L), c(1L, 1L, 1L, 1L, 2L, 2L, 2L), c(1L, 1L, 1L, 1L, 1L, 
    1L, 1L, 3L), c(1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L), c(1L, 1L, 
    1L, 1L, 1L, 1L, 1L, 1L, 2L)))[[as.character(n_clique)]]
    )
  } else {
    # heuristic: return equal-sized cliques, but add 1s to compensate if n_clique is not divisible by X
    # (e.g., for n_clique = 12, we have 6x2, 4x3, 2x5+2x1, 2x6)
    max_number <- floor(n_clique/2)
    candidates <- get_set_for_subset_problem(n_clique)
    candidates <- candidates[candidates <= max_number] 
    subsets_init <- lapply(1:max_number, function(x) candidates[candidates == x])[-1] # first entry is the 1s, which we don't use here
    diffs <- lapply(lapply(subsets_init, sum), "-", n_clique) # get difference of sum to each 
    lapply(subsets_init, function(x) c(x, rep(1, n_clique - sum(x))))
  }
}

get_set_for_subset_problem <- function(n_clique) {
  init <- rep(1:n_clique, n_clique:1)
  init_list <- mapply(FUN = rep, 1:n_clique, n_clique:1)
  cumsums <- lapply(1:n_clique, function(x) cumsum(init[init == x]))
  valid_cumsums <- lapply(cumsums, function(x) x[x <= n_clique])
  indices <- lapply(valid_cumsums, function(x) 1:length(x))
  set <- unlist(mapply("[", init_list, indices)) # set of numbers that can add up to n_clique
  set <- set[!is.na(set)] # some cleanup
  set
}

## get a random (match for a) combination of fitting cliques rather than the first, which is returned by match()
random_match <- function(combination, frequencies) {
  frequencies <- c(frequencies) # comes in as table, not vector
  df <- data.frame(frequencies)
  df$original_order <- 1:nrow(df)
  df <- df[sample(1:nrow(df)), ]
  df$original_order[match_only_once(combination, df$frequencies)]
}

# match() will return the same index repeatedly, but we can't use that because it is "used" once found
# so match each value individually sequentially and remove indices afterwards.
match_only_once <- function(combination, frequencies) {
  matches <- rep(NA, length(combination))
  for (i in seq_along(combination)) {
    matches[i] <- match(combination[i], frequencies)
    frequencies[matches[i]] <- NA
  }
  matches
}

library("anticlust")
library("tinytest")

must_link <- c(1, 1, 2, 2, 3, 3, 3, 4, 5, 6)

ml_indices <- anticlust:::get_must_link_indices(must_link)
singletons <- anticlust:::get_singletons(ml_indices)
cliques <- anticlust:::get_cliques(ml_indices)
expect_true(all(singletons == 8:10))
expect_true(all(cliques[[1]] == 1:2))
expect_true(all(cliques[[2]] == 3:4))
expect_true(all(cliques[[3]] == 5:7))

## Test generating exchange partners

init <- anticlust:::merged_cluster_to_original_cluster(
  anticlust:::init_must_link_groups(
    length(must_link), 
    IDs_initial = must_link, 
    IDs_reduced = anticlust:::get_must_link_indices(must_link), 
    target_groups = c(5, 5)
  ), must_link
)

# the last 3 elements can serve as exchange partners for a clique of size 3
ep <- anticlust:::get_exchange_partners_singletons(singletons, init[singletons], 3)
expect_true(all(sort(ep$sample_ids) == 8:10))

# now test for clique of size 2 (any combination of the last 3 works)
ep <- anticlust:::get_exchange_partners_singletons(singletons, init[singletons], 2)
expect_true(all(ep$sample_ids %in% 8:10))

## test generating exchange partners from cliques (and singletons)
# (this does not work because clique must be larger than 2)
expect_error(anticlust:::get_exchange_partners_clique(cliques, 1, init, must_link))

## Third clique has size of 3. It must be matched against positions 1 and 2, and (8, 9, or 10)

anticlust:::get_exchange_partners_clique(cliques, 3, init, must_link)[["sample_ids"]]
tt <- replicate(1000, anticlust:::get_exchange_partners_clique(cliques, 3, init, must_link)[["sample_ids"]])

# We have two cliques of size 2. (1) Indices 1, 2; (2) Indices 3, 4.
# These two cliques must be partnered with (a) the clique of size 3, (b) the 3 singletons.

# Test that only those matches occur are matched
expect_true(all(apply(tt, 2, function(x) x == c(1:2, 8) | x == c(1:2, 9) | x == c(1:2, 10) | x == c(3:4, 8) | x == c(3:4, 9) | x == c(3:4, 10))))

## But not always c(1:2, 8) or c(3:4, 8)! (which we get when using `base::match()` rather than `anticlust:::random_match()`)
expect_true(any(!tt == c(c(1:2, 8))))
expect_true(any(!tt == c(c(3:4, 8))))

# Test random match function. The match should always include 5 and 1, but 2, 3 or 4 is arbitrary.
combination <- c(1, 2, 3)
frequencies <- c(2, 1, 1, 1, 3)
match(combination, frequencies) # always c(2, 1, 5); we also like to have c(3, 1, 5), c(4, 1, 5)
# However, positions 1 and 5 must always be matched
expect_true(all(replicate(100, 1 %in% anticlust:::random_match(combination, frequencies))))
expect_false(all(replicate(100, 2 %in% anticlust:::random_match(combination, frequencies))))
expect_true(all(replicate(100, 5 %in% anticlust:::random_match(combination, frequencies))))


## Test that a 4 clique can be exchanged with 2, 2 ### FAILS!

must_link <- c(1, 1, 1, 1, 5, 2, 2, 3, 3, 4)

ml_indices <- anticlust:::get_must_link_indices(must_link)
cliques <- anticlust:::get_cliques(ml_indices)
init <- anticlust:::merged_cluster_to_original_cluster(
  anticlust:::init_must_link_groups(
    length(must_link), 
    IDs_initial = must_link, 
    IDs_reduced = anticlust:::get_must_link_indices(must_link), 
    target_groups = c(5, 5)
  ), must_link
)

exchanges <- anticlust:::get_exchange_partners_clique(cliques, 1, init, must_link)
expect_true(!is.null(exchanges$cluster_id))
expect_true(all(exchanges$sample_ids == 6:9))

## Test that a 5 clique can be exchanged with 2, 2, 1

must_link <- c(1, 1, 1, 1, 1, 2, 2, 3, 3, 4)

ml_indices <- anticlust:::get_must_link_indices(must_link)
cliques <- anticlust:::get_cliques(ml_indices)
init <- anticlust:::merged_cluster_to_original_cluster(
  anticlust:::init_must_link_groups(
    length(must_link), 
    IDs_initial = must_link, 
    IDs_reduced = anticlust:::get_must_link_indices(must_link), 
    target_groups = c(5, 5)
  ), must_link
)

exchanges <- anticlust:::get_exchange_partners_clique(cliques, 1, init, must_link)
expect_true(!is.null(exchanges$cluster_id))
expect_true(all(exchanges$sample_ids == 6:10))

## Test stuff

N <- 100
must_link <- sample(N, N, replace = TRUE)

ml_indices <- anticlust:::get_must_link_indices(must_link)
cliques <- anticlust:::get_cliques(ml_indices)

init <- tryCatch(
  init <- anticlust:::merged_cluster_to_original_cluster(
    anticlust:::init_must_link_groups(
      length(must_link), 
      IDs_initial = must_link, 
      IDs_reduced = anticlust:::get_must_link_indices(must_link), 
      target_groups = c(N/2, N/2)
    ), must_link
  ),
  error = function(e) e
)

if (!"simpleError" %in% class(init)) {
  exchanges <- anticlust:::get_exchange_partners_clique(
    cliques, 
    which(sapply(cliques, function(x) length(x) > 2))[1], init, must_link
  )
  # there must at least be one clique among the exchange partners:
  expect_true(any(duplicated(must_link[exchanges$sample_ids])))
}


########################################
### Now do some general testing on the must-link constraints. E.g., are they valid after anticlustering?

# Function that test if constraints are valid
must_link_constraints_valid <- function(cl, must_link) {
  same <- as.numeric(names(table(must_link)[table(must_link) > 1]))
  all_good <- c()
  for (i in same) {
    all_good <- c(all_good, all(cl[must_link == i] == cl[must_link == i][1]))
  }
  all(all_good)
}

N <- 10

must_link <- sample(N, size = N, replace = TRUE)


tt <- tryCatch(
  anticlustering(1:N, K = 2, must_link = must_link),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  tt2 <- anticlustering(1:N, K = 2, must_link = must_link, method = "2PML")
  
  expect_true(must_link_constraints_valid(tt, must_link))
  expect_true(must_link_constraints_valid(tt2, must_link))
  
  # test constructing merged clusters from must-link constraints and its reversal
  expect_true(all(
    tt == anticlust:::merged_cluster_to_original_cluster(
      anticlust:::original_cluster_to_merged_cluster(tt, must_link), 
      must_link
    )
  ))
}


## now a random larger data set

K <- 20
group_sizes <- 12
N <- group_sizes * K # here 128
M <- 5
data <- matrix(rnorm(N * M), ncol = M)
distances <- as.matrix(dist(data))

# Generate random patient IDs
must_link <- sample(N, replace = TRUE)


tt1 <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "2PML"),
  error = function(e) e
)

if (!"simpleError" %in% class(tt1)) {
  expect_true(must_link_constraints_valid(tt1, must_link))
}


tt2 <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "2PML", repetitions = 40),
  error = function(e) e
)

if (!"simpleError" %in% class(tt2)) {
  expect_true(must_link_constraints_valid(tt2, must_link))
}

tt0 <- tryCatch(
  anticlustering(distances, K, must_link = must_link),# exchange method / Phase 1 of 2PML without ensured local optimality
  error = function(e) e
)

if (!"simpleError" %in% class(tt0)) {
  diversity_objective(distances, tt0)
  diversity_objective(distances, tt1)
  diversity_objective(distances, tt2)
}

library(anticlust)
library(tinytest)

## Test "vanilla" bin packing

n_batches <- 20
n <- 50
capacities <- rep(10, n_batches)
weights <- sample(5, size = n, replace = TRUE)

opt <- tryCatch(
  anticlust:::optimal_binpacking_(capacities, weights),
  error = function(e) e
)

error <- "simpleError" %in% class(opt)

if (!error) {
  table(opt)
  sums <- tapply(weights, opt, sum)
  sums_all <- rep(0, length(capacities))
  sums_all[as.numeric(names(sums))] <- sums
  # are capacity limits not exceeded?
  expect_true(all(sums_all <= capacities))
}




########################
## Test bin packing as used in must-link anticlustering

must_link_fulfilled <- function(ID, groups_must_link) {
  same <- as.numeric(names(table(ID)[table(ID) > 1]))
  all_good <- c()
  for (i in same) {
    all_good <- c(all_good, all(groups_must_link[ID == i] == groups_must_link[ID == i][1]))
  }
  all(all_good)
}

N <- 120
K <- 10
capacities <- rep(N/K, K)
must_link <- sample(N, replace = TRUE)

start <- Sys.time()
opt <- tryCatch(
  anticlust:::optimal_binpacking_(capacities, table(must_link)),
  error = function(e) e
)
Sys.time() - start

error <- "simpleError" %in% class(opt)

if (!error) {
  table(opt)

  ## This is how we would use optimal bin packing in must-link anticlustering
  # Users cannot explicitly request the optimal method currently, so
  # I emulate how it is done in anticlust
  init <- anticlust:::get_init_assignments(N, ID = must_link, target_groups = capacities, method = "optimal")
  init <- anticlust:::add_unassigned_elements(capacities, init, N, length(capacities))
  # Output of bin packing + filling is as expected?
  expect_true(all(table(init) == capacities))
  # Must-link constraints are met?
  expect_true(must_link_fulfilled(must_link, init))
  
  ## TODO ALWAYS USE TRYCATCH WITH MUST LINK CONSTRAINTS IN CASE THEY CANNOT BE FULFILLED
  # (and then the test cases would fail!)
  groups_must_link <- anticlustering(
    rnorm(N), K = K, must_link = must_link
  )
  expect_true(must_link_fulfilled(must_link, groups_must_link))
  
}

library("anticlust")
library("tinytest")

must_link <- c(1, 1, 2, 2, 3, 3, 3, 4, 5, 6)

ml_indices <- anticlust:::get_must_link_indices(must_link)
singletons <- anticlust:::get_singletons(ml_indices)
cliques <- anticlust:::get_cliques(ml_indices)
expect_true(all(singletons == 8:10))
expect_true(all(cliques[[1]] == 1:2))
expect_true(all(cliques[[2]] == 3:4))
expect_true(all(cliques[[3]] == 5:7))

## Test generating exchange partners

init <- anticlust:::merged_cluster_to_original_cluster(
  anticlust:::init_must_link_groups(
    length(must_link), 
    IDs_initial = must_link, 
    IDs_reduced = anticlust:::get_must_link_indices(must_link), 
    target_groups = c(5, 5)
  ), must_link
)

# the last 3 elements can serve as exchange partners for a clique of size 3
ep <- anticlust:::get_exchange_partners_singletons(singletons, init[singletons], 3)
expect_true(all(sort(ep$sample_ids) == 8:10))

# now test for clique of size 2 (any combination of the last 3 works)
ep <- anticlust:::get_exchange_partners_singletons(singletons, init[singletons], 2)
expect_true(all(ep$sample_ids %in% 8:10))

## test generating exchange partners from cliques (and singletons)
# (this does not work because clique must be larger than 2)
expect_error(anticlust:::get_exchange_partners_clique(cliques, 1, init, must_link))

## Third clique has size of 3. It must be matched against positions 1 and 2, and (8, 9, or 10)

anticlust:::get_exchange_partners_clique(cliques, 3, init, must_link)[["sample_ids"]]
tt <- replicate(1000, anticlust:::get_exchange_partners_clique(cliques, 3, init, must_link)[["sample_ids"]])

# We have two cliques of size 2. (1) Indices 1, 2; (2) Indices 3, 4.
# These two cliques must be partnered with (a) the clique of size 3, (b) the 3 singletons.

# Test that only those matches occur are matched
expect_true(all(apply(tt, 2, function(x) x == c(1:2, 8) | x == c(1:2, 9) | x == c(1:2, 10) | x == c(3:4, 8) | x == c(3:4, 9) | x == c(3:4, 10))))

## But not always c(1:2, 8) or c(3:4, 8)! (which we get when using `base::match()` rather than `anticlust:::random_match()`)
expect_true(any(!tt == c(c(1:2, 8))))
expect_true(any(!tt == c(c(3:4, 8))))

# Test random match function. The match should always include 5 and 1, but 2, 3 or 4 is arbitrary.
combination <- c(1, 2, 3)
frequencies <- c(2, 1, 1, 1, 3)
match(combination, frequencies) # always c(2, 1, 5); we also like to have c(3, 1, 5), c(4, 1, 5)
# However, positions 1 and 5 must always be matched
expect_true(all(replicate(100, 1 %in% anticlust:::random_match(combination, frequencies))))
expect_false(all(replicate(100, 2 %in% anticlust:::random_match(combination, frequencies))))
expect_true(all(replicate(100, 5 %in% anticlust:::random_match(combination, frequencies))))


## Test that a 4 clique can be exchanged with 2, 2 ### FAILS!

must_link <- c(1, 1, 1, 1, 5, 2, 2, 3, 3, 4)

ml_indices <- anticlust:::get_must_link_indices(must_link)
cliques <- anticlust:::get_cliques(ml_indices)
init <- anticlust:::merged_cluster_to_original_cluster(
  anticlust:::init_must_link_groups(
    length(must_link), 
    IDs_initial = must_link, 
    IDs_reduced = anticlust:::get_must_link_indices(must_link), 
    target_groups = c(5, 5)
  ), must_link
)

exchanges <- anticlust:::get_exchange_partners_clique(cliques, 1, init, must_link)
expect_true(!is.null(exchanges$cluster_id))
expect_true(all(exchanges$sample_ids == 6:9))

## Test that a 5 clique can be exchanged with 2, 2, 1

must_link <- c(1, 1, 1, 1, 1, 2, 2, 3, 3, 4)

ml_indices <- anticlust:::get_must_link_indices(must_link)
cliques <- anticlust:::get_cliques(ml_indices)
init <- anticlust:::merged_cluster_to_original_cluster(
  anticlust:::init_must_link_groups(
    length(must_link), 
    IDs_initial = must_link, 
    IDs_reduced = anticlust:::get_must_link_indices(must_link), 
    target_groups = c(5, 5)
  ), must_link
)

exchanges <- anticlust:::get_exchange_partners_clique(cliques, 1, init, must_link)
expect_true(!is.null(exchanges$cluster_id))
expect_true(all(exchanges$sample_ids == 6:10))

## Test stuff

N <- 100
must_link <- sample(N, N, replace = TRUE)

ml_indices <- anticlust:::get_must_link_indices(must_link)
cliques <- anticlust:::get_cliques(ml_indices)

init <- tryCatch(
  init <- anticlust:::merged_cluster_to_original_cluster(
    anticlust:::init_must_link_groups(
      length(must_link), 
      IDs_initial = must_link, 
      IDs_reduced = anticlust:::get_must_link_indices(must_link), 
      target_groups = c(N/2, N/2)
    ), must_link
  ),
  error = function(e) e
)

if (!"simpleError" %in% class(init)) {
  exchanges <- anticlust:::get_exchange_partners_clique(
    cliques, 
    which(sapply(cliques, function(x) length(x) > 2))[1], init, must_link
  )
  # there must at least be one clique among the exchange partners:
  expect_true(any(duplicated(must_link[exchanges$sample_ids])))
}


########################################
### Now do some general testing on the must-link constraints. E.g., are they valid after anticlustering?

# Function that test if constraints are valid
must_link_constraints_valid <- function(cl, must_link) {
  same <- as.numeric(names(table(must_link)[table(must_link) > 1]))
  all_good <- c()
  for (i in same) {
    all_good <- c(all_good, all(cl[must_link == i] == cl[must_link == i][1]))
  }
  all(all_good)
}

N <- 10

must_link <- sample(N, size = N, replace = TRUE)


tt <- tryCatch(
  anticlustering(1:N, K = 2, must_link = must_link),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  tt2 <- tryCatch(
    anticlustering(1:N, K = 2, must_link = must_link, method = "2PML"),
    error = function(e) e
  )
  if (!"simpleError" %in% class(tt2)) {
    expect_true(must_link_constraints_valid(tt, must_link))
    expect_true(must_link_constraints_valid(tt2, must_link))
    
    # test constructing merged clusters from must-link constraints and its reversal
    expect_true(all(
      tt == anticlust:::merged_cluster_to_original_cluster(
        anticlust:::original_cluster_to_merged_cluster(tt, must_link), 
        must_link
      )
    ))
  }
}


## now a random larger data set

K <- 20
group_sizes <- 12
N <- group_sizes * K # here 128
M <- 5
data <- matrix(rnorm(N * M), ncol = M)
distances <- as.matrix(dist(data))

# Generate random patient IDs
must_link <- sample(N, replace = TRUE)


tt1 <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "2PML"),
  error = function(e) e
)

if (!"simpleError" %in% class(tt1)) {
  expect_true(must_link_constraints_valid(tt1, must_link))
}


tt2 <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "2PML", repetitions = 40),
  error = function(e) e
)

if (!"simpleError" %in% class(tt2)) {
  expect_true(must_link_constraints_valid(tt2, must_link))
}

tt0 <- tryCatch(
  anticlustering(distances, K, must_link = must_link),# exchange method / Phase 1 of 2PML without ensured local optimality
  error = function(e) e
)

if (!"simpleError" %in% class(tt0)) {
  diversity_objective(distances, tt0)
  diversity_objective(distances, tt1)
  diversity_objective(distances, tt2)
}

#' Solve exact equal-sized cluster editing
#'
#' @param data A N x N distance matrix or N x M features table.
#' @param K How many clusters are to be created.
#' @param solver "glpk" or "symphony"
#'
#' @return A vector representing the clustering.
#'
#' @noRd
#' 
#'
balanced_cluster_editing <- function(data, K, solver) {
  if (!is_distance_matrix(data)) {
    data <- as.matrix(dist(data))
  }
  ilp <- anticlustering_ilp(data, K)
  solution <- solve_ilp(ilp, "min", solver)
  ilp_to_groups(solution, nrow(data))
}

#' Anticlustering
#'
#' Partition a pool of elements into groups (i.e., anticlusters) with
#' the aim of creating high within-group heterogeneity and high
#' between-group similarity.  Anticlustering is accomplished by
#' maximizing instead of minimizing a clustering objective function.
#' Implements anticlustering methods as described in Papenberg and
#' Klau (2021; <doi:10.1037/met0000301>), Brusco et al. 
#' (2020; <doi:10.1111/bmsp.12186>), Papenberg (2024; 
#' <doi:10.1111/bmsp.12315>), Papenberg, Wang, et al. (2025; 
#' <doi:10.1016/j.crmeth.2025.101137>), Papenberg, Breuer, et al. (2025; 
#' <doi:10.1017/psy.2025.10052>), and Yang et al. (2022; <doi:10.1016/j.ejor.2022.02.003>). 
#'
#' @param x The data input. Can be one of two structures: (1) A
#'     feature matrix where rows correspond to elements and columns
#'     correspond to variables (a single numeric variable can be
#'     passed as a vector). (2) An N x N matrix dissimilarity matrix;
#'     can be an object of class \code{dist} (e.g., returned by
#'     \code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
#'     where the entries of the upper and lower triangular matrix
#'     represent pairwise dissimilarities. 
#' @param K How many anticlusters should be created. Alternatively:
#'     (a) A vector describing the size of each group, or (b) a vector
#'     of length \code{nrow(x)} describing how elements are assigned
#'     to anticlusters before the optimization starts.
#' @param objective The objective to be maximized. The options
#'     "diversity" (default; previously called "distance", which is
#'     still supported), "average-diversity", "variance", "kplus" and "dispersion" are
#'     natively supported. May also be a user-defined function. See
#'     Details.
#' @param method One of "exchange" (default) , "local-maximum",
#'     "brusco", "ilp", "2PML", or "3phase".  See Details.
#' @param preclustering Boolean. Should a preclustering be conducted
#'     before anticlusters are created? Defaults to \code{FALSE}. See
#'     Details.
#' @param categories A vector, data.frame or matrix representing one
#'     or several categorical variables whose distribution should be similar 
#'     between groups. See Details.
#' @param repetitions The number of times a search heuristic is
#'     initiated when using \code{method = "exchange"}, \code{method =
#'     "local-maximum"}, \code{method = "brusco"}, \code{method = "3phase"}, 
#'     or \code{method = "2PML"}. In the end, the best objective found across 
#'     the repetitions is returned.
#' @param standardize Boolean. If \code{TRUE} and \code{x} is a
#'     feature matrix, the data is standardized through a call to
#'     \code{\link{scale}} before the optimization starts. This
#'     argument is silently ignored if \code{x} is a distance matrix.
#' @param cannot_link A 2 column matrix where each row has the indices 
#'     of two elements that must not be assigned to the same anticluster. 
#'     Alternatively a vector of length \code{nrow(x)}; elements having 
#'     the same value in this vector cannot be assigned to the same anticluster.
#' @param must_link A numeric vector of length \code{nrow(x)}. Elements having 
#'     the same value in this vector are assigned to the same anticluster.
#' @param blocks A vector, data.frame or matrix representing one
#'     or several categorical variables. Anticlustering is done sequentially 
#'     within blocks. See details.
#'
#' @return A vector of length N that assigns a group (i.e, a number
#'     between 1 and \code{K}) to each input element.
#'
#' @importFrom Matrix sparseMatrix
#' @importFrom stats as.dist dist sd
#' 
#' @useDynLib anticlust, .registration = TRUE
#'
#' @export
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' @details
#'
#' This function is used to solve anticlustering. That is, the data
#' input is divided into \code{K} groups in such a way that elements
#' within groups are heterogeneous and the different groups are
#' similar. Anticlustering is accomplished by maximizing instead of
#' minimizing a clustering objective function. The maximization of
#' five objectives is natively supported (other
#' functions can also defined by the user as described below):
#' 
#' \itemize{
#'   \item{the diversity, setting 
#'         \code{objective = "diversity"} (this is the default objective)}
#'   \item{the average diversity, which normalizes the diversity by cluster size, 
#'        setting \code{objective = "average-diversity"}}
#'   \item{the k-means (or "variance") objective, setting \code{objective = "variance"}}
#'   \item{the k-plus objective, an extension of the k-means objective,
#'         setting \code{objective = "kplus"}}
#'   \item{the dispersion, which is the minimum distance between 
#'         any two elements within the same cluster (setting 
#'         \code{objective = "dispersion"})}
#' }
#'
#' The k-means objective is the within-group variance---that is, the
#' sum of the squared distances between each element and its cluster
#' center (see \code{\link{variance_objective}}). K-means
#' anticlustering focuses on minimizing differences with regard to the
#' means of the input variables (that is, the columns in \code{x}), but it ignores any other distribution
#' characterstics such as the variance / standard deviation. K-plus anticlustering
#' (using \code{objective = "kplus"}) is an extension of the k-means criterion that also
#' minimizes differences with regard to the standard
#' deviations between groups (for details see \code{\link{kplus_anticlustering}}). K-plus
#' anticlustering can also be extended towards higher order moments such as skew and kurtosis; 
#' to consider these additional distribution characteristics, use the function
#' \code{\link{kplus_anticlustering}}. Setting \code{objective = "kplus"} in 
#' \code{anticlustering} function will only consider means 
#' and standard deviations (in my experience, this is what users usually want). 
#' It is strongly recommended to set the argument \code{standardize = TRUE} 
#' when using the k-plus objective.
#' 
#' The "diversity" objective is the sum of pairwise
#' distances of elements within the same groups (see
#' \code{\link{diversity_objective}}). Hence, anticlustering using the diversity 
#' criterion maximizes between-group similarity
#' by maximizing within-group heterogeneity (represented as the sum of all pairwise distances). 
#' In the publication that introduced
#' the \code{anticlust} package (Papenberg & Klau, 2021), we used the term "anticluster 
#' editing" to refer to the maximization of the diversity, because the reversed 
#' procedure - minimizing the diversity - is also known as "cluster editing".
#' If the diversity is computed on the basis of the Euclidean distance (which is the default
#' behaviour if \code{x} is a feature matrix), the diversity is an all rounder objective that 
#' tends to equate all distribution characteristics between groups (such as means, variances, ...). 
#' Note that the equivalence of within-group heterogeneity and between-group similarity only
#' holds for equal-sized groups. For unequal-sized groups, it is recommended to
#' use a different objective when striving for overall between-group similarity,
#' e.g., the k-plus objective or the \code{"average-diversity"}. The average diversity
#' was introduced in version 0.8.6, and it is more useful if groups are not 
#' equal-sized. The average diversity normalizes the sum of intra-cluster distances 
#' by group size. If all groups are equal-sized, it is equivalent to the 
#' regular diversity. 
#' 
#' The "dispersion" is the minimum distance between any two elements
#' that are part of the same cluster; maximization of this objective
#' ensures that any two elements within the same group are as
#' dissimilar from each other as possible. Applications that require
#' high within-group heterogeneity often require to maximize the
#' dispersion. Oftentimes, it is useful to also consider the diversity
#' and not only the dispersion; to optimize both objectives at the
#' same time, see the function
#' \code{\link{bicriterion_anticlustering}}.
#'
#' In the standard case, groups of equal size are generated. Adjust
#' the argument \code{K} to create groups of different size (see
#' Examples).
#' 
#' 
#' \strong{Data input}
#' 
#' The data input \code{x} is usually a feature matrix where each each row
#' is a "case" and each column is a "variable". The default
#' \code{objective = "diversity"} or the \code{objective = "dispersion"} are computed
#' using a dissimilarity matrix as input, however. The default behaviour for 
#' feature matrices is to compute the Euclidean distance as the 
#' basic unit of these objectives, using a call to \code{\link{dist}}. 
#' The combination \code{objective = "diversity"} + 
#' Euclidean distance was evaluated by us (e.g., Papenberg and Klau, 2021; 
#' Papenberg, 2024; Papenberg, Wang, et al., 2025) and is generally considered 
#' to be a good default for equal-sized groups. For unequal-sized groups, prefer 
#' \code{objective = "average-diversity"}. 
#' 
#' If a different type of dissimilarity measure is preferred (e.g., the Gower distance), 
#' you may pass a self-generated dissimilarity matrix via the argument \code{x}. This dissimilarity
#' matrix can be an object of class \code{dist} (e.g., returned by \code{\link{dist}} 
#' or \code{\link{as.dist}}) or a \code{matrix} where the entries of the upper 
#' and lower triangular matrix represent pairwise dissimilarities. If the input
#' is a matrix rather than of class \code{dist}, the function internally tests if
#' the input is a dissimilarity matrix (because a \code{matrix} could also be a 
#' feature matrix). It assumes the input is a dissimilarity matrix if the 
#' upper and lower triangular consists of the same values, and if the diagonal 
#' only consists of the same values. If the input is a \code{data.frame},
#' this function always assumes that it is a feature matrix. In a \code{data.frame},
#' each column must be either \code{numeric} or \code{factor} (the latter for 
#' categorical variables).
#' 
#' As of version 0.8.12, this function supports handling of missing values (\code{NA}). 
#' In this case the function \code{\link[stats]{dist}} handles the \code{NA}s when converting
#' features to pairwise distances. Note that \code{\link[stats]{dist}} always handles
#' missing values even for objectives that are usually computed using a feature 
#' matrix as input (i.e., \code{objective = "variance"} and \code{objective = "kplus"}).
#' In this case, we exploit an equivalence between the \code{objective = "variance"} and
#' the \code{objective = "average-diversity"}: When the squared Euclidean distance
#' \code{objective = "variance"} and the \code{objective = "average-diversity"}
#' are actually equivalent, so we can optimize the \code{objective = "average-diversity"}
#' in order to perform k-means (and thus, k-plus) anticlustering. See Papenberg, 
#' Breuer, et al. (2025) for details.
#' 
#' 
#' \strong{Algorithms for anticlustering}
#'
#' By default, a heuristic method is employed for anticlustering: the
#' exchange method (\code{method = "exchange"}). First, elements are
#' randomly assigned to anticlusters (It is also possible to
#' explicitly specify the initial assignment using the argument
#' \code{K}; in this case, \code{K} has length \code{nrow(x)}.) Based
#' on the initial assignment, elements are systematically swapped
#' between anticlusters in such a way that each swap improves the
#' objective value. For an element, each possible swap with elements
#' in other clusters is simulated; then, the one swap is performed
#' that improves the objective the most, but a swap is only conducted
#' if there is an improvement at all. This swapping procedure is
#' repeated for each element. When using \code{method =
#' "local-maximum"}, the exchange method does not terminate after the
#' first iteration over all elements; instead, the swapping continues
#' until a local maximum is reached. This method corresponds to the algorithm 
#' "LCW" by Weitz and Lakshminarayanan (1998). This means that after the
#' exchange process has been conducted once for each data point, the
#' algorithm restarts with the first element and proceeds to conduct
#' exchanges until the objective cannot be improved.
#'
#' When setting \code{preclustering = TRUE}, only the \code{K - 1}
#' most similar elements serve as exchange partners for each element,
#' which can speed up the optimization (more information
#' on the preclustering heuristic follows below). If the \code{categories} argument
#' is used, only elements having the same value in \code{categories} serve as exchange
#' partners.
#' 
#' Using \code{method = "brusco"} implements the local bicriterion
#' iterated local search (BILS) heuristic by Brusco et al. (2020) and
#' returns the partition that best optimized either the diversity or
#' the dispersion during the optimization process. The function
#' \code{\link{bicriterion_anticlustering}} can also be used to run
#' the algorithm by Brusco et al., but it returns multiple partitions
#' that approximate the optimal pareto efficient set according to both
#' objectives (diversity and dispersion). Thus, to fully utilize the
#' BILS algorithm, use the function
#' \code{\link{bicriterion_anticlustering}}.
#' 
#' Since version 0.8.12, \code{method = "3phase"} calls the three phase search algorithm 
#' for anticlustering by Yang et al. (2022). For anticlust, we implemented some
#' changes of their default procedure, in particular they used a maximum "time"
#' that the algorithm runs before terminating. For consistency with our other 
#' methods in anticlust, we are however using a maximum number of repetitions. 
#' Our default is 50 repetitions, which may be increased (but not decreased!) 
#' with the \code{repetitions} argument. 
#'
#' \strong{Optimal anticlustering}
#'
#' Usually, heuristics are employed to tackle anticlustering problems,
#' and their performance is generally very satisfying.  However,
#' heuristics do not investigate all possible group assignments and
#' therefore do not (necessarily) find the
#' "globally optimal solution", i.e., a partitioning that has the best
#' possible value with regard to the objective that is optimized.  Enumerating
#' all possible partitions in order to find the best solution,
#' however, quickly becomes impossible with increasing N, and
#' therefore it is not possible to find a global optimum this
#' way. Because all anticlustering problems considered here are also
#' NP-hard, there is also no (known) clever algorithm that might
#' identify the best solution without considering all possibilities -
#' at least in the worst case. Integer linear programming (ILP) is an
#' approach for tackling NP hard problems that nevertheless tries to
#' be clever when finding optimal solutions: It does not necessarily
#' enumerate all possibilities but is still guaranteed to return the
#' optimal solution. Still, for NP hard problems such as
#' anticlustering, ILP methods will also fail at some point (i.e.,
#' when N increases).
#'
#' \code{anticlust} implements optimal solution algorithms via integer
#' linear programming. In order to use the ILP methods, set
#' \code{method = "ilp"}. The integer linear program optimizing the
#' diversity was described in Papenberg & Klau, (2021; (8) -
#' (12)). It can also be used to optimize the k-means and k-plus objectives,
#' but you actually have to use the function \code{\link{optimal_anticlustering}}
#' for these objectives. The documentation of the function
#' \code{\link{optimal_dispersion}} and \code{\link{optimal_anticlustering}}
#' contain more information on the optimal anticlustering algorithms.
#' 
#' \strong{Categorical variables}
#'
#' There are several ways to balance categorical variables among anticlusters (also 
#' see the package vignette "Using categorical variables with anticlustering").
#' The first way is to treat them as "hard constraints" via the argument 
#' \code{categories} (see Papenberg & Klau, 2021). If done so, balancing the 
#' categorical variable is accomplished via \code{categorical_sampling} through
#' a stratified split before the anticlustering optimization. After that, the 
#' balance is never changed when the algorithm runs (hence, it is a "hard constraint"). 
#' When \code{categories} has multiple columns (i.e., there are multiple 
#' categorical variables), each combination of categories is treated as a
#'  distinct category by the exchange method (i.e., the multiple columns
#' are "merged" into a single column). This behaviour may lead
#' to less than optimal results on the level of each single categorical variable.
#' In this case, it may be useful to treat the categorical variables as part of 
#' the numeric data, i.e., the first argument \code{x} via binary coding 
#' (e.g. using \code{\link{categories_to_binary}}). The examples show how to do this 
#' when using the bicriterion algorithm by Bruso et al. Using the argument 
#' \code{categories} is only available for the classical exchange procedures, 
#' that is, for \code{method = "exchange"} and \code{method = "local-maximum"}. 
#' 
#' As of version 0.8.12, the data input \code{x} can contain variables of type \code{factor}
#' to represent categorical variables. In this case, they are internally converted to 
#' binary via \code{\link{categories_to_binary}}, which no longer needs to be done 
#' manually by users (and which can be tricky in particular when combining categorical 
#' and numeric variables, and when using the k-plus objective). However note that the internal
#' conversion always uses \code{use_combinations = FALSE}, which may not always 
#' what users need. In this case, I still recommend using \code{\link{categories_to_binary}} 
#' manually.
#' 
#' The argument \code{blocks} (available as of version 0.8.13) is similar 
#' to \code{categories}. Anticlustering is repeated within each stratum
#' (i.e., each levels or combination of levels of \code{blocks} -- as in \code{categories}
#' mulitple variables are merged into a single variable to form strata). Anticlustering
#' in later blocks "remembers" the assignment of previous blocks. This way, 
#' overall balance as well as balance within blocks is maximized. This is unlike 
#' when using \code{categories}, which only maximizes overall balance, but not 
#' specifically within strata. 
#' 
#' \strong{Anticlustering with constraints}
#' 
#' Versions 0.8.6 and 0.8.7 of anticlust introduced the possibility to induce
#' cannot-link and must-link constraints with anticlustering with 
#' the arguments \code{cannot_link} and \code{must_link}, respectively.
#' Cannot-link constraints ensure that pairs of items are assigned to different
#' clusters. They are given as a 2-column matrix, where each row has the indices
#' of two elements, which must not be assigned to the same cluster. It is possible
#' that a set of cannot-link constraints cannot be fulfilled. To verify whether
#' the constraints cannot be fulfilled (and to actually assign elements 
#' while respecting the constraints), a graph coloring algorithm algorithm is used.
#' This algorithm is is actually the same method as used in \code{\link{optimal_dispersion}}. 
#' The graph coloring algorithm uses an ILP solver and it greatly profits (that is,
#' it may be much faster) from the Rsymphony package, which is not installed as 
#' a necessary dependency with anticlust. It is therefore recommended to 
#' manually install the Rsymphony package, which is then automatically 
#' selected as solver when using the \code{must_link} argument. If you have
#' access to the gurobi solver and have the gurobi R package installed, it will
#' be selected as solver (which is even faster than Symphony). As of version 0.8.11, 
#' it is also possible to use \code{cannot_link} as a vector. In this case it 
#' is ensured that elements having the same value in \code{cannot_link} are not
#' linked in the same cluster. The methods for handling \code{cannot_link} constraints
#' have been described in Papenberg, Breuer, et al. (2025). 
#' 
#' Must-link constraints are passed as a single vector of length \code{nrow(x)}.
#' Positions that have the same numeric index are assigned to the same anticluster 
#' (if the constraints can be fulfilled). When including must-link constraints, 
#' \code{method = "2PML"} performs a specialized search heuristic that potentially
#' yields better results than \code{method = "local-maximum"}. The must-link 
#' functionality and the 2PML algorithm was introduced in Papenberg, Wang, et al. (2025).
#' 
#' The examples illustrate the usage of the \code{must_link} and \code{cannot_link}
#' arguments. Currently, the different kinds of constraints (arguments \code{must_link}, 
#' \code{cannot_link}, and \code{categories}) cannot be used together, but this 
#' may change in future versions. 
#' 
#' \strong{Preclustering}
#' 
#' A useful heuristic for anticlustering is to form small groups of
#' very similar elements and assign these to different groups. This
#' logic is used as a preprocessing when setting \code{preclustering =
#' TRUE}. That is, before the anticlustering objective is optimized, a
#' cluster analysis identifies small groups of similar elements (pairs
#' if \code{K = 2}, triplets if \code{K = 3}, and so forth). The
#' optimization of the anticlustering objective is then conducted
#' under the constraint that these matched elements cannot be assigned
#' to the same group. When using the exchange algorithm, preclustering
#' is conducted using a call to \code{\link{matching}}. When using
#' \code{method = "ilp"}, the preclustering optimally finds groups of
#' minimum pairwise distance by solving the integer linear program
#' described in Papenberg and Klau (2021; (8) - (10), (12) - (13)).
#' Note that when combining preclustering restrictions with \code{method = "ilp"},
#' the anticlustering result is no longer guaranteed to be globally optimal, but
#' only optimal given the preclustering restrictions.
#' 
#' 
#' \strong{Optimize a custom objective function}
#' 
#' It is possible to pass a \code{function} to the argument
#' \code{objective}. See \code{\link{dispersion_objective}} for an
#' example. If \code{objective} is a function, the exchange method
#' assigns elements to anticlusters in such a way that the return
#' value of the custom function is maximized (hence, the function
#' should return larger values when the between-group similarity is
#' higher). The custom function has to take two arguments: the first
#' is the data argument, the second is the clustering assignment. That
#' is, the argument \code{x} will be passed down to the user-defined
#' function as first argument. \strong{However, only after}
#' \code{\link{as.matrix}} has been called on \code{x}. This implies
#' that in the function body, columns of the data set cannot be
#' accessed using \code{data.frame} operations such as
#' \code{$}. Objects of class \code{dist} will be converted to matrix
#' as well.
#' 
#' 
#' @examples
#'
#' # Use default method ("exchange") and the default diversity criterion, also include
#' # a categorical variable via argument `categories`:
#' anticlusters <- anticlustering(
#'   schaper2019[, 3:6],
#'   K = 3,
#'   categories = schaper2019$room
#' )
#' # Compare feature means and standard deviations by anticluster
#' mean_sd_tab(schaper2019[, 3:6], anticlusters)
#' # Verify that the "room" is balanced across anticlusters:
#' table(anticlusters, schaper2019$room)
#' 
#' # Use multiple starts of the algorithm to improve the objective and
#' # optimize the k-means criterion ("variance")
#' anticlusters <- anticlustering(
#'   schaper2019[, 3:6],
#'   objective = "variance",
#'   K = 3,
#'   categories = schaper2019$room,
#'   method = "local-maximum", # better search algorithm
#'   repetitions = 20 # multiple restarts of the algorithm 
#' )
#' # Compare means and standard deviations by anticluster
#' mean_sd_tab(schaper2019[, 3:6], anticlusters)
#' 
#' # Use different group sizes and optimize the extended k-means
#' # criterion ("kplus")
#' anticlusters <- anticlustering(
#'   schaper2019[, 3:6],
#'   objective = "kplus",
#'   K = c(24, 24, 48),
#'   categories = schaper2019$room,
#'   repetitions = 20,
#'   method = "local-maximum",
#'   standardize = TRUE # ususally recommended
#' )
#' 
#' # Use cannot_link constraints: Element 1 must not be linked with elements 2 to 10:
#' cl_matrix <- matrix(c(rep(1, 9), 2:10), ncol = 2)
#' cl <- anticlustering(
#'   schaper2019[, 3:6],
#'   K = 10,
#'   cannot_link = cl_matrix
#' )
#' all(cl[1] != cl[2:10])
#' 
#' # Use cannot_link constraints: Element 1 must be linked with elements 2 to 10.
#' # Element 11 must be linked with elements 12-20.
#' must_link <- rep(NA, nrow(schaper2019))
#' must_link[1:10] <- 1
#' must_link[11:20] <- 2
#' cl <- anticlustering(
#'   schaper2019[, 3:6],
#'   K = 3,
#'   must_link = must_link
#' )
#' cl[1:10]
#' cl[11:20]
#' 
#' # Use the heuristic by Brusco et al. (2020) for k-plus anticlustering
#' # Include categorical variable as part of the optimization criterion rather 
#' # than the argument categories!
#' anticlusters <- anticlustering(
#'   cbind(
#'     kplus_moment_variables(schaper2019[, 3:6], 2), 
#'     categories_to_binary(schaper2019$room)
#'   ),
#'   objective = "variance", # k-plus anticlustering because of the input above!
#'   K = 3,
#'   repetitions = 20,
#'   method = "brusco"
#' )
#' 
#' mean_sd_tab(schaper2019[, 3:6], anticlusters)
#' table(anticlusters, schaper2019$room)
#' 
#' @references
#' 
#' Brusco, M. J., Cradit, J. D., & Steinley, D. (2020). Combining
#' diversity and dispersion criteria for anticlustering: A bicriterion
#' approach. British Journal of Mathematical and Statistical
#' Psychology, 73, 275-396. https://doi.org/10.1111/bmsp.12186
#' 
#' Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. 
#' (2025). Extending the Bicriterion Approach for Anticlustering: 
#' Exact and Hybrid Approaches. Advance online publication. Psychometrika.
#' 
#' Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
#' data sets into equivalent parts. Psychological Methods, 26(2), 
#' 161–174. https://doi.org/10.1037/met0000301.
#' 
#' Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for 
#' Maximizing Between-Group Similarity. British Journal of Mathematical and 
#' Statistical Psychology, 77(1), 80-102. https://doi.org/10.1111/bmsp.12315
#' 
#' Papenberg, M., Wang, C., Diop, M., Bukhari, S. H., Oskotsky, B., Davidson, B. R., 
#' Vo, K. C., Liu, B., Irwin, J. C., Combes, A., Gaudilliere, B., Li, J., Stevenson, D. K., 
#' Klau, G. W., Giudice, L. C., Sirota, M., & Oskotsky, T. T. (2025). Anticlustering
#' for sample allocation to minimize batch effects. Cell Reports Methods, 5(8), 
#' 101137. https://doi.org/10.1016/j.crmeth.2025.101137
#'
#' Späth, H. (1986). Anticlustering: Maximizing the variance criterion.
#' Control and Cybernetics, 15, 213-218.
#' 
#' Weitz, R. R., & Lakshminarayanan, S. (1998). An empirical comparison of 
#' heuristic methods for creating maximally diverse groups. Journal of the 
#' Operational Research Society, 49(6), 635-646. https://doi.org/10.1057/palgrave.jors.2600510
#' 
#' Yang, X., Cai, Z., Jin, T., Tang, Z., & Gao, S. (2022). A three-phase search 
#' approach with dynamic population size for solving the maximally diverse grouping 
#' problem. European Journal of Operational Research, 302(3), 925-953. 
#' https://doi.org/10.1016/j.ejor.2022.02.003
#'

anticlustering <- function(x, K, objective = "diversity", method = "exchange",
                           preclustering = FALSE, categories = NULL, 
                           repetitions = NULL, standardize = FALSE, cannot_link = NULL,
                           must_link = NULL, blocks = NULL) {


  ## Get data into required format
  input_validation_anticlustering(x, K, objective, method, preclustering, 
                                  categories, repetitions, standardize, cannot_link,
                                  must_link, blocks)

  input_is_a_feature_matrix <- !is_distance_matrix(x) || inherits(x, "data.frame")
  
  # Preclustering and categorical constraints are both processed in the
  # variable `categories` after this step:
  # Preclustering must be done first because it needs the raw features and not (potentially) kplus features, which
  # would be appended in the next step via get_anticlustering_features()
  categories <- get_categorical_constraints(x, K, preclustering, categories)
  
  # Convert input into usable features for anticlustering, if feature matrix is passed and not custom objective used
  if (input_is_a_feature_matrix && !is.function(objective)) {
    x <- get_anticlustering_features(x, objective, standardize)
    if (any(rowSums(is.na(x)) == ncol(x))) {
      stop("Some observations only consist of NA, I cannot deal with this.")
    }
    if (objective == "kplus") {
      objective <- "variance" # now has the kplus variables, it is now standard kmeans
    }
  }
  x <- to_matrix(x)
  N <- nrow(x)

  NUMBER_OF_ANTICLUSTERS <- length(table(initialize_clusters(N, K, NULL)))
  TARGET_GROUPS <- table(initialize_clusters(N, K, NULL))
  
  ## Some data handling; in particular, determine whether we need a distance matrix even though we might not have one
  need_distance_matrix <- FALSE
  if (!is.function(objective)) { # when custom objective is passed, users need to ensure the data is correct
    if (objective == "distance") { # for compatibility with very old version...
      objective <- "diversity"
    }
    need_distance_matrix <- objective %in% c("diversity", "average-diversity", "dispersion") # this case is clear - objectives are computed from distances
    # some algorithms always use distance matrix even for kmeans/kplus that are usually computed from the features directly
    need_distance_matrix <- need_distance_matrix | method %in% c("3phase", "brusco")
    need_distance_matrix <- need_distance_matrix | argument_exists(cannot_link)
    need_distance_matrix <- need_distance_matrix | argument_exists(must_link)
    need_distance_matrix <- need_distance_matrix | sum(is.na(x)) > 0 # use dist() to deal with NAs.
  }

  if (need_distance_matrix) {
    kmeans_type_objective <- objective == "variance"
    x <- convert_to_distances(x, squared = kmeans_type_objective)
    if (kmeans_type_objective) { # when using distance matrix, kmeans/kplus are equivalent to average diversity.
      objective <- "average-diversity"
    }
  }
  
  ####### Data handling is done; now call the required algorithm
  
  if (argument_exists(blocks)) {
    return(blocked_anticlustering(
      x, objective = objective, 
      method = method, 
      categories = categories, K = K, 
      cannot_link = cannot_link, 
      blocks = blocks
    ))
  }
  
  if (method == "3phase") {
    if (length(K) == NUMBER_OF_ANTICLUSTERS) {
      clusters <- K
    } else {
      clusters <- NULL
    }
    return(
      three_phase_search_anticlustering(
        x, K = NUMBER_OF_ANTICLUSTERS, N = N, clusters = clusters,
        number_iterations = min(repetitions, 50)
      )
    )
  }
  
  ## Exact method using ILP
  if (method == "ilp") {
    if (objective == "dispersion") {
      return(optimal_dispersion(x, K)$groups)
    }
    return(exact_anticlustering(x, K, preclustering, cannot_link))
  }
  
  if (argument_exists(cannot_link)) {
    cannot_link <- as.matrix(cannot_link)
    init <- initialize_clusters(N, K, NULL) # only used if clustering is already given
    if (length(K) != N) {
      if (NCOL(cannot_link) == 1) {# cannot_link is an ID/grouping vector
        if (max(table(c(cannot_link))) > K) {
          stop("Cannot-link constraints cannot be fulfilled.")
        }
        init <- t(replicate(max(1, repetitions), categorical_sampling(c(cannot_link), K = K)))
      } else if (NCOL(cannot_link) == 2) {
        init <- optimal_cannot_link(N, NUMBER_OF_ANTICLUSTERS, table(init), cannot_link, repetitions)
      }
    }

    return(cannot_link_anticlustering(
      x = x, 
      init_clusters = init,
      cannot_link = cannot_link,
      objective = objective,
      method = method
    ))
  }
  
  if (argument_exists(must_link)) {
    return(
      must_link_anticlustering(
        x, 
        K, must_link = must_link, 
        method = method, 
        objective = "diversity", # always uses diversity computation in optimization algorithm, even for other objectives (that are implemented through changes in data)
        repetitions = repetitions
      )
    )
  }

  # BILS by Brusco et al.:
  if (method == "brusco") {
    average_diversity <- ifelse(objective == "average-diversity", TRUE, FALSE)
    return(bicriterion_anticlustering(x, K, repetitions, average_diversity = average_diversity, return = paste0("best-", objective)))
  }
  
  # Some special cases must be considered now:
  # (a) Is a user defined objective function passed? 
  # (b) Do we need "repeated" anticlustering (i.e., calling the standard exchange method multiple times via
  #     local maximum search and/or multiple initializations)
  # (c) Is the repeated optimization implemented in C, or do we just call the exchange method repeatedly from R?
  
  is_objective_user_defined <- inherits(objective, "function")
  repeated_anticlustering_needed <- method == "local-maximum" || (method == "exchange" && argument_exists(repetitions))
  repeated_anticlustering_has_c_implementation <- !inherits(objective, "function") && objective %in% c("diversity", "average-diversity")
  
  # Diversity anticlustering has C implementation for repeated anticlustering, consider this special case:
  if (repeated_anticlustering_needed && !repeated_anticlustering_has_c_implementation) { 
    repetitions <- ifelse(!argument_exists(repetitions), 1, repetitions)
      return(repeat_anticlustering(x, K, objective, categories, method, repetitions))
  }
  # Most generic exchange method for user defined objectives:
  if (is_objective_user_defined) {
    return(exchange_method(x, K, objective, categories))
  }

  # Redirect to specialized fast exchange methods for diversity, dispersion, kmeans/kplus objectives:
  local_maximum <- ifelse(method == "local-maximum", TRUE, FALSE)
  if (argument_exists(repetitions) && repetitions > 1) { # this can only be the case for diversity objective, based on the logic above
    repetitions <- t(simplify2array(get_multiple_initial_clusters(N, K, categories, repetitions))) - 1
  } else if (argument_exists(repetitions) && repetitions == 1) {
    repetitions <- NULL
  }
  c_anticlustering(x, K, categories, objective, local_maximum = local_maximum, init_partitions = repetitions)
}

# Function that processes input and returns the data set that the
# optimization is conducted on as matrix (for exchange method)
# Returned matrix either represents distances or features.
to_matrix <- function(data) {
  if (!is_distance_matrix(data)) {
    data <- as.matrix(data)
    return(data)
  }
  as.matrix(as.dist(data))
}

# Determines if preclustering constraints or categorical constraints
# are present. Returns a grouping vector if one or both constraints 
# have been passed, or NULL if none is required
get_categorical_constraints <- function(data, K, preclustering, categories) {
  if (preclustering == TRUE) {
    N <- nrow(data)
    matches <- matching(data, p = K, match_within = categories, sort_output = FALSE)
    # deal with NA in matches
    return(replace_na_by_index(matches))
  }
  if (argument_exists(categories)) {
    return(merge_into_one_variable(categories))
  }
  NULL
}

replace_na_by_index <- function(matches) {
  na_matches <- is.na(matches)
  NAs <- sum(na_matches) 
  if (NAs == 0) {
    return(matches)
  }
  max_group <- max(matches, na.rm = TRUE)
  matches[na_matches] <- max_group + 1:NAs 
  matches
}

# Convert user input to useful feature matrix
get_anticlustering_features <- function(x, objective, standardize) {
  x <- data.frame(x)
  is_categorical <- sapply(x, class) == "factor"
  if (sum(is_categorical) > 0) {
    categorical_variables <- categories_to_binary(x[, is_categorical])
  } else {
    categorical_variables <- NULL
  } 
  if (sum(!is_categorical) > 0) {
    numeric_variables <- as.matrix(x[, !is_categorical])
    if (objective == "kplus") {
      numeric_variables <- kplus_moment_variables(numeric_variables, 2, FALSE)
    }
  } else {
    numeric_variables <- NULL
  }
  
  all_ <- cbind(numeric_variables, categorical_variables)
  if (standardize) {
    all_ <- scale(all_)
  }
  all_
}

library("anticlust")

conditions <- expand.grid(m = 1:4, p = 2)
for (k in 1:nrow(conditions)) {
  m_features <- conditions[k, "m"]
  K <- conditions[k, "p"]
  n_elements <- K * 5 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  distances <- dist(features)
  
  ac_feat <- anticlustering(features, K = K, preclustering = FALSE,
                            method = "ilp")
  ac_dist <- anticlustering(distances, K = K,
                            preclustering = FALSE,
                            method = "ilp")
  expect_equal(diversity_objective(distances, ac_feat),
               diversity_objective(features, ac_feat))
  expect_equal(diversity_objective(distances, ac_dist),
               diversity_objective(features, ac_dist))
  expect_equal(diversity_objective(distances, ac_feat),
               diversity_objective(distances, ac_dist))
}


conditions <- expand.grid(m = 1:4, p = 2)
for (k in 1:nrow(conditions)) {
  m_features <- conditions[k, "m"]
  K <- conditions[k, "p"]
  n_elements <- K * 5 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  distances <- dist(features)
  
  ac_feat <- anticlustering(features, K = K, preclustering = TRUE,
                            method = "ilp")
  ac_dist <- anticlustering(distances, K = K,
                            preclustering = TRUE,
                            method = "ilp")
  expect_equal(diversity_objective(distances, ac_feat),
               diversity_objective(features, ac_feat))
  expect_equal(diversity_objective(distances, ac_dist),
               diversity_objective(features, ac_dist))
  expect_equal(diversity_objective(distances, ac_feat),
               diversity_objective(distances, ac_dist))
  
  # ensure that preclusters are balanced between anticlusters
  preclusters <- balanced_clustering(features, K = n_elements / K, method = "ilp")
  expect_true(all(table(ac_feat, preclusters) == 1))
}


conditions <- expand.grid(m = 1:4, p = 2:4)
for (k in 1:nrow(conditions)) {
  m_features <- conditions[k, "m"]
  K <- conditions[k, "p"]
  n_elements <- K * 5 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  distances <- dist(features)
  
  ## Use a fixed seed to compare the random sampling method based on
  ## features and distance input
  rnd_seed <- sample(10000, size = 1)
  
  set.seed(rnd_seed)
  ac_feat <- anticlustering(features, K = K)
  set.seed(rnd_seed)
  ac_dist <- anticlustering(distances, K = K)
  
  expect_equal(diversity_objective(distances, ac_feat),
               diversity_objective(features, ac_feat))
  expect_equal(diversity_objective(distances, ac_dist),
               diversity_objective(features, ac_dist))
  expect_equal(diversity_objective(distances, ac_feat),
               diversity_objective(distances, ac_dist))
}

# At the end, just test that the input works without error when preclustering is enabled
anticlustering(features, K = K, preclustering = TRUE)
anticlustering(distances, K = K, preclustering = TRUE)

library(anticlust)
library(tinytest)
#set.seed(1234)
# Generate some example data, N = 128, M = 5
K <- 20
group_sizes <- 12
N <- group_sizes * K # here 128
M <- 5
data <- matrix(rnorm(N * M), ncol = M)
distances <- as.matrix(dist(data))

# Function that test if constraints are valid
must_link_constraints_valid <- function(cl, must_link) {
  same <- as.numeric(names(table(must_link)[table(must_link) > 1]))
  all_good <- c()
  for (i in same) {
    all_good <- c(all_good, all(cl[must_link == i] == cl[must_link == i][1]))
  }
  all(all_good)
}

# Generate random patient IDs
must_link <- sample(N, replace = TRUE)

# Fulfilling the must-link constraints may fail, even though the data generating 
# process here makes it very unlikely. Still we should wrap the code in a tryCatch
# block to make sure. 

tt <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "exchange"),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  expect_true(must_link_constraints_valid(tt, must_link))
}

tt <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "exchange", repetitions = 2),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  expect_true(must_link_constraints_valid(tt, must_link))
}

tt <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "local-maximum"),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  expect_true(must_link_constraints_valid(tt, must_link))
}


tt <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "local-maximum", repetitions = 2),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  expect_true(must_link_constraints_valid(tt, must_link))
}

## Expect errors

expect_error(
  anticlustering(distances, K = K, must_link = must_link, categories = sample(LETTERS, size = N, replace = TRUE)),
  pattern = "categories"
)
expect_error(
  anticlustering(distances, K = K, must_link = must_link, preclustering = TRUE),
  pattern = "preclustering"
)
expect_error(
  anticlustering(data, K = K, must_link = "A"),
  pattern = "length"
)
expect_error(
  anticlustering(data, K = K, must_link = must_link, objective = "kplus"),
  pattern = "diversity"
)
expect_error(
  anticlustering(data, K = K, must_link = must_link, objective = "variance"),
  pattern = "diversity"
)
expect_error(
  anticlustering(data, K = K, must_link = must_link, objective = "dispersion"),
  pattern = "diversity"
)

expect_error(
  anticlustering(data, K = K, must_link = must_link, cannot_link = rbind(1:2, 2:3)),
  pattern = "cannot-link"
)
expect_error(
  anticlustering(data, K = K, must_link = matrix(NA))
)

library("anticlust")

set.seed(123)
N <- sample(40:80, size = 1)
M <- sample(1:5, size = 1)
K <- sample(2:5, size = 1)
features <- matrix(rnorm(N * M), ncol = M)
clusters <- anticlust:::initialize_clusters(N, K, NULL)
cl1 <- anticlustering(features, clusters, objective = "variance")
cl2 <- anticlustering(features, clusters, variance_objective)
expect_true(all(cl1 == cl2))

# now also use a categorical constraint
set.seed(71207)
C <- sample(2:5, size = 1)
categories <- sample(C, size = N, replace = TRUE)
clusters <- anticlust:::initialize_clusters(N, K, categories)
cl1 <- anticlustering(
  features, 
  clusters, 
  objective = "variance", 
  categories = categories
)
cl2 <- anticlustering(
  features, 
  clusters, 
  objective = variance_objective, 
  categories = categories
)
expect_true(all(cl1 == cl2))

# Include a test for the edge cases (i.e., categories with 1 or N members)
categories[1] <- C + 1 
clusters <- anticlust:::initialize_clusters(N, K, categories)
cl1 <- anticlustering(
  features, 
  clusters, 
  objective = "variance", 
  categories = categories
)
cl2 <- anticlustering(
  features, 
  clusters, 
  objective = variance_objective, 
  categories = categories
)
expect_true(all(cl1 == cl2))

# Include a test for the edge cases (i.e., categories with 1 or N members)
categories <- rep(1, N)
clusters <- anticlust:::initialize_clusters(N, K, categories)
cl1 <- anticlustering(
  features, 
  clusters, 
  objective = "variance", 
  categories = categories
)
cl2 <- anticlustering(
  features, 
  clusters, 
  objective = variance_objective, 
  categories = categories
)
expect_true(all(cl1 == cl2))

# Use diversity objective
set.seed(123)
clusters <- anticlust:::initialize_clusters(N, K, NULL)
cl1 <- anticlustering(features, clusters, objective = "diversity")
cl2 <- anticlustering(features, clusters, diversity_objective)
expect_true(all(cl1 == cl2))

# now also use a categorical constraint
set.seed(71207)
C <- sample(2:5, size = 1)
categories <- sample(C, size = N, replace = TRUE)
clusters <- anticlust:::initialize_clusters(N, K, categories)
cl1 <- anticlustering(
  features, 
  clusters, 
  objective = "diversity", 
  categories = categories
)
cl2 <- anticlustering(
  features, 
  clusters, 
  objective = diversity_objective, 
  categories = categories
)
expect_true(all(cl1 == cl2))

# now also use preclustering
set.seed(71207)
C <- sample(2:5, size = 1)
categories <- sample(C, size = N, replace = TRUE)
clusters <- anticlust:::initialize_clusters(N, K, categories)
cl1 <- anticlustering(
  features, 
  clusters, 
  objective = "diversity", 
  categories = categories
)
cl2 <- anticlustering(
  features, 
  clusters, 
  objective = diversity_objective, 
  categories = categories
)
expect_true(all(cl1 == cl2))

# test dispersion objective
set.seed(123)
N <- 50 # number of elements
M <- 2  # number of variables per element
K <- 2  # number of clusters
random_data <- matrix(rnorm(N * M), ncol = M)
random_clusters <- sample(rep_len(1:K, N))

# Maximize the dispersion 
optimized_clusters <- anticlustering(
  random_data,
  K = random_clusters, 
  objective = dispersion_objective
)
dispersion_objective(random_data, optimized_clusters)

optimized_clusters2 <- anticlustering(
  random_data,
  K = random_clusters, 
  objective = "dispersion"
)
expect_true(all(optimized_clusters == optimized_clusters2))

library("anticlust")

# fast exchange and exchange functions yield the same results - kmeans"
# for K = 2
clusters <- rep(1:2, 5)
features <- matrix(rnorm(20), ncol = 2)
ac <- anticlustering(features, K = clusters, objective = "variance")
ac_fast <- fast_anticlustering(features, clusters, exchange_partners = rep(list(1:10), 10))
expect_equal(all(ac == ac_fast), TRUE)

## for K = 3
clusters <- rep(1:3, 5)
features <- matrix(rnorm(30), ncol = 2)
ac <- anticlustering(features, K = clusters, objective = "variance")
ac_fast <- fast_anticlustering(features, clusters, exchange_partners = rep(list(1:15), 15))
expect_equal(all(ac == ac_fast), TRUE)

## for K = 4
clusters <- rep(1:4, 5)
features <- matrix(rnorm(20), ncol = 1)
ac <- anticlustering(features, K = clusters, objective = "variance")
ac_fast <- fast_anticlustering(features, clusters, exchange_partners = rep(list(1:20), 20))
expect_equal(all(ac == ac_fast), TRUE)


# fast exchange and exchange functions yield the same results - anticluster editing (max diversity)
# for K = 2
K <- 2
clusters <- rep(1:K, 5)
features <- matrix(rnorm(K * 10), ncol = 2)
ac <- anticlustering(features, K = clusters, objective = diversity_objective)
ac_fast <- anticlust:::fast_exchange_dist(as.matrix(dist(features)), clusters, NULL)
expect_equal(all(ac == ac_fast), TRUE)

# for K = 3
K <- 3
clusters <- rep(1:K, 5)
features <- matrix(rnorm(K * 10), ncol = 2)
ac <- anticlustering(features, K = clusters, objective = diversity_objective)
ac_fast <- anticlust:::fast_exchange_dist(as.matrix(dist(features)), clusters, NULL)
expect_equal(all(ac == ac_fast), TRUE)

# for K = 4
K <- 4
clusters <- rep(1:K, 5)
features <- matrix(rnorm(K * 10), ncol = 2)
ac <- anticlustering(features, K = clusters, objective = diversity_objective)
ac_fast <- anticlust:::fast_exchange_dist(as.matrix(dist(features)), clusters, NULL)
expect_equal(all(ac == ac_fast), TRUE)

## Test preclustering restrictions
features <- schaper2019[, 3:6]
clusters <- categorical_sampling(matching(features, 2), K = 2)
ac <- anticlustering(features, K = clusters, objective = diversity_objective)
ac_fast <- anticlustering(features, K = clusters, objective = "distance")
expect_equal(all(ac == ac_fast), TRUE)

## Test categorical restrictions
features <- schaper2019[, 3:6]
ac <- anticlustering(features, K = rep_len(1:2, nrow(schaper2019)),
                     objective = diversity_objective,
                     categories = schaper2019$room)
ac_fast <- anticlustering(features, K = rep_len(1:2, nrow(schaper2019)),
                          objective = "distance",
                          categories = schaper2019$room)
expect_equal(all(ac == ac_fast), TRUE)



# same results for local updating and recomputing - different group sizes"
N <- 60 
M <- 2
clusters <- rep(1:3, c(30, 15, 15))
features <- matrix(rnorm(N * M), ncol = M)
ac1 <- anticlustering(features, K = clusters, objective = "variance")
ac2 <- fast_anticlustering(features, clusters, exchange_partners = rep(list(1:N), N))
ac3 <- anticlustering(features, clusters, objective = "variance")
expect_true(all(ac1 == ac2))
expect_true(all(ac2 == ac3))

ac4 <- anticlustering(features, K = clusters, objective = diversity_objective)
ac5 <- anticlustering(features, clusters)
expect_true(all(ac4 == ac5))

library("anticlust")
library("tinytest")

# Test that using factor variables as input now works

gr <- anticlustering(iris, K = 3, standardize = TRUE)
mean_sd_tab(iris[-5], gr)
table(gr, iris$Species)

gr <- anticlustering(iris, K = 3, standardize = TRUE, method = "3phase")
mean_sd_tab(iris[-5], gr)
table(gr, iris$Species)

gr <- anticlustering(iris, K = 5, standardize = TRUE, method = "3phase", objective = "kplus")
mean_sd_tab(iris[-5], gr)
table(gr, iris$Species)

gr <- anticlustering(iris$Species, K = 5, method = "3phase",  objective = "variance")
table(gr, iris$Species)
expect_true(all(table(gr, iris$Species) == 10))

library("anticlust")
# Preclustering works with all criteria

features <- schaper2019[, 3:6]
matches <- matching(features, p = 3)

for (obj in c("variance", "kplus", "dispersion", "diversity")) {
  anticlusters <- anticlustering(
    features, 
    K = 3, 
    objective = obj,
    preclustering = TRUE
  )
  
  expect_true(all(table(matches, anticlusters) == 1))
}

# kplus_anticlustering must work as well 
anticlusters <- kplus_anticlustering(features, K = 3, preclustering = TRUE)

expect_true(all(table(matches, anticlusters) == 1))

library("anticlust")

# heuristic clustering works for different inputs
for (m in 1:4) {
  m_features <- m
  ## vary number of anticlusters
  for (p in 2:5) {
    n_elements <- p * 3 # n-elements must be multiplier of p
    features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
    clusters <- balanced_clustering(features, K = p)
    expect_equal(anticlust:::legal_number_of_clusters(features, clusters), NULL)
    clusters <- balanced_clustering(dist(features), K = p)
    expect_equal(anticlust:::legal_number_of_clusters(features, clusters), NULL)
  }
}

# heuristic anticlustering produces expected output
conditions <- expand.grid(m = 1:4, p = 2:4, objective = c("distance", "variance"),
                          preclustering = c(TRUE, FALSE))
for (i in 1:nrow(conditions)) {
  m_features <- conditions[i, "m"]
  p_anticlusters <- conditions[i, "p"]
  n_elements <- p_anticlusters * 5 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  
  ## First: heuristic preclustering
  preclusters <- NULL
  if (conditions$preclustering[i] == TRUE) {
    n_preclusters <- n_elements / p_anticlusters
    preclusters <- balanced_clustering(features, K = n_preclusters)
    ## Legal number of preclusters?
    expect_equal(anticlust:::legal_number_of_clusters(features, preclusters), NULL)
    ## Expected number of preclusters?
    expect_equal(as.numeric(table(preclusters)[1]), n_elements / n_preclusters)
  }
  
  ## Now anticlustering:
  obj_function <- ifelse(conditions$objective[i] == "distance",
                         diversity_objective, variance_objective)
  anticlusters <- anticlustering(
    features,
    K = p_anticlusters,
    categories = preclusters,
    objective = obj_function
  )
  ## Legal number of anticlusters?
  expect_equal(anticlust:::legal_number_of_clusters(features, anticlusters), NULL)
  ## Expected number of anticlusters?
  expect_equal(as.numeric(table(anticlusters)[1]), n_elements / p_anticlusters)
  if (conditions$preclustering[i] == TRUE) {
    # Are no preclustered elements together?
    tab <- table(anticlusters, preclusters)
    expect_equal(all(tab == 1), TRUE)
  }
}

library("anticlust")


# Test feature input:
data <- schaper2019[, 2:6]
data[1:3,1] <- NA
tt <- anticlustering(data, K = 3, standardize = TRUE, objective = "kplus")
mean_sd_tab(data[-1], tt)
table(tt, data$room, useNA = "ifany")

# Test distance input
dists <- dist(anticlust:::get_anticlustering_features(data, standardize = TRUE, objective = "kplus"))
sum(is.na(dists))
tt <- anticlustering(dists, K = 3)
mean_sd_tab(data[-1], tt)
table(tt, data$room, useNA = "ifany")

# Test distance input (but NAs in distances, which is not allowed!)
dists <- dist(schaper2019[, 3:6])
dists[1] <- NA
expect_error(
  anticlustering(dists, K = 3), 
  pattern = "No NA allowed in distance matrix input."
)

# Test distance input (but NAs in distances in matrix form, which is not allowed!)
dists <- dist(schaper2019[, 3:6])
dists[1] <- NA
expect_error(
  anticlustering(as.matrix(dists), K = 3), 
  pattern = "No NA allowed in distance matrix input."
)

library("anticlust")

# standardize argument only works for intended input
expect_error(
  anticlustering(
    rnorm(10),
    K = 2,
    standardize = 123
  )
)

expect_error(
  anticlustering(
    rnorm(10),
    K = 2,
    standardize = "foof"
  )
)

expect_error(
  anticlustering(
    rnorm(10),
    K = 2,
    standardize = c(TRUE, FALSE)
  )
)

expect_error(
  anticlustering(
    rnorm(10),
    K = 2,
    standardize = NA
  )
)

expect_true(length(
  anticlustering(
    rnorm(10),
    K = 2,
    standardize = TRUE
  )) == 10)

expect_true(length(
  anticlustering(
    rnorm(10),
    K = 2,
    standardize = FALSE
  )) == 10)

library("anticlust")

# Test standard usage: K = number of groups
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 3,
  categories = schaper2019$room
)
expect_true(all(table(anticlusters) == c(32, 32, 32)))

expect_true(
  all(table(anticlusters, schaper2019$room) == matrix(rep(16, 6), ncol = 2))
)

# Test K = group size
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = c(48, 24, 24),
  categories = schaper2019$room
)
expect_true(all(table(anticlusters) == c(48, 24, 24)))

# somewhat more extreme values:
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = c(80, 8, 8),
  categories = schaper2019$room
)

expect_true(
  all(table(anticlusters, schaper2019$room) == matrix(c(40, 4, 4, 40, 4, 4), ncol = 2))
)

# Test K = a priori assignment of cases
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = anticlusters,
  categories = schaper2019$room
)
expect_true(
  all(table(anticlusters, schaper2019$room) == matrix(c(40, 4, 4, 40, 4, 4), ncol = 2))
)

# Test problematic cases

expect_error(
  anticlustering(
    schaper2019[, 3:6],
    K = c(2, 4),
    categories = schaper2019$room
  ), 
  pattern = "misspecified"
)

expect_error(
  anticlustering(
    schaper2019[, 3:6],
    K = c(1, 2, 3, 4, 5, 100000),
    categories = schaper2019$room
  ), 
  pattern = "misspecified"
)

# too many groups
expect_error(
  anticlustering(
    schaper2019[, 3:6],
    K = nrow(schaper2019)
  ), 
  pattern = "smaller than"
)

# preclustering not possible
expect_error(
  anticlustering(
    schaper2019[, 3:6],
    K = c(24, 24, 48),
    preclustering = TRUE
  ), 
  pattern = "preclustering"
)

library("anticlust")

N <- 60
M <- 2
features <- matrix(rnorm(N * M), ncol = M)
## iterate over number of categories & number of anticlusters
for (K in 2:4) {
  for (C in 2:4) {
    categories <- sample(rep_len(1:C, N))
    ac <- categorical_sampling(categories = categories, K = K)
    tab <- table(categories, ac)
    ## At most 1 deviation between categories in anticlusters
    expect_equal(all(abs(tab - tab[[1]]) <= 1), TRUE)
  }
}


## The following test assumes that the categorical variables are
## balanced (otherwise testing is hard)

M <- 2
## iterate over number of categories & number of anticlusters
for (K in 2:3) {
  for (C1 in 2:3) {
    for (C2 in 2:3) {
      ## 1. Choose appropriate N that allows for a balanced assignment
      N <- (K * C1 * C2)^2
      features <- matrix(rnorm(N * M), ncol = M)
      ## 2. Ensure that the categories are actually balanced
      categories1 <- sort(rep_len(1:C1, N))
      categories2 <- categorical_sampling(categories1, K)
      frame_together <- ifelse(sample(2, size = 1) <= 2, data.frame, cbind)
      categories <- frame_together(categories1, categories2)
      ## Random order to cath potential problem in the implementation
      ## that might be due to a reliance on sorted input (We don't
      ## want that)
      categories <- categories[sample(nrow(categories)), ]
      vectorized_categories <- factor(do.call(paste0, as.list(categories)))
      ## Critical test: did it work to create balanced groups?
      tab <- table(vectorized_categories)
      expect_equal(all(tab == tab[[1]]), TRUE)
      
      
      ## 3. Are the categories balanced across anticlusters?
      ac <- categorical_sampling(K = K, categories = anticlust:::merge_into_one_variable(categories))
      tab1 <- table(categories[, 1], ac)
      tab2 <- table(categories[, 2], ac)
      tab3 <- table(categories[, 1], categories[, 2], ac)
      ## At most 1 deviation between categories in anticlusters
      expect_equal(all(abs(tab1 - tab1[[1]]) <= 1), TRUE)
      expect_equal(all(abs(tab2 - tab2[[1]]) <= 1), TRUE)
      expect_equal(all(abs(tab3 - tab3[[1]]) <= 1), TRUE)
    }
  }
}

N <- 60
M <- 2
features <- matrix(rnorm(N * M), ncol = M)
## iterate over number of categories & number of anticlusters
for (K in 2:4) {
  preclusters <- balanced_clustering(features, K = N / K)
  ac <- categorical_sampling(K = K, categories = preclusters)
  tab <- table(ac, preclusters)
  expect_equal(all(tab == 1), TRUE)
}


features <- schaper2019[, 3:6]
K <- 3
categories <- schaper2019$room
# Anticluster editing
ac <- anticlustering(
  features,
  K = K,
  categories = categories
)
expect_true(all(table(ac, categories) == 16))

# Anticluster editing, preclustering
ac <- anticlustering(
  features,
  K = K,
  categories = categories,
  preclustering = TRUE
)
expect_true(all(table(ac, categories) == 16))

# Anticluster editing, distance input
ac <- anticlustering(
  dist(features),
  K = K,
  categories = categories
)
expect_true(all(table(ac, categories) == 16))

# Anticluster editing, distance input, preclustering
ac <- anticlustering(
  dist(features),
  K = K,
  categories = categories,
  preclustering = TRUE
)
expect_true(all(table(ac, categories) == 16))

# K-means anticlustering 
ac <- anticlustering(
  features,
  K = K,
  categories = categories
)
expect_true(all(table(ac, categories) == 16))

# K-means anticlustering, preclustering
ac <- anticlustering(
  features,
  K = K,
  categories = categories,
  preclustering = TRUE
)
expect_true(all(table(ac, categories) == 16))

# Fast k-means anticlustering 
ac <- fast_anticlustering(
  features,
  K = K,
  categories = categories
)
expect_true(all(table(ac, categories) == 16))

# Fast k-means anticlustering, reduced number of exchange partners
ac <- fast_anticlustering(
  features,
  K = K,
  categories = categories,
  k_neighbours = 10
)
expect_true(all(table(ac, categories) == 16))


# kplus_anticlustering must work as well 
features <- schaper2019[, 3:6]
anticlusters <- kplus_anticlustering(features, K = 3, categories = schaper2019$room)

expect_true(all(table(schaper2019$room, anticlusters) == 16))

# does preclustering work at the same time as well?

anticlusters <- kplus_anticlustering(features, K = 3, categories = schaper2019$room, preclustering = TRUE)
matches <- matching(features, p = 3, match_within = schaper2019$room)
expect_true(all(table(schaper2019$room, anticlusters) == 16))
expect_true(all(table(matches, anticlusters) == 1))

library("anticlust")

# high level equal sized clustering function runs through
conditions <- expand.grid(m = 1:4, p = 2:4)
for (k in 1:nrow(conditions)) {
  m_features <- conditions[k, "m"]
  n_clusters <- conditions[k, "p"]
  n_elements <- n_clusters * 5 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  clusters_exact <- balanced_clustering(features, K = n_clusters, method = "ilp")
  clusters_heuristic <- balanced_clustering(features, K = n_clusters,
                                            method = "centroid")
  ## Check that output is valid
  expect_equal(anticlust:::legal_number_of_clusters(features, clusters_exact), NULL)
  expect_equal(anticlust:::legal_number_of_clusters(features, clusters_heuristic), NULL)
  
  ## Assert that exact solution has lowest objective (for distance
  ## criterion), allowing for numeric imprecision of ILP solver
  obj_exact     <- anticlust:::diversity_objective_(clusters_exact, features)
  obj_heuristic <- anticlust:::diversity_objective_(clusters_heuristic, features)
  expect_equal(round(obj_exact, 10) <= round(obj_heuristic, 10), TRUE)
}


# high level anticlustering function runs through
conditions <- expand.grid(m = 1:4, p = 2:3)
for (k in 1:nrow(conditions)) {
  m_features <- conditions[k, "m"]
  n_clusters <- conditions[k, "p"]
  n_elements <- n_clusters * 3 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  anticlusters_exact <- anticlustering(features, K = n_clusters,
                                       method = "ilp",
                                       preclustering = FALSE)
  anticlusters_heuristic <- anticlustering(features,
                                           K = n_clusters)
  ## Check that output is valid
  expect_equal(anticlust:::legal_number_of_clusters(features, anticlusters_exact), NULL)
  expect_equal(anticlust:::legal_number_of_clusters(features, anticlusters_heuristic), NULL)
  ## Assert that exact solution has highest objective (for distance
  ## criterion), allowing for numeric imprecision of ILP solver
  obj_exact     <- anticlust:::diversity_objective_(anticlusters_exact, features)
  obj_heuristic <- anticlust:::diversity_objective_(anticlusters_heuristic, features)
  expect_equal(round(obj_exact, 10) >= round(obj_heuristic, 10), TRUE)
}


# all argument combinations run through
conditions <- expand.grid(preclustering = c(TRUE, FALSE),
                          method = c("ilp", "exchange"))
# Set up matrix to store the objective values obtained by different methods
storage <- matrix(ncol = 2, nrow = 2)
colnames(storage) <- c("ilp", "exchange")
rownames(storage) <- c("preclustering", "no_preclustering")

criterion <- "distance"
n_elements <- 12
features <- matrix(rnorm(n_elements * 2), ncol = 2)
n_anticlusters <- 2

for (i in 1:nrow(conditions)) {
  method <- conditions$method[i]
  preclustering <- conditions$preclustering[i]
  anticlusters <- anticlustering(features, K = n_anticlusters,
                                 objective = criterion,
                                 method = method,
                                 preclustering = preclustering)
  obj <- anticlust:::diversity_objective_(anticlusters, features)
  rowname <- ifelse(preclustering, "preclustering", "no_preclustering")
  storage[rowname, method] <- obj
}
## Exact solution must be best:
expect_equal(all(round(storage["no_preclustering", "ilp"], 10) >= round(c(storage), 10)), TRUE)

library(anticlust)
library(tinytest)

set.seed(123)

N <- 12
M <- 5
K <- 2
dat <- matrix(rnorm(N * M), ncol = M)
distances <- dist(dat)

result_cluster1 <- three_phase_search_anticlustering(distances, K, N)
result_cluster2 <- anticlustering(distances, K=K, method="local-maximum", repetitions = 10)
result_cluster3 <- optimal_anticlustering(distances, objective = "diversity", K=K, solver = "lpSolve")

diversity1 <- diversity_objective(distances, result_cluster1)
diversity2 <- diversity_objective(distances, result_cluster2)
diversity3 <- diversity_objective(distances, result_cluster3)

expect_true(diversity3 >= diversity1)
expect_true(diversity3 >= diversity2)

### Test more clusters ###

N <- 12
M <- 2
K <- 3

dat <- matrix(rnorm(N * M), ncol = M)
distances <- dist(dat)

result_cluster1 <- anticlust:::three_phase_search_anticlustering(distances, K, N)
result_cluster2 <- anticlustering(distances, K=K, method="local-maximum", repetitions = 10)
result_cluster3 <- optimal_anticlustering(distances, objective = "diversity", K=K, solver = "lpSolve", time_limit = 20)

diversity1 <- diversity_objective(distances, result_cluster1)
diversity2 <- diversity_objective(distances, result_cluster2)
diversity3 <- diversity_objective(distances, result_cluster3)

expect_true(diversity3 >= diversity1)
expect_true(diversity3 >= diversity2)


## test cluster vector with different sizes

N2 <- 11
M2 <- 2
K2 <- 3
clusters <- c(3,3,5)
dat2 <- matrix(rnorm(N2 * M2), ncol = M2)
distances2 <- dist(dat2)

result_cluster <- anticlust:::three_phase_search_anticlustering(distances2, K2, N2, clusters=clusters)
table_cluseters <- table(result_cluster)
table_cluseters
expect_true(all(table_cluseters == clusters))

## test cluster vector swith higher cluster size

N2 <- 140
M2 <- 2
K2 <- 4
clusters <- c(20,40,30,50)
dat2 <- matrix(rnorm(N2 * M2), ncol = M2)
distances2 <- dist(dat2)

result_cluster <- anticlust:::three_phase_search_anticlustering(distances2, K2, N2, clusters=clusters)
table_cluseters <- table(result_cluster)
table_cluseters
expect_true(all(table_cluseters == clusters))


## test cluster vector throughs error

N2 <- 11
M2 <- 2
K2 <-2
clusters <- c(3,3,4)
dat2 <- matrix(rnorm(N2 * M2), ncol = M2)
distances2 <- dist(dat2)

expect_error(anticlust:::three_phase_search_anticlustering(distances2, K2, N2, clusters=clusters))


# Test problematic cases

expect_error(
  anticlust:::three_phase_search_anticlustering(distances2, K2, N2, clusters=c(1)),
  pattern = "len"
)

expect_error(
  anticlust:::three_phase_search_anticlustering(distances2, K2, N2, alpha=-1),
  pattern = "alpha must be greater than 0"
)

expect_error(
  anticlust:::three_phase_search_anticlustering(distances2, K2, N2, eta_max=1.2),
  pattern = "must be integer"
)

expect_error(
  anticlust:::three_phase_search_anticlustering(distances2, K2, N2, objective = "something"),
  pattern = "Argument objective can either be set to 'diversity' or 'dispersion"
)


## Larger data set

N <- 200
M <- 5
K <- 10

data <- matrix(rnorm(N*M), ncol = M)

g1 <- anticlustering(data, K = K)
g2 <- anticlustering(data, K = K, method = "local-maximum")
g3 <- three_phase_search_anticlustering(data, K, N)

diversity_objective(data, g1)
diversity_objective(data, g2)
diversity_objective(data, g3)

### Test dispersion ###

N <- 90
M <- 20
K <- 3

dat <- matrix(rnorm(N * M), ncol = M)

result_cluster1 <- three_phase_search_anticlustering(dat, K, N, objective = "dispersion", number_iterations = 10)
result_cluster2 <- anticlustering(dat, K = K, objective = "dispersion", method = "brusco", repetitions = 10)

optimal_dispersion <- optimal_dispersion(dat, K=K, solver = "symphony")$dispersion
dispersion_3phase <- dispersion_objective(dat, result_cluster1)
dispersion_bils <- dispersion_objective(dat, result_cluster2)
optimal_dispersion
dispersion_3phase
dispersion_bils

expect_true(dispersion_3phase <= optimal_dispersion)
expect_true(dispersion_bils <= optimal_dispersion)

library("anticlust")

# argument combinations run through (no preclustering / categories)
methods <- c("exchange", "local-maximum")
conditions <- expand.grid(
  method = methods,
  repetitions = c(1, 5),
  objective = c("variance", "diversity"),
  input = c("distance", "features")
)
illegal_conditions <- conditions$objective == "variance" & conditions$input == "distance"
conditions <- conditions[!illegal_conditions, ]
# for several repetitions, we cannot expect that local maximum always outperforms
# exchange method (latter may have better initial state)
illegal_conditions <- conditions$repetitions > 1 & conditions$method == "local-maximum"
conditions <- conditions[!illegal_conditions, ]

# Set up matrix to store the objective values obtained by different methods
K <- sample(2:6, size = 1)
N <- K * sample(5:10, size = 1)
M <- sample(1:4, size = 1)
features <- matrix(rnorm(N * M), ncol = M)
distances <- dist(features)
objs <- rep(NA, nrow(conditions)) # store objectives
clusters <- sample(rep_len(1:K, length.out = N))

for (i in 1:nrow(conditions)) {
  if (conditions$input[i] == "distance") {
    data <- distances
  } else {
    data <- features
  }
  
  anticlusters <- anticlustering(
    features, 
    K = clusters,
    objective = conditions$objective[i],
    method = conditions$method[i],
    repetitions = conditions$repetitions[i]
  )
  
  if (conditions$objective[i] == "diversity") {
    obj_function <- diversity_objective
  } else {
    obj_function <- variance_objective
  }
  
  objs[i] <- obj_function(data, anticlusters)
}

# Ensure that repetition exchange method is at least as good as single exchange
conditions$obj_value <- objs
# make wide format
wide1 <- reshape(
  conditions[conditions$method != "local-maximum", ], 
  direction = "wide", 
  timevar = "repetitions",
  idvar = colnames(conditions)[c(1, 3, 4)]
)
# test that repetitions improves results
expect_true(all(wide1[, 5] >= wide1[, 4]))

wide2 <- reshape(
  conditions[conditions$repetitions == 1, ], 
  direction = "wide", 
  timevar = "method",
  idvar = colnames(conditions)[2:4]
)
# test that local maximum search tend to improve results
expect_true(all(wide2[, 5] >= wide2[, 4]))


# repeated exchange method works with preclustering / categories
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 3,
  objective = "variance",
  categories = schaper2019$room,
  repetitions = 2,
  method = "local-maximum"
)
expect_true(all(table(anticlusters, schaper2019$room) == 16))

# Preclustering
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  objective = "variance",
  preclustering = TRUE,
  categories = schaper2019$room,
  repetitions = 2,
  method = "local-maximum"
)
expect_true(all(table(anticlusters, schaper2019$room) == 12))

# Test that matched preclusters are assigned to different anticlusters
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  objective = "variance",
  preclustering = TRUE,
  method = "local-maximum"
)
matches <- matching(schaper2019[, 3:6], p = 4)
expect_true(all(table(matches, anticlusters) == 1))

anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  objective = "variance",
  preclustering = TRUE,
  method = "local-maximum",
  repetitions = 3
)
matches <- matching(schaper2019[, 3:6], p = 4)
expect_true(all(table(matches, anticlusters) == 1))

# Test that matched preclusters are assigned to different anticlusters
# even when a categorical constraint is induced
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  objective = "variance",
  K = 2,
  preclustering = TRUE,
  method = "local-maximum",
  repetitions = 5,
  categories = schaper2019$room
)
matches <- matching(
  schaper2019[, 3:6], 
  p = 2, 
  match_within = schaper2019$room
)
expect_true(all(table(anticlusters, matches)))


# repeated exchange method works with different sized groups
N <- nrow(schaper2019)
n1 <- sample(seq(20, 40, 2), size = 1)
n2 <- sample(seq(20, 40, 2), size = 1)
n3 <- N - n2 - n1
stopifnot((n1 + n2 + n3) == N)

# no repetitions:
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = c(n1, n2, n3),
  categories = schaper2019$room,
  objective = "variance"
)
tab <- table(anticlusters, schaper2019$room)
expect_true(all(tab[, 1] == tab[, 2]))

# only local maximum:
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = c(n1, n2, n3),
  categories = schaper2019$room,
  method = "local-maximum",
  objective = "variance"
)
tab <- table(anticlusters, schaper2019$room)
expect_true(all(tab[, 1] == tab[, 2]))

# only repetitions:
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = c(n1, n2, n3),
  categories = schaper2019$room,
  repetitions = 5,
  objective = "variance"
)
tab <- table(anticlusters, schaper2019$room)
expect_true(all(tab[, 1] == tab[, 2]))

# repetitions and local-maximum search:
anticlusters <- anticlustering(
  dist(schaper2019[, 3:6]), # use distances as input at least once here
  K = c(n1, n2, n3),
  categories = schaper2019$room,
  repetitions = 5,
  method = "local-maximum"
)
tab <- table(anticlusters, schaper2019$room)
expect_true(all(tab[, 1] == tab[, 2]))

library("anticlust")

# Test that k-plus criterion is computed correctly
set.seed(123)

init <- anticlust:::initialize_clusters(96, 3, NULL)

df <- schaper2019[, 3:6]

groups1 <- anticlustering(
  df,
  K = init,
  objective = "kplus"
)

groups2 <- anticlustering(
  kplus_moment_variables(df, T = 2, standardize = FALSE), 
  K = init,
  objective = "variance"
)

expect_true(all(groups1 == groups2))

# add test with standardize argument
groups3 <- anticlustering(
  df,
  K = init,
  objective = "kplus",
  standardize = TRUE
)

groups4 <- anticlustering(
  kplus_moment_variables(df, T = 2), 
  K = init,
  objective = "variance"
)

expect_true(all(groups3 == groups4))

library("anticlust")

x <- schaper2019[, 3:6]
val1 <- optimal_dispersion(x, K = 3)$dispersion
val2 <- dispersion_objective(x, anticlustering(x, K = 3, method = "brusco", objective = "dispersion"))
expect_true(val1 >= val2) 


# Output format is as expected for optimal_dispersion()
N <- 30
M <- 5
K <- 3
data <- matrix(rnorm(N*M), ncol = M)
distances <- dist(data)

opt <- optimal_dispersion(distances, K = K)

groups_heuristic <- anticlustering(
  distances, 
  K = K,
  method = "local-maximum", 
  objective = "dispersion", 
  repetitions = 10
)
expect_true(dispersion_objective(distances, opt$groups) >= dispersion_objective(distances, groups_heuristic))

expect_true(all(sort(table(opt$groups)) == sort(table(anticlust:::initialize_clusters(N, K, NULL)))))


# test for unequal group sizes (groups differ by 1 at most)
N <- 28
M <- 5
K <- 3
data <- matrix(rnorm(N*M), ncol = M)
distances <- dist(data)

opt <- optimal_dispersion(distances, K = K)

groups_heuristic <- anticlustering(
  distances, 
  K = K,
  method = "local-maximum", 
  objective = "dispersion", 
  repetitions = 10
)
expect_true(dispersion_objective(distances, opt$groups) >= dispersion_objective(distances, groups_heuristic))
expect_true(all(sort(table(opt$groups)) == sort(table(anticlust:::initialize_clusters(N, K, NULL)))))


# now "truly" unequal group sizes 
K <- c(10, 12, 6)
opt <- optimal_dispersion(distances, K = K)

groups_heuristic <- anticlustering(
  distances, 
  K = K,
  method = "local-maximum", 
  objective = "dispersion", 
  repetitions = 10
)

expect_true(dispersion_objective(distances, opt$groups) >= dispersion_objective(distances, groups_heuristic))
expect_true(all(sort(table(opt$groups)) == sort(K)))

# another grouping, more extreme case
K <- c(2, 2, 24)
opt <- optimal_dispersion(distances, K = K)

groups_heuristic <- anticlustering(
  distances, 
  K = K,
  method = "local-maximum", 
  objective = "dispersion", 
  repetitions = 10
)

expect_true(dispersion_objective(distances, opt$groups) >= dispersion_objective(distances, groups_heuristic))
expect_true(all(sort(table(opt$groups)) == sort(K)))

library(anticlust)
library(tinytest)

set.seed(123)

N <- 12
M <- 5
K <- 2
dat <- matrix(rnorm(N * M), ncol = M)
distances <- dist(dat)

result_cluster1 <- three_phase_search_anticlustering(distances, K, N)
result_cluster2 <- anticlustering(distances, K=K, method="local-maximum", repetitions = 10)
result_cluster3 <- optimal_anticlustering(distances, objective = "diversity", K=K, solver = "lpSolve")

diversity1 <- diversity_objective(distances, result_cluster1)
diversity2 <- diversity_objective(distances, result_cluster2)
diversity3 <- diversity_objective(distances, result_cluster3)

expect_true(diversity3 >= diversity1)
expect_true(diversity3 >= diversity2)

### Test more clusters ###

N <- 12
M <- 2
K <- 3

dat <- matrix(rnorm(N * M), ncol = M)
distances <- dist(dat)

result_cluster1 <- anticlust:::three_phase_search_anticlustering(distances, K, N)
result_cluster2 <- anticlustering(distances, K=K, method="local-maximum", repetitions = 10)
result_cluster3 <- optimal_anticlustering(distances, objective = "diversity", K=K, solver = "lpSolve", time_limit = 20)

diversity1 <- diversity_objective(distances, result_cluster1)
diversity2 <- diversity_objective(distances, result_cluster2)
diversity3 <- diversity_objective(distances, result_cluster3)

expect_true(diversity3 >= diversity1)
expect_true(diversity3 >= diversity2)


## test cluster vector with different sizes

N2 <- 11
M2 <- 2
K2 <- 3
clusters <- c(3,3,5)
dat2 <- matrix(rnorm(N2 * M2), ncol = M2)
distances2 <- dist(dat2)

result_cluster <- anticlust:::three_phase_search_anticlustering(distances2, K2, N2, clusters=clusters)
table_cluseters <- table(result_cluster)
table_cluseters
expect_true(all(table_cluseters == clusters))

## test cluster vector swith higher cluster size

N2 <- 140
M2 <- 2
K2 <- 4
clusters <- c(20,40,30,50)
dat2 <- matrix(rnorm(N2 * M2), ncol = M2)
distances2 <- dist(dat2)

result_cluster <- anticlust:::three_phase_search_anticlustering(distances2, K2, N2, clusters=clusters)
table_cluseters <- table(result_cluster)
table_cluseters
expect_true(all(table_cluseters == clusters))


## test cluster vector throughs error

N2 <- 11
M2 <- 2
K2 <-2
clusters <- c(3,3,4)
dat2 <- matrix(rnorm(N2 * M2), ncol = M2)
distances2 <- dist(dat2)

expect_error(anticlust:::three_phase_search_anticlustering(distances2, K2, N2, clusters=clusters))


# Test problematic cases

expect_error(
  anticlust:::three_phase_search_anticlustering(distances2, K2, N2, clusters=c(1)),
  pattern = "len"
)

expect_error(
  anticlust:::three_phase_search_anticlustering(distances2, K2, N2, alpha=-1),
  pattern = "alpha must be greater than 0"
)

expect_error(
  anticlust:::three_phase_search_anticlustering(distances2, K2, N2, eta_max=1.2),
  pattern = "must be integer"
)

expect_error(
  anticlust:::three_phase_search_anticlustering(distances2, K2, N2, objective = "something"),
  pattern = "Argument objective can either be set to 'diversity' or 'dispersion"
)


## Larger data set

N <- 200
M <- 5
K <- 10

data <- matrix(rnorm(N*M), ncol = M)

g1 <- anticlustering(data, K = K)
g2 <- anticlustering(data, K = K, method = "local-maximum")
g3 <- three_phase_search_anticlustering(data, K, N)

diversity_objective(data, g1)
diversity_objective(data, g2)
diversity_objective(data, g3)

### Test dispersion ###

N <- 90
M <- 20
K <- 3

dat <- matrix(rnorm(N * M), ncol = M)

result_cluster1 <- three_phase_search_anticlustering(dat, K, N, objective = "dispersion", number_iterations = 10)
result_cluster2 <- anticlustering(dat, K = K, objective = "dispersion", method = "brusco", repetitions = 10)

optimal_dispersion <- optimal_dispersion(dat, K=K, solver = "symphony")$dispersion
dispersion_3phase <- dispersion_objective(dat, result_cluster1)
dispersion_bils <- dispersion_objective(dat, result_cluster2)
optimal_dispersion
dispersion_3phase
dispersion_bils

expect_true(dispersion_3phase <= optimal_dispersion)
expect_true(dispersion_bils <= optimal_dispersion)


## Also test this algorithm via interface in anticlustering()

data <- schaper2019[, 3:6]

a1 <- anticlustering(data, K = 3, standardize = TRUE, objective = "kplus")
a2 <- anticlustering(data, K = 3, method = "3phase", standardize = TRUE, objective = "kplus")
mean_sd_tab(data, a1)
mean_sd_tab(data, a2)

library(anticlust)
library(tinytest)


table(brunel2025$target_word_emotionality)

blocks <- brunel2025$target_word_emotionality

features <- brunel2025[, c("valence_target_word", "arousal_target_word")]

groups <- anticlustering(features, K = 2, objective = "kplus", blocks = blocks, standardize = TRUE, method = "local-maximum")
tab <- table(blocks, groups)

# is condition blocked?
expect_true(all(abs(tab[ ,1] - tab[, 2]) <= 1))

# non blocked assignment
groups2 <- anticlustering(features, K = 2, objective = "kplus", categories = blocks, standardize = TRUE, method = "local-maximum")

# verify balance, overall and within blocks, and compare to non-blocked assignment:

# overall
mean_sd_tab(features, groups)
mean_sd_tab(features, groups2)

# within blocks
mean_sd_tab(features[blocks == "Neutral", ], groups[blocks == "Neutral"])
mean_sd_tab(features[blocks == "Neutral", ], groups2[blocks == "Neutral"])

mean_sd_tab(features[blocks == "Positive", ], groups[blocks == "Positive"])
mean_sd_tab(features[blocks == "Positive", ], groups2[blocks == "Positive"])

mean_sd_tab(features[blocks == "Negative", ], groups[blocks == "Negative"])
mean_sd_tab(features[blocks == "Negative", ], groups2[blocks == "Negative"])


# useful results with distance matrix?
groups <- anticlustering(dist(features), K = 2, blocks = blocks, method = "local-maximum")
mean_sd_tab(features[blocks == "Negative", ], groups[blocks == "Negative"])
groups <- anticlustering(dist(features)^2, K = 2, blocks = blocks, method = "local-maximum")
mean_sd_tab(features[blocks == "Negative", ], groups[blocks == "Negative"])
# yep, seems to work with distance input

## Test wrapper function anticlustering() with different input specifications
anticlustering(features, K = 2, objective = "diversity", blocks = blocks, standardize = FALSE)

anticlustering(features, K = 2, objective = "diversity", blocks = blocks)

anticlustering(features, K = 2, objective = "variance", blocks = blocks)

anticlustering(features, K = 2, objective = "kplus", blocks = blocks, categories = brunel2025$sentence_emotionality)

## use other (random) data input

N <- 1000
blocks <- sample(1:10, size = N, replace = TRUE)
foo <- anticlustering(1:1000, K = 10, blocks = blocks, objective = "variance")

tab <- table(blocks, foo)
expect_true(all(abs(tab[ ,1] - tab[, 2]) <= 1))



## verify that some combinations of arguments throw errors
N <- 10
K <- 2
data <- rnorm(N)
must_link <- c(1, 1, 1, 2, 2, 2, 3, 4, 5, 6)
blocks <- rep(1:K, each = N/K)
expect_error(
  anticlustering(data, K = 2, blocks = blocks, must_link = must_link),
  pattern = "must-link"
)

expect_error(
  anticlustering(data, K = 2, objective = "kplus", blocks = blocks, method ="3phase"), 
  pattern = "exchange"
)
expect_error(
  anticlustering(data, K = 2, objective = "kplus", blocks = blocks, preclustering = TRUE),
  pattern = "preclustering"
)

expect_error(
  anticlustering(data, K = 2, objective = "kplus", blocks = blocks, repetitions = 10), 
  pattern = "repetitions"
)

# use cannot-link constraints
N <- 100
data <- rnorm(N)
blocks <- rep(1:K, each = N/K)
cannot_link <- cbind(1, 2)
tt <- anticlustering(data, K = 2, blocks = blocks, cannot_link = cannot_link, objective = "variance")
expect_true(tt[1] != tt[2])
tapply(data, tt, mean) |> round(2)


cannot_link <- cbind(c(1, 2, 3), c(2, 3, 1))
tt <- anticlustering(data, K = 3, blocks = blocks, cannot_link = cannot_link, objective = "variance")
expect_true(tt[1] != tt[2])
expect_true(tt[2] != tt[3])
expect_true(tt[1] != tt[3])
tapply(data, tt, mean) |> round(2)

library(anticlust)
library(tinytest)

# Test if cannot_link constraint works
a <- anticlustering(rnorm(6), K = 2, cannot_link = cbind(1, 2))
expect_true(a[1] != a[2])

a <- anticlustering(rnorm(6), K = 2, cannot_link = cbind(1, 2), method = "ilp")
expect_true(a[1] != a[2])

a <- anticlustering(rnorm(6), K = 2, cannot_link = cbind(1, 2), method = "brusco")
expect_true(a[1] != a[2])


# now make it harder: find instance where two items are linked in optimal solution, 
# and then insert cannot-link constraint!

N <- 14
M <- 3
data <- matrix(rnorm(N*M), ncol = M)
a <- anticlustering(data, K = 2, method = "ilp")
one <- which(a == 1)[1]
other <- which(a == 1)[2]
diversity_objective(data, a)

b <- anticlustering(data, K = 2, cannot_link = cbind(one, other))
expect_true(b[one] != b[other])
expect_true(diversity_objective(data, a) >= diversity_objective(data, b))

b <- anticlustering(data, K = 2, cannot_link = cbind(one, other), method = "ilp")
expect_true(b[one] != b[other])
expect_true(diversity_objective(data, a) >= diversity_objective(data, b))

b <- anticlustering(data, K = 2, cannot_link = cbind(one, other), method = "brusco")
expect_true(b[one] != b[other])
expect_true(a[one] == a[other]) # this is necessary, but maybe we appreciate the reminder.
expect_true(diversity_objective(data, a) >= diversity_objective(data, b))



# Now: use multiple repetitions!

N <- 14
M <- 3
data <- matrix(rnorm(N*M), ncol = M)
a <- anticlustering(data, K = 2, method = "ilp")
one <- which(a == 1)[1]
other <- which(a == 1)[2]
diversity_objective(data, a)

b <- anticlustering(data, K = 2, cannot_link = cbind(one, other), repetitions = 10)
expect_true(b[one] != b[other])
expect_true(diversity_objective(data, a) >= diversity_objective(data, b))

b <- anticlustering(data, K = 2, cannot_link = cbind(one, other), method = "brusco", repetitions = 10)
expect_true(b[one] != b[other])
expect_true(a[one] == a[other]) # this is necessary, but maybe we appreciate the reminder.
expect_true(diversity_objective(data, a) >= diversity_objective(data, b))


## Test for larger data sets:
N <- 140
M <- 5
data <- matrix(rnorm(N*M), ncol = M)
b <- anticlustering(data, K = 2, cannot_link = cbind(1:2, 2:3), repetitions = 10, method = "local-maximum")
expect_true(b[1] != b[2])
expect_true(b[2] != b[3])
b <- anticlustering(data, K = 2, cannot_link = rbind(1:2, 2:3), method = "brusco", repetitions = 10, objective = "variance")
expect_true(b[1] != b[2])
expect_true(b[2] != b[3])

expect_true(table(b)[1] == table(b)[2])

# different sized groups:
b <- anticlustering(data, K = c(100, 40), cannot_link = rbind(1:2, 2:3), method = "brusco", repetitions = 10)
expect_true(b[1] != b[2])
expect_true(b[2] != b[3])

expect_true(sort(table(b))[1] == 40)
expect_true(sort(table(b))[2] == 100)

# more constraints, different sized groups
b <- anticlustering(data, K = c(100, 40), cannot_link = rbind(1:2, 2:3, 4:5, 6:7), method = "brusco", repetitions = 10)
expect_true(b[1] != b[2])
expect_true(b[2] != b[3])
expect_true(b[4] != b[5])
expect_true(b[6] != b[7])

expect_true(sort(table(b))[1] == 40)
expect_true(sort(table(b))[2] == 100)

# more groups, different sized groups
groups <- c(20, 20, 40, 30, 10, 10, 10)
b <- anticlustering(data, K = groups, cannot_link = rbind(1:2, 2:3, 4:5, 6:7), method = "brusco", repetitions = 10)
expect_true(b[1] != b[2])
expect_true(b[2] != b[3])
expect_true(b[4] != b[5])
expect_true(b[6] != b[7])

expect_true(all(sort(groups) == sort(table(b))))

# Test that all argument combinations go through without error

N <- 30
M <- 5
K <- 5
data <- matrix(rnorm(N*M), ncol = M)

combinations <- expand.grid(
  objective = c("kplus", "variance", "diversity", "average-diversity"),
  repetitions = c(1, 10),
  method = c("brusco", "local-maximum", "exchange"),
  standardize = c(TRUE, FALSE)
)


for (i in 1:nrow(combinations)) {
  objective <- combinations[i, "objective"]
  method <- combinations[i, "method"]
  repetitions <- combinations[i, "repetitions"]
  standardize <- combinations[i, "standardize"]
  a <- anticlustering(data, K = K, objective = objective, method = method, repetitions = repetitions, standardize = standardize) # Once without cannot-link constraints
  b <- anticlustering(data, K = K, objective = objective, method = method, repetitions = repetitions, standardize = standardize, cannot_link = rbind(1:2, 2:3))
  expect_true(b[1] != b[2])
  expect_true(b[2] != b[3])
}

## Expect errors

expect_error(
  anticlustering(data, K = K, cannot_link = rbind(1:2, 2:3), categories = sample(LETTERS, size = N, replace = TRUE)),
  pattern = "categories"
)
expect_error(
  anticlustering(data, K = K, cannot_link = rbind(1:2, 2:3),preclustering = TRUE),
  pattern = "preclustering"
)

expect_error(
  anticlustering(data, K = K, cannot_link = rbind(1:3, 5:7)),
  pattern = "columns"
)

# Test cannot-link constraints via vector

cannot_link <- sample(N/2, size = N, replace = TRUE)
K <- max(table(cannot_link))
a <- anticlustering(data, K = K, cannot_link = cannot_link)

expect_true(all(table(a, cannot_link) <= 1))

expect_error(anticlustering(data, K = K-1, cannot_link = cannot_link))

library("anticlust")

# all levels of heuristicism work and that exact approach has best objective
conditions <- expand.grid(m = 1:4, p = 2)
for (k in 1:nrow(conditions)) {
  m_features <- conditions[k, "m"]
  p_anticlusters <- conditions[k, "p"]
  n_elements <- p_anticlusters * 5 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  ## Traverse through levels of heuristicism
  obj_values <- rep(NA, 4)
  anti_list <- list()
  for (i in c(TRUE, FALSE)) {
    anticlusters <- anticlust:::exact_anticlustering(as.matrix(dist(features)),
                                         p_anticlusters, preclustering = i, NULL)
    anti_list[[i + 1]] <- anticlusters
    # Allow for some numeric imprecision of ILP solver:
    obj_values[i + 1]  <- round(anticlust:::diversity_objective_(anticlusters, features), 10)
  }
  ## Exact solution must have maximum objective
  expect_equal(which.max(obj_values), 1)
}

# Solving ILP works as expected for max dispersion
N <- 30
K <- 3
M <- 5
dat <- matrix(rnorm(N*M), ncol = M)

ILP <- anticlustering(dat, K = K, objective = "dispersion", method = "ilp")
HEURISTIC <- anticlustering(dat, K = K, objective = "dispersion", method = "local-maximum")

expect_true(dispersion_objective(dat, ILP) >= dispersion_objective(dat, HEURISTIC))
expect_true(all(table(ILP) == N/K))

library("anticlust")

# k-means and diversity are equivalent for equal sized groups
N <- 50
M <- 5
K <- 5
set.seed(123)
data <- matrix(rnorm(N * M), ncol = M)
init <- anticlust:::initialize_clusters(N, K, NULL)

# k-means
g1 <- anticlustering(data, K = init, objective = "variance")
g2 <- anticlustering(dist(data)^2, K = init)
# this can go wrong due to floating point errors, therefore I use a seed
expect_true(all(g1 == g2))

# k-plus 
g1 <- anticlustering(data, K = init, objective = "kplus")
g2 <- anticlustering(dist(kplus_moment_variables(data, 2, FALSE))^2, K = init)
expect_true(all(g1 == g2))

# unequal-sized groups
# k-means
K <- c(5, 10, 15, 20)
init <- anticlust:::initialize_clusters(N, K, NULL)
g1 <- anticlustering(data, K = init, objective = "variance")
g2 <- anticlustering(dist(data)^2, K = init)
expect_false(all(g1 == g2))

# needs "average" diversity
g3 <- anticlustering(dist(data)^2, K = init, objective = "average-diversity")
expect_true(all(g1 == g3))

# k-plus 
g1 <- anticlustering(data, K = init, objective = "kplus")
g2 <- anticlustering(dist(kplus_moment_variables(data, 2, FALSE))^2, K = init)
expect_false(all(g1 == g2))

# needs "average" diversity
g3 <- anticlustering(dist(kplus_moment_variables(data, 2, FALSE))^2, K = init, objective = "average-diversity")
expect_true(all(g1 == g3))

library(anticlust)
library(tinytest)
#set.seed(1234)
# Generate some example data, N = 128, M = 5
K <- 20
group_sizes <- 12
N <- group_sizes * K # here 128
M <- 5
data <- matrix(rnorm(N * M), ncol = M)
distances <- as.matrix(dist(data))

# Function that test if constraints are valid
must_link_constraints_valid <- function(cl, must_link) {
  same <- as.numeric(names(table(must_link)[table(must_link) > 1]))
  all_good <- c()
  for (i in same) {
    all_good <- c(all_good, all(cl[must_link == i] == cl[must_link == i][1]))
  }
  all(all_good)
}

# Generate random patient IDs
must_link <- sample(N, replace = TRUE)

# Fulfilling the must-link constraints may fail, even though the data generating 
# process here makes it very unlikely. Still we should wrap the code in a tryCatch
# block to make sure. 

tt <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "exchange"),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  expect_true(must_link_constraints_valid(tt, must_link))
}

tt <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "exchange", repetitions = 2),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  expect_true(must_link_constraints_valid(tt, must_link))
}

tt <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "local-maximum"),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  expect_true(must_link_constraints_valid(tt, must_link))
}


tt <- tryCatch(
  anticlustering(distances, K, must_link = must_link, method = "local-maximum", repetitions = 2),
  error = function(e) e
)

if (!"simpleError" %in% class(tt)) {
  expect_true(must_link_constraints_valid(tt, must_link))
}

## Expect errors

expect_error(
  anticlustering(distances, K = K, must_link = must_link, categories = sample(LETTERS, size = N, replace = TRUE)),
  pattern = "categories"
)
expect_error(
  anticlustering(distances, K = K, must_link = must_link, preclustering = TRUE),
  pattern = "preclustering"
)
expect_error(
  anticlustering(data, K = K, must_link = "A"),
  pattern = "length"
)
expect_error(
  anticlustering(data, K = K, must_link = must_link, objective = "dispersion"),
  pattern = "diversity"
)

expect_error(
  anticlustering(data, K = K, must_link = must_link, cannot_link = rbind(1:2, 2:3)),
  pattern = "cannot-link"
)
expect_error(
  anticlustering(data, K = K, must_link = matrix(NA))
)

# no error for unequal-sized groups with diversity
tryCatch(anticlustering(1:100, K = c(10, 10, 80), must_link = sample(100, replace = TRUE)), error = function(e) e)
# no error for unequal-sized groups with k-means/k-plus
expect_error(
  anticlustering(1:100, K = c(10, 10, 80), must_link = sample(100, replace = TRUE), objective = "variance"),
  pattern = "equal-sized"
)
expect_error(
  anticlustering(1:100, K = c(10, 10, 80), must_link = sample(100, replace = TRUE), objective = "kplus"),
  pattern = "equal-sized"
)


#try k-plus objective with schaper data set

data <- schaper2019[, 3:6]
must_link <- rep(NA, nrow(data))
must_link[1:5] <- 1
gr <- anticlustering(data, K = 4, objective = "kplus", repetitions = 100, must_link = must_link)
expect_true(all(gr[1:5] == gr[1]))
mean_sd_tab(data, gr) #nice!

#' Solve the ILP formulation of anticluster editing
#'
#' @param ilp An object representing the ILP formulation of the
#'     instance, returned by \code{anticlustering_ilp}
#' @param objective A string identifying whether the objective function
#'     of the ILP should be maximized ("max") or minimized
#'     ("min"). Maximizing creates similar groups (i.e., solves
#'     anticlustering), minimizing creates distinct clusters (i.e.,
#'     solves weighted cluster editing).
#' @param time_limit time limit given to solver, in seconds
#'
#' @return A `list` with two entries: `x` is the vector of optimal
#'     coefficients for all decision variables. `obj` is the optimal
#'     objective value.
#'
#' @noRd

solve_ilp <- function(ilp, objective = "max", solver = NULL, time_limit = NULL) {

  if (is.null(solver)) {
    solver <- find_ilp_solver()
  }

  if (solver == "glpk") {
    return(solve_ilp_glpk(ilp, objective, time_limit))
  } else if (solver == "symphony") {
    return(solve_ilp_symphony(ilp, objective, time_limit))
  } else if (solver == "lpSolve") {
    return(solve_ilp_lpSolve(ilp, objective, time_limit))
  } else if (solver == "gurobi") {
    return(solve_ilp_gurobi(ilp, objective, time_limit))
  }
}

solve_ilp_glpk <- function(ilp, objective, time_limit) {
  ilp_solution <- Rglpk::Rglpk_solve_LP(
    obj = ilp$obj_function,
    mat = ilp$constraints,
    dir = ilp$equalities,
    rhs = ilp$rhs,
    types = "B",
    max = objective == "max",
    control = list(tm_limit = ifelse(is.null(time_limit), 0, time_limit * 1000))
  )
  
  # return the optimal value and the variable assignment
  ret_list <- list() 
  ret_list$x <- ilp_solution$solution
  ret_list$obj <- ilp_solution$optimum
  ret_list$status <- ilp_solution$status
  ## name the decision variables
  names(ret_list$x) <- colnames(ilp$constraints)
  ret_list
}

solve_ilp_symphony <- function(ilp, objective, time_limit) {
  
  ilp_solution <- Rsymphony::Rsymphony_solve_LP(
    obj = ilp$obj_function,
    mat = ilp$constraints,
    dir = ilp$equalities,
    rhs = ilp$rhs,
    types = "B",
    max = objective == "max",
    time_limit = ifelse(is.null(time_limit), -1, time_limit)
  )
  
  # return the optimal value and the variable assignment
  ret_list <- list() 
  ret_list$x <- ilp_solution$solution
  ret_list$obj <- ilp_solution$objval
  ret_list$status <- ilp_solution$status
  ## name the decision variables
  names(ret_list$x) <- colnames(ilp$constraints)
  ret_list
}

solve_ilp_lpSolve <- function(ilp, objective, time_limit) {
  ilp_solution <- lpSolve::lp(
    objective,
    ilp$obj_function,
    as.matrix(ilp$constraints),
    ilp$equalities,
    ilp$rhs,
    all.bin = TRUE,
    timeout = ifelse(is.null(time_limit), 0, time_limit)
  )
  # return the optimal value and the variable assignment
  ret_list <- list() 
  ret_list$x <- ilp_solution$solution
  ret_list$obj <- ilp_solution$objval
  ret_list$status <- ilp_solution$status
  ## name the decision variables
  names(ret_list$x) <- colnames(ilp$constraints)
  ret_list
}

solve_ilp_gurobi <- function(ilp, objective, time_limit) {
  
  ilp$equalities[ilp$equalities == "=="] <- "=" # gurobi uses 1 '='

  ## build model
  model <- list()
  model$A          <- ilp$constraints
  model$obj        <- ilp$obj_function
  model$modelsense <- objective
  model$rhs        <- ilp$rhs
  model$sense      <- ilp$equalities
  model$vtype     <- "B"
  
  ## solve
  if (argument_exists(time_limit)) {
    ilp_solution <- gurobi::gurobi(model, params = list(LogToConsole = 0, TimeLimit = time_limit, MIPGap = 0))
  } else {
    ilp_solution <- gurobi::gurobi(model, params = list(LogToConsole = 0, MIPGap = 0))
  }
  ret_list <- list() 
  ret_list$x <- ilp_solution$x
  ret_list$obj <- ilp_solution$objval
  ret_list$status <- ifelse(ilp_solution$status == "OPTIMAL", 0, 1)
  ## name the decision variables
  if (!is.null(ilp_solution$x)) { # gurobi does not return solution if not optimal
    names(ret_list$x) <- colnames(ilp$constraints)
  }
  ret_list
}

# Function to find a solver package
find_ilp_solver <- function() {
  if (requireNamespace("gurobi", quietly = TRUE)) {
    return("gurobi")
  } else if (requireNamespace("lpSolve", quietly = TRUE)) {
    return("lpSolve")
  } else if (requireNamespace("Rglpk", quietly = TRUE)) {
    return("glpk")
  } else if (requireNamespace("Rsymphony", quietly = TRUE)) {
    return("symphony")
  }
  check_if_solver_is_available() # throws an error here!
}

library("anticlust")

# objective value for variance criterion is computed correctly
for (m in 1:4) {
  m_features <- m
  ## vary number of anticlusters
  for (p in 2:5) {
    p_anticlusters <- p
    n_elements <- p * 3 # n-elements must be multiplier of p
    features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
    cl <- kmeans(features, p_anticlusters)
    obj_kmeans <- cl$tot.withinss
    obj_mine  <- variance_objective(features, cl$cluster)
    expect_equal(obj_kmeans, obj_mine)
  }
}

# objective value for distance criterion is computed correctly
conditions <- expand.grid(m = 1:4, p = 2:3)
for (k in 1:nrow(conditions)) {
  m_features <- conditions[k, "m"]
  p_anticlusters <- conditions[k, "p"]
  n_elements <- p_anticlusters * 3 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  distances <- as.matrix(dist(features))
  ilp <- anticlust:::anticlustering_ilp(distances, p_anticlusters)
  solution <- anticlust:::solve_ilp(ilp, "min")
  anticlusters <- anticlust:::ilp_to_groups(solution, n_elements)
  expect_equal(solution$obj, anticlust:::diversity_objective_(anticlusters, features))
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
set.seed(123)

## ----setup--------------------------------------------------------------------
library(anticlust)

## -----------------------------------------------------------------------------
data(penguins)
# First exclude cases with missing values
df <- na.omit(palmerpenguins::penguins)
head(df)
nrow(df)

## -----------------------------------------------------------------------------
numeric_vars <- df[, c("bill_length_mm", "bill_depth_mm", "flipper_length_mm", "body_mass_g")]
groups <- anticlustering(
  numeric_vars, 
  K = 3,
  categories = df$sex
)

## -----------------------------------------------------------------------------
table(groups, df$sex)

## -----------------------------------------------------------------------------
groups <- anticlustering(
  numeric_vars, 
  K = 3,
  categories = df$species
)

table(groups, df$species)

## -----------------------------------------------------------------------------
groups <- anticlustering(
  numeric_vars, 
  K = 3,
  categories = df[, c("species", "sex")]
)

table(groups, df$sex)
table(groups, df$species) 

## -----------------------------------------------------------------------------
all_features <- data.frame(numeric_vars, df[, c("species", "sex")])

## -----------------------------------------------------------------------------
groups <- anticlustering(
  all_features,
  K = 3,
  method = "local-maximum", 
  standardize = TRUE
)
table(groups, df$sex)
table(groups, df$species)

## -----------------------------------------------------------------------------
binary_categories <- categories_to_binary(df[, c("species", "year")], use_combinations = FALSE)
data_input <- data.frame(binary_categories, numeric_vars)
groups <- anticlustering(
  data_input,
  K = 3,
  method = "local-maximum",
  standardize = TRUE
)
table(groups, df$year, df$species)

## -----------------------------------------------------------------------------
binary_categories <- categories_to_binary(df[, c("species", "year")], use_combinations = TRUE)
data_input <- data.frame(binary_categories, numeric_vars)
groups <- anticlustering(
  data_input,
  K = 3,
  method = "local-maximum",
  standardize = TRUE
)
table(groups, df$year, df$species)


## -----------------------------------------------------------------------------
groups1 <- anticlustering(
  numeric_vars, 
  K = 3,
  categories = df$sex, 
  standardize = TRUE,
  objective = "kplus", 
  method = "local-maximum"
)
groups2 <- anticlustering(
  numeric_vars, 
  K = 3,
  blocks = df$sex, 
  standardize = TRUE,
  objective = "kplus", 
  method = "local-maximum"
)

table(df$sex, groups1)
table(df$sex, groups2)

## -----------------------------------------------------------------------------
knitr::kable(mean_sd_tab(numeric_vars[df$sex == "female", ], groups1[df$sex == "female"]), row.names = TRUE) # categories argument
knitr::kable(mean_sd_tab(numeric_vars[df$sex == "female", ], groups2[df$sex == "female"]), row.names = TRUE) # blocks argument

knitr::kable(mean_sd_tab(numeric_vars[df$sex == "male", ], groups1[df$sex == "male"]), row.names = TRUE) # categories argument
knitr::kable(mean_sd_tab(numeric_vars[df$sex == "male", ], groups2[df$sex == "male"]), row.names = TRUE) # blocks argument

#' Number of equal sized partitions
#'
#' @param K How many partitions
#' @param N How many elements
#'
#' @return The number of partitions
#'
#' @export
#' 
#' @examples
#' n_partitions(20, 2)
#'
n_partitions <- function(N, K) {
  if (N %% K != 0) {
    stop("N = ", N, " cannot be evenly split into K = ", K, " sets.")
  }
  x <- rep(N / K, K)
  set_size <- c(sum(x), sum(x) - cumsum(x))
  result   <- 1
  for (i in 1:(length(x)-1)) {
    result <- result * choose(set_size[i], x[i])
  }
  result / factorial(K)
}

library("anticlust")

# generate_partitions function correctly removes redundant partitions
for (K in 2:4) {
  for (N in 4:10) {
    if (N %% K != 0) {
      next
    }
    permutations <- generate_partitions(N, K, TRUE)
    partitions   <- generate_partitions(N, K, FALSE)
    # remove duplicates from permutations
    permutations <- lapply(permutations, anticlust:::order_cluster_vector)
    permutations <- permutations[!duplicated(permutations)]
    expect_equal(permutations, partitions)
  }
}

# analytical solution and generative functions result in same number of partitions
K <- 2
for (N in seq(4, 18, 2)) {
  partitions <- generate_partitions(N, K, FALSE)
  analytical_n <- n_partitions(N, K)
  expect_equal(analytical_n, length(partitions))
}
K <- 3
for (N in seq(6, 12, 3)) {
  partitions <- generate_partitions(N, K, FALSE)
  analytical_n <- n_partitions(N, K)
  expect_equal(analytical_n, length(partitions))
}
K <- 4
for (N in c(8, 12)) {
  partitions <- generate_partitions(N, K, FALSE)
  analytical_n <- n_partitions(N, K)
  expect_equal(analytical_n, length(partitions))
}
K <- 5
for (N in c(5, 10)) {
  partitions <- generate_partitions(N, K, FALSE)
  analytical_n <- n_partitions(N, K)
  expect_equal(analytical_n, length(partitions))
}

#' Matching
#'
#' Conduct K-partite or unrestricted (minimum distance) matching to
#' find pairs or groups of similar elements. By default, finding
#' matches is based on the Euclidean distance between data points, but
#' a custom dissimilarity measure can also be employed.
#'
#' @param x The data input. Can be one of two structures: (1) A feature
#'     matrix where rows correspond to elements and columns correspond
#'     to variables (a single numeric variable can be passed as a
#'     vector). (2) An N x N matrix dissimilarity matrix; can be an
#'     object of class \code{dist} (e.g., returned by
#'     \code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
#'     where the entries of the upper and lower triangular matrix
#'     represent pairwise dissimilarities.
#' @param p The size of the groups; the default is 2, in which case
#'     the function returns pairs.
#' @param match_between An optional vector, \code{data.frame} or
#'     matrix representing one or several categorical constraints. If
#'     passed, the argument \code{p} is ignored and matches are sought
#'     between elements of different categories.
#' @param match_within An optional vector, \code{data.frame} or matrix
#'     representing one or several categorical constraints. If passed,
#'     matches are sought between elements of the same category.
#' @param match_extreme_first Logical: Determines if matches are first
#'     sought for extreme elements first or for central
#'     elements. Defaults to \code{TRUE}.
#' @param target_group Currently, the options "none",
#'     smallest" and "diverse" are supported. See Details.
#' @param sort_output Boolean. If \code{TRUE} (default), the output clusters 
#'     are sorted by similarity. See Details.
#'     
#'
#' @return An integer vector encoding the matches. See Details for
#'     more information.
#'
#'
#' @details
#' 
#' If the data input \code{x} is a feature matrix, matching is based
#' on the Euclidean distance between data points. If the argument
#' \code{x} is a dissimilarity matrix, matching is based on the
#' user-specified dissimilarities. To find matches, the algorithm
#' proceeds by selecting a target element and then searching its
#' nearest neighbours. Critical to the behaviour or the algorithm is
#' the order in which target elements are selected. By default, the
#' most extreme elements are selected first, i.e., elements with the
#' highest distance to the centroid of the data set (see
#' \code{\link{balanced_clustering}} that relies on the same
#' algorithm). Set the argument \code{match_extreme_first} to
#' \code{FALSE}, to enforce that elements close to the centroid are
#' first selected as targets.
#' 
#' If the argument \code{match_between} is passed and the groups
#' specified via this argument are of different size, target elements
#' are selected from the smallest group by default (because in this
#' group, all elements can be matched). However, it is also possible
#' to specify how matches are selected through the option
#' \code{target_group}. When specifying \code{"none"}, matches are
#' always selected from extreme elements, irregardless of the group
#' sizes (or from central elements first if \code{match_extreme_first
#' = FALSE}). With option \code{"smallest"}, matches are selected from
#' the smallest group. With option \code{"diverse"}, matches are
#' selected from the most heterogenous group according to the sum of
#' pairwise distances within groups.
#' 
#' The output is an integer vector encoding which elements have been
#' matched. The grouping numbers are sorted by similarity. That is,
#' elements with the grouping number »1« have the highest intra-group
#' similarity, followed by 2 etc (groups having the same similarity
#' index are still assigned a different grouping number,
#' though). Similarity is measured as the sum of pairwise (Euclidean)
#' distances within groups (see \code{\link{diversity_objective}}). To 
#' prevent sorting by similarity (this is some extra computational burden),
#' set \code{sort_output = FALSE}. Some unmatched elements may be \code{NA}. 
#' This happens if it is not
#' possible to evenly split the item pool evenly into groups of size
#' \code{p} or if the categories described by the argument
#' \code{match_between} are of different size.
#' 
#' @note It is possible to specify grouping restrictions via
#'     \code{match_between} and \code{match_within} at the same time.
#' 
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' @examples
#'
#' # Find triplets
#' N <- 120
#' lds <- data.frame(f1 = rnorm(N), f2 = rnorm(N))
#' triplets <- matching(lds, p = 3)
#' plot_clusters(
#'   lds,
#'   clusters = triplets,
#'   within_connection = TRUE
#' )
#'
#' # Bipartite matching with unequal-sized groups:
#' # Only selects matches for some elements
#' N <- 100
#' data <- matrix(rnorm(N), ncol = 1)
#' groups <- sample(1:2, size = N, replace = TRUE, prob = c(0.8, 0.2))
#' matched <- matching(data[, 1], match_between = groups)
#' plot_clusters(
#'   cbind(groups, data), 
#'   clusters = matched, 
#'   within_connection = TRUE
#' )
#' 
#' # Match objects from the same category only
#' matched <- matching(
#'   schaper2019[, 3:6], 
#'   p = 3, 
#'   match_within = schaper2019$room
#' )
#' head(table(matched, schaper2019$room))
#' 
#' # Match between different plant species in the »iris« data set
#' species <- iris$Species != "versicolor"
#' matched <- matching(
#'   iris[species, 1], 
#'   match_between = iris[species, 5]
#' )
#' # Adjust `match_extreme_first` argument
#' matched2 <- matching(
#'   iris[species, 1], 
#'   match_between = iris[species, 5],
#'   match_extreme_first = FALSE
#' )
#' # Plot the matching results
#' user_par <- par("mfrow")
#' par(mfrow = c(1, 2))
#' data <- data.frame(
#'   Species = as.numeric(iris[species, 5]),
#'   Sepal.Length = iris[species, 1]
#' )
#' plot_clusters(
#'   data,
#'   clusters = matched,
#'   within_connection = TRUE,
#'   main = "Extreme elements matched first"
#' )
#' plot_clusters(
#'   data,
#'   clusters = matched2,
#'   within_connection = TRUE,
#'   main = "Central elements matched first"
#' )
#' par(mfrow = user_par)
#' 
#' 
#' @export
#'

matching <- function(
  x, 
  p = 2, 
  match_between = NULL,
  match_within = NULL,
  match_extreme_first = TRUE,
  target_group = NULL,
  sort_output = TRUE
) {
  
  input_validation_matching(x, p, match_between, match_within, match_extreme_first, target_group)
  
  data <- to_matrix(x)
  match_between <- merge_into_one_variable(match_between)
  target_group <- get_target_group(data, match_between, target_group)
  if (argument_exists(match_within)) {
    cl <- match_within(data, p, match_between, match_within, match_extreme_first, target_group)
  } else {
    cl <- nn_centroid_clustering(data, p, match_between, match_extreme_first, target_group)
  }
  # Before returning: order the group numbers by objective - most similar 
  # matches have lower indices
  if (sort_output) {
    return(sort_by_objective(cl, data))
  }
  cl
}

# Determine from which group target elements are selected
get_target_group <- function(data, match_between, target_group) {
  
  if (!argument_exists(match_between)) { # there is no target group
    return(FALSE)
  }
  tab <- table(match_between)
  # next: target group was not specified by user
  if (!argument_exists(target_group)) {
    if (all(tab == tab[1]))  {
      # if all groups are of same size, there is no target group
      return(FALSE)
    } 
    # smallest group is target group
    return(which.min(tab))
  } 
  # last: user specified target group ("smallest", "diverse", "none")
  if (target_group == "smallest") {
    return(which.min(tab))
  } else if (target_group == "diverse") {
    return(which.max(diversity_objective_by_group(match_between, data)))
  } else if (target_group == "none") {
    return(FALSE)
  } else {
    stop("argument `target_group` must be one of 'smallest', 'diverse' or 'none'")
  }
}


# conduct a matching for each category if `match_within` is passed
match_within <- function(data, p, match_between, match_within, match_extreme_first, target_group) {
  match_within <- merge_into_one_variable(match_within)
  N <- nrow(data)
  cl <- rep(NA, N)
  c <- length(unique(match_within))
  for (i in 1:c) {
    tmp_data <- subset_data_matrix(data, match_within == i)
    cl_tmp <- nn_centroid_clustering(
      tmp_data, 
      p, 
      match_between[match_within == i], 
      match_extreme_first
    )
    # ensure that different cluster numbers are given to different groups
    cl[match_within == i] <- ifelse(is.na(cl_tmp), NA, paste0(cl_tmp, "_", i))
  }
  to_numeric(cl)
}

# After matching was conducted, reorder the group labels by similarity
sort_by_objective <- function(cl, data, N) {
  N <- nrow(data)
  selected <- (1:N)[!is.na(cl)]
  cl_sub <- cl[selected]
  cl_sub <- order_cluster_vector(cl_sub)
  objectives <- diversity_objective_by_group(cl_sub, subset_data_matrix(data, selected))
  # recode original matching labels according to objective
  N <- length(cl_sub)
  # sorry for this code, but it works and it was really complicated
  one <- data.frame(match = cl_sub, order_matches = 1:length(cl_sub))
  two <- data.frame(match = order(objectives), objective_id = 1:length(objectives))
  merged <- merge(one, two)
  new <- rep(NA, N)
  new[merged$order_matches] <- merged$objective_id
  cl[!is.na(cl)] <- new
  cl
}

library("anticlust")

# some exported function cannot be used with NA (anticlustering now can!)
data(iris)
iris[1, 1] <- NA
expect_error(
  matching(iris[, -5], p = 3)
)
expect_error(
  balanced_clustering(iris[, -5], K = 3)
)

# some exported functions cannot be used with non-numeric input (anticlustering now can!)
data(iris)
expect_error(
  matching(iris, p = 3)
)
expect_error(
  balanced_clustering(iris, K = 3)
)

#' Objective value for the variance criterion
#'
#' Compute the k-means variance objective for a given clustering.
#'
#' @param x A vector, matrix or data.frame of data points. Rows
#'     correspond to elements and columns correspond to features. A
#'     vector represents a single feature.
#' @param clusters A vector representing (anti)clusters (e.g., returned
#'     by \code{\link{anticlustering}} or
#'     \code{\link{balanced_clustering}})
#'
#' @return The total within-cluster variance
#'
#' @details
#'
#' The variance objective is given by the sum of the squared
#' errors between cluster centers and individual data points. It is the
#' objective function used in k-means clustering, see
#' \code{\link{kmeans}}.
#'
#'
#' @references
#'
#' Jain, A. K. (2010). Data clustering: 50 years beyond k-means.
#' Pattern Recognition Letters, 31, 651–666.
#'
#' Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
#' data sets into equivalent parts. Psychological Methods, 26(2), 
#' 161–174. https://doi.org/10.1037/met0000301.
#'
#' Späth, H. (1986). Anticlustering: Maximizing the variance criterion.
#' Control and Cybernetics, 15, 213–218.
#'
#' @export
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' @examples
#'
#' data(iris)
#' ## Clustering
#' clusters <- balanced_clustering(
#'   iris[, -5],
#'   K = 3
#' )
#' # This is low:
#' variance_objective(
#'   iris[, -5],
#'   clusters
#' )
#' ## Anticlustering
#' anticlusters <- anticlustering(
#'   iris[, -5],
#'   K = 3,
#'   objective = "variance"
#' )
#' # This is higher:
#' variance_objective(
#'   iris[, -5],
#'   anticlusters
#' )
#' 
#' # Illustrate variance objective
#' N <- 18
#' data <- matrix(rnorm(N * 2), ncol = 2)
#' cl <- balanced_clustering(data, K = 3)
#' plot_clusters(data, cl, illustrate_variance = TRUE)

variance_objective <- function(x, clusters) {
  x <- as.matrix(x)
  validate_input(x, "features", objmode = "numeric")
  validate_input(clusters, "clusters", len = nrow(x), not_na = TRUE)
  variance_objective_(clusters, x)
}

# Internal function - no input handling
variance_objective_ <- function(clusters, data) {
  ## 1. Compute cluster centers
  centers <- cluster_centers(data, clusters)
  ## 2. For each item, compute distance to each cluster center
  distances <- dist_from_centers(data, centers, squared = TRUE)
  ## 3. Use two-column matrix to select relevant distances
  distances <- distances[cbind(1:nrow(distances), clusters)]
  sum(distances)
}


# Compute cluster centers
#
# @param features A data matrix of element features
# @param clusters A numeric vector indicating cluster membership of
#     each element
#
# @return A matrix of cluster centers. Rows represent clusters and
#   columns represent features
#

cluster_centers <- function(features, clusters) {
  centers <- by(features, clusters, colMeans, na.rm = TRUE)
  do.call(rbind, as.list(centers)) # as.list for the case of only one feature
}

# Determine distances of n data points to m cluster centers
#
#
# @param features A vector, matrix or data.frame of data points. If a
#     matrix or data.frame is passed, rows correspond to items and
#     columns to features.
# @param centers A matrix of cluster centers. Each row corresponds to a
#     cluster and each column corresponds to a feature (this format is,
#     for example, returned by the function `stats::kmeans` through the
#     element `centers`).
# @param squared Boolean - compute the squared euclidean distance?
#
# @return A data matrix; columns represent clusters
#     and contain the distance to the respective cluster for each item.
#
# @details
# This code was published in Leisch (2006).
#
# @references
# Leisch (2006). A Toolbox for K-Centroids Cluster Analysis. Computational
# Statistics and Data Analysis, 51(2), 526–544.
#
dist_from_centers <- function(features, centers, squared) {
  z <- matrix(0, nrow = nrow(features), ncol = nrow(centers))
  for (k in 1:nrow(centers)) {
    if (squared)
      z[,k] <- colSums((t(features) - centers[k,])^2, na.rm = TRUE)
    else
      z[,k] <- sqrt( colSums((t(features) - centers[k,])^2, na.rm = TRUE) )
  }
  z
}

library("anticlust")

# local updating works correctly for k-means anticlustering

M <- 2
N <- 60
K <- 2
clusters <- sample(rep(1:K, c(40, 20)))
features <- matrix(rnorm(N * M), ncol = M)

# Swap two items and check objective
to_swap <- sample(1:K, size = 2)
swap1 <- which(clusters == to_swap[1])[1]
swap2 <- which(clusters == to_swap[2])[1]

centers <- anticlust:::cluster_centers(features, clusters)

# Compute clusters after swap
local_update_centers <- anticlust:::update_centers(
  centers, 
  features, 
  swap1, 
  swap2, 
  clusters[swap1], 
  clusters[swap2], 
  table(clusters)
)

# swap clusters, and recompute cluster centers
tmp <- clusters[swap1]
clusters[swap1] <- clusters[swap2]
clusters[swap2] <- tmp

centers_after_swap <- anticlust:::cluster_centers(features, clusters)

expect_equal(local_update_centers, centers_after_swap)

library("anticlust")


## What data structures does k-means use?

# 1. Data matrix of features - row is element, column is feature.
#    n rows, m features
# 2. A vector of anticluster assignments
#    n elements, p unique elements (= number of different anticlusters)
# 3. Matrix of anticluster centers. Each row corresponds to a
#    cluster and each column corresponds to a feature.
#    p rows, m columns
# 4. Matrix, distances from points to anticluster centers,
#    p columns, n rows


## A standard (or squared) euclidean distance between two data points
## Currently, this function is only used in the test files.
euc_dist <- function(x1, x2, squared = FALSE) {
  if (squared)
    return(sum((x1 - x2)^2))
  sqrt(sum((x1 - x2)^2))
}


# computation of cluster centers and distances to centers is correct
## Vary number of features
for (m in 1:4) {
  m_features <- m
  ## vary number of anticlusters
  for (p in 2:5) {
    p_anticlusters <- p
    n_elements <- p * 3 # n-elements must be multiplier of p
    features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
    anticlusters <- rep(1:p_anticlusters, n_elements/p_anticlusters)
    centers  <- anticlust:::cluster_centers(features, anticlusters)
    ## Check output of cluster_centers
    expect_equal(inherits(centers, "matrix"), TRUE)
    # Matrix of distances should have p rows
    expect_equal(dim(centers)[1], p_anticlusters)
    # Matrix of distances should have m columns
    expect_equal(dim(centers)[2], m_features)
    ## Check that means are computed correctly for all features
    for (i in 1:m) {
      expect_equal(as.numeric(centers[, i]),
                   as.numeric(tapply(features[, i], anticlusters, mean)))
    }
    
    ## Now also check if distances to cluster centers are computed correctly
    distances <- anticlust:::dist_from_centers(features, centers, squared = FALSE)
    expect_equal(inherits(centers, "matrix"), TRUE)
    expect_equal(dim(distances)[1], n_elements)
    expect_equal(dim(distances)[2], p_anticlusters)
    ## Check distance output for all points
    for (f in 1:n_elements) {
      for (c in 1:p_anticlusters) {
        expect_equal(distances[f, c], euc_dist(features[f, ], centers[c, ]))
      }
    }
  }
}

#' Validating the arguments passed to `anticlustering`
#' 
#' @return NULL
#'
#' @noRd
input_validation_anticlustering <- function(x, K, objective, method,
                                          preclustering, categories,
                                          repetitions, standardize = FALSE, cannot_link = NULL,
                                          must_link = NULL, blocks = NULL) {
  
  ## Validate feature input
  validate_data_matrix(x)
  
  is_df <- inherits(x, "data.frame")
  if (is_df && !is.function(objective)) {
    x <- get_anticlustering_features(x, objective, standardize) # this is the input that must be considered
  }
  
  x <- as.matrix(x)
  N <- nrow(x)
  unequal_group_sizes <- (length(K) != 1) && (sd(table(initialize_clusters(N, K, NULL))) != 0)
  initial_grouping_passed <- length(K) == N

  validate_input(
    method, "method", len = 1,
    input_set = c("ilp", "exchange", "heuristic", "local-maximum", "brusco", "2PML", "3phase"), 
    not_na = TRUE, not_function = TRUE
  )
  
  if (method == "3phase") {
    if (is.function("objective")) stop("objective can only be a function with method = 'exchange' or method = 'local-maximum'.")
    if (objective %in% "dispersion") {
      stop("objective = dispersion does not work with the 3 phase search algorithm.")
    }
    if (argument_exists(categories)) {
      stop("Using the argument categories does not work with the 3 phase search algorithm.")
    }
    if (unequal_group_sizes && objective %in% c("average-diversity", "variance", "kplus")) {
      stop("The three phase algorithm does not support the average diversity, variance and k-plus objective for unequal-sized groups.")
    }
    if (initial_grouping_passed) {
      stop("The 3 phase algorithm does not handle passing an initial grouping via argument `K`.")
    }
  }
  
  if (argument_exists(must_link)) {
    validate_input(must_link, "must_link", not_function = TRUE, len = N)
    must_link <- as.matrix(must_link)

    if (argument_exists(categories)) {
      stop("\nCombining the `categories` argument together with must-link constraints \n",
           "is currently not supported; use the categorical variables as part of the first argument\n",
           "`x` instead (see the vignette on categorical variables).")
    }
    
    validate_input(objective, "objective", input_set = c("diversity","kplus", "variance"), not_na = TRUE, len = 1) 
    validate_input(method, "method", input_set = c("local-maximum", "exchange", "2PML"), not_na = TRUE, len = 1) 
    
    if (unequal_group_sizes && objective %in% c("kplus", "variance")) {
      stop("K-means and k-plus anticlustering are only possible with equal-sized groups when must-link constraints are used.")
    }
    
    if (ncol(must_link) != 1) {
      stop("Argument must_link must be a vector.")
    }

    if (isTRUE(preclustering)) {
      stop("It is not possible to combine preclustering with must-link constraints.")
    }
    if (argument_exists(cannot_link)) {
      stop("Currently, it is not possible to use both cannot-link and must-link constraints.")
    }
  }
  
  if (argument_exists(cannot_link)) {
    cannot_link <- as.matrix(cannot_link)
    validate_input(cannot_link, "cannot_link", 
                   objmode = "numeric", must_be_integer = TRUE, 
                   greater_than = 0, smaller_than = NROW(x)+1, 
                   not_na = TRUE, not_function = TRUE)
    if (ncol(cannot_link) > 2) {
      stop("Argument cannot_link must have 2 columns.")
    }
    if (objective == "dispersion") {
      stop("objective = dispersion does not work with cannot_link constraints.")
    }
    if (!method %in% c("brusco", "exchange", "local-maximum", "ilp")) {
      stop("Cannot-link constraints currently work with method = 'brusco', 'exchange', 'local-maximum', or 'ilp'.")
    }
    if (argument_exists(categories)) {
      stop("\nCombining the `categories` argument together with cannot-link constraints \n",
           "is currently not supported; use the categorical variables as part of the first argument\n",
           "`x` instead (see the vignette on categorical variables).")
    }
    if (isTRUE(preclustering)) {
      stop("It is not possible to combine preclustering with cannot-link constraints.")
    }
  }
  
  validate_input(standardize, "standardize", objmode = "logical", len = 1,
                 input_set = c(TRUE, FALSE), not_na = TRUE, not_function = TRUE)
  
  if (argument_exists(repetitions)) {
    validate_input(repetitions, "repetitions", objmode = "numeric", len = 1, 
                   greater_than = 0, must_be_integer = TRUE, not_na = TRUE,
                   not_function = TRUE)
  }
  
  ## Merge categories variable so that `length` can be applied:
  if (argument_exists(categories)) {
    categories <- merge_into_one_variable(categories)
    validate_input(categories, "categories", not_function = TRUE, len = N)
  }
  if (argument_exists(blocks)) {
    blocks <- merge_into_one_variable(blocks)
    validate_input(blocks, "blocks", not_function = TRUE, len = N)
    if (argument_exists(must_link)) stop("Currently, using must-link constraints is not possible within blocks.")
    if (!method %in% c("exchange", "local-maximum")) {
      stop("Blocked anticlustering currently only works with method = 'exchange' or 'local-maximum'.")
    }
    if (argument_exists(repetitions)) {
      stop("Cannot use multiple repetitions with blocks.")
    }
    if (preclustering) {
      stop("Cannot use preclustering with blocks.")
    }
  }

  validate_input(preclustering, "preclustering", len = 1,
                 input_set = c(TRUE, FALSE), not_na = TRUE, not_function = TRUE)

  if (any(is.na(x)) && preclustering) {
    stop("Cannot use preclustering when there is NAs in the input.")
  }
  
  if (method == "2PML") {
    if (!argument_exists(must_link)) {
      stop("Method 2PML only works with must-link constraints.")
    }
  }
  
  if (method == "brusco") {
    if (argument_exists(categories)) {
      stop("It is not possible to use the algorithm by Brusco et al. with categorical restrictions.")
    }
    if (preclustering == TRUE) {
      stop("It is not possible to use the algorithm by Brusco et al. with preclustering restrictions.")
    }
  }
  
                 
  # Allow that K is an initial assignment of elements to clusters
  validate_input(K, "K", objmode = "numeric", must_be_integer = TRUE, not_na = TRUE, not_function = TRUE)
  if (length(K) == 1) {
    validate_input(K, "K", greater_than = 1, smaller_than = N)
  } else {
    if (length(K) != N && sum(K) != N) {
      stop("Argument `K` is misspecified.")
    }
    if (method == "ilp") {
      stop("Pass the number of groups via argument `K` when using method = 'ilp'.")
    }
    if (preclustering == TRUE) {
      stop("Cannot only use preclustering when argument `K` is the number of groups.")
    }
    if (method == "ilp") {
      stop("The argument `K` should indicate the number of groups when using `method = 'ilp'`.")
    }
    if (argument_exists(categories)) {
      if (length(categories) != length(K) && sum(K) != N) {
        stop("Length of arguments `categories` and `K` differ.")
      }
    }
  }

  if (argument_exists(categories) && length(categories) != N) {
    stop("The length of the `categories` argument is not equal to the the number of input elements.")
  }

  if (length(K) == 1 && N %% K != 0) {
    if (method == "ilp") {
      stop("K must be a divider of the number of elements when using the ILP method. ",
           "(Try out method = 'exchange'.)")
    }
  }



  if (method == "ilp") {
    check_if_solver_is_available()
    if (!objective %in% c("distance", "diversity", "dispersion")) {
      stop("The ILP method is only available for the diversity and dispersion objectives.")
    }
    if (argument_exists(repetitions)) {
      stop("Do not use argument `repetitions` when using the ILP method.")
    }
  }

  if (!inherits(objective, "function")) {
    validate_input(objective, "objective", input_set = c(
      "distance", 
      "diversity", 
      "average-diversity", 
      "dispersion",
      "variance", 
      "kplus"
      ), 
      len = 1, not_na = TRUE)
    if (objective %in% c("variance", "kplus") && method == "ilp") {
      stop("You cannot use this function to optimally maximize the kmeans or kplus criterion. Use optimal_anticlustering().")
    }
    if (objective %in% c("variance", "kplus") && is_distance_matrix(x) && !is_df) {
        stop("You cannot use a distance matrix with the objective 'variance' or 'kplus'.")
    }
  }

  if (argument_exists(categories) && method == "ilp") {
    stop("The ILP method cannot incorporate categorical restrictions.")
  }
  return(invisible(NULL))
}


#' A function for input validation
#'
#' @param obj The object that undergoes validation
#' @param argument_name A string indicating the name of the object
#'   This name is used when an error is thrown so the user
#'   is informed on the cause of the error.
#' @param len Optional numeric vector for objects having a length
#'   (mostly for vectors). Tests via `NROW`, so can also test matrix-like 
#'   objects.
#' @param greater_than Optional scalar indicating if numeric input has
#'   to be greater than a specified number.
#' @param must_be_integer Optional logical vector indicating if numeric
#'   input has to be integer.
#' @param groupsize Optional argument how many groups a grouping variable
#'   consist of.
#' @param input_set Optional argument specifying a set of values an
#'   argument can take.
#' @param objmode The required mode of \code{obj}
#' @param not_na Boolean to indicate whether NA input is forbidden
#'   (TRUE means that NA is not allowed)
#' @param smaller_than A value for numeric input that must not be exceeded
#'
#' @return NULL
#'
#' @noRd

validate_input <- function(obj, argument_name, len = NULL, greater_than = NULL,
                           must_be_integer = FALSE, groupsize = NULL, input_set = NULL, 
                           objmode = NULL, not_na = FALSE, smaller_than = NULL, 
                           not_function = NULL) {
                           
                           
  self_validation(argument_name, len, greater_than,
                  must_be_integer, groupsize, input_set,
                  objmode, not_na, not_function)

  argument_name <- paste("Argument", argument_name)
  
  if (argument_exists(not_function) && not_function == TRUE) {
    if (inherits(obj, "function")) {
      stop(argument_name, "must not be a function")
    }
  }
   
  
  ## - Check length of input
  if (argument_exists(len)) {
    if (NROW(obj) != len) {
      stop(argument_name, " must have length ", len)
    }
  }

  ## - Check mode of input
  if (argument_exists(objmode)) {
    if (mode(obj) != objmode) {
      stop(argument_name, " must be ", objmode, ", but is ", mode(obj))
    }
  }
  
  ## - Check if input has to be greater than some value
  if (argument_exists(greater_than)) {
    if (any(obj <= greater_than)) {
      stop(argument_name, " must be greater than ", greater_than)
    }
  }
  
  ## - Check if input has to be greater than some value
  if (argument_exists(smaller_than)) {
    if (any(obj >= smaller_than)) {
      stop(argument_name, " must be smaller than ", smaller_than)
    }
  }
  
  if (not_na == TRUE) {
    if (sum(is.na(obj) >= 1)) {
      stop(argument_name, " must not be NA, but contains NA")
    }
  }
  
  ## - Check if input has to be integer
  if (must_be_integer == TRUE && any(obj %% 1 != 0)) {
    stop(argument_name, " must be integer")
  }
  ## - Check if correct number of groups is provided
  if (argument_exists(groupsize)) {
    if (length(table(obj)[table(obj) != 0]) != groupsize) {
      stop(argument_name, " must consist of exactly ", groupsize,
           " groups with more than 0 observations.")
    }
  }

  ## - Check if argument matches a predefined input set
  if (argument_exists(input_set)) {
    if (!obj %in% input_set) {
      stop(argument_name, " can either be set to '",
           paste(input_set, collapse = "' or '"), "'")
    }
  }

  return(invisible(NULL))
}

## Validate feature input
validate_data_matrix <- function(x, NA_allowed = TRUE) {
  x <- data.frame(x)
  modes <- sapply(x, mode)
  if (any(modes != "numeric")) {
    stop("Your data (the first argument `x`) should only consist of numeric and factor variables.")
  }
  if (any(is.na(x)) && !NA_allowed) {
    stop("Missing values are not allowed.")
  }
}

## Validate input for the `validate_input` function (these errors are
## not for users, but only for developers)
self_validation <- function(argument_name, len, greater_than,
                            must_be_integer, groupsize,
                            input_set, objmode, not_na, not_function) {

  if (argument_exists(not_function)) {
    stopifnot(not_function %in% c(TRUE, FALSE))
  }
                            
  if (argument_exists(len)) {
    stopifnot(class(len) %in% c("numeric", "integer"))
    stopifnot(length(len) == 1)
    stopifnot(len >= 0)
    stopifnot(len %% 1 == 0)
  }

  if (argument_exists(greater_than)) {
    stopifnot(length(greater_than) == 1)
    stopifnot(class(greater_than) %in% c("numeric", "integer"))
  }

  stopifnot(length(must_be_integer) == 1)
  stopifnot(must_be_integer %in% c(TRUE, FALSE))
  stopifnot(length(not_na) == 1)
  stopifnot(not_na %in% c(TRUE, FALSE))

  if (argument_exists(groupsize)) {
    stopifnot(mode(groupsize) == "numeric")
    stopifnot(length(groupsize) == 1)
  }

  if ((argument_exists(input_set) && !argument_exists(len)) ||
      (argument_exists(input_set) && len != 1))  {
    stop("If an input set is passed, argument len must be 1 ",
         "(this message should not be seen by users of the package).")
  }

  if (argument_exists(objmode)) {
    stopifnot(mode(objmode) == "character")
    stopifnot(length(objmode) == 1)
  }

  return(invisible(NULL))
}

argument_exists <- function(arg) {
  !is.null(arg)
}


# Test that the number of features is a multiplier of unique
# anticlusters. I think this function is only used in test cases.
legal_number_of_clusters <- function(features, clusters) {
  ## 1. correct number of clusters assignments?
  if (length(clusters) != nrow(features))
    stop("The number of cluster assignments and the number of features have to match.")
  n_anticlusters <- length(unique(clusters))
  ## 2. More than 1 disctinct cluster?
  if (n_anticlusters <= 1)
    stop("There have to be at least two different anticlusters")
  ## 3. Do all clusters occur equally often?
  if (!all(table(clusters) == table(clusters)[1]))
    stop("Each clusters must occur equally often")
  ## 4. Probably redundant to 3:
  if (nrow(features) %% n_anticlusters != 0)
    stop("The number of elements is not a multiplier of the number of anticlusters")
  invisible(NULL)
}


input_validation_matching <- function(
  x, p, 
  match_between, 
  match_within, 
  match_extreme_first, 
  target_group
) {
  validate_data_matrix(x, FALSE)
  N <- nrow(as.matrix(x))
  validate_input(
    p, "p", 
    len = 1, 
    must_be_integer = TRUE, 
    not_na = TRUE
  )
  if (argument_exists(match_between)) {
    validate_input(
      match_between, 
      "match_between", 
      len = N
    )
  }
  if (argument_exists(match_within)) {
    validate_input(
      match_within, 
      "match_within", 
      len = N
    )
  }
  validate_input(
    match_extreme_first, 
    "match_extreme_first", 
    len = 1, 
    input_set = c(TRUE, FALSE), 
    not_na = TRUE
  )
  if (argument_exists(target_group)) {
    validate_input(
      target_group, 
      "target_group", 
      len = 1, 
      input_set = c("smallest", "diverse", "none"), 
      not_na = TRUE
    )
  }
}

# Check if a solver package can be used
check_if_solver_is_available <- function() {
  glpk_available <- requireNamespace("Rglpk", quietly = TRUE)
  symphony_available <- requireNamespace("Rsymphony", quietly = TRUE)
  lpSolve_available <- requireNamespace("lpSolve", quietly = TRUE)
  gurobi_available <- requireNamespace("gurobi", quietly = TRUE)
  no_solver_available <- !glpk_available && !symphony_available && !lpSolve_available && !gurobi_available
  
  if (no_solver_available) {
    stop("\n\nAn exact method was requested, but no ILP solver is ",
         "available. For example, you could install the GNU linear programming kit: \n\n",
         "- On windows, visit ",
         "http://gnuwin32.sourceforge.net/packages/glpk.htm \n\n",
         "- Use homebrew to install it on mac, 'brew install glpk' \n\n",
         "- 'sudo apt install libglpk-dev' on Ubuntu ",
         "\n\nThen, install the Rglpk package via ",
         "`install.packages('Rglpk')`.\n \n Another possibilty is to install the R package 'Rsymphony' ",
         "and the SYMPHONY ILP solver (https://github.com/coin-or/SYMPHONY), or the R package `lpSolve`.")
  }
  return(invisible(NULL))
}

library("anticlust")

# test of input works as expected
features <- matrix(rnorm(12), nrow = 4)
## False number of anticlusters:
expect_error(anticlust:::legal_number_of_clusters(features, 1))
## Only one distinct anticluster:
expect_error(anticlust:::legal_number_of_clusters(features, rep(1, nrow(features))))
## Anticlusters do not occur equally often:
expect_error(anticlust:::legal_number_of_clusters(features, c(1, 1, 1, 2)))
correct <- anticlust:::legal_number_of_clusters(matrix(rnorm(12), nrow = 4), c(1, 2, 1, 2))
expect_equal(correct, NULL)

#' Plot similarity objective by cluster
#' 
#' @param x The data input. Can be one of two structures: (1) A data matrix
#'     where rows correspond to elements and columns correspond to
#'     features (a single numeric feature can be passed as a vector). (2)
#'     An N x N matrix dissimilarity matrix; can be an object of class
#'     \code{dist} (e.g., returned by \code{\link{dist}} or
#'     \code{\link{as.dist}}) or a \code{matrix} where the entries of
#'     the upper and lower triangular matrix represent the pairwise
#'     dissimilarities. 
#' @param groups A grouping vector of length N, usually the output
#'     of \code{\link{matching}}.
#' 
#' @return The diversity (sum of distances) by group.
#' 
#' @details Plots the sum of pairwise distances by group.
#' 
#' @export
#' 
#' @examples
#' 
#' # Match elements and plot similarity by match
#' N <- 100
#' lds <- data.frame(f1 = rnorm(N), f2 = rnorm(N))
#' pairs <- matching(lds, p = 2)
#' plot_similarity(lds, pairs)
#' 
#' @seealso
#'
#' \code{\link{diversity_objective}}
#'
#' @author 
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#' 

plot_similarity <- function(x, groups) {
  x <- as.matrix(x)
  select <- !is.na(groups)
  x <- subset_data_matrix(x, select)
  groups <- groups[select]
  diversity <- diversity_objective_by_group(groups, x)
  plot(diversity, ylab = "Sum of distances", xlab = "Group")
  return(invisible(diversity))
}

#' Visualize a cluster analysis
#'
#' @param features A data.frame or matrix representing the features that
#'     are plotted. Must have two columns.
#' @param clusters A vector representing the clustering
#' @param within_connection Boolean. Connect the elements within each
#'     clusters through lines? Useful to illustrate a graph structure.
#' @param between_connection Boolean. Connect the elements between each
#'     clusters through lines? Useful to illustrate a graph structure.
#'     (This argument only works for two clusters).
#' @param illustrate_variance Boolean. Illustrate the variance criterion
#'     in the plot?
#' @param show_axes Boolean, display values on the x and y-axis? Defaults
#'     to `FALSE`.
#' @param xlab The label for the x-axis
#' @param ylab The label for the y-axis
#' @param xlim The limits for the x-axis
#' @param ylim The limits for the y-axis
#' @param main The title of the plot
#' @param cex The size of the plotting symbols, see \code{\link{par}}
#' @param cex.axis The size of the values on the axes
#' @param cex.lab The size of the labels of the axes
#' @param lwd The width of the lines connecting elements.
#' @param lty The line type of the lines connecting elements
#'    (see \code{\link{par}}).
#' @param frame.plot a logical indicating whether a box should be drawn
#'    around the plot.
#' @param cex_centroid The size of the cluster center symbol (has an
#'    effect only if \code{illustrate_variance} is \code{TRUE})
#'
#' @export
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' @importFrom grDevices palette rainbow
#' @importFrom graphics lines par plot points
#'
#' @details In most cases, the argument \code{clusters} is a vector
#'   returned by one of the functions \code{\link{anticlustering}},
#'   \code{\link{balanced_clustering}} or \code{\link{matching}}. 
#'   However, the plotting function can also be used to plot the results 
#'   of other cluster functions such as \code{\link{kmeans}}. This function
#'   is usually just used to get a fast impression of the results of an 
#'   (anti)clustering assignment, but limited in its functionality. 
#'   It is useful for depicting the intra-cluster connections using 
#'   argument \code{within_connection}.
#'
#' @examples
#'
#' N <- 15
#' features <- matrix(runif(N * 2), ncol = 2)
#' K <- 3
#' clusters <- balanced_clustering(features, K = K)
#' anticlusters <- anticlustering(features, K = K)
#' user_par <- par("mfrow")
#' par(mfrow = c(1, 2))
#' plot_clusters(features, clusters, main = "Cluster editing", within_connection = TRUE)
#' plot_clusters(features, anticlusters, main = "Anticluster editing", within_connection = TRUE)
#' par(mfrow = user_par)
#' 
plot_clusters <- function(features, clusters, within_connection = FALSE,
                          between_connection = FALSE, illustrate_variance = FALSE,
                          show_axes = FALSE, xlab = NULL, ylab = NULL,
                          xlim = NULL, ylim = NULL,
                          main = "", cex = 1.2,
                          cex.axis = 1.2, cex.lab = 1.2, lwd = 1.5,
                          lty = 2, frame.plot = FALSE, cex_centroid = 2) {
  features <- as.matrix(features)
  if (!argument_exists(xlim)) {
    xlim <- c(min(features[, 1]), max(features[, 1]))
  }
  if (!argument_exists(ylim)) {
    ylim <- c(min(features[, 2]), max(features[, 2]))
  }
  if (ncol(features) != 2)
    stop("Can only plot two features")
  if (nrow(features) != length(clusters))
    stop("There must be a cluster for each element - nrow(features) != length(clustering)")

  ## Just in case anybody passes something weird (i.e., cluster numbers
  ## do not start at 1):
  clusters <- to_numeric(clusters)
  # Replace NAs in clustering vector
  clusters[is.na(clusters)] <- seq(
    max(clusters, na.rm = TRUE) + 1, length.out = sum(is.na(clusters))
  )

  ## Specify axis labels if they cannot be assumed from data matrix or argument
  if (is.null(xlab))
    xlab <- ifelse(is.null(colnames(features)[1]), "Feature 1", colnames(features)[1])
  if (is.null(ylab))
    ylab <- ifelse(is.null(colnames(features)[2]), "Feature 2", colnames(features)[2])

  x <- features[, 1]
  y <- features[, 2]

  ## Set colors for clusters
  K <- length(unique(clusters))
  if (K < 8) {
    col <- palette()[2:(K+1)]
    pch <- c(15:18, 3, 4, 8)[1:K]
    pch <- pch[clusters]
  } else {
    col <- rainbow(K)
    pch <- 19
  }

  col <- col[clusters]
  
  # Draw plot
  axt <- "n"
  if (show_axes == TRUE)
    axt <- "s"
  plot(x, y, las = 1, cex.axis = cex.axis, cex.lab = cex.lab,
       col = col, xlab = xlab, ylab = ylab, cex = cex, bg = col,
       xaxt = axt, yaxt = axt, pch = pch, main = main,
       frame.plot = frame.plot, xlim = xlim, ylim = ylim)
  ## Draw graph structure
  if (within_connection == TRUE) {
    draw_all_cliques(x, y, clusters, cols = col, lwd = lwd, lty = lty)
  }
  if (between_connection == TRUE) {
    if (K == 2) {
      draw_between_cliques(x, y, clusters, lwd = lwd, lty = lty)
      ## redraw points so that the lines do notlay the points
      points(x, y, cex = cex, pch = pch, col = col)
    } else {
      warning("Connections between elements of different clusters were ",
              "not drawn because there were more than two clusters.")
    }
  }
  if (illustrate_variance) {
    illustrate_variance(features, clusters, col, lwd, lty, cex_centroid)
  }
}

# Note: length(cols) must be == length(x)
draw_all_cliques <- function(x, y, assignment, lwd = 1.5, lty = 1, cols) {
  n_cliques <- length(unique(assignment))
  for (i in 1:n_cliques) {
    draw_clique(x[assignment == i], y[assignment == i],
                col = cols[assignment == i], lwd = lwd,
                lty = lty)
  }
}


# draw edges connecting all elements of a clique
draw_clique <- function(x, y, col, lwd = 1.5, lty = 1) {
  if (length(x) == 1) {
    return(invisible(NULL))
  }
  connections <- connections_within_clique(x, y)
  cords <- data.frame(x = x, y = y)
  cords <- cords[c(t(connections)), ]
  startpoints <- seq(1, nrow(cords) - 1, 2)
  for (i in startpoints) {
    start <- i
    end   <- i + 1
    lines(c(cords$x[start], cords$x[end]),
          c(cords$y[start], cords$y[end]),
            col = col, lwd = lwd, lty = lty)
  }
}

connections_within_clique <- function(x, y) {
  nitems <- length(x)
  connections <- expand.grid(1:nitems, 1:nitems)
  connections <- connections[connections[, 1] < connections[, 2], ]
  return(connections)
}

## Function to illustrate distances between elements of different cliques
## (does not work in general, but only for K = 2)
draw_between_cliques <- function(x, y, assignment, lwd = 1.5,
                                 lty = 2, col = "darkgrey") {
  selfx <- x[assignment == 1] ## first cluster
  selfy <- y[assignment == 1]
  otherx <- x[assignment != 1] ## second cluster
  othery <- y[assignment != 1]
  for (j in seq_along(selfx)) {
    for (k in seq_along(otherx)) {
      lines(c(selfx[j], otherx[k]), c(selfy[j], othery[k]),
            col = col, lwd = lwd, lty = lty)
    }
  }
}

illustrate_variance <- function(features, clusters, cols, lwd, lty, cex_centroid) {
  K <- length(unique(clusters))
  centers <- cluster_centers(features, clusters)
  # only K colors; in the order implied by "clusters" so that the color
  # scheme is consistent with the points
  cols <- unique(cols[order(clusters)])
  for (i in 1:K) {
    feats <- features[clusters == i, ]
    draw_between_cliques(
      c(feats[, 1], centers[i, 1]),
      c(feats[, 2], centers[i, 2]),
      c(rep(1, nrow(feats)), 2),
      lwd = lwd,
      col = cols[i],
      lty = lty
    )
  }
  points(centers, cex = cex_centroid, pch = 24,
         col = "black", bg = cols, lwd = cex_centroid + 1)
}

#' (Anti)cluster editing "diversity" objective
#'
#' Compute the diversity for a given clustering.
#'
#' @param x The data input. Can be one of two structures: (1) A data matrix
#'     where rows correspond to elements and columns correspond to
#'     features (a single numeric feature can be passed as a vector). (2)
#'     An N x N matrix dissimilarity matrix; can be an object of class
#'     \code{dist} (e.g., returned by \code{\link{dist}} or
#'     \code{\link{as.dist}}) or a \code{matrix} where the entries of
#'     the upper and lower triangular matrix represent the pairwise
#'     dissimilarities.
#' @param clusters A vector representing (anti)clusters (e.g.,
#'     returned by \code{\link{anticlustering}}).
#'
#' @return The cluster editing objective
#'
#' @details
#'
#' The objective function used in (anti)cluster editing is the
#' diversity, i.e., the sum of the pairwise distances between elements
#' within the same groups. When the input \code{x} is a feature
#' matrix, the Euclidean distance is computed as the basic distance
#' unit of this objective.
#'
#'
#' @examples
#'
#' data(iris)
#' distances <- dist(iris[1:60, -5])
#' ## Clustering
#' clusters <- balanced_clustering(distances, K = 3)
#' # This is low:
#' diversity_objective(distances, clusters)
#' ## Anticlustering
#' anticlusters <- anticlustering(distances, K = 3)
#' # This is higher:
#' diversity_objective(distances, anticlusters)
#'
#' @export
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' @references
#'
#' Brusco, M. J., Cradit, J. D., & Steinley, D. (2020). Combining
#' diversity and dispersion criteria for anticlustering: A bicriterion
#' approach. British Journal of Mathematical and Statistical
#' Psychology, 73, 275-396. https://doi.org/10.1111/bmsp.12186
#' 
#' Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
#' data sets into equivalent parts. Psychological Methods, 26(2), 
#' 161–174. https://doi.org/10.1037/met0000301.
#'
#'

diversity_objective <- function(x, clusters) {
  x <- as.matrix(x)
  validate_input(x, "x", objmode = "numeric")
  validate_input(clusters, "clusters", len = nrow(x), not_na = TRUE)
  diversity_objective_(clusters, x)
}

# other order of the arguments, needed for some internal handling
diversity_objective_ <- function(clusters, x) {
  sum(diversity_objective_by_group(clusters, x))
}

# for local-maximum and multiple repetition method with "average diversity" objective:
weighted_diversity_objective_ <- function(x, clusters, frequencies) {
  sum(diversity_objective_by_group(clusters, x) / frequencies)
}

# Compute distance objective by cluster
# param data: distance matrix or feature matrix
# param cl: cluster assignment
diversity_objective_by_group <- function(cl, data) {
  if (is_distance_matrix(data)) {
    objectives <- sapply(
      sort(unique(cl)), 
      function(x) sum(as.dist(data[cl == x, cl == x]))
    )
  } else {
    objectives <- sapply(
      sort(unique(cl)), 
      function(x) sum(dist(data[cl == x, , drop = FALSE]))
    )
  }
  objectives
}

### Test solvers for maximum diversity
library(anticlust)
library(tinytest)
N <- 12
M <- 5
K <- 2

data <- matrix(rnorm(N*M), ncol = M)
start <- Sys.time()
lpsolve <- optimal_anticlustering(data, K, objective = "diversity", solver = "lpSolve")
Sys.time() - start
val1 <- diversity_objective(data, lpsolve)

start <- Sys.time()
glpk <- optimal_anticlustering(data, K, objective = "diversity", solver = "glpk")
Sys.time() - start
val2 <- diversity_objective(data, glpk)

start <- Sys.time()
symphony <- optimal_anticlustering(data, K, objective = "diversity", solver = "symphony") # symphony is slowest for max diversity!
Sys.time() - start
val3 <- diversity_objective(data, symphony)

expect_equal(val1, val2)
expect_equal(val1, val3)

if (requireNamespace("gurobi", quietly = TRUE)) {
  gurobi <- optimal_anticlustering(data, K, objective = "diversity", solver = "gurobi") 
  val4 <- diversity_objective(data, gurobi)
  expect_equal(val1, val4)
}

### Test solvers for maximum dispersion

x <- schaper2019[, 3:6]
start <- Sys.time()
val1 <- optimal_dispersion(x, K = 3, solver = "lpSolve")$dispersion # here lpSolve is slower
Sys.time() - start
start <- Sys.time()
val2 <- optimal_dispersion(x, K = 3, solver = "glpk")$dispersion
Sys.time() - start
start <- Sys.time()
val3 <- optimal_dispersion(x, K = 3, solver = "symphony")$dispersion
Sys.time() - start
expect_equal(val1, val2)
expect_equal(val1, val3)

if (requireNamespace("gurobi", quietly = TRUE)) {
  val4 <- optimal_dispersion(x, K = 3, solver = "gurobi")$dispersion
  expect_equal(val1, val4)
}



### Test solvers for balanced clustering (i.e., reversed maximum - minimum - diversity)

N <- 20
M <- 5
K <- 4

data <- matrix(rnorm(N*M), ncol = M)

lpsolve <- balanced_clustering(data, K, method = "ilp", solver = "lpSolve")
val1 <- diversity_objective(data, lpsolve)

glpk <- balanced_clustering(data, K, method = "ilp", solver = "glpk")
val2 <- diversity_objective(data, glpk)

symphony <- balanced_clustering(data, K, method = "ilp", solver = "symphony")
val3 <- diversity_objective(data, symphony)

expect_equal(val1, val2)
expect_equal(val1, val3)

if (requireNamespace("gurobi", quietly = TRUE)) {
  gurobi <- balanced_clustering(data, K, method = "ilp", solver = "gurobi")
  val4 <- diversity_objective(data, gurobi)
  expect_equal(val1, val4)
}

### Test solvers for weighted cluster editing (i.e., reversed max diversity without group restrictions)

features <- swiss[sample(nrow(swiss), size = 20), ]
features <- swiss
distances <- dist(scale(features))
agreements <- ifelse(as.matrix(distances) < median(distances), 1, -1)

start <- Sys.time()
lpsolve <- wce(agreements, solver = "lpSolve")
Sys.time() - start
val1 <- diversity_objective(agreements, lpsolve)

start <- Sys.time()
glpk <- wce(agreements, solver = "glpk")
Sys.time() - start
val2 <- diversity_objective(agreements, glpk)

start <- Sys.time()
symphony <- wce(agreements, solver = "symphony")
Sys.time() - start
val3 <- diversity_objective(agreements, symphony)

expect_equal(val1, val2)
expect_equal(val1, val3)

if (requireNamespace("gurobi", quietly = TRUE)) {
  gurobi <- wce(agreements, solver = "gurobi")
  val4 <- diversity_objective(agreements, gurobi)
  expect_equal(val1, val4)
}

library(tinytest)
library(anticlust)

# Since version (0.8.7) it is possible that a singleton cluster exists in anticlustering().
# This test files tries to make sure that this does not cause any problems (prior to 
# that, requesting a singleton cluster just threw an error because it can also 
# be left out completely from anticlustering. However, with the implementation of 
# must-link constraints, it is now possible that we "accidentally" have a singleton 
# without intention of the user cluster)

features <- schaper2019[, 3:6]
cluster_sizes <- c(1, 1, 1, 49, nrow(features) - (49+1+1+1))
tt <- anticlustering(
  features,
  K = cluster_sizes
)
expect_true(all(sort(table(tt)) == sort(cluster_sizes)))

tt <- anticlustering(
  features,
  K = cluster_sizes,
  objective = "dispersion"
)
expect_true(all(sort(table(tt)) == sort(cluster_sizes)))

tt <- anticlustering(
  features,
  K = cluster_sizes,
  objective = "variance"
)
expect_true(all(sort(table(tt)) == sort(cluster_sizes)))

tt <- anticlustering(
  features,
  K = cluster_sizes,
  objective = "kplus"
)
expect_true(all(sort(table(tt)) == sort(cluster_sizes)))



# Ensure that diversity_objective is computed correctly when there are singleton clusters

expect_true(
  sum(anticlust:::diversity_objective_by_group(tt, features)) == diversity_objective(features, tt)
)

#############

# C implementation has same result as R implementation when there are singletons?
init <- sample(tt)

tt <- anticlustering(
  features,
  K = init
)

uu <- anticlustering(
  features, 
  K = init,
  objective = diversity_objective
)

expect_true(all(tt == uu))


### Do these tests for other objectives:

# Variance

init <- sample(tt)

tt <- anticlustering(
  features,
  K = init,
  objective = "variance"
)

uu <- anticlustering(
  features, 
  K = init,
  objective = variance_objective
)

expect_true(all(tt == uu))

# Dispersion

init <- sample(tt)

tt <- anticlustering(
  features,
  K = init,
  objective = "dispersion"
)

uu <- anticlustering(
  features, 
  K = init,
  objective = dispersion_objective
)

expect_true(all(tt == uu))


# - Test for singleton created via must-link constraint
N <- nrow(features)
K <- 12
must_link <- rep(NA, N)
must_link[1:(N/K)] <- 1
cl <- anticlustering(features, K = K, must_link = must_link)
expect_true(all(cl[1] == cl[1:(N/K)]))
expect_true(all(cl[1] != cl[((N/K)+1):N]))

# now make it impossible to fulfill the constraints:
must_link[1:((N/K)+1)] <- 1
expect_error(
  anticlustering(features, K = K, must_link = must_link),
  pattern = "must-link"
)

### Test solvers for maximum diversity
library(anticlust)
library(tinytest)
N <- 12
M <- 5
K <- 2

data <- matrix(rnorm(N*M), ncol = M)
start <- Sys.time()
lpsolve <- optimal_anticlustering(data, K, objective = "diversity", solver = "lpSolve")
Sys.time() - start
val1 <- diversity_objective(data, lpsolve)

start <- Sys.time()
glpk <- optimal_anticlustering(data, K, objective = "diversity", solver = "glpk")
Sys.time() - start
val2 <- diversity_objective(data, glpk)

start <- Sys.time()
symphony <- optimal_anticlustering(data, K, objective = "diversity", solver = "symphony") # symphony is slowest for max diversity!
Sys.time() - start
val3 <- diversity_objective(data, symphony)

expect_equal(val1, val2)
expect_equal(val1, val3)

if (requireNamespace("gurobi", quietly = TRUE)) {
  gurobi <- optimal_anticlustering(data, K, objective = "diversity", solver = "gurobi") 
  val4 <- diversity_objective(data, gurobi)
  expect_equal(val1, val4)
}

### Test solvers for maximum dispersion

x <- schaper2019[, 3:6]
start <- Sys.time()
val1 <- optimal_dispersion(x, K = 3, solver = "lpSolve")$dispersion # here lpSolve is slower
Sys.time() - start
start <- Sys.time()
val2 <- optimal_dispersion(x, K = 3, solver = "glpk")$dispersion
Sys.time() - start
start <- Sys.time()
val3 <- optimal_dispersion(x, K = 3, solver = "symphony")$dispersion
Sys.time() - start
expect_equal(val1, val2)
expect_equal(val1, val3)

if (requireNamespace("gurobi", quietly = TRUE)) {
  val4 <- optimal_dispersion(x, K = 3, solver = "gurobi")$dispersion
  expect_equal(val1, val4)
}



### Test solvers for balanced clustering (i.e., reversed maximum - minimum - diversity)

N <- 20
M <- 5
K <- 4

data <- matrix(rnorm(N*M), ncol = M)

start <- Sys.time()
lpsolve <- balanced_clustering(data, K, method = "ilp", solver = "lpSolve")
Sys.time() - start

val1 <- diversity_objective(data, lpsolve)

start <- Sys.time()
glpk <- balanced_clustering(data, K, method = "ilp", solver = "glpk")
Sys.time() - start
val2 <- diversity_objective(data, glpk)

start <- Sys.time()
symphony <- balanced_clustering(data, K, method = "ilp", solver = "symphony")
Sys.time() - start
val3 <- diversity_objective(data, symphony)

expect_equal(val1, val2)
expect_equal(val1, val3)

if (requireNamespace("gurobi", quietly = TRUE)) {
  gurobi <- balanced_clustering(data, K, method = "ilp", solver = "gurobi")
  val4 <- diversity_objective(data, gurobi)
  expect_equal(val1, val4)
}

### Test solvers for weighted cluster editing (i.e., reversed max diversity without group restrictions)

features <- swiss[sample(nrow(swiss), size = 20), ]
features <- swiss
distances <- dist(scale(features))
agreements <- ifelse(as.matrix(distances) < median(distances), 1, -1)

start <- Sys.time()
lpsolve <- wce(agreements, solver = "lpSolve")
Sys.time() - start
val1 <- diversity_objective(agreements, lpsolve)

start <- Sys.time()
glpk <- wce(agreements, solver = "glpk")
Sys.time() - start
val2 <- diversity_objective(agreements, glpk)

start <- Sys.time()
symphony <- wce(agreements, solver = "symphony")
Sys.time() - start
val3 <- diversity_objective(agreements, symphony)

expect_equal(val1, val2)
expect_equal(val1, val3)

if (requireNamespace("gurobi", quietly = TRUE)) {
  gurobi <- wce(agreements, solver = "gurobi")
  val4 <- diversity_objective(agreements, gurobi)
  expect_equal(val1, val4)
}

library("anticlust")

# Test standard usage: K = number of groups
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 3,
  method = "brusco"
)
expect_true(all(table(anticlusters) == c(32, 32, 32)))

# Different group sizes 
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = c(48, 24, 24),
  method = "brusco"
)
expect_true(all(table(anticlusters) == c(48, 24, 24)))

# no categorical restrictions
expect_error(
  anticlusters <- anticlustering(
    schaper2019[, 3:6],
    K = 4,
    categories = schaper2019$room,
    method = "brusco"
  ), 
  pattern = "categorical"
)

# no categorical restrictions
expect_error(
  anticlusters <- anticlustering(
    schaper2019[, 3:6],
    K = 4,
    preclustering = TRUE,
    method = "brusco"
  ), 
  pattern = "preclustering"
)

# Test that bicriterion function works as intended -- equal sized groups
bc <- bicriterion_anticlustering(schaper2019[, 3:6], K = 3, R = 20)
tab <- apply(bc, 1, table)
tab <- data.frame(tab)
tab <- tab[order(tab[, 1], decreasing = TRUE), ]
expect_true(all(tab == c(32, 32, 32)))

# Ensure that vector is returned when using "return" argument
expect_true(
  is.vector(bicriterion_anticlustering(schaper2019[, 3:6], K = 3, return = "best-diversity"))
)

expect_true(
  is.vector(bicriterion_anticlustering(schaper2019[, 3:6], K = 3, return = "best-dispersion"))
)

expect_true(is.vector(
  anticlustering(schaper2019[, 3:6], K = 2, method = "brusco")
))

# For dispersion objective
expect_true(is.vector(
  anticlustering(schaper2019[, 3:6], K = 2, method = "brusco", objective = "dispersion")
))

# Ensure that random seeds work with bicriterion algorithm (this is important
# because the algorithm uses random number generation in C, should be 
# reproducible from R!)
set.seed(1)
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  method = "brusco"
)
set.seed(1)
anticlusters2 <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  method = "brusco"
)
expect_true(all(anticlusters == anticlusters2))

# Other seed = different results
set.seed(3)
anticlusters3 <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  method = "brusco"
)
expect_true(!all(anticlusters == anticlusters3))

# Same with the dispersion criterion
# Ensure that random seeds work with bicriterion algorithm
set.seed(1)
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  method = "brusco",
  objective = "dispersion"
)
set.seed(1)
anticlusters2 <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  method = "brusco",
  objective = "dispersion"
)
expect_true(all(anticlusters == anticlusters2))

# Other seed = different results
set.seed(3)
anticlusters3 <- anticlustering(
  schaper2019[, 3:6],
  K = 4,
  method = "brusco",
  objective = "dispersion"
)
expect_true(!all(anticlusters == anticlusters3))



## Test return argument

N <- 100
M <- 5
K <- 10
data <- matrix(rnorm(N*M), ncol = M)

set.seed(123)
a <- bicriterion_anticlustering(
  data,
  K = K,
  R = c(10, 10)
)

set.seed(123)
b <- bicriterion_anticlustering(
  data,
  K = K,
  R = c(10, 10)
)

expect_true(identical(a, b))

set.seed(123)
c <- bicriterion_anticlustering(
  data,
  K = K,
  R = c(10, 10),
  return = "best-diversity"
)

set.seed(123)
e <- bicriterion_anticlustering(
  data,
  K = K,
  R = c(10, 10),
  return = "best-dispersion"
)

# Is the 1 solution from the pareto set?
n <- nrow(a)
expect_true(duplicated(rbind(a, c))[n+1])
expect_true(duplicated(rbind(a, e))[n+1])


## Test that computation of the return arguments is correct! 
# Use k-means anticlustering + different sized groups

kmeans_distances <- as.matrix(dist(data)^2)
set.seed(123)
a <- bicriterion_anticlustering(
  kmeans_distances,
  K = c(10, 10, 20, 20, 40),
  R = c(10, 10),
  average_diversity = TRUE
)

set.seed(123)
b <- bicriterion_anticlustering(
  kmeans_distances,
  K = c(10, 10, 20, 20, 40),
  R = c(10, 10),
  average_diversity = TRUE
)

expect_identical(a, b)

set.seed(123)
c <- bicriterion_anticlustering(
  kmeans_distances,
  K = c(10, 10, 20, 20, 40),
  R = c(10, 10),
  average_diversity = TRUE,
  return = "best-average-diversity"
)

n <- nrow(a)
expect_true(duplicated(rbind(a, b))[n+1])

# manually compute objectives:
average_diversities <- c()
diversities <- c()
dispersions <- c()
for (i in 1:nrow(a)) {
  average_diversities[i] <- anticlust:::weighted_diversity_objective_(kmeans_distances, a[i, ], table(a[i, ]))
  diversities[i] <- diversity_objective(kmeans_distances, a[i, ])
}

# Compare with "actual" k-means objective computations
expect_equal(average_diversities, apply(a, 1, variance_objective, x = data)) # numeric imprecision is okay here!

# -> weighted_diversity computation is correct!

# Does BILS function return the partition that maximizes the average diversity?
expect_true(all(a[which.max(average_diversities), ] == c))



# prevent inconsistent calls (average_diversity + return)
expect_error(bicriterion_anticlustering(
  dist(data)^2,
  K = c(10, 10, 20, 20, 40),
  R = c(10, 10),
  average_diversity = FALSE,
  return = "best-average-diversity"
))

expect_error(bicriterion_anticlustering(
  dist(data)^2,
  K = c(10, 10, 20, 20, 40),
  R = c(10, 10),
  average_diversity = TRUE,
  return = "best-diversity"
))

#' Generate all partitions of same cardinality
#'
#' @param N The total N. \code{K} has to be dividble
#'     by \code{N}.
#' @param K How many partitions
#' @param generate_permutations If TRUE, all permutations are returned,
#'     resulting in duplicate partitions.
#'
#' @return A list of all partitions (or permutations if
#' \code{generate_permutations} is \code{TRUE}).
#'
#' @details
#'
#' In principle, anticlustering can be solved to optimality by
#' generating all possible partitions of N items into K groups.
#' The example code below illustrates how to do this.
#' However, this approach only works for small N because the
#' number of partitions grows exponentially with N.
#'
#' The partition c(1, 2, 2, 1)
#' is the same as the partition c(2, 1, 1, 2) but they correspond
#' to different permutations of the elements [1, 1, 2, 2]. If the argument
#' \code{generate_permutations} is \code{TRUE}, all permutations are
#' returned. To solve balanced anticlustering exactly, it is sufficient
#' to inspect all partitions while ignoring duplicated permutations.
#'
#' @examples
#'
#' ## Generate all partitions to solve k-means anticlustering
#' ## to optimality.
#'
#' N <- 14
#' K <- 2
#' features <- matrix(sample(N * 2, replace = TRUE), ncol = 2)
#' partitions <- generate_partitions(N, K)
#' length(partitions) # number of possible partitions
#'
#' ## Create an objective function that takes the partition
#' ## as first argument (then, we can use sapply to compute
#' ## the objective for each partition)
#' var_obj <- function(clusters, features) {
#'   variance_objective(features, clusters)
#' }
#'
#' all_objectives <- sapply(
#'   partitions,
#'   FUN = var_obj,
#'   features = features
#' )
#'
#' ## Check out distribution of the objective over all partitions:
#' hist(all_objectives) # many large, few low objectives
#' ## Get best k-means anticlustering objective:
#' best_obj <- max(all_objectives)
#' ## It is possible that there are multiple best solutions:
#' sum(all_objectives == best_obj)
#' ## Select one best partition:
#' best_anticlustering <- partitions[all_objectives == best_obj][[1]]
#' ## Look at mean for each partition:
#' by(features, best_anticlustering, function(x) round(colMeans(x), 2))
#'
#'
#' ## Get best k-means clustering objective:
#' min_obj <- min(all_objectives)
#' sum(all_objectives == min_obj)
#' ## Select one best partition:
#' best_clustering <- partitions[all_objectives == min_obj][[1]]
#'
#' ## Plot minimum and maximum objectives:
#' user_par <- par("mfrow")
#' par(mfrow = c(1, 2))
#' plot_clusters(
#'   features,
#'   best_anticlustering,
#'   illustrate_variance = TRUE,
#'   main = "Maximum variance"
#' )
#' plot_clusters(
#'   features,
#'   best_clustering,
#'   illustrate_variance = TRUE,
#'   main = "Minimum variance"
#' )
#' par(mfrow = user_par)
#'
#' @references
#'
#' Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
#' data sets into equivalent parts. Psychological Methods, 26(2), 
#' 161–174. https://doi.org/10.1037/met0000301.
#'
#' @export
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'

generate_partitions <- function(N, K, generate_permutations = FALSE) {
  if (N %% K != 0) {
    stop("K must be a divider of N.")
  }
  partitions <- 1:K
  init <- rep(partitions, each = N / K)
  anticlusters <- init
  all_ <- list(anticlusters)
  i <- 2
  while (!all(next_permutation(anticlusters) == init)) {
    anticlusters <- next_permutation(anticlusters)
    if (anticlusters[1] == max(anticlusters) &&
        generate_permutations == FALSE) {
      break
    }
    if (!is.unsorted(match(partitions, anticlusters)) ||
        generate_permutations == TRUE) {
      all_[[i]] <- anticlusters
      i <- i + 1
    }
  }
  return(all_)
}

#' Get the next permutation 
#'
#' Each permutation is computed on basis of a passed permutation where
#' lexicographic ordering of the permutations is used to determine the
#' next "higher" permutation. This is an adaption of the
#' `next_permutation` function in C++.
#'
#' @param permutation A vector of elements.
#'
#' @return The next higher permutation of the elements in vector
#'     `permutation` with regard to its lexicographic ordering.
#'
#' 
#' @author Martin Papenberg \email{martin.papenberg@@hhu.de}
#' 
#' @noRd
#' 
next_permutation <- function(permutation) {
  n    <- length(permutation)
  last <- permutation[n]
  i    <- n
  while(last <= permutation[i-1]) {
    last <- permutation[i-1]
    i    <- i - 1
    ## if lexicographic order is already at the maximum:
    if (i-1 == 0) return(sort(permutation))
  }
  ## this algorithm divides the input in a head and a tail; the tail
  ## is monotonically decreasing
  head <- permutation[1:(i-1)]
  tail <- permutation[i:length(permutation)]
  ## which element in the tail is the smallest element that is larger
  ## than the last element in the head?
  larger_values  <- tail[tail > head[length(head)]]
  ## last element of the head:
  final_head <- head[length(head)]
  ## replace last element of head by smallest larger value in tail
  head[length(head)] <- min(larger_values)  
  ## replace smallest larger value in tail by final head element
  tail[max(which(tail == min(larger_values)))] <- final_head
  ## reverse tail before returning
  return(c(head, rev(tail)))
}

# Optimal algorithm for one dimensional bin packing

optimal_binpacking_ <- function(capacities, weights, solver = NULL, time_limit = NULL) {
  
  n_batches <- length(capacities)
  n <- length(weights)
  
  # We get names for the decision variables to keep track of everything
  variables <- get_col_names(n_batches, n)
  
  # Decision variables: Is item i assigned to batch j (so there are 
  # N x C constraints for N items and C batches)
  
  # Two types of constraints:
  # a. Capacity of batches must not be exceeded (1 constraint per batch)
  constraints1 <- matrix(0, ncol = length(variables), nrow = n_batches)
  colnames(constraints1) <- variables
  for (i in 1:nrow(constraints1)) {
    constraints1[i, grepl(paste0("b", i, "_"), variables)] <- weights
  }
  
  # b. Each item is filled into exactly one bin (1 constraint per item)
  constraints2 <- matrix(0, ncol = length(variables), nrow = n)
  colnames(constraints2) <- variables
  for (i in 1:nrow(constraints2)) {
    constraints2[i, grepl(paste0("item_", i, "_"), variables)] <- 1
  }
  
  dirs <- c(rep("<=", nrow(constraints1)), rep("==", nrow(constraints2)))
  rhs <- c(capacities, rep(1, nrow(constraints2)))
  
  ilp <- list()
  ilp$constraints  <- rbind(constraints1, constraints2)
  ilp$equalities   <- dirs
  ilp$rhs          <- rhs
  ilp$obj_function <- rep(1, ncol(constraints1))
  
  ilp_solution <- solve_ilp(ilp, objective = "min", solver = solver, time_limit = time_limit)
  if (ilp_solution$status != 0) {
    stop("The constraints cannot be fulfilled (really).")
  }
  target_bins <- variables[ilp_solution$x == 1]
  # we have to extract the actual target bin from the strings, like this:
  as.integer(gsub("b", "", sapply(strsplit(target_bins, "_"), "[", 1)))
}

get_col_names <- function(n_batches, n) {
  combs <- expand.grid(1:n_batches, 1:n)
  paste0("b", combs[,1], "_item_", combs[, 2], "_")
}

# Function that repeatedly calls anticlustering and returns best results
repeat_anticlustering <- function(x, K, objective, categories, method, repetitions) {
  
  N <- nrow(x)
  
  # Create initial cluster assignment for each `repetition`
  if (argument_exists(repetitions) && repetitions > 1) {
    clusters <- get_multiple_initial_clusters(N, K, categories, repetitions)
  } else {
    clusters <- list(initialize_clusters(N, K, categories))
  }
  
  if (inherits(objective, "function")) {
    obj_function <- objective
  } else if (objective == "variance") {
    obj_function <- variance_objective
  } else if (objective == "diversity" || objective == "distance") {
    obj_function <- diversity_objective
  } else if (objective == "dispersion") {
    obj_function <- dispersion_objective
  } else if (objective == "average-diversity") {
    obj_function <- function(clusters, x) {
      weighted_diversity_objective_(clusters, x, table(clusters))
    }
  }
  
  if (method == "local-maximum") {
    candidate_solutions <- lapply(
      clusters,
      local_maximum_anticlustering,
      data = x, 
      objective = objective, 
      obj_function = obj_function,
      categories = categories
    )
  } else if (method == "exchange") {
    candidate_solutions <- lapply(
      clusters,
      anticlustering_,
      data = x, 
      objective = objective, 
      categories = categories
    )
    candidate_solutions
  }
  
  # Define objective function that can be used by `lapply` on many clusterings
  obj_function_ <- function(clusters, x) {
    obj_function(x, clusters)
  }
  # Get best of all solutions
  objs <- lapply(
    candidate_solutions,
    obj_function_,
    x = x
  )
  candidate_solutions[[which.max(objs)]]
}

get_multiple_initial_clusters <- function(N, K, categories, repetitions) {
  
  clusters <- list(initialize_clusters(N, K, categories))
  
  if (argument_exists(categories)) {
    clusters_repetitions <- as.list(data.frame(as.matrix(
      replicate(repetitions - 1, categorical_sampling(categories, K))
    )))
  } else {
    clusters_repetitions <- as.list(data.frame(as.matrix(
      replicate(repetitions - 1, sample(clusters[[1]]))
    )))
  }
    
  clusters_repetitions <- unname(clusters_repetitions)
  merge_lists(list(clusters, clusters_repetitions))
}


# Function that calls anticlustering exchange method until a local 
# maximum is reached (i.e., no possible swap improves the objective)
local_maximum_anticlustering <- function(
  clusters, data, objective, obj_function, categories) {
  
  old_obj <- obj_function(data, clusters)
  has_improved <- TRUE
  
  while (has_improved) {
    new_clusters <- anticlustering(
      x = data, 
      K = clusters, 
      objective = objective, 
      categories = categories, 
      repetitions = NULL,
      method = "exchange",
      preclustering = FALSE
    )
    new_obj <- obj_function(data, new_clusters)
    if (new_obj > old_obj) {
      clusters <- new_clusters
      old_obj <- new_obj
    } else {
      has_improved <- FALSE
    }
  }
  clusters
}

# `anticlustering()` to be used repeatedly by `lapply()`
anticlustering_ <- function(clusters, data, objective, categories) {
  anticlustering(
    data, 
    K = clusters, 
    objective = objective, 
    method = "exchange", 
    categories = categories, 
    repetitions = NULL,
    preclustering = FALSE
  )
}

#' Optimal ("exact") algorithms for anticlustering
#' 
#' Wrapper function that gives access to all optimal algorithms for anticlustering 
#' that are available in anticlust.
#' 
#'@param x The data input. Can be one of two structures: (1) A feature
#'     matrix where rows correspond to elements and columns correspond
#'     to variables (a single numeric variable can be passed as a
#'     vector). (2) An N x N matrix dissimilarity matrix; can be an
#'     object of class \code{dist} (e.g., returned by
#'     \code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
#'     where the entries of the upper and lower triangular matrix
#'     represent pairwise dissimilarities.
#' @param K How many anticlusters should be created or alternatively:
#'     (a) A vector describing the size of each group (the latter
#'     currently only works for \code{objective = "dispersion")}.
#' @param objective The anticlustering objective, can be "diversity",
#'     "variance", "kplus" or "dispersion".
#' @param solver Optional. The solver used to obtain the optimal
#'     method.  Currently supports "glpk", "symphony", "lpSolve" and "gurobi". 
#'     See details.
#' @param time_limit Time limit in seconds, given to the solver.
#'    Default is there is no time limit.
#'
#' @return A vector of length N that assigns a group (i.e, a number
#'     between 1 and \code{K}) to each input element.
#' 
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#' 
#' @details 
#' 
#' This is a wrapper for all optimal methods supported in anticlust
#' (currently and in the future).  As compared to
#' \code{\link{anticlustering}}, it allows to specify the solver to
#' obtain an optimal solution, a time limit, and it can be used to obtain optimal
#' solutions for all supported anticlustering objectives (variance,
#' diversity, k-plus, dispersion). For the objectives "variance",
#' "diversity" and "kplus", the optimal ILP method in Papenberg and
#' Klau (2021) is used, which maximizes the sum of all pairwise
#' intra-cluster distances (given user specified number of clusters,
#' for equal-sized clusters).  To employ k-means anticlustering
#' (i.e. set \code{objective = "variance"}), the squared Euclidean
#' distance is used. For k-plus anticlustering, the squared Euclidean
#' distance based on the extended k-plus data matrix is used (see
#' \code{\link{kplus_moment_variables}}).  For the diversity (and the
#' dispersion), the Euclidean distance is used by default, but any
#' user-defined dissimilarity matrix is possible. The equivalence
#' of the k-means/k-plus objectives and the diversity, which enables 
#' using the ILP approach by Papenberg and Klau for the k-means/k-plus 
#' objectives, was discussed in Papenberg, Breuer, et al. (2025). 
#' Also see the examples. 
#' 
#' 
#' The dispersion is solved optimal using the algorithm OptDispF presented in 
#' Papenberg, Breuer, et al. (2025). Also see \code{\link{optimal_dispersion}}.
#' 
#' The optimal methods make use of "solvers" that actually implement
#' the algorithm for finding optimal solutions. The package anticlust
#' supports four solvers:
#' 
#' \itemize{
#'   \item{The default solver lpSolve (<https://sourceforge.net/projects/lpsolve/>).}
#'   \item{GNU linear programming kit (<http://www.gnu.org/software/glpk/>), 
#'   available from the package Rglpk and requested using \code{solver = "glpk"}.
#'   The R package Rglpk has to be installed manually if this solver should be used.}
#'   \item{The Symphony solver (<https://github.com/coin-or/SYMPHONY>),
#'   available from the package Rsymphony and requested using \code{solver = "symphony"}.
#'   (The package Rsymphony has to be installed manually if this solver should be used).}
#'   \item{The commercial gurobi solver, see https://www.gurobi.com/downloads/.}
#' }
#' 
#' In general, the commercial gurobi solver is best at solving exact anticlustering
#' problems. If you have a license to use it, use it!
#' Among the open source alternatives, for the maximum dispersion problem, 
#' it seems that the Symphony
#' solver is fastest, while the lpSolve solver seems to be good for
#' maximum diversity. However, note that in general the dispersion can
#' be solved optimally for much larger data sets than the diversity.
#' 
#' If a \code{time_limit} is set and the function cannot find in the optimal
#' objective in the given time, it will throw an error.
#' 
#' @references 
#' 
#' Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
#' data sets into equivalent parts. Psychological Methods, 26(2), 
#' 161–174. https://doi.org/10.1037/met0000301.
#' 
#' Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. (2025). 
#' Extending the Bicriterion Approach for Anticlustering: Exact and Hybrid 
#' Approaches. Psychometrika. Advance online publication. https://doi.org/10.1017/psy.2025.10052

#' @export
#' 
#' @examples 
#' 
#' data <- matrix(rnorm(24), ncol = 2)
#' 
#' # These calls are equivalent for k-means anticlustering:
#' optimal_anticlustering(data, K = 2, objective = "variance")
#' optimal_anticlustering(dist(data)^2, K = 2, objective = "diversity")
#' 
#' # These calls are equivalent for k-plus anticlustering:
#' optimal_anticlustering(data, K = 2, objective = "kplus")
#' optimal_anticlustering(dist(kplus_moment_variables(data, 2))^2, K = 2, objective = "diversity")
#' 
optimal_anticlustering <- function(x, K, objective, solver = NULL, time_limit = NULL) {
  
  validate_input_optimal_anticlustering(x, K, objective, solver, time_limit)
  input_validation_anticlustering(
    x, K, objective = objective, method = "exchange", # 'method' is actually a lie to make it work
    preclustering = FALSE, categories = NULL, 
    repetitions = NULL, 
    standardize = FALSE
  )
  
  if (!argument_exists(solver)) {
    solver <- find_ilp_solver()
  }
  
  if (objective == "kplus") {
    x <- kplus_moment_variables(x, 2)
  }
  
  x <- convert_to_distances(x)
  if (objective %in% c("kplus", "variance")) {
    x <- x^2
    objective <- "diversity"
  }

  if (objective == "diversity") {
    ilp <- anticlustering_ilp(x, K)
    solution <- solve_ilp(ilp, solver = solver, time_limit = time_limit)
    if (solution$status != 0) {
      stop("Could not find the optimal objective in the given time limit.")
    }
    return(ilp_to_groups(solution, nrow(x)))
  } else {
    return(optimal_dispersion(x, K, solver, time_limit = time_limit)$groups)
  }
  
}
  
validate_input_optimal_anticlustering <- function(x, K, objective, solver, time_limit) {
  # x
  validate_data_matrix(x)
  # K
  validate_input(K, "K", objmode = "numeric", must_be_integer = TRUE, not_na = TRUE, not_function = TRUE)
  N <- nrow(as.matrix(x))
  if (!(length(K) == 1 || sum(K) == N)) {
    stop("Argument `K` is misspecified.")
  } 
  if (objective != "dispersion") {
    if (!all(K == K[1]) || length(K) == 1 && N%%K != 0) {
      stop("The selected method can only be used for equal-sized groups")
    }
  }
  
  # Objective
  validate_input(
    objective, "objective", objmode = "character", len = 1,
    input_set = c("variance", "diversity", "kplus", "dispersion"), 
    not_na = TRUE, 
    not_function = TRUE
  )

  # Solver
  if (argument_exists(solver)) {
    validate_input(solver, "solver", objmode = "character", len = 1,
                  input_set = c("glpk", "symphony", "lpSolve", "gurobi"), not_na = TRUE, not_function = TRUE)
    if (solver == "glpk") {
      if (!requireNamespace("Rglpk", quietly = TRUE)) {
        stop("The package Rglpk must be installed to use `solver = glpk`.\n", 
             "Type install.packages('Rglpk') in the R console to install it.")
      }
    } else if (solver == "symphony") {
      if (!requireNamespace("Rsymphony", quietly = TRUE)) {
        stop("The package Rsymphony must be installed to use `solver = symphony`.\n", 
             "Type install.packages('Rsymphony') in the R console to install it.")
      }
    } else if (solver == "gurobi") {
      if (!requireNamespace("gurobi", quietly = TRUE)) {
        stop("The package gurobi must be installed to use `solver = gurobi`.")
      }
    }
  }
  # time_limit 
  if (argument_exists(time_limit)) {
    validate_input(
      time_limit, 
      "time_limit",
      objmode = "numeric",
      len = 1, not_na = TRUE, 
      not_function = TRUE, 
      greater_than = 0,
      must_be_integer = TRUE
    )
  }
}

library(tinytest)
library(anticlust)

N <- 12
M <- 5
K <- 2
dat <- matrix(rnorm(N * M), ncol = M)

# just make sure that all runs okay
optimal_anticlustering(dat, K, "diversity", solver = "glpk")
optimal_anticlustering(dat, K, "kplus", solver = "glpk")
optimal_anticlustering(dat, K, "variance", solver = "glpk")
optimal_anticlustering(dat, K, "dispersion", solver = "glpk") 

expect_error(
  optimal_anticlustering(dat, K = 5, "diversity", solver = "glpk"),
  pattern = "equal-sized groups"
)

expect_error(
  optimal_anticlustering(dat, K = c(2, 2, 2, 6), "diversity", solver = "glpk"),
  pattern = "equal-sized groups"
)

# it should work for dispersion however: 
optimal_anticlustering(dat, K = 5, "dispersion", solver = "glpk")
optimal_anticlustering(dat, K = c(2, 2, 2, 6), "dispersion", solver = "glpk")

expect_error(
  optimal_anticlustering(dat, K = K, "diversity", solver = "FOO"),
  pattern = "Argument solver"
)

expect_error(
  optimal_anticlustering(dist(dat), K = K, "variance", solver = "glpk"),
  pattern = "objective 'variance'"
)

expect_error(
  optimal_anticlustering(dist(dat), K = K, "kplus", solver = "glpk"),
  pattern = "objective 'variance'"
)

#' Solve balanced K-anticluster editing exactly using ILP
#'
#' @param distances A N x N matrix representing the
#'     pairwise dissimilarities between all N elements. Cannot an be an
#'     object of class \code{dist} (i.e., has been converted to matrix
#'     before this function is called)
#' @param K How many anticlusters should be created.
#' @param preclustering Boolean, should a preclustering be conducted
#'     before anticlusters are created.
#'
#' @return A vector representing the anticluster affiliation of
#'     elements.
#' @importFrom utils combn
#'
#' @noRd

exact_anticlustering <- function(data, K, preclustering, cannot_link) {
  
  distances <- convert_to_distances(data)
  N <- nrow(distances)

  if (preclustering == TRUE) {
    ilp <- anticlustering_ilp(distances, N / K)
    solution <- solve_ilp(ilp, "min")
    preclusters <- ilp_to_groups(solution, N)
    ## Fix distances - ensures that the most similar items are assigned
    ## to different groups
    distances <- edit_distances(distances, preclusters, value = (sum(distances) + 1) * (-1))
    ## Edit ILP - objective function and group sizes
    ilp$obj_function <- vectorize_weights(distances)$costs
    ilp$rhs <- c(rep(1, choose(N, 3) * 3),
                 rep((N / K) - 1, N))
    ## Solve edited ILP
    solution <- solve_ilp(ilp)
    assignment <- ilp_to_groups(solution, N)
    return(assignment)
  }
  
  if (argument_exists(cannot_link)) {
    distances[cleanup_cannot_link_indices(cannot_link)] <- -(sum(distances) + 1)
  }

  ## Here the ILP is created without adjusting distances; i.e., true
  ## exact anticlustering
  ilp <- anticlustering_ilp(distances, K)
  solution <- solve_ilp(ilp)
  ilp_to_groups(solution, N)
}

# Ensure that a distance matrix is passed
convert_to_distances <- function(data, squared = FALSE) {
  if (!is_distance_matrix(data)) {
    distances <- dist(data)
    if (squared) distances <- distances^2
  } else {
    distances <- data
  }
  as.matrix(distances)
}

#' Construct the ILP represenation of a cluster editing instance
#' 
#' (Can be used to solve anticlustering and clustering problems)
#'
#' @param distances An n x n matrix representing the
#'     distances between items
#' @param K The number of groups to be created
#' @param group_restriction If FALSE, there is no restriction on the
#'     group size, leading to a normal weighted cluster editing formulation
#'
#' @return A list representing the ILP formulation of the instance
#'
#' @noRd
#'

anticlustering_ilp <- function(distances, K, group_restriction = TRUE) {

  # Initialize some constant variables:
  equality_signs <- equality_identifiers()
  n        <- nrow(distances)
  group_size     <- n / K
  costs          <- vectorize_weights(distances)

  # Specify the number of triangular constraints:
  n_tris <- choose(n, 3) * 3

  # Construct ILP constraint matrix
  constraints <- sparse_constraints(n, costs$pair)
  colnames(constraints) <- costs$pair

  ## Directions of the constraints:
  equalities <- c(rep(equality_signs$l, n_tris),
                  rep(equality_signs$e, n))

  # Right-hand-side of ILP
  rhs <- c(rep(1, n_tris), rep(group_size - 1, n))

  # Objective function of the ILP
  obj_function <- costs$costs

  # Give names to all objects for inspection purposes
  names(obj_function) <- colnames(constraints)

  # For normal cluster editing, remove group constraints:
  if (group_restriction == FALSE) {
    rhs <- rhs[1:n_tris]
    equalities <- equalities[1:n_tris]
    constraints <- constraints[1:n_tris, ]
  }

  ## return instance
  instance              <- list()
  instance$n_groups     <- K
  instance$group_size   <- group_size
  instance$distances    <- distances
  instance$costs        <- costs
  instance$constraints  <- constraints
  instance$equalities   <- equalities
  instance$rhs          <- rhs
  instance$obj_function <- obj_function

  return(instance)
}

# Return identifiers for equality relationships
#
# @return A list of three elements containing strings representing
#     equality (e), lower (l), and greater (g) relationships
#
equality_identifiers <- function() {
  equal_sign <- "=="
  lower_sign <- "<="
  greater_sign <- ">="
  list(e = equal_sign, l = lower_sign, g = greater_sign)
}


# Convert matrix of distances into vector of distances
#
# @param distances A distance matrix of class matrix or dist
#
# @return A data.frame having the following columns: `costs` - the
#     distances in vectorized form; `i` the first index of the item
#     pair that is connected; `j` the second index of the item pair
#     that is connected; `pair` A string of form "xi_j" identifying the
#     item pair
vectorize_weights <- function(distances) {
  # Problem: I have matrix of costs but need vector for ILP.
  # Make vector of costs in data.frame (makes each cost identifiable)
  costs <- expand.grid(1:ncol(distances), 1:nrow(distances))
  colnames(costs) <- c("i", "j")
  costs$costs <- c(distances)
  # remove redundant or self distances:
  costs <- costs[costs$i < costs$j, ]
  costs$pair <- paste0("x", costs$i, "_", costs$j)
  rownames(costs) <- NULL
  return(costs)
}

# Construct a sparse matrix representing the ILP constraints
#
# @param n The number of elements
# @param pair_names A character vector of names representing the item
#     pairs (e.g.: "x1_2", ..., "x100_123"). Is read out from the data frame
#     have `costs` (i.e., costs$pair is passed as pair_names)
# @return A sparse matrix representing the left-hand side of the ILP (A
#     in Ax ~ b)
#
sparse_constraints <- function(n, pair_names) {
  tri <- triangular_constraints(n, pair_names)
  gr  <- group_constraints(n, pair_names)
  Matrix::sparseMatrix(c(tri$i, gr$i), c(tri$j, gr$j), x = c(tri$x, gr$x))
}

# Indices for sparse matrix representation of triangular constraints
triangular_constraints <- function(n, pair_names) {
  triangular_constraints <- choose(n, 3)
  coef_per_constraint <- 3
  # number of coefficients in constraint matrix:
  col_indices <- matrix(ncol = triangular_constraints, nrow = coef_per_constraint * 3)
  row_indices <- rep(1:(triangular_constraints*3), each = 3)
  xes <- rep(c(-1, 1, 1, 1, -1, 1, 1, 1, -1), triangular_constraints)
  # generate all item triplets
  triplets <- t(combn(1:n, 3))
  # get the respective column indices in constraint matrix
  ds <- c(t(data.frame(
    paste0("x", triplets[, 1], "_", triplets[, 2]),
    paste0("x", triplets[, 1], "_", triplets[, 3]),
    paste0("x", triplets[, 2], "_", triplets[, 3])
  )))
  col_indices <- matrix(match(ds, pair_names), nrow = 3)
  col_indices <- rbind(col_indices, col_indices, col_indices)
  list(i = row_indices, j = c(col_indices), x = xes)
}

# Indices for sparse matrix representation of group constraints
group_constraints <- function(n, pair_names) {
  coef_per_constraint <- (n - 1)
  group_coefficients <- coef_per_constraint * n
  row_indices <- rep((1:n) + (3 * choose(n, 3)), each = coef_per_constraint)
  col_indices <- matrix(ncol = n, nrow = coef_per_constraint)
  xes         <- rep(1, length = group_coefficients)
  for (i in 1:n) {
    connections <- all_connections(i, n)
    pairs <- paste0("x", connections[, 1], "_", connections[, 2])
    col_indices[, i] <- match(pairs, pair_names)
  }
  return(list(i = row_indices, j = c(col_indices), x = xes))
}

# Find all connections of an element
# @param i The point for which the connections are sought (must be in
#     1...n)
# @param n The number of points
# @return A `data.frame` with two columns representing start and end
#     point.  The second column always contains the "larger number"
#     point.
all_connections <- function(i, n) {
  connections <- expand.grid(i, setdiff(1:n, i))
  # put lower index in first column
  wrongorder <- connections[, 1] > connections[, 2]
  temp <- connections[, 2][wrongorder]
  connections[, 2][wrongorder] <- connections[, 1][wrongorder]
  connections[, 1][wrongorder] <- temp
  return(connections)
}

#test setup of ILP
library("anticlust")

#anticlustering_ilp
# ILP is set up as expected
conditions <- expand.grid(m = 1:4, p = 2:4)
for (k in 1:nrow(conditions)) {
  m_features <- conditions[k, "m"]
  p_anticlusters <- conditions[k, "p"]
  n_elements <- p_anticlusters * 2 # n must be multiplier of p
  features <- matrix(rnorm(n_elements * m_features), ncol = m_features)
  distances <- as.matrix(dist(features))
  ilp <- anticlust:::anticlustering_ilp(distances, p_anticlusters)
  expect_equal(nrow(ilp$constraints), choose(n_elements, 3) * 3 + n_elements)
  expect_equal(sum(is.na(ilp$constraints)), 0)
  
  ## Test that distance matrix is correctly transfered into objective vector
  distances <- as.matrix(distances)
  costs <- ilp$costs
  for (i in 1:n_elements) {
    for (j in 1:n_elements) {
      if (i >= j) next
      expect_equal(costs[costs$i == i & costs$j == j, ]$costs, distances[i, j])
    }
  }
}

#' Get binary representation of categorical variables
#'
#' @param categories A vector, data.frame or matrix representing one
#'     or several categorical variables
#' @param use_combinations Logical, should the output also include columns representing
#'    the combination / interaction of the categories (defaults to \code{FALSE}).
#'
#' @return A matrix encoding the categorical variable(s) in binary form. 
#'
#' @details
#' 
#' The conversion of categorical variables to binary variables is done via
#' \code{\link[stats]{model.matrix}}. Since version 0.8.9, each category
#' of a categorical variable is coded by a separate variable ('one hot' encoding).
#' So we do not use 'dummy' coding, which is often used to encode predictors 
#' in statistical analysis. Dummy coding uses a reference category that is not
#' explicitly modelled via a separate variable. This implies that there is
#' a different distance to the reference category than among the other 
#' categories, which is unwarranted in anticlustering (thanks to Gunnar Klau
#' for noting this). See examples. 
#' 
#' This function can be used to include categorical variables as part of the 
#' optimization criterion in anticlustering, rather than including them as hard constraints as done when using the 
#' argument \code{categories} in \code{\link{anticlustering}} (or \code{\link{fast_anticlustering}}). 
#' This way, categorical variables are treated as numeric variables, 
#' which can be useful when there are several
#' categorical variables or when the group sizes are unequal (or both).
#' See examples. Please see the vignette 'Using categorical variables with anticlustering'
#' for more information on this approach.
#' 
#' Since version 0.8.12, this function treats NA in the categorical input variables
#' as a distinct category. This implies that different observations having an NA are treated
#' as having the same value on this variable. If this is not desired, you need to do some other
#' preprocessing of the NAs before calling this function.
#' 
#' @importFrom stats as.formula model.matrix contrasts
#'
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' @references
#' 
#' Papenberg, M., Wang, C., Diop, M., Bukhari, S. H., Oskotsky, B., Davidson, B. R., 
#' Vo, K. C., Liu, B., Irwin, J. C., Combes, A., Gaudilliere, B., Li, J., Stevenson, D. K., 
#' Klau, G. W., Giudice, L. C., Sirota, M., & Oskotsky, T. T. (2025). Anticlustering
#' for sample allocation to minimize batch effects. Cell Reports Methods, 5(8), 
#' 101137. https://doi.org/10.1016/j.crmeth.2025.101137
#' 
#' @export
#'
#' @examples
#' 
#' # Illustrate why dummy encoding is not appropriate for anticlustering.
#' # Use 3 elements of the iris data set, with different levels of 'Species'
#' input <- iris[c(1, 51, 101), "Species", drop = FALSE]
#' input
#' # Default dummy encoding: 
#' (dummy <- model.matrix(~ . , input))
#' dist(dummy) 
#' # Distance between versicolor and virginica is larger than among setosa and 
#' # versicolor, and setosa and virginica. This would bias the anticlustering 
#' # computation.
#' (one_hot <- categories_to_binary(input))
#' dist(one_hot) # all the same distances 
#' dist(one_hot)^2 # or use squared Euclidean distance (corresponds to Manhattan distance in this case)
#' dist(one_hot, method = "manhattan")
#' 
#' # We could also get one-hot encoding directly via model.matrix(), which is 
#' # used in categories_to_binary():
#' model.matrix(~ . -1, input) # suppresses the 'intercept'
#' 
#' # Use Schaper data set for anticlustering example
#' data(schaper2019)
#' features <- schaper2019[, 3:6]
#' K <- 3
#' N <- nrow(features) 
#' 
#' # - Generate data input for k-means anticlustering -
#' # We conduct k-plus anticlustering by first generating k-plus variables, 
#' # and also include the categorical variable as "numeric" input for the 
#' # k-means optimization (rather than as input for the argument categories)
#' 
#' input_data <- cbind(
#'   kplus_moment_variables(features, T = 2), 
#'   categories_to_binary(schaper2019$room) 
#' )
#' 
#' kplus_groups <- anticlustering(
#'   input_data, 
#'   K = K,
#'   objective = "variance",
#'   method = "local-maximum", 
#'   repetitions = 10
#' )
#' mean_sd_tab(features, kplus_groups)
#' table(kplus_groups, schaper2019$room) # argument categories was not used!
#' 
#'  

categories_to_binary <- function(categories, use_combinations = FALSE) {
  validate_input(use_combinations, "use_combinations", objmode = "logical", len = 1,
                 input_set = c(TRUE, FALSE), not_na = TRUE, not_function = TRUE)
  categories <- data.frame(categories)
  categories <- as.data.frame(lapply(categories, factor, exclude = NULL))
  colnames(categories) <- paste0("X", 1:ncol(categories))
  ## Does some variable only have 1 value??? Allow this by converting to numeric (i.e., 1!)
  for (i in 1:ncol(categories)) {
    if (length(levels(categories[, i])) == 1) {
      categories[, i] <- 1
    }
  }
  combine_by <- ifelse(use_combinations, " * ", " + ")
  formula_string <- paste("~", paste(colnames(categories), collapse = combine_by), collapse = "")
  model.matrix(
    as.formula(formula_string), 
    data = categories,
    contrasts.arg = lapply(categories, contrasts, contrasts=FALSE) # this ensures that each level of the category has a binary variable
  )[ ,-1, drop = FALSE]
}

library(anticlust)
library(tinytest)

# test that one-hot encoding is done correctly

generate_categorical_data <- function(N, M, P, prob = NULL) {
  data <- data.frame(matrix(sample(P, replace = TRUE, size = N*M, prob = prob), ncol = M))
  # wandel Datensatz in Faktor um (jede Spalte)
  as.data.frame(lapply(data, as.factor))
}

# old version of categories_to_binary(), that clearly uses one-hot encoding:
one_hot_encoding <- function(categories, use_combinations = FALSE) {
  categories <- data.frame(categories)
  combine_by <- ifelse(use_combinations, " * ", " + ")
  formula_string <- paste("~", paste(colnames(categories), collapse = combine_by), collapse = "")
  model.matrix(
    as.formula(formula_string),
    data = categories,
    contrasts.arg = lapply(categories, contrasts, contrasts=FALSE) # this ensures that each level of the category has a binary variable
  )[ ,-1, drop = FALSE]
}


d <- generate_categorical_data(20, 4, 5)

one_hot1 <- one_hot_encoding(d)
dim(one_hot1)
one_hot2 <- anticlust::categories_to_binary(d)
dim(one_hot2)

expect_equal(one_hot1, one_hot2)

#' Cluster dispersion 
#' 
#' Compute the dispersion objective for a given clustering (i.e., the
#' minimum distance between two elements within the same cluster).
#' 
#' @param x The data input. Can be one of two structures: (1) A
#'     feature matrix where rows correspond to elements and columns
#'     correspond to variables (a single numeric variable can be
#'     passed as a vector). (2) An N x N matrix dissimilarity matrix;
#'     can be an object of class \code{dist} (e.g., returned by
#'     \code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
#'     where the entries of the upper and lower triangular matrix
#'     represent pairwise dissimilarities.
#' @param clusters A vector representing (anti)clusters (e.g.,
#'     returned by \code{\link{anticlustering}}).
#'     
#' @details 
#' 
#' The dispersion is the minimum distance between two elements within
#' the same cluster. When the input \code{x} is a feature matrix, the
#' Euclidean distance is used as the distance unit. Maximizing the
#' dispersion maximizes the minimum heterogeneity within clusters and
#' is an anticlustering task.
#' 
#' @examples 
#' 
#' N <- 50 # number of elements
#' M <- 2  # number of variables per element
#' K <- 5  # number of clusters
#' random_data <- matrix(rnorm(N * M), ncol = M)
#' random_clusters <- sample(rep_len(1:K, N))
#' dispersion_objective(random_data, random_clusters)
#' 
#' # Maximize the dispersion 
#' optimized_clusters <- anticlustering(
#'   random_data,
#'   K = random_clusters, 
#'   objective = dispersion_objective
#' )
#' dispersion_objective(random_data, optimized_clusters)
#' 
#' @export
#'
#' @references
#'
#' Brusco, M. J., Cradit, J. D., & Steinley, D. (2020). Combining
#' diversity and dispersion criteria for anticlustering: A bicriterion
#' approach. British Journal of Mathematical and Statistical
#' Psychology, 73, 275-396. https://doi.org/10.1111/bmsp.12186

#' 

dispersion_objective <- function(x, clusters) {
  x <- convert_to_distances(x)
  validate_input(x, "x", objmode = "numeric")
  validate_input(clusters, "clusters", len = nrow(x), not_na = TRUE)
  dispersion_objective_(clusters, x)
}

dispersion_objective_ <- function(clusters, distances) {
  objectives <- sapply(
    sort(unique(clusters)), 
    # min() gives warning for clusters of length 0 (but its behaviour is as intended)
    function(x) suppressWarnings(min(as.dist(distances[clusters == x, clusters == x])))
  )
  min(objectives)
}

#' Solve anticlustering using the modified exchange method
#'
#' @param data the data -- a N x N dissimilarity matrix or a N x M
#'     table of item features
#' @param K The number of cluster or an initial cluster assignment
#' @param obj_function the objective function. Takes as first argument
#'     a cluster assignment and as second argument the data set `data`
#'     (`data` is matrix, no `data.frame`).
#' @param categories A vector representing preclustering/categorical constraints
#'
#' @return The anticluster assignment
#'
#' @noRd
#'
#'

exchange_method <- function(data, K, obj_function, categories) {

  clusters <- initialize_clusters(NROW(data), K, categories)
  N <- nrow(data)
  best_total <- obj_function(data, clusters)
  for (i in 1:N) {
    # cluster of current item
    exchange_partners <- get_exchange_partners(clusters, i, categories)
    ## Do not use this item if there are zero exchange partners
    if (length(exchange_partners) == 0) {
      next
    }
    # container to store objectives associated with each exchange of item i:
    comparison_objectives <- rep(NA, length(exchange_partners))
    for (j in seq_along(exchange_partners)) {
      ## Swap item i with all legal exchange partners and check out objective
      comparison_objectives[j] <- update_objective_generic(
        data,
        clusters,
        i,
        exchange_partners[j],
        obj_function
      )
    }
    ## Do the swap if an improvement occured
    best_this_round <- max(comparison_objectives)
    if (best_this_round > best_total) {
      # Which element has to be swapped
      swap <- exchange_partners[comparison_objectives == best_this_round][1]
      # Swap the elements
      clusters <- cluster_swap(clusters, i, swap)
      # Update best solution
      best_total <- best_this_round
    }
  }
  clusters
}

# Update the objective value - generic version taking any objective function
# param data: the data (distances/features)
# param clusters: a vector of clusters
# param i, j: the to be swapped items (indexes)
# param obj_function: The function that computes the objective value
# return: The objective after a hypothetical swap
update_objective_generic <- function(data, clusters, i, j, obj_function) {
  tmp_clusters <- cluster_swap(clusters, i, j)
  obj_function(data, tmp_clusters)
}

# Swap two items and return new arranged clusters
#
# param clusters: a vector of clusters
# param i, j: the to be swapped items (indexes of the items)
# return: the cluster vector after swapping items i and j
cluster_swap <- function(clusters, i, j) {
  group_i <- clusters[i]
  clusters[i] <- clusters[j]
  clusters[j] <- group_i
  clusters
}

# For a given item i, get all feasible exchange partners
#
# param clusters: a vector of clusters
# param i: the to be swapped item (index of the item)
# param categories: A vector of categories inducing constraints wrt
#                   which items are feasible exchange partners
# return: The indexes of the feasible exchange partners
get_exchange_partners <- function(clusters, i, categories) {
  N <- length(clusters)
  # are there categorical variables?
  if (!is.null(categories)) {
    # only exchange within the same group
    allowed_category <- categories == categories[i]
  } else {
    allowed_category <- rep(TRUE, N) # no constraint
  }

  ## Feasible exchange partners:
  # (a) are in different anticluster
  # (b) are in the same category/precluster
  exchange_partners <- (clusters != clusters[i]) & allowed_category
  (1:N)[exchange_partners]
}


# Initialize a cluster assignment.
# The initial assignment is either:
# (a) already passed via the K argument
# (b) a completely random assignment
# (c) a random assignment that balances the categories across anticlusters
#     (if categories are preclusters: each preclustered item is assigned
#      to a different anticlusters)
# param N: The number of cases
# param K: The number of clusters or an initial clustering assignment
# param categories: A vector of categories
# return: The initialized clusters
initialize_clusters <- function(N, K, categories) {
  if (length(K) == N) {
    K <- to_numeric(K)
    return(K) # K is already an anticluster assignment
  }
  ## Initial assignment based on categorical constraints
  ## (categorical constraints may be preclustering constraints)
  if (argument_exists(categories)) {
    return(categorical_sampling(categories, K))
  }
  ## Initial random assignment unrestricted:
  if (sum(K) == N) {
    return(sample(rep(1:length(K), K)))
  }
  sample(rep_len(1:K, length.out = N))
}

#' Ratings for 96 words
#'
#' A stimulus set that was used in experiments by Schaper, Kuhlmann and
#' Bayen (2019a; 2019b). The item pool consists of 96 German words. 
#' Each word represents an object that is either 
#' typically found in a bathroom or in a kitchen. 
#'
#' @format A data frame with 96 rows and 7 variables
#' \describe{
#'   \item{item}{The name of an object (in German)}
#'   \item{room}{The room in which the item is typically found; can be
#'       'kitchen' or 'bathroom'}
#'   \item{rating_consistent}{How expected would it
#'       be to find the \code{item} in the typical \code{room}}
#'   \item{rating_inconsistent}{How expected would it
#'       be to find the \code{item} in the atypical \code{room}}
#'   \item{syllables}{The number of syllables in the object name}
#'   \item{frequency}{A value indicating the relative frequency of the
#'       object name in German language (lower values indicate higher
#'       frequency)}
#'   \item{list}{Represents the set affiliation of the \code{item} as
#'       realized in experiments by Schaper et al.}
#' }
#'
#' @source
#' Courteously provided by Marie Lusia Schaper and Ute Bayen.
#'
#' @examples
#'
#' head(schaper2019)
#' features <- schaper2019[, 3:6]
#'
#' # Optimize the variance criterion
#' # (tends to maximize similarity in feature means)
#' anticlusters <- anticlustering(
#'   features,
#'   K = 3,
#'   objective = "variance",
#'   categories = schaper2019$room,
#'   method = "exchange"
#' )
#'
#' # Means are quite similar across sets:
#' by(features, anticlusters, function(x) round(colMeans(x), 2))
#' # Check differences in standard deviations:
#' by(features, anticlusters, function(x) round(apply(x, 2, sd), 2))
#' # Room is balanced between the three sets:
#' table(Room = schaper2019$room, Set = anticlusters)
#'
#' # Maximize the diversity criterion
#' ac_dist <- anticlustering(
#'   features,
#'   K = 3,
#'   objective = "diversity",
#'   categories = schaper2019$room,
#'   method = "exchange"
#' )
#' # With the distance criterion, means tend to be less similar,
#' # but standard deviations tend to be more similar:
#' by(features, ac_dist, function(x) round(colMeans(x), 2))
#' by(features, ac_dist, function(x) round(apply(x, 2, sd), 2))
#'
#'
#' @references
#'
#' Schaper, M. L., Kuhlmann, B. G., & Bayen, U. J. (2019a). Metacognitive expectancy
#' effects in source monitoring: Beliefs, in-the-moment experiences, or both? Journal
#' of Memory and Language, 107, 95–110. https://doi.org/10.1016/j.jml.2019.03.009
#'
#' Schaper, M. L., Kuhlmann, B. G., & Bayen, U. J. (2019b). Metamemory expectancy illusion
#' and schema-consistent guessing in source monitoring. Journal of Experimental
#' Psychology: Learning, Memory, and Cognition, 45, 470.
#' https://doi.org/10.1037/xlm0000602
#'
#'
"schaper2019"

#' Use cannot-link constraints with anticlustering
#' 
#' @param x An N x M feature matrix or N x N distance matrix
#' @param init_clusters vector of initial clusters or matrix with N columns where each row is an initial partition
#'   (usually a structure returned by optimal_dispersion() in the slot $groups)
#' @param cannot_link A 2 column matrix containing the indices of elements
#'     that must not be assigned to the same anticluster. (e.g., a matrix returned by
#'     optimal_dispersion() in the slot $edges)
#' @param objective "diversity", "average-diversity", "variance" or "kplus" (not dispersion!)
#' @param method "local-maximum" or "exchange", "brusco"
#' 
#' 
#' @noRd
cannot_link_anticlustering <- function(x, init_clusters, cannot_link, objective, method) {
  
  cannot_link <- as.matrix(cannot_link)

  if (ncol(cannot_link) == 1) { # cannot_link is ID/grouping vector
    cannot_link <- to_numeric(c(cannot_link))
    cannot_link <- get_partners(cannot_link)
  }

  if (objective == "kplus") {
    x <- kplus_moment_variables(x, 2)
    objective <- "variance"
  }
  
  if (objective == "variance") {
    x <- convert_to_distances(x)^2
  } else if (grepl("diversity", objective)) {
    x <- convert_to_distances(x)
  } 
  frequencies <- table(init_clusters)
  if (objective == "variance") {
    objective <- "average-diversity"
  }
  stopifnot(grepl("diversity", objective)) # must be diversity or average-diversity here

  ## special case of only one init partition...
  init_clusters <- as.matrix(init_clusters)
  if (ncol(init_clusters) == 1) {
    K <- c(init_clusters)
    init_clusters_bils <- K # unfortunately, these functions have different interfaces...
    init_clusters <- NULL
  } else {
    K <- init_clusters[1, ]
    init_clusters_bils <- init_clusters
    init_clusters <- init_clusters - 1 # -1 for C
  }
  
  if (method == "brusco") {
    return(BILS_CANNOT_LINK(x, init_clusters_bils, cannot_link, objective))
  }
  
  # For LCW: Set cannot-link distances to large negative value so they cannot be linked
  x[cleanup_cannot_link_indices(cannot_link)] <- -(sum(x) + 1)

  c_anticlustering(
    x, 
    K = K, 
    categories = NULL, 
    objective = objective, 
    local_maximum = ifelse(method == "local-maximum", TRUE, FALSE),
    exchange_partners = NULL,
    init_partitions = init_clusters
  )
}


# Wrapper for BILS method that preserves optimal dispersion and potentially uses multiple initial partitions
BILS_CANNOT_LINK <- function(distances, init_clusters, cannot_link, objective) {
  N <- nrow(distances)
  multiple_partitions_as_input <- is.matrix(init_clusters)
  cl_matrix <- matrix(1, ncol = N, nrow = N)
  cl_matrix[cleanup_cannot_link_indices(cannot_link)] <- -1
  bicriterion_anticlustering(
    distances, 
    K = if (multiple_partitions_as_input) init_clusters[1, ] else init_clusters,
    R = if (multiple_partitions_as_input) rep(ceiling(nrow(init_clusters)/2), 2) else c(1, 1),
    init_partitions = if (multiple_partitions_as_input) init_clusters[1:ceiling(nrow(init_clusters)/2),] else NULL,
    dispersion_distances = cl_matrix,
    average_diversity = objective == "average-diversity",
    return = "best-dispersion"
  )
}

cleanup_cannot_link_indices <- function(cannot_link) {
  cannot_link <- rbind(cannot_link, t(apply(cannot_link, 1, rev))) # use (1, 2) and (2, 1)
  cannot_link[!duplicated(cannot_link), ] # but do not use (1, 2), (2, 1), (1, 2) and (2, 1)
} 

# helper funciton to convert the category labels to pairwise constraints
get_partners <- function(x) {
  n_clusters <- length(unique(x))
  indices <- lapply(1:n_clusters, function(i) get_partners_(x, i))
  indices <- do.call(rbind, indices)
  indices[indices[, 1] != indices[, 2], ]
}

get_partners_ <- function(x, i) {
  if (sum(x==i) == 1) {
    return(NULL)
  }
  t(combn(which(x == i), 2))
}

#' Maximize dispersion for K groups
#' 
#' @param x The data input. Can be one of two structures: (1) A
#'     feature matrix where rows correspond to elements and columns
#'     correspond to variables (a single numeric variable can be
#'     passed as a vector). (2) An N x N matrix dissimilarity matrix;
#'     can be an object of class \code{dist} (e.g., returned by
#'     \code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
#'     where the entries of the upper and lower triangular matrix
#'     represent pairwise dissimilarities.
#' @param K The number of groups or a vector describing the size of
#'     each group.
#' @param solver Optional argument; currently supports "lpSolve", 
#'     "glpk", "symphony", and "gurobi". See \code{\link{optimal_anticlustering}}.
#' @param max_dispersion_considered Optional argument used for early
#'     stopping. If the dispersion found is equal to or exceeds this
#'     value, a solution having the previous best dispersion is
#'     returned.
#' @param min_dispersion_considered Optional argument used for
#'     speeding up the algorithm computation.  If passed, the
#'     dispersion is optimized starting from this value instead the
#'     global minimum distance.
#' @param npartitions The number of groupings that are returned, each
#'     having an optimal dispersion value (defaults to 1).
#' @param time_limit Time limit in seconds, given to the solver.
#'    Default is there is no time limit.
#'
#' @export
#' 
#' @return A list with four elements:  
#'    \item{dispersion}{The optimal dispersion}
#'    \item{groups}{An assignment of elements to groups (vector)}
#'    \item{edges}{A matrix of 2 columns. Each row contains the indices of 
#'    elements that had to be investigated to find the dispersion (i.e., each pair
#'    of elements cannot be part of the same group in order to achieve maximum 
#'    dispersion).}
#'    \item{dispersions_considered}{All distances that were tested until the dispersion was found.}
#' 
#' @details
#'
#'   The dispersion is the minimum distance between two elements
#'   within the same group. This function implements an optimal method
#'   to maximize the dispersion. If the data input \code{x} is a
#'   feature matrix and not a dissimilarity matrix, the pairwise
#'   Euclidean distance is used. It uses the algorithm OptDispF presented 
#'   Papenberg, Breuer, et al. (2025). It was originally developed as part of 
#'   Max Diekhoff's Bachelor thesis at the Computer Science Department
#'   at Heinrich Heine University Düsseldorf.
#'
#'   To find out which items are not allowed to be grouped in the same
#'   cluster for maximum dispersion, the algorithm sequentially builds
#'   instances of a graph coloring problem, using an integer linear
#'   programming (ILP) representation. It is possible to specify the ILP solver via the argument
#'   \code{solver} (See \code{\link{optimal_anticlustering}} for more
#'   information on this argument). Optimally solving the maximum
#'   dispersion problem is NP-hard for K > 2 and therefore
#'   computationally infeasible for larger data sets. For K = 3 and K
#'   = 4, it seems that this approach scales up to several 100
#'   elements, or even > 1000 for K = 3 (at least when using the
#'   Symphony or gurobi solver).  For larger data sets, use the heuristic
#'   approaches in \code{\link{anticlustering}} or
#'   \code{\link{bicriterion_anticlustering}}. However, note that for
#'   K = 2, the optimal approach is usually much faster than the
#'   heuristics.
#'   
#'   In the output, the element \code{edges} defines which elements
#'   must be in separate clusters in order to achieve maximum
#'   dispersion. All elements not listed here can be changed
#'   arbitrarily between clusters without reducing the dispersion.  If
#'   the maximum possible dispersion corresponds to the minimum
#'   dispersion in the data set, the output elements \code{edges} and
#'   \code{groups} are set to \code{NULL} because all possible
#'   groupings have the same value of dispersion.  In this case the
#'   output element \code{dispersions_considered} has length 1.
#'   
#'   If a \code{time_limit} is set and the function cannot find in the optimal
#'   dispersion in the given time, it will throw an error.
#'   
#'
#' @note If the SYMPHONY solver is used, an unfortunate "message" is
#'     printed to the console when this function terminates:
#' 
#' sym_get_col_solution(): No solution has been stored!
#' 
#' This message is no reason to worry and instead is a direct result
#' of the algorithm finding the optimal value for the dispersion.
#' Unfortunately, this message is generated in the C code underlying
#' the SYMPHONY library (via the printing function \code{printf}),
#' which cannot be prevented in R.
#'
#' @author
#' 
#' Max Diekhoff
#' 
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#' 
#' @seealso \code{\link{dispersion_objective}} \code{\link{anticlustering}}
#' 
#' @references 
#' 
#' Diekhoff (2023). Maximizing dispersion for anticlustering. Retrieved from 
#' https://www.cs.hhu.de/fileadmin/redaktion/Fakultaeten/Mathematisch-Naturwissenschaftliche_Fakultaet/Informatik/Algorithmische_Bioinformatik/Bachelor-_Masterarbeiten/2831963_ba_ifo_AbschlArbeit_klau_mapap102_madie120_20230203_1815.pdf  
#' 
#' Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. (2025). 
#' Extending the Bicriterion Approach for Anticlustering: Exact and Hybrid 
#' Approaches. Psychometrika. Advance online publication. https://doi.org/10.1017/psy.2025.10052
#' 
#' @examples
#' 
#' N <- 30
#' M <- 5
#' K <- 3
#' data <- matrix(rnorm(N*M), ncol = M)
#' distances <- dist(data)
#' 
#' opt <- optimal_dispersion(distances, K = K)
#' opt
#' 
#' # Compare to bicriterion heuristic:
#' groups_heuristic <- anticlustering(
#'   distances, 
#'   K = K,
#'   method = "brusco", 
#'   objective = "dispersion", 
#'   repetitions = 100
#' )
#' c(
#'   OPT = dispersion_objective(distances, opt$groups),
#'   HEURISTIC = dispersion_objective(distances, groups_heuristic)
#' )
#' 
#' # Different group sizes are possible:
#' table(optimal_dispersion(distances, K = c(15, 10, 5))$groups)
#' 
#' # Induce cannot-link constraints by maximizing the dispersion:
#' solvable <- matrix(1, ncol = 6, nrow = 6)
#' solvable[2, 1] <- -1
#' solvable[3, 1] <- -1
#' solvable[4, 1] <- -1
#' solvable <- as.dist(solvable)
#' solvable
#' 
#' # An optimal solution has to put item 1 in a different group than 
#' # items 2, 3 and 4 -> this is possible for K = 2
#' optimal_dispersion(solvable, K = 2)$groups
#' 
#' # It no longer works when item 1 can also not be linked with item 5:
#' # (check out output!)
#' unsolvable <- as.matrix(solvable)
#' unsolvable[5, 1] <- -1
#' unsolvable <- as.dist(unsolvable)
#' unsolvable
#' optimal_dispersion(unsolvable, K = 2)
#' # But:
#' optimal_dispersion(unsolvable, K = c(2, 4)) # group sizes, not number of groups
#' 


optimal_dispersion <- function(
    x, K, 
    solver = NULL, 
    max_dispersion_considered = NULL, 
    min_dispersion_considered = NULL,
    npartitions = 1,
    time_limit = NULL) {
  
  validate_input_optimal_anticlustering(x, K, "dispersion", solver, time_limit)
  
  if (!argument_exists(solver)) {
    solver <- find_ilp_solver()
  }

  validate_input(npartitions, "npartitions", objmode = "numeric", must_be_integer = TRUE, not_na = TRUE, not_function = TRUE, greater_than = 0)

  if (is.null(max_dispersion_considered)) {
    max_dispersion_considered <- Inf
  } else {
    validate_input(max_dispersion_considered, "max_dispersion_considered", 
                   objmode = "numeric", len = 1, not_na = TRUE, not_function = TRUE)
  }
  
  distances <- convert_to_distances(x)
  diag(distances) <- Inf
  N <- nrow(distances)
  
  if (argument_exists(min_dispersion_considered)) {
    validate_input(min_dispersion_considered, "min_dispersion_considered", 
                   objmode = "numeric", len = 1, not_na = TRUE, not_function = TRUE)
    sorted_unique_distances <- sort(unique(distances))
    dispersion <- sorted_unique_distances[which(sorted_unique_distances == min_dispersion_considered)[1] - 1]
  }
  
  # `target_groups` is primarily needed for unequal sized groups
  target_groups <- sort(table(initialize_clusters(N, K, NULL)), decreasing = TRUE)
  K <- length(target_groups)
  dispersion_found <- FALSE
  # Data frame to keep track of previous nearest neighbours (init as NULL)
  all_nns <- NULL

  # Placeholders to store data needed for retrieving the anticlusters
  last_solution <- NULL
  all_nns_last <- NULL
  all_nns_reordered_last <- NULL
  dispersions_considered <- NULL
  time_limit_exceeded <- FALSE
  counter <- 1
  MINIMUM_DISTANCE <- min(distances)
  start <- Sys.time()
  while (!dispersion_found) {
    if (is.null(min_dispersion_considered) || counter > 1) {
      dispersion <- min(distances)
    }
    if (dispersion >= max_dispersion_considered) {
      break
    }
    ids_of_nearest_neighbours <- which(distances <= dispersion, arr.ind = TRUE)
    all_nns <- rbind(all_nns, remove_redundant_edges(ids_of_nearest_neighbours))
    # Reorder edge labels so that they start from 1 to C, where C is the number
    # of relevant edges (Better for creating K-coloring ILP).
    all_nns_reordered <- reorder_edges(all_nns)
    ilp <- k_coloring_ilp(all_nns_reordered, N, K, target_groups)
    solution <- solve_ilp(ilp, objective = "min", solver = solver, time_limit = time_limit)
    dispersion_found <- solution$status != 0
    if (argument_exists(time_limit) && (as.numeric(difftime(Sys.time(), start, units = "s")) > time_limit)) {
      stop("Could not find the optimal dispersion in the given time limit.")
    }
    if (!dispersion_found){
      last_solution <- solution
      all_nns_last <- all_nns
      all_nns_reordered_last <- all_nns_reordered
      dispersions_considered <- c(dispersions_considered, dispersion)
      
    }
    counter <- counter + 1
    # Take out distances that have been investigated to proceed
    distances[ids_of_nearest_neighbours] <- Inf 
  }
  if (dispersion == MINIMUM_DISTANCE) { # no improvement for dispersion is possible; first test fails
    return(
      list(
        dispersion = MINIMUM_DISTANCE, 
        groups = NULL,
        edges = NULL, 
        dispersions_considered = MINIMUM_DISTANCE
      )
    )
  }
  # Calculate anticlusters from the last iteration with a K-coloring
  groups <- t(sapply(
    1:npartitions,
    repeat_grouping,
    result_value = last_solution$obj,
    result_x = last_solution$x,
    all_nns = all_nns_last,
    all_nns_reordered = all_nns_reordered_last,
    N = N,
    K = K, 
    target_groups = target_groups
  ))
  if (npartitions == 1) {
    groups <- c(groups)
  } else {
    if (any(duplicated(groups))) { # warn if there are duplicate partitions
      warning("Some of the returned partitions are duplicates (i.e. argument 'npartitions' was > 1).")
    }
  }
  
  group_fixated <- graph_coloring_to_group_vector(
    all_nns_reordered = all_nns_reordered_last, 
    result_x = last_solution$x, 
    K = K, 
    all_nns = all_nns_last, 
    N = N
  )
  
  return(
    list(
      dispersion = dispersion, 
      groups = groups,
      groups_fixated = group_fixated,
      edges = unname(all_nns_last), # rownames can be quite ugly here
      dispersions_considered = c(dispersions_considered, dispersion)
    )
  )
}

k_coloring_ilp <- function(all_nns_reordered, N, K, target_groups) {
  # Initialize some constant variables
  nr_of_nodes <- max(all_nns_reordered)
  nr_of_edges <- nrow(all_nns_reordered)
  nr_of_x_variables <- nr_of_nodes * K
  equality_signs <- equality_identifiers()
  
  # Construct ILP constraint matrix
  constraints <- sparse_constraints_dispersion(all_nns_reordered, nr_of_nodes, nr_of_edges, K)
  colnames(constraints) <- constraint_names(nr_of_nodes, K)
  
  # Directions of the constraints:
  equalities <- c(rep(equality_signs$e, nr_of_nodes),
                  rep(equality_signs$l, (nr_of_edges +1) * K))
  
  # Right-hand-side of ILP
  rhs <- c(rep(1, nr_of_nodes), rep(0, nr_of_edges * K), target_groups)
  
  # Objective function of the ILP
  obj_function <- c(rep(1, K), rep(0, nr_of_x_variables))
  
  # Give names to all objects for inspection purposes
  names(obj_function) <- colnames(constraints)
  
  instance <- list()
  instance$constraints  <- constraints
  instance$equalities   <- equalities
  instance$rhs          <- rhs
  instance$obj_function <- obj_function
  
  return(instance)

}

# Construct a sparse matrix representing the ILP constraints
#
# @param all_nns_reordered A data frame where every row defines an edge in the graph
# @param nr_of_nodes The number of nodes in the graph
# @param K The maximum number of colors to be used
#
# @return A sparse matrix representing the left-hand side of the ILP (A in Ax ~ b)
#
sparse_constraints_dispersion <- function(all_nns_reordered, nr_of_nodes, nr_of_edges, K) {
  node <- node_constraints(nr_of_nodes, K)
  edge <- edge_constraints(all_nns_reordered, nr_of_nodes, nr_of_edges, K)
  group <- group_constraints_dispersion(nr_of_nodes, nr_of_edges, K)
  Matrix::sparseMatrix(c(node$i, edge$i, group$i), c(node$j, edge$j, group$j), x = c(node$x, edge$x, group$x))
}

# Indices for sparse matrix representation of node constraints
node_constraints <- function(nr_of_nodes, K) {
  col_indices <- rep(1:nr_of_nodes, each=K)
  row_start <- K+1
  row_end <- (nr_of_nodes+1)*K
  row_indices <- row_start:row_end
  xes <- rep(1, nr_of_nodes * K)
  list(i = col_indices, j = row_indices, x = xes)
}

# Indices for sparse matrix representation of edge constraints
edge_constraints <- function(all_nns, nr_of_nodes, nr_of_edges, K) {
  col_start <- nr_of_nodes+1
  col_end <- nr_of_nodes+(nr_of_edges * K)
  col_indices <- rep(col_start:col_end, each=3)
  row_indices <- c()
  for(i in 1:nr_of_edges){
    u <- all_nns[i,1]
    v <- all_nns[i,2]
    current_rows <- c(rep(c(0, u * K, v * K), K))
    row_indices <- c(row_indices, current_rows + rep(1:K, each=3))
  }
  xes <- rep(c(-1,1,1), nr_of_edges * K)
  list(i = col_indices, j = row_indices, x = xes)
}

# Indices for sparse matrix representation of group size constraints
group_constraints_dispersion <- function(nr_of_nodes, nr_of_edges, K)  {
  col_start <- nr_of_nodes+(nr_of_edges * K)+1
  col_end <- nr_of_nodes+(nr_of_edges * K)+K
  col_indices <- rep(col_start:col_end, each=nr_of_nodes)
  row_indices <- (rep(1:nr_of_nodes) * K) + rep(1:K, each=nr_of_nodes) 
  xes <- rep(1, nr_of_nodes * K)
  list(i = col_indices, j = row_indices, x = xes)
}


# Return variable names to have meaningful column names for the objective function and constraint matrix
#
# @param nr_of_nodes The number of nodes in the graph
# @param K The maximum number of colors to be used
#
# @return A vector with w_l (l=1,...,K) and x_i_j (i=1,...,nr_of_nodes, j=1,...,K)
#
constraint_names <- function(nr_of_nodes, K) {
  w_l <- rep("w_1", 1)
  for(l in 2:K){
    w_l <- c(w_l, paste0("w_", l))
  }
  
  x_j_i <- expand.grid(1:K, 1:nr_of_nodes)
  colnames(x_j_i) <- c("j", "i")
  x_j_i$variables <- paste0("x_",x_j_i$i,"_",x_j_i$j)
  
  return(c(w_l, x_j_i$variables))
}

# dummy function for calling groups_from_k_coloring_mapping() several times using sapply
repeat_grouping <- function(X, result_value, result_x, all_nns, all_nns_reordered, N, K, target_groups) {
  groups_from_k_coloring_mapping(result_value, result_x, all_nns, all_nns_reordered, N, K, target_groups)
}

# Restore a grouping from the solved ilp
groups_from_k_coloring_mapping <- function(result_value, result_x, all_nns, all_nns_reordered, N, K, target_groups) {
  groups_new <- graph_coloring_to_group_vector(all_nns_reordered, result_x, K, all_nns, N)
  # now we have the original indices, assign remaining elements randomly to groups
  # how many are not yet assigned
  add_unassigned_elements(target_groups, groups_new, N, K)
}

# After initial assignment, fill the rest randomly
add_unassigned_elements <- function(target_groups, init, N, K) {
  if (sum(!is.na(init)) == N) {
    return(init)  # groups are already full, no unassigned elements!
  }
  K <- length(target_groups)
  table_assigned <- table(init)
  # assign elements that have no group (unfortunately, this "simple" task is quite difficult in general)
  if (length(table_assigned) != length(target_groups)) {
    table_assigned <- data.frame(K = 1:K, size = 0)
    df <- as.data.frame(table(init, useNA = "always"))
    together <- merge(table_assigned, df, by.x = "K", by.y = "init", all = TRUE)
    together$K <- as.numeric(together$K)
    together <- together[order(together$K), ]
    table_assigned <- ifelse(is.na(together$Freq), 0, together$Freq)[-(K+1)]
  }
  freq_not_assigned <- target_groups - table_assigned
  init[is.na(init)] <- sample_(rep(1:K, freq_not_assigned))
  stopifnot(all(table(init) == target_groups))
  # now sort labels by group size (so that each time this function is called, we get the same output of table())
  new_labels <- order(table(init), decreasing = TRUE)
  as.numeric(as.character(factor(init, levels = 1:K, labels = new_labels)))
}


graph_coloring_to_group_vector <- function(all_nns_reordered, result_x, K, all_nns, N) {
  # Retrieve assigned colors from the x_v,i ILP variables
  mapping <- rep(NA, max(all_nns_reordered))
  matrix <- matrix(result_x[-(1:K)], nrow=K)
  # sort labeling by frequency of occurrence
  matrix <- matrix[order(rowSums(matrix), decreasing = TRUE), ]
  for(l in 1:ncol(matrix)){
    mapping[l] <- which.max(matrix[,l]) # mapping is the groups, now retrieve original indices
  }
  
  df <- data.frame(
    mapping = mapping[c(all_nns_reordered)],
    tmp_id = c(all_nns_reordered),
    id = c(all_nns)
  )
  # create vector with groupings
  groups_new <- rep(NA, N)
  groups_new[df$id] <- df$mapping
  groups_new
}

# Function that takes an edge list with any integer indices and remaps
# these indices to 1, ..., C (where C is the number of nodes) while 
# preserving the original order of indices.
reorder_edges <- function(edgelist) {
  dims <- dim(edgelist)
  edgelist <- as.numeric(as.factor(edgelist))
  dim(edgelist) <- dims
  edgelist
}

remove_redundant_edges <- function(df) {
  df <- t(apply(df, 1, sort))
  df[!duplicated(df), ]
}

# Function to solve optimal cannot_link constraints, used for the argument 
# cannot_link in anticlustering().
# First function only returns a grouping for all elements that are fixated via the constraints.
# Second function returns full groupings 
optimal_cannot_link_reduced <- function(N, K, target_groups, cannot_link, repetitions) {
  all_nns_reordered <- reorder_edges(cannot_link)
  ilp <- k_coloring_ilp(all_nns_reordered, N, K, target_groups)
  # select solver: gurobi > symphony > lpSolve > Glpk
  if (requireNamespace("gurobi", quietly = TRUE)) {
    solver <- "gurobi"
  } else if (requireNamespace("Rsymphony", quietly = TRUE)) {
    solver <- "symphony"
  } else {
    solver <- find_ilp_solver()
  }
  solution <- solve_ilp(ilp, solver = solver)
  if (solution$status != 0) {
    stop("The cannot-link constraints cannot be fulfilled.")
  }
  groups_fixated <- graph_coloring_to_group_vector(all_nns_reordered, solution$x, K, cannot_link, N)
  groups_fixated
}

optimal_cannot_link <- function(N, K, target_groups, cannot_link, repetitions) {
  repetitions <- ifelse(is.null(repetitions), 1, repetitions)
  groups_fixated <- optimal_cannot_link_reduced(N, K, target_groups, cannot_link, repetitions)
  if (repetitions > 1) {
    groups <- t(replicate(repetitions, add_unassigned_elements(target_groups, groups_fixated, N, K)))
  } else {
    groups <- add_unassigned_elements(target_groups, groups_fixated, N, K)
  }
  groups
}

#' Merge several grouping variable into one
#'
#' @param categories A vector, data.frame or matrix that represents
#'     one or several categorical constraints.
#'
#' @return A vector representing the group membership (or the combination
#'     of group memberships) as one variable
#'
#' @noRd
#'

merge_into_one_variable <- function(categories) {
  if (is.null(categories)) {
    return(NULL)
  }
  categories <- data.frame(categories)
  categories <- factor(do.call(paste, as.list(categories)))
  # sort as numeric to get consistent return value
  order_cluster_vector(to_numeric(categories))
}

#' Converting any categorical input to integer
#' 
#' @param x a vector
#' 
#' @return A vector of \code{length(x)} where each element is integer
#' 
#' @examples 
#' iris$Species
#' to_numeric(iris$Species)
#' 
#' @noRd
#' 
to_numeric <- function(x) {
  as.numeric(factor(x))
}

# Order a clustering vector
# 
# For a clustering vector, ensure that the first cluster that occurs
# in the vector is 1, the next 2, etc ... until K
#
# param clusters: A clustering vector with elements 1, ..., K indicating cluster membership
# return: A clustering vector in order
order_cluster_vector <- function(clusters) {
  unique_clusters <- unique(clusters)
  # deal with NA
  unique_clusters <- unique_clusters[!is.na(unique_clusters)]
  K <- length(unique_clusters)
  clusters <- factor(clusters, levels = unique_clusters, labels = 1:K)
  as.numeric(clusters)
}

library("anticlust")

# Matching function behaves correctly with `p` argument
# generate some random data
m <- sample(1:4, size = 1)
n <- sample(10:100, size = 1)
data <- matrix(rnorm(n * m), ncol = m)
p <- sample(2:5, size = 1)

# feature input
# test that matches are of size p
matches <- matching(data, p = p)
expect_true(all(table(matches) == p))
# non-fitting elements have NA
expect_equal(sum(is.na(matches)), n %% p)
# are all matches objective sorted by distance objective?
objectives <- sapply(
  1:max(matches, na.rm = TRUE), 
  function(x) sum(dist(data[!is.na(matches) & matches == x, ]))
)
expect_true(!is.unsorted(objectives))

# repeat the above for distance input
data <- as.matrix(dist(data))
matches <- matching(data, p = p)
expect_true(all(table(matches) == p))
expect_equal(sum(is.na(matches)), n %% p)
objectives <- sapply(
  1:max(matches, na.rm = TRUE), 
  function(x) sum(as.dist(data[!is.na(matches) & matches == x, !is.na(matches) & matches == x]))
)
expect_true(!is.unsorted(objectives))

# Matching function behaves correctly with `match_between` argument
# generate some random data
m <- sample(1:4, size = 1)
n <- sample(20:100, size = 1)
data <- matrix(rnorm(n * m), ncol = m)
p <- sample(2:5, size = 1)
groups <- sample(1:p, size = n, replace = TRUE)
while (any(table(groups) < 2)) {
  groups <- sample(1:p, size = n, replace = TRUE)
}

# feature input
# test that matches are of size 
matches <- matching(data, match_between = groups)
expect_true(all(table(matches) == p))
# non-fitting elements have NA
n_matched <- (min(table(groups)) * p) # how many elements were matched
expect_equal(sum(is.na(matches)), n - n_matched)
# are all matches objective sorted by distance objective?
objectives <- sapply(
  1:max(matches, na.rm = TRUE), 
  function(x) sum(dist(data[!is.na(matches) & matches == x, ]))
)
expect_true(!is.unsorted(objectives))

# repeat the above for distance input
data <- as.matrix(dist(data))
matches <- matching(data, match_between = groups)
expect_true(all(table(matches) == p))
# non-fitting elements have NA
n_matched <- (min(table(groups)) * p) # how many elements were matched
expect_equal(sum(is.na(matches)), n - n_matched)
objectives <- sapply(
  1:max(matches, na.rm = TRUE), 
  function(x) sum(as.dist(data[!is.na(matches) & matches == x, !is.na(matches) & matches == x]))
)
expect_true(!is.unsorted(objectives))


# Matching function behaves correctly with `match_within` argument
# generate some random data
m <- sample(1:4, size = 1)
n <- sample(40:100, size = 1)
data <- matrix(rnorm(n * m), ncol = m)
p <- sample(2:5, size = 1)
n_groups <- sample(2:5, size = 1)
groups <- sample(1:n_groups, size = n, replace = TRUE)
while (any(table(groups) < p) || length(unique(groups)) != n_groups) {
  groups <- sample(1:p, size = n, replace = TRUE)
}

# feature input
# test that matches are of size p
matches <- matching(data, p = p, match_within = groups)
expect_true(all(table(matches) == p))
# test that all matches are within a category
tab <- table(matches, groups)
expect_true(all(apply(tab, 1, function(x) sum(x == p)) == 1))
expect_true(all(apply(tab, 1, function(x) sum(x == 0)) == (n_groups - 1)))
# non-fitting elements have NA
not_matched <- sum(table(groups) %% p) # how many elements were matched
expect_equal(sum(is.na(matches)), not_matched)
# are all matches objective sorted by distance objective?
objectives <- sapply(
  1:max(matches, na.rm = TRUE), 
  function(x) sum(dist(data[!is.na(matches) & matches == x, ]))
)
expect_true(!is.unsorted(objectives))

# repeat the above for distance input
# test that matches are of size p
matches <- matching(dist(data), p = p, match_within = groups)
expect_true(all(table(matches) == p))
tab <- table(matches, groups)
expect_true(all(apply(tab, 1, function(x) sum(x == p)) == 1))
expect_true(all(apply(tab, 1, function(x) sum(x == 0)) == (n_groups - 1)))
not_matched <- sum(table(groups) %% p) 
expect_equal(sum(is.na(matches)), not_matched)
objectives <- sapply(
  1:max(matches, na.rm = TRUE), 
  function(x) sum(dist(data[!is.na(matches) & matches == x, ]))
)
expect_true(!is.unsorted(objectives))

# Matching behaves correctly when combining `match_within` and `match_between`
# generate some random data
p <- sample(2:5, size = 1)
tab <- p - 1
while (any(tab < p)) {
  m <- sample(1:4, size = 1)
  n <- sample(20:200, size = 1)
  data <- matrix(rnorm(n * m), ncol = m)
  n_groups_within <- sample(2:5, size = 1)
  groups_between <- sample(1:p, size = n, replace = TRUE)
  groups_within <- sample(1:n_groups_within, size = n, replace = TRUE)
  tab <- table(groups_between, groups_within)
}

# feature input
# test that matches are of size p
matches <- matching(
  data, 
  p = p, 
  match_between = groups_between,
  match_within = groups_within
)
expect_true(all(table(matches) == p))
# predict number of matches
tab <- table(groups_within, groups_between)
# for each group_within, there are as many matches as the minimum groups_between
n_matches <- sum(apply(tab, 1, min))
expect_equal(n_matches, max(matches, na.rm = TRUE))
# all not-matched elements have NA
expect_equal(n - (n_matches * p), sum(is.na(matches)))
# check the balancing across all grouping variables
tab <- table(matches, groups_within, groups_between)
expect_true(all(tab %in% c(0, 1)))
expect_true(all(apply(tab, 3, rowSums) == 1))

# feature input
# repeat the same for distance input
matches <- matching(
  dist(data), 
  p = p, 
  match_between = groups_between,
  match_within = groups_within
)
expect_true(all(table(matches) == p))
# predict number of matches
tab <- table(groups_within, groups_between)
# for each group_within, there are as many matches as the minimum groups_between
n_matches <- sum(apply(tab, 1, min))
expect_equal(n_matches, max(matches, na.rm = TRUE))
# all not-matched elements have NA
expect_equal(n - (n_matches * p), sum(is.na(matches)))
# check the balancing across all grouping variables
tab <- table(matches, groups_within, groups_between)
expect_true(all(tab %in% c(0, 1)))
expect_true(all(apply(tab, 3, rowSums) == 1))


# this test is monster code; it tests that the first target element
# is selected as it should be and paired with its nearest neighbour,
# for different input. Fun fact: I made a lot of mistakes implementing
# this test, but the code itself was fine all the time ... 

# Algorithm matches the element it should match
# generate some random data
m <- sample(1:4, size = 1)
n <- sample(40:100, size = 1)
data <- matrix(rnorm(n * m), ncol = m)
p <- sample(2:5, size = 1)

for (most_extreme in c(TRUE, FALSE)) {
  if (most_extreme == TRUE) {
    FUN <- which.max
  } else {
    FUN <- which.min
  }
  for (dat in c("features", "distance")) {
    if (dat == "distances") {
      df <- dist(data)
      distances <- as.matrix(data)
    } else {
      df <- data
      distances <- as.matrix(dist(data))
    }
    # vary if I use p or matches_between as input
    for (j in c("p", "groups")) {
      if (j == "p") {
        matches <- matching(df, p = p, match_extreme_first = most_extreme)
        first_target <- FUN(anticlust:::distances_from_centroid(data))
        target_group <- FALSE
      } else {
        groups <- anticlust:::to_numeric(sample(1:p, size = n, replace = TRUE))
        while (any(table(groups) < 2) || length(unique(table(groups))) == 1) {
          groups <- sample(1:p, size = n, replace = TRUE)
        }
        groups <- anticlust:::merge_into_one_variable(groups)
        matches <- matching(df, match_between = groups, match_extreme_first = most_extreme)
        centroid_distances <- anticlust:::distances_from_centroid(data)
        smallest_group <- which.min(table(groups))
        # select index of element in smallest group based on distance to centroid
        target_group <- groups == smallest_group
        centroid_distances[!target_group] <- NA
        first_target <- FUN(centroid_distances)
      }
      target_distances <- distances[, first_target]
      target_distances[target_group] <- NA # only select from «other» groups
      closest <- min(target_distances[-first_target], na.rm = TRUE)
      closest_neighbours <- which(target_distances == closest)
      # ensure that at least one of the neighbours with minimum distance
      # is in the same match
      expect_true(closest_neighbours %in% which(matches == matches[first_target]))
    }
  }
}

#' C implementation of anticlustering
#' 
#' @param data An N x M data matrix or N x N distance matrix.
#' @param K The number of clusters or an initial assignment of the elements 
#'     to clusters.
#' @param categories A vector, data.frame or matrix representing one
#'     or several categorical constraints. 
#' 
#' @noRd
#' 
c_anticlustering <- function(data, K, categories = NULL, objective, exchange_partners = NULL, local_maximum = FALSE, init_partitions = NULL) {
  
  clusters <- initialize_clusters(NROW(data), K, categories)

  # Ensure that `clusters` is a vector of integers between 0 and K-1
  clusters <- to_numeric(clusters) - 1
  # As convenience arguments for C, also pass 
  # a. the number of clusters (`K`)
  # b. how often each cluster occurs (`frequencies`)
  # c. the number of rows (`N`) and columns (`M`) in `data`
  K <- length(unique(clusters))
  frequencies <- table(clusters)
  N <- NROW(data)
  M <- NCOL(data)

  # Case: initial partitions may be passed as a matrix (rows = partitions; cols = elements)
  if (!argument_exists(init_partitions)) {
    R <- 1
    use_init_partitions <- 0
    init_partitions <- 0
  } else {
    R <- nrow(init_partitions)
    if (min(init_partitions) > 0) {
      stop("You forgot 0 indexing in C again.")
    }
    use_init_partitions <- 1
  }

  if (argument_exists(categories)) {
    USE_CATEGORIES <- TRUE
    categories <- merge_into_one_variable(categories) - 1
    N_CATS <- length(unique(categories))
    CAT_frequencies <- table(categories)
  } else {
    USE_CATEGORIES <- FALSE
    categories <- 0
    N_CATS <- 0
    CAT_frequencies <- 0
  }
  
  # Call C implementation of anticlustering
  if (objective == "variance") {
    results <- .C(
      "kmeans_anticlustering", 
      as.double(data),
      as.integer(N),
      as.integer(M),
      as.integer(K),
      as.integer(frequencies),
      clusters = as.integer(clusters),
      as.integer(USE_CATEGORIES),
      as.integer(N_CATS),
      as.integer(CAT_frequencies),
      as.integer(categories),
      mem_error = as.integer(0),
      PACKAGE = "anticlust"
    )
  } else if (objective %in% c("diversity", "distance", "average-diversity")) {
    if (objective != "average-diversity") {
      frequencies <- rep_len(1, K)
    }

    results <- .C(
      "distance_anticlustering", 
      as.double(convert_to_distances(data)),
      as.integer(N),
      as.integer(K),
      as.integer(frequencies),
      clusters = as.integer(clusters),
      as.integer(USE_CATEGORIES),
      as.integer(N_CATS),
      as.integer(CAT_frequencies),
      as.integer(categories),
      as.integer(local_maximum),
      as.integer(R),
      as.integer(use_init_partitions),
      as.integer(t(init_partitions)),
      mem_error = as.integer(0),
      PACKAGE = "anticlust"
    )
  } else if (objective == "dispersion") {
    results <- .C(
      "dispersion_anticlustering", 
      as.double(convert_to_distances(data)),
      as.integer(N),
      as.integer(K),
      clusters = as.integer(clusters),
      as.integer(USE_CATEGORIES),
      as.integer(N_CATS),
      as.integer(CAT_frequencies),
      as.integer(categories),
      mem_error = as.integer(0),
      PACKAGE = "anticlust"
    )
  } else if (objective == "fast-kmeans") {
    results <- .C(
      "fast_kmeans_anticlustering",
      as.double(data),
      as.integer(N),
      as.integer(M),
      as.integer(K),
      as.integer(frequencies),
      clusters = as.integer(clusters),
      as.integer(exchange_partners),
      as.integer(nrow(exchange_partners)),
      PACKAGE = "anticlust"
    )
    results[["mem_error"]] <- 0
  }

  if (results[["mem_error"]] == 1) {
    stop("Could not allocate enough memory.")
  }
  results[["clusters"]] + 1
}

#' Fast anticlustering
#'
#' Increasing the speed of (k-means / k-plus) anticlustering by (1) 
#' conducting fewer exchanges during the optimization and (2) using an alternative
#' formulation of the k-means objective. Makes anticlustering applicable to 
#' quite large data sets.
#'
#' @param x A numeric vector, matrix or data.frame of data points.
#'     Rows correspond to elements and columns correspond to
#'     features. A vector represents a single numeric feature.
#' @param K How many anticlusters should be created. Alternatively:
#'     (a) A vector describing the size of each group, or (b) a vector
#'     of length \code{nrow(x)} describing how elements are assigned
#'     to anticlusters before the optimization starts.
#' @param k_neighbours The number of nearest neighbours that serve as
#'     exchange partner for each element. See details.
#' @param categories A vector, data.frame or matrix representing one
#'     or several categorical constraints.
#' @param exchange_partners Optional argument. A list of length
#'     \code{NROW(x)} specifying for each element the indices of the
#'     elements that serve as exchange partners. If used, this
#'     argument overrides the \code{k_neighbours} argument. See
#'     examples.
#'
#' @importFrom RANN nn2
#'
#' @seealso
#'
#' \code{\link{anticlustering}}
#' 
#' \code{\link{kplus_moment_variables}}
#' 
#' \code{\link{categories_to_binary}}
#'
#' \code{\link{variance_objective}}
#' 
#' \code{\link{generate_exchange_partners}}
#'
#' @export
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' 
#' @references
#' 
#' 
#' Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
#' data sets into equivalent parts. Psychological Methods, 26(2), 
#' 161–174. https://doi.org/10.1037/met0000301.
#' 
#' Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for 
#' Maximizing Between-Group Similarity. British Journal of Mathematical and 
#' Statistical Psychology, 77(1), 80--102. https://doi.org/10.1111/bmsp.12315
#' 
#' Späth, H. (1986). Anticlustering: Maximizing the variance criterion.
#' Control and Cybernetics, 15, 213-218.
#'
#'
#' @details
#'
#' This function was created to make anticlustering applicable to
#' large data sets (e.g., several 100,000 elements). It optimizes the
#' k-means objective because computing all pairwise distances as is
#' done when optimizing the "diversity" (i.e., the default in
#' \code{\link{anticlustering}}) is not feasible for very large data
#' sets (for about N > 20000 on my personal computer). Using
#' \code{fast_anticlustering} for k-plus anticlustering is also
#' possible by applying \code{\link{kplus_moment_variables}} on the
#' input (and possibly by using the argument \code{exchange_partners},
#' see Examples).
#' 
#' The function \code{fast_anticlustering} employs a speed-optimized
#' exchange method, which is basically equivalent to \code{method =
#' "exchange"} in \code{\link{anticlustering}}, but may reduce the number
#' of exchanges that are investigated for each input element. The number of 
#' exchange partners per element has to be set using the argument \code{k_neighbours}. By
#' default, it is set to \code{Inf}, meaning that all possible swaps are
#' tested. If \code{k_neighbours} is set differently (which is usually recommended when running 
#' this function), the default behaviour is to generate exchange partners using a
#' nearest neighbour search (using the function \code{\link[RANN]{nn2}}
#' from the \code{RANN} package). Using more exchange partners can improve the quality of
#' the results, but also increase run time. Note that for very large
#' data sets, anticlustering generally becomes "easier" (even a random
#' split may yield satisfactory results), so using few exchange
#' partners is usually not a problem. 
#' 
#' It is possible to specify custom exchange partners using the
#' argument \code{exchange_partners} instead of relying on the default
#' nearest neighbour search.  When using \code{exchange_partners}, it
#' is not necessary that each element has the same number of exchange
#' partners; this is why the argument \code{exchange_partners} has to
#' be a \code{list} instead of a \code{matrix} or
#' \code{data.frame}. Exchange partners can for example be generated
#' by \code{\link{generate_exchange_partners}} (see Examples), but a
#' custom list may also be used. Note that categorical constraints
#' induced via \code{categories} may not be respected during the
#' optimization if the \code{exchange_partners} argument allows
#' exchanges between members of different categories, so care must be
#' taken when combining the arguments \code{exchange_partners} and
#' \code{categories}.
#' 
#' In \code{anticlustering(..., objective = "variance")}, the run time
#' of computing the k-means objective is in O(M N), where N is the
#' total number of elements and M is the number of variables. This is
#' because the variance is computed as the sum of squared distances
#' between all data points and their cluster centers.  The function
#' \code{fast_anticlustering} uses a different - but equivalent -
#' formulation of the k-means objective, where the re-computation of
#' the objective only depends and M but no longer on N. In
#' particular, this variant of k-means anticlustering minimizes 
#' the weighted sum of squared distances between
#' cluster centroids and the overall data centroid; the distances
#' between all individual data points and their cluster center are not
#' computed (Späth, 1986). Using the different objective formulation 
#' reduces the run time by an
#' order of magnitude and makes k-means anticlustering applicable to
#' very large data sets (even in the millions). For a fixed number of
#' exchange partners (specified using the argument
#' \code{k_neighbours}), the approximate run time of
#' \code{fast_anticlustering} is in O(M N). The algorithm
#' \code{method = "exchange"} in \code{\link{anticlustering}} with
#' \code{objective = "variance"} has a run time of O(M N^3). 
#' Thus, \code{fast_anticlustering} can improve the run time
#' by two orders of magnitude as compared to the standard exchange
#' algorithm. The nearest neighbour search, which is done in the
#' beginning usually does not strongly contribute to the overall
#' run time. It is nevertheless possible to suppress the nearest
#' neighbour search by using the \code{exchange_partners} argument.
#'
#' When setting the \code{categories} argument, exchange partners
#' (i.e., nearest neighbours) will be generated from the same
#' category. Note that when \code{categories} has multiple columns,
#' each combination of these categories is treated as a distinct
#' category by the exchange method. You can also use
#' \code{\link{categories_to_binary}} to potentially improve results
#' for several categorical variables, instead of using the argument
#' \code{categories}.
#'
#' @examples
#'
#' ## Use fewer or more exchange partners to adjust speed (vs. quality tradeoff)
#' features <- iris[, - 5]
#' N <- nrow(features)
#' init <- sample(rep_len(1:3, N)) # same starting point for all calls:
#' groups1 <- fast_anticlustering(features, K = init) # default: all exchanges
#' groups2 <- fast_anticlustering(features, K = init, k_neighbours = 20) 
#' groups3 <- fast_anticlustering(features, K = init, k_neighbours = 2)
#' 
#' variance_objective(features, groups1)
#' variance_objective(features, groups2)
#' variance_objective(features, groups3)
#'
#' # K-plus anticlustering is straight forward when sticking with the default
#' # for k_neighbours
#' kplus_anticlusters <- fast_anticlustering(
#'   kplus_moment_variables(features, T = 2), 
#'   K = 3
#' )
#' mean_sd_tab(features, kplus_anticlusters)
#' 
#' # Some care is needed when applying k-plus using with this function 
#' # while using a reduced number of exchange partners generated in the 
#' # nearest neighbour search. Then we:
#' # 1) Use kplus_moment_variables() on the numeric input
#' # 2) Generate custom exchange_partners because otherwise nearest 
#' #    neighbours are internally selected based on the extended k-plus 
#' #    variables returned by kplus_moment_variables() 
#' #    (which does not really make sense)
#' kplus_anticlusters <- fast_anticlustering(
#'   kplus_moment_variables(features, T = 2), 
#'   K = 3,
#'   exchange_partners = generate_exchange_partners(120, features = features, method = "RANN")
#'  )
#' mean_sd_tab(features, kplus_anticlusters)
#' # Or we use random exchange partners: 
#' kplus_anticlusters <- fast_anticlustering(
#'   kplus_moment_variables(features, T = 2), 
#'   K = 3,
#'   exchange_partners = generate_exchange_partners(120, N = nrow(features), method = "random")
#' )
#' mean_sd_tab(features, kplus_anticlusters)
#' 
#' 
#' # Working on several 1000 elements is very fast (Here n = 10000, m = 2)
#' data <- matrix(rnorm(10000 * 2), ncol = 2)
#' start <- Sys.time()
#' groups <- fast_anticlustering(data, K = 5, k_neighbours = 5)
#' Sys.time() - start 
#'

fast_anticlustering <- function(x, K, k_neighbours = Inf, categories = NULL, 
                                exchange_partners = NULL) {
  input_validation_anticlustering(
    x, K, "variance", "exchange", FALSE, categories, NULL
  )
  categories <- merge_into_one_variable(categories)
  x <- as.matrix(x)
  N <- nrow(x)
  
  if (argument_exists(exchange_partners)) {
    validate_exchange_partners(exchange_partners, N)
  } else {
    if (!isTRUE(k_neighbours == Inf)) {
      validate_input(k_neighbours, "k_neighbours", objmode = "numeric", len = 1,
                     must_be_integer = TRUE, greater_than = 0, not_na = TRUE)
    }
    exchange_partners <- all_exchange_partners(x, k_neighbours, categories)
  }
  exchange_partners <- cleanup_exchange_partners(exchange_partners, N)

  c_anticlustering(
    x, initialize_clusters(N, K, categories), 
    categories = NULL, objective = "fast-kmeans", 
    exchange_partners - 1
  )
}

#' Get exchange partners for k-means anticlustering 
#'
#' @details
#'
#' Computes the k nearest neighbours for each input element using
#' RANN::nn2. If no nearest neighbours are required, argument 
#' `k_neighbours` will be `Inf`. May compute nearest neighbors within
#' categories. 
#' 
#' @return A list of length `N`. Each element is a vector
#'    of exchange partners (that may be nearest neighbors).
#'
#' @noRd


all_exchange_partners <- function(features, k_neighbours, categories) {
  # Case 1: no nearest neighbor search needed
  if (is.infinite(k_neighbours)) { 
    return(all_exchange_partners_(nrow(features), categories))
  }
  # Case 2: NN search needed
  return(nearest_neighbours(features, k_neighbours, categories))
}

# Generate all possible exchange partners 
all_exchange_partners_ <- function(N, categories) {
  # Case 1: Exchange partners are from the same category
  if (argument_exists(categories)) {
    return(list_idx_by_category(categories))
  }
  # Case 2: Everyone is potential exchange partner
  rep(list(1:N), N) 
}

# convert list of exchange partners to matrix for C; 
# when doing that, possibly "fill" empty entries with "N+1" (in C -> N), if some
# list elements are shorter than others (if categorical variables have been used and are
# unevenly distributed)
cleanup_exchange_partners <- function(exchange_partners, N) {
  max_exchanges_partners <- max(lengths(exchange_partners))
  exchange_partners <- lapply(exchange_partners, function(x) c(x[1:length(x)], rep(N+1, max(0, max_exchanges_partners - length(x)))))
  # remove potential NAs
  exchange_partners <- lapply(exchange_partners, function(x) x[!is.na(x)]) 
  exchange_partners <- unname(t(t(as.data.frame(exchange_partners))))
  # `exchange_partners` is passed as matrix to C; there it is converted to a 1-dimensional "vector".
  # Here we pass it as a matrix where elements = columns; cols = exchange partners.
  # Usually it would be the other way around, e.g. RANN returns the standard "N x k_neighbours"
  # format, but I pass a "k_neighbours x N" matrix to C, which is more easily handled there.
  exchange_partners
}

validate_exchange_partners <- function(exchange_partners, N) {
  if (!inherits(exchange_partners, "list")) {
    stop("Argument `exchange_partners` must be a list.")
  }
  if (length(exchange_partners) != N) {
    stop("If used, argument `exchange_partners` must have the same length as you have data points.")
  }
  if (any(lengths(exchange_partners)) > N) {
    stop("Argument `exchange_partners` has problems: Some elements have more exchange partners than there are data points.")
  }
  unlisted <- unlist(exchange_partners)
  if (any(unlisted) > N) {
    stop("Argument `exchange_partners` has problems: The maximum index is larger than the number of elements.")
  }
  if (any(unlisted) < 1) {
    stop("Argument `exchange_partners` has problems: Some indices are lower than 1.")
  }
  if (any(as.integer(unlisted) != unlisted)) {
    stop("Argument `exchange_partners` has problems: Only integer indices are allowed.")
  }
}


#' Get exchange partners for fast_anticlustering()
#' 
#' @param n_exchange_partners The number of exchange partners per
#'     element
#' @param N The number of elements for which exchange partners; can be
#'     \code{NULL} if \code{features} is passed (it is ignored if
#'     \code{features} is passed).
#' @param features The features for which nearest neighbours are
#'     sought if \code{method = "RANN"}.  May be NULL if random
#'     exchange partners are generated.
#' @param method Currently supports "random" (default), "RANN" and
#'     "restricted_random". See details.
#' @param categories A vector, data.frame or matrix representing one
#'     or several categorical constraints.
#'
#' @return A list of length \code{N}. Is usually used as input to the
#'     argument \code{exchange_partners} in
#'     \code{\link{fast_anticlustering}}.  Then, the i'th element of
#'     the list contains the indices of the exchange partners that are
#'     used for the i'th element.
#' 
#' @export
#' 
#' @details
#' 
#' The \code{method = "RANN"} generates exchange partners using a
#' nearest neighbour search via \code{\link[RANN]{nn2}} from the
#' \code{RANN} package; \code{methode = "restricted_random"} generates
#' random exchange partners but ensures that for each element, no
#' duplicates are generated and that the element itself does not occur
#' as exchange partner (this is the slowest method, and I would not
#' recommend it for large N); \code{method = "random"} (default) does
#' not impose these restrictions and generates unrescricted random
#' partners (it may therefore generate duplicates and the element
#' itself as exchange partner).
#' 
#' When setting the \code{categories} argument and using \code{method
#' = "RANN"}, exchange partners (i.e., nearest neighbours) will be
#' generated from the same category; \code{methode =
#' "restricted_random"} will also adhere to categorical constraints
#' induced via \code{categories} (i.e. each element only receives
#' exchange partners from the same category as itself); \code{methode
#' = "random"} cannot incoorporate categorical restrictions.
#' 
#' 
#' @examples
#' 
#' # Restricted random method generates no duplicates per element and cannot return 
#' # the element itself as exchange partner
#' generate_exchange_partners(5, N = 10, method = "restricted_random")
#' # "random" simply randomizes with replacement and without restrictions
#' # (categorical restrictions are also not possible; is much faster for large data sets)
#' generate_exchange_partners(5, N = 10, method = "random")
#' # May return less than 5 exchange partners if there are not enough members 
#' # of the same category: 
#' generate_exchange_partners(
#'   5, N = 10, 
#'   method = "restricted_random", 
#'   categories = cbind(schaper2019$room, schaper2019$frequency)
#' )
#' # using nearest neighbour search (unlike RANN::nn2, this does not 
#' # return the ID of the element itself as neighbour)
#' generate_exchange_partners(5, features = schaper2019[, 3:5], method = "RANN")[1:3]
#' # compare with RANN directly:
#' RANN::nn2(schaper2019[, 3:5], k = 6)$nn.idx[1:3, ] # note k = 6
#' 
generate_exchange_partners <- function(n_exchange_partners, N = NULL, features = NULL, method = "random", categories = NULL) {
  if (argument_exists(features)) {
      validate_data_matrix(features)
      N <- NROW(features)
  }
  categories <- merge_into_one_variable(categories)
  validate_input(n_exchange_partners, "n_exchange_partners", objmode = "numeric", len = 1,
                 must_be_integer = TRUE, greater_than = 0, not_na = TRUE)
  validate_input(method, "method", objmode = "character", len = 1,
                 not_na = TRUE, not_function = TRUE, input_set = c("RANN", "random", "restricted_random"))
  if (argument_exists(features) && method == "RANN") {
    return(nearest_neighbours(features, n_exchange_partners, categories))
  } else if (method == "restricted_random") {
    init_list <- NULL
    if (argument_exists(categories)) {
      init_list <- list_idx_by_category(categories)
    }
    return(random_exchange_partners(N, n_exchange_partners, init_list))
  } else if (method == "random" && !argument_exists(categories)) {
    return(completely_random_exchange_partners(N, n_exchange_partners))
  }
  else {
    stop("Argument method must be 'RANN' or 'random' or 'restricted_random'.\n'",
         "When method is 'RANN', the argument 'features' must be used.\n'",
         "When method is 'random', you cannot use the argument 'categories'.")
  }
}

list_idx_by_category <- function(categories) {
  category_ids <- lapply(1:max(categories), function(i) which(categories == i))
  category_ids[categories]
}

# Generate exchange partners via nearest neighbor search using RANN::nn2
nearest_neighbours <- function(features, k_neighbours, categories) {
  if (!argument_exists(categories)) {
    idx <- matrix_to_list(RANN::nn2(features, k = min(k_neighbours + 1, nrow(features)))$nn.idx)
  } else {
    # compute nearest neighbors within each category
    nns <- list()
    # track the indices when dividing by category. problem is:
    # in nearest neighbour search, there is no guarantee that 
    # the nearest neighbour is the element itself; if it were, 
    # restoring original order would be easy
    new_order <- list()
    for (i in 1:max(categories)) {
      tmp_indices <- which(categories == i)
      new_order[[i]] <- tmp_indices
      tmp_features <- features[tmp_indices, , drop = FALSE]
      tmp_nn <- RANN::nn2(tmp_features, k = min(k_neighbours + 1, nrow(tmp_features)))$nn.idx
      # get original index per category
      nns[[i]] <- which(categories == i)[tmp_nn]
      # restore matrix structure, gets lost 
      dim(nns[[i]]) <- dim(tmp_nn)
    }
    # per category, convert matrix to list
    idx_list <- lapply(nns, matrix_to_list)
    # in the end, merge all lists into 1 list
    idx <- merge_lists(idx_list)
    # restore original order after dividing by category
    original_order <- order(unlist(new_order))
    idx <- idx[original_order]
  }
  # return as list
  remove_self(idx)
}

# Convert a matrix to list - each row becomes list element
matrix_to_list <- function(x) {
  as.list(as.data.frame(t(x)))
}

# Merge a list of lists into one list
merge_lists <- function(list_of_lists) {
  do.call(c, list_of_lists)
}

completely_random_exchange_partners <- function(N, n_exchange_partners) {
  ret <- matrix_to_list(matrix(sample(N, size = N * n_exchange_partners, replace = TRUE), ncol = n_exchange_partners))
  unname(ret)
}

# generate list of random exchange partners; will not generate itself as exchange partner
# and has no duplicates for each element
random_exchange_partners <- function(N, n_exchange_partners, init_partners = NULL) {
  if (is.null(init_partners)) {
    init_partners <- rep(list(1:N), N) 
  }
  init_partners <- remove_self(init_partners)
  partners <- lapply(init_partners, function(x) sample_(x)[1:n_exchange_partners])
  # possibly remove NAs
  lapply(partners, function(x) x[!is.na(x)]) 
}

# remove in a list for each element the ID of the list elements position 
remove_self <- function(idx_list) {
  N <- length(idx_list)
  lapply(1:N, function(i) idx_list[[i]][idx_list[[i]] != i])
}




#### THIS IS AN OLD AN OBSOLETE R IMPLEMENTATION; I AM KEEPING IT FOR THE TEST FILES
#' Solve anticlustering using the fast exchange method
#'
#' @param data the data -- an N x M table of item features
#' @param clusters An initial cluster assignment
#' @param all_exchange_partners A list of exchange partners
#'
#' @return The anticluster assignment
#'
#' @noRd
#'

fast_exchange_ <- function(data, clusters, all_exchange_partners) {
  N <- nrow(data)
  best_total <- variance_objective_(clusters, data)
  centers <- cluster_centers(data, clusters)
  distances <- dist_from_centers(data, centers, squared = TRUE)
  
  ## frequencies of each cluster are required for updating cluster centers:
  tab <- c(table(clusters))
  for (i in 1:N) {
    # cluster of current item
    cluster_i <- clusters[i]
    # get exchange partners for item i
    exchange_partners <- all_exchange_partners[[i]]
    # exchange partners are not in the same cluster:
    exchange_partners <- exchange_partners[clusters[exchange_partners] != clusters[i]]
    # Sometimes an exchange cannot take place
    if (length(exchange_partners) == 0) {
      next
    }
    # container to store objectives associated with each exchange of item i:
    comparison_objectives <- rep(NA, length(exchange_partners))
    for (j in seq_along(exchange_partners)) {
      ## Swap item i with all legal exchange partners and check out objective
      # (a) Determine clusters of to-be-swapped elements
      tmp_clusters <- clusters
      tmp_swap <- exchange_partners[j]
      cluster_j <- tmp_clusters[tmp_swap]
      # (b) Swap the elements
      tmp_clusters[i] <- cluster_j
      tmp_clusters[tmp_swap] <- cluster_i
      # (c) Update cluster centers after swap
      tmp_centers <- update_centers(centers, data, i, tmp_swap, cluster_i, cluster_j, tab)
      # (d) Update distances from centers after swap
      tmp_distances <- update_distances(data, tmp_centers, distances, cluster_i, cluster_j)
      # (e) Compute objective after swap
      comparison_objectives[j] <- sum(tmp_distances[cbind(1:nrow(tmp_distances), tmp_clusters)])
    }
    ## If an improvement of the objective occured, do the swap
    best_this_round <- max(comparison_objectives)
    if (best_this_round > best_total) {
      # which element has to be swapped
      swap <- exchange_partners[comparison_objectives == best_this_round][1]
      # Update cluster centers
      centers <- update_centers(centers, data, i, swap, clusters[i], clusters[swap], tab)
      # Update distances
      distances <- update_distances(data, centers, distances, cluster_i, clusters[swap])
      # Actually swap the elements - i.e., update clusters
      clusters[i] <- clusters[swap]
      clusters[swap] <- cluster_i
      # Update best solution
      best_total <- best_this_round
    }
  }
  clusters
}

#' Recompute distances from cluster centers after swapping two elements
#' @param distances distances from cluster centers per element (old)
#' @param cluster_i the cluster of element i
#' @param cluster_j the cluster of element j
#' @return The new distances
#' @noRd
update_distances <- function(features, centers, distances, cluster_i, cluster_j) {
  for (k in c(cluster_i, cluster_j)) {
    distances[, k] <- colSums((t(features) - centers[k,])^2)
  }
  distances
}

#' Update a cluster center after swapping two elements
#'
#' @param centers The current cluster centers
#' @param features The features
#' @param i the index of the first element to be swapped
#' @param j the index of the second element to be swapped
#' @param cluster_i the cluster of element i
#' @param cluster_j the cluster of element j
#' @param tab A table of the cluster frequencies
#'
#' @details
#'
#' This should make the fast exchange method much faster, because
#' most time is spent on finding the cluster centers. After swapping
#' only two elements, it should be possible to update the two centers
#' very fast
#' @noRd

update_centers <- function(centers, features, i, j, cluster_i, cluster_j, tab) {
  ## First cluster: item i is removed, item j is added
  centers[cluster_i, ] <- centers[cluster_i, ] - (features[i, ] / tab[cluster_i]) + (features[j, ] / tab[cluster_i])
  ## Other cluster: item j is removed, item i is added
  centers[cluster_j, ] <- centers[cluster_j, ] + (features[i, ] / tab[cluster_j]) - (features[j, ] / tab[cluster_j])
  centers
}

############ END OLD OBSOLETE IMPLEMENTATION

library("anticlust")

# different options for preclustering have the same result - variance objective
for (M in 1:4) {
  for (K in 2:5) {
    N <- K * 3
    features <- matrix(rnorm(N * M), ncol = M)
    
    # determine a random seed to make the exchange method reproducible
    seed <- sample(1000, size = 1)
    set.seed(seed)
    
    ## First option
    # Set `preclustering = TRUE`
    ac1 <- anticlustering(
      features,
      K = K,
      objective = "variance",
      preclustering = TRUE
    )
    
    ## Second option
    # Call `balanced_clustering` and use output as `categories` argument
    preclusters <- balanced_clustering(
      features,
      K = N / K
    )
    set.seed(seed)
    ac2 <- anticlustering(
      features,
      K = K,
      objective = "variance",
      categories = preclusters
    )
    
    ## Third option
    # Use `fast_anticlustering` function
    set.seed(seed)
    ac3 <- fast_anticlustering(
      features,
      K = categorical_sampling(preclusters, K),
      exchange_partners = anticlust:::nearest_neighbours(features, N-1, preclusters)
    )
    
    expect_true(all(ac1 == ac2))
    expect_true(all(ac3 == ac3))
    
  }
}


# different options for preclustering have the same result - distance objective
for (M in 1:4) {
  for (K in 2:5) {
    N <- K * 3
    features <- matrix(rnorm(N * M), ncol = M)
    
    # determine a random seed to make the exchange method reproducible
    seed <- sample(1000, size = 1)
    set.seed(seed)
    
    ## First option
    # Set `preclustering = TRUE`
    ac1 <- anticlustering(
      features,
      K = K,
      objective = "distance",
      preclustering = TRUE
    )
    
    ## Second option
    # Call `balanced_clustering` and use output as `categories` argument
    preclusters <- balanced_clustering(
      features,
      K = N / K
    )
    set.seed(seed)
    ac2 <- anticlustering(
      features,
      K = K,
      objective = "distance",
      categories = preclusters
    )
    
    ## Third option
    # Use distance input (and categories for preclusters because the
    # preclustering algorithm is slightly different for distance input)
    set.seed(seed)
    ac3 <- anticlustering(
      dist(features),
      K = K,
      objective = "distance",
      categories = preclusters
    )
    
    expect_equal(all(ac1 == ac2), TRUE)
    expect_equal(all(ac2 == ac3), TRUE)
  }
}

library("anticlust")

# Exchange partners are generated correctly - no restriction
clusters <- c(1, 2, 1, 2, 1, 2)
i <- 1
group_i <- clusters[i]
N <- length(clusters)
categories <- NULL
partners <- anticlust:::get_exchange_partners(clusters, i, categories)
expect_equal(all(partners == c(2, 4, 6)), TRUE)

# Exchange partners are generated correctly - categorical restriction
clusters <- c(1, 2, 1, 2, 1, 2)
i <- 1
group_i <- clusters[i]
N <- length(clusters)
categories <- c(1, 1, 1, 1, 2, 2)
partners <- anticlust:::get_exchange_partners(clusters, i, categories)
expect_equal(all(partners == c(2, 4)), TRUE)

# Exchange partners are generated correctly - preclustering restriction
clusters <- c(1, 2, 1, 2, 1, 2)
i <- 1
group_i <- clusters[i]
N <- length(clusters)
# Add a preclustering restriction (this is now just the same as
# categorical restrictions and does not really make sense here)
preclusters <- c(1, 1, 2, 2, 3, 3)
partners <- anticlust:::get_exchange_partners(clusters, i, preclusters)
expect_equal(partners == 2, TRUE)

# Exchange partners are generated correctly - preclustering and categorical restriction
clusters <- c(1, 2, 1, 2, 1, 2)
i <- 1
group_i <- clusters[i]
N <- length(clusters)
## Use both restrictions
categories <- c(1, 1, 1, 1, 2, 2)
preclusters <- c(1, 2, 2, 1, 3, 3)
constraints <- anticlust:::merge_into_one_variable(cbind(categories, preclusters))
partners <- anticlust:::get_exchange_partners(clusters, i, constraints)
expect_equal(partners == 4, TRUE)

# function that tests that all exchange partners are from the same category
test_idx <- function(idx, categories) {
  if (!is.null(categories)) {
    for (i in 1:length(idx)) {
      category_of_partners <- categories[idx[[i]]]
      # test that all partners have the same category
      expect_true(all(category_of_partners == category_of_partners[1]))
      # test that all partners actually have the category of item `i`
      expect_true(category_of_partners[1] == categories[i])
    }
  }
}

# Exchange partners are generatede correctly for fast k-means method

# generate random categories and data
N <- 1000
M <- 2
k_neighbours <- 10
features <- matrix(rnorm(N * M), ncol = M)
categories <- sample(1:4, size = N, replace = TRUE)

### Using categorical restrictions
# Case 1: restricted number of exchange partners (i.e., nearest neighbour search)

partners <- anticlust:::nearest_neighbours(
  features = features,
  k_neighbours = k_neighbours,
  categories = categories
)
test_idx(partners, categories)
## ensure correct order of the output 
expect_true(all(lengths(partners) == k_neighbours))

# Case 2: no restriction on number of exchange partners
partners <- anticlust:::list_idx_by_category(categories)
test_idx(partners, categories)

### Not using categorical restrictions
# Case 1: restricted number of exchange partners (i.e., nearest neighbour search)
categories <- anticlust:::to_numeric(schaper2019$room)
k_neighbours <- k_neighbours
partners <- anticlust:::nearest_neighbours(
  features = features,
  k_neighbours = k_neighbours,
  categories = NULL
)
expect_true(all(lengths(partners) == k_neighbours))

# remove_self() function works as intended in exchange partner generation
expect_true(
  identical(
    anticlust:::remove_self(list(c(2, 1), c(3, 2), c(4, 3))), list(2, 3, 4)
  )
)

library("anticlust")
library("RANN")

# New k-means C implementation yields same results as other implementations
  
  set.seed(123)
  M <- 5
  N <- 180
  K <- 4
  features <- matrix(rnorm(N*M), ncol = M)
  init <- sample(rep_len(1:K, nrow(features)))
  

  # fast_anticlustering() does the same as anticlustering()?
  ac_exchangeC <- fast_anticlustering(
    features, K = init, exchange_partners = rep(list(1:N), N) # -> everyone is exchange partner
  )
  ac_exchangeC2 <- anticlustering(features, K = init, objective = "variance")

  expect_true(all(ac_exchangeC == ac_exchangeC2))
  
  # new C fast_anticlustering() implementation does the same as old R implementation?
  ac_exchangeC <- fast_anticlustering(features, K = init, k_neighbours = 20)
  ac_exchangeR <- anticlust:::fast_exchange_(features, init, anticlust:::nearest_neighbours(features, 20, NULL))
  expect_true(all(ac_exchangeC == ac_exchangeR))
  
  # Also test with categorical restrictions
  features <- as.matrix(schaper2019[, 3:6])
  categories <- schaper2019$room
  N <- nrow(features)
  ac_exchangeC <- fast_anticlustering(features, K = 4, k_neighbours = N-1, categories = categories)
  ac_exchangeC2 <- anticlustering(features, K = 4, objective = "variance", categories = categories)
  expect_true(all(table(ac_exchangeC, categories)  == table(ac_exchangeC2, categories)))
  
  # Use reduced exchange partners
  init <- anticlust:::initialize_clusters(N, 3, categories)
  ac_exchangeC <- fast_anticlustering(features, K = init, k_neighbours = 20, categories = categories)
  ac_exchangeR <- anticlust:::fast_exchange_(features, init, anticlust:::nearest_neighbours(features, 20, anticlust:::to_numeric(categories)))
  expect_true(all(ac_exchangeC == ac_exchangeR))
  
  # What if `k_neighbours` is greater than the number of elements in the group with fewest members?
  categories <- anticlust:::merge_into_one_variable(cbind(schaper2019$syllables, schaper2019$room))
  init <- categorical_sampling(categories, K = 3)
  ac_exchangeC <- fast_anticlustering(features, K = init, k_neighbours = 10, categories = categories)
  ac_exchangeR <- anticlust:::fast_exchange_(features, init, anticlust:::nearest_neighbours(features, 10, categories))
  expect_true(all(table(ac_exchangeC, categories) == table(ac_exchangeR, categories)))
  
  # test custom exchange partners
  N <- nrow(features)
  # random exchange partners
  k_neighbours <- 10
  exchange_partners <- lapply(rep(N, N), function(x) sample(1:x)[1:k_neighbours])
  # R and C implementation the same?
  ac_exchangeC <- fast_anticlustering(features, K = init, exchange_partners = exchange_partners)
  ac_exchangeR <- anticlust:::fast_exchange_(features, init, exchange_partners)
  expect_true(all(ac_exchangeC == ac_exchangeR))
  
  # Remove some exchange partners to test if an unequal number of exchange partners works
  exchange_partners[[1]] <- exchange_partners[[1]][-(1:3)]
  exchange_partners[[10]] <- exchange_partners[[10]][-2]
  exchange_partners[[23]] <- exchange_partners[[23]][-(10:7)]
  ac_exchangeC <- fast_anticlustering(features, K = init, exchange_partners = exchange_partners)
  ac_exchangeR <- anticlust:::fast_exchange_(features, init, exchange_partners)
  expect_true(all(ac_exchangeC == ac_exchangeR))
  
  # Does the argument `exchange_partners` really do the same as `k_neighbours`?
  nn_exchange_partners <- RANN::nn2(features, k = k_neighbours + 1)$nn.idx #+1 because the element itself is also returned by nn2
  ac1 <- fast_anticlustering(features, K = init, exchange_partners = anticlust:::matrix_to_list(nn_exchange_partners))
  ac2 <- fast_anticlustering(features, K = init, k_neighbours = k_neighbours)
  expect_true(all(ac1 == ac2))
  
  # ensure that using exchange_partners does the same as not using it, when combining with new function generate_exchange_partners()
  ac1 <- fast_anticlustering(features, K = init, exchange_partners = generate_exchange_partners(5, features = features, method = "RANN"))
  ac2 <- fast_anticlustering(features, K = init, k_neighbours = 5)
  expect_true(all(ac1 == ac2))
  
  # again, using a different data set
  features <- iris[,-5]
  init <- anticlust:::initialize_clusters(K = 75, N = nrow(features), NULL)
  n_exchange_partners <- 10
  groups1 <- fast_anticlustering(
    features,
    K = init,
    exchange_partners = generate_exchange_partners(n_exchange_partners, features = features, method = "RANN")
  )
  # compare with using nearest neighbours as exchange partners:
  groups2 <- fast_anticlustering(features, K = init, k_neighbours = n_exchange_partners)
  expect_true(all(groups1 == groups2))
  
  ## Test that new default for k_neighbours works (and if N < 20, it still works!)
  expect_length(fast_anticlustering(1:6, K = 2), 6)
  expect_length(fast_anticlustering(1:20, K = 2), 20)
  expect_length(fast_anticlustering(1:21, K = 2), 21)

# nearest neighbor centroid clustering

# data = feature data frame
# K = size of the small groups (not the number of clusters!)
# groups = vector of length nrow(data); if passed, K-partite matching is conducted,
# i.e., each element is clustered with elements that are in *other* groups than 
# the element itself. Consists of integers 1, ..., K
nn_centroid_clustering <- function(
  data, K, 
  groups = NULL, 
  match_extreme_first = TRUE, 
  target_group = FALSE
) {
  data <- as.matrix(data)
  N <- nrow(data)
  distances <- distances_from_centroid(data)

  if (argument_exists(groups)) {
    K <- length(unique(groups))
  }

  # some book-keeping variables for the loop
  cluster_id <- 1
  idx <- 1:nrow(data)
  clusters <- rep(NA, nrow(data))
  while (data_fits(data, K) && no_group_empty(groups, K)) {
    # Get target item
    target <- get_target(distances, groups, target_group, match_extreme_first)
    # Compute nearest neighbors for target element
    clustered <- get_nearest_neighbours(data, target, K, groups)
    clusters[idx[clustered]] <- cluster_id
    # Remove matched elements:
    data <- subset_data_matrix(data, -clustered)
    distances <- distances[-clustered]
    groups <- groups[-clustered]
    idx  <- idx[-clustered]
    # increase cluster id for next elements
    cluster_id <- cluster_id + 1
  }
  order_cluster_vector(clusters)
}

# simple test: does a data frame still fit into pieces of K units
data_fits <- function(data, K)  {
  nrow(data) > (K-1)
}

# for k-partite matching: test if no group is empty 
no_group_empty <- function(groups, K) {
  !(argument_exists(groups) && any(!(1:K %in% groups)))
}

# Find target element for which neighbours are sought
# param distances: distances to a cluster centroid
# param groups: grouping vector (%in% 1, ..., K)
# param target_group: id of the group from which target is selected 
#    (may be FALSE instead of an ID)
# param match_extreme_first: boolean
get_target <- function(distances, groups, target_group, match_extreme_first) {
  if (!target_group) {
    # no target group specified: select an item from all elements
    if (match_extreme_first) {
      return(which.max(distances))
    } else {
      return(which.min(distances))
    }
  }
  # return a member from the target group as target
  ids_target <- which(groups == target_group)
  ordered_distances <- order(distances, decreasing = match_extreme_first)
  ordered_distances[ordered_distances %in% ids_target][1]
}

# Get nearest neighbours for a current element
# param data: the data
# param target: the index of the element for which nearest neighbors are sought
# param K: The number of nearest neighbours (including the element itself!)
# param groups = vector of length nrow(data); elements are clustered with elements in *other* groups 
# return: The indices of the current element as vector. !! The first
#   index is the element itself !!
get_nearest_neighbours <- function(data, target, K, groups) {
  
  # k-partite clustering
  if (argument_exists(groups)) {
    groups <- to_numeric(groups)
    current_group <- groups[target]
    other_groups <- (1:K)[-current_group]
    current_element <- data[target, , drop = FALSE]
    # get nearest neighbour in each other group
    nns <- rep(NA, K-1)
    for (i in seq_along(other_groups)) {
      if (is_distance_matrix(data)) {
        # get indices of elements in other groups
        idx_other <- which(groups == other_groups[i])
        most_similar <- order(current_element)
        nns[i] <- most_similar[most_similar %in% idx_other][1]
      } else {
        tmp_data <- data[groups == other_groups[i], , drop = FALSE]
        nn <- c(nn2(tmp_data, current_element, 1)$nn.idx)
        nns[i] <- which(groups == other_groups[i])[nn]
      }
    }
    nns <- c(target, nns)
    return(nns)
  }
  
  # normal clustering - no grouping restrictions
  if (is_distance_matrix(data)) {
    nns <- order(data[target, ])[1:K]
  } else {
    nns <- nn2(data, data[target, , drop = FALSE], K)$nn.idx
  }
  nns
}

# Compute the distances from centroid of a data set.
# Centroid is either computed in euclidean space or taken as a central element
distances_from_centroid <- function(x) {
  if (is_distance_matrix(x)) {
    distances <- as.matrix(x)
    maxima <- apply(distances, 1, max)
    centroid <- which.min(maxima)
    return(distances[centroid, ])
  }
  centroid <- t(as.matrix(colMeans(x)))
  c(dist_from_centers(x, centroid, squared = FALSE))
}


#' Is a matrix a legal distance matrix
#'
#' @param m a Matrix
#' @return TRUE if `m` is distance matrix, FALSE otherwise
#' @noRd
is_distance_matrix <- function(m) {
  m <- as.matrix(m)
  if (nrow(m) != ncol(m)) {
    return(FALSE)
  }
  # this function needs to test if there is NAs in the data input
  # this is because I do not rely on class to identify if the input is a distance matrix (which is probably bad)
  any_na <- any(is.na(m))
  if (any_na) m[is.na(m)] <- -9999
  lower <- m[lower.tri(m)]
  m <- t(m)
  upper <- m[lower.tri(m)]
  is_dist <- all(lower == upper) && all(diag(m) == diag(m)[1])
  if (is_dist && any_na) stop("No NA allowed in distance matrix input.")
  is_dist
}

# Subset a distance or feature data matrix (not knowing which one)
subset_data_matrix <- function(data, selection) {
  if (is_distance_matrix(data)) {
    data <- data[selection, selection, drop = FALSE]
  } else {
    data <- data[selection, , drop = FALSE]
  }
  data
}

#' Solve anticluster editing using a fast exchange method
#'
#' @param data An N x N dissimilarity matrix or N x M features matrix.
#' @param K The number of cluster or an initial cluster assignment
#' @param categories A vector representing categorical constraints
#' @param preclusters A vector representing preclustering constraints
#'
#' @return The anticluster assignment
#'
#' @noRd
#'

fast_exchange_dist <- function(data, K, categories) {
  distances <- convert_to_distances(data)  
  clusters <- initialize_clusters(nrow(distances), K, categories)
  best_total <- diversity_objective_(clusters, distances)
  distances[upper.tri(distances)] <- 0
  diag(distances) <- 0
  ## Matrix for indexing the elements in the distance matrix
  selected <- selection_matrix_from_clusters(clusters)
  N <- nrow(distances)
  for (i in 1:N) {
    exchange_partners <- get_exchange_partners(clusters, i, categories)
    ## Skip if there are zero exchange partners
    if (length(exchange_partners) == 0) {
      next
    }
    # container to store objectives associated with each exchange of item i:
    comparison_objectives <- rep(NA, length(exchange_partners))
    ## Swap item i with all legal exchange partners and store objectives
    for (j in seq_along(exchange_partners)) {
      comparison_objectives[j] <- update_objective_distance(
        distances = distances,
        selection_matrix = selected,
        i = i,
        j = exchange_partners[j],
        current_objective = best_total
      )
    }
    # Do the swap if an improvement occured
    best_this_round <- max(comparison_objectives)
    if (best_this_round > best_total) {
      # Which element has to be swapped
      swap <- exchange_partners[comparison_objectives == best_this_round][1]
      # Swap the elements in clustering vector
      clusters <- cluster_swap(clusters, i, swap)
      # Swap the elements in selection matrix
      selected <- swap_items(selected, i, swap)
      # Update best solution
      best_total <- best_this_round
    }
  }
  clusters
}

# Update objective value after swapping two items
#
# param distances: The N x N distance matrix
# param selection_matrix: An N x N matrix encoding the currently selected items
#                (generated via `selection_matrix_from_clusters`)
# param i, j: The items to be swapped
# param current_objective: The best found objective
# return: the objective value after swapping
update_objective_distance <- function(distances, selection_matrix, i, j, current_objective) {
  tmp_selection <- swap_items(selection_matrix, i, j)
  old_distances <- itemwise_distance_sum(distances, selection_matrix, i, j)
  new_distances <- itemwise_distance_sum(distances, tmp_selection, i, j)
  current_objective - old_distances + new_distances
}

# Create a boolean matrix for indexing in a distance matrix
# Inverse function for `clusters_from_selection_matrix`
#
# param clusters: A clustering vector with elements 1, ..., K indicating cluster membership
# return: A N x N matrix where TRUE in cell [i,j] indicates that elements
#     i and j are in the same (anti)cluster
selection_matrix_from_clusters <- function(clusters) {
  n <- length(clusters)
  K <- length(unique(clusters))
  selected <- matrix(FALSE, nrow = n, ncol = n)
  change_thing <- function(k) {
    in_same_cluster <- which(clusters == k)
    combis <- expand.grid(in_same_cluster, in_same_cluster)
    selected[as.matrix(combis)] <<- TRUE
  }
  lapply(1:K, change_thing)
  selected
}

# Convert a selection matrix to clustering vector
# Inverse function for `selection_matrix_from_clusters`
#
# param selection_matrix: A N x N matrix where TRUE in cell [i,j]
#     indicates that elements i and j are in the same (anti)cluster
# return: A clustering vector with elements 1, ..., K indicating cluster membership
clusters_from_selection_matrix <- function(selection_matrix) {
  N <- nrow(selection_matrix)
  clusters <- rep(NA, )
  # assign first cluster to set and all its connections
  clusters[1] <- 1
  clusters[which(selection_matrix[1, ] == TRUE)] <- 1
  # Now iterate over the remaining elements
  for (i in 2:N) {
    # test if item i is already in a cluster
    if (!is.na(clusters[i])) {
      next
    }
    next_cluster_index <- max(clusters, na.rm = TRUE) + 1
    clusters[i] <- next_cluster_index
    clusters[which(selection_matrix[i, ] == TRUE)] <- next_cluster_index
  }
  clusters
}

## Swap items in a boolean matrix for indexing.
swap_items <- function(selected, i, j) {
  tmp1 <- selected[i, ]
  tmp2 <- selected[, i]
  selected[i, ] <- selected[j, ]
  selected[, i] <- selected[, j]
  selected[j, ] <- tmp1
  selected[, j] <- tmp2
  selected[i, j] <- FALSE
  selected[j, i] <- FALSE
  selected
}

# (not that) important: distances has 0 in diagonal and upper triangular
itemwise_distance_sum <- function(distances, selected, i, j) {
  n <- nrow(distances)
  sum(
    distances[i, 1:(i-1)][selected[i, 1:(i-1)]],
    distances[i:n, i][selected[i:n, i]],
    distances[j, 1:(j-1)][selected[j, 1:(j-1)]],
    distances[j:n, j][selected[j:n, j]]
  )
}

library("anticlust")

# constructing selection matrix from clusterin vector and reversing works
for (M in 1:4) {
  for (K in 2:5) {
    for (i in 3:8) {
      N <- K * i
      clusters <- sample(rep(1:K, N/K))
      selected <- anticlust:::selection_matrix_from_clusters(clusters)
      clusters2 <- anticlust:::clusters_from_selection_matrix(selected)
      expect_true(all(anticlust:::order_cluster_vector(clusters) == clusters2))
    }
  }
}

library("anticlust")

for (M in 1:4) {
  for (K in 2:5) {
    for (i in 3:8) {
      N <- K * i
      features <- matrix(rnorm(N * M), ncol = M)
      distances <- as.matrix(dist(features))
      clusters <- sample(rep(1:K, N/K))
      
      objective <- anticlust:::diversity_objective_(clusters, distances)
      objective2 <- anticlust:::diversity_objective_(clusters, features)
      expect_equal(objective, objective2)
      
      # Swap two items and check objective
      to_swap <- sample(1:K, size = 2)
      swap1 <- which(clusters == to_swap[1])[1]
      swap2 <- which(clusters == to_swap[2])[1]
      
      # Method 1: As in specialized exchange method for anticluster editing
      selected <- anticlust:::selection_matrix_from_clusters(clusters)
      obj1 <- anticlust:::update_objective_distance(distances, selected, swap1, swap2, objective)
      
      # Method 2: As in generic exchange method
      obj2 <- anticlust:::update_objective_generic(distances, clusters, swap1, swap2, diversity_objective)
      expect_equal(obj1, obj2)
      # note, this DOES NOT always work:
      # expect_true(obj1 == obj2) # floating points =/
    }
  }
}

# Return a clustering vector based on ILP solution for cluster editing
#
# @param solution The solution of the instance returned by
#     `solve_ilp`
#
# @return A clustering vector
#
ilp_to_groups <- function(solution, N) {
  # These steps are taken:
  # 1. Create matrix of inter-item connections from vector of inter-item connections
  # 2. Create clustering vector from matrix; this is returned
  selection_matrix <- matrix(ncol = N, nrow = N)
  selection_matrix[upper.tri(selection_matrix)] <- solution$x
  # the following line copies the upper triangular into the lower; diagonal = 0
  selection_matrix <- as.matrix(as.dist(t(selection_matrix)))
  clusters_from_selection_matrix(selection_matrix)
}


# Edit distances
#
# Based on a clustering, distances between elements of the same cluster
# are set to a specified value. This is used to enforce or forbid pairs
# of elements to be part of the same anticluster when the distance
# matrix is passed to the ILP solver.  By considering a preclustering
# of items, very similar items will not be assigned to the same group
# when the fixed distance object is used to create the ILP formulation
# of the item assignment instance.
#
# @param distances A N x N matrix of between-item-distances.
#
# @param clustering A vector representing a clustering; objects that
#     are part of the same cluster are assigned a new distance.
# @param value The value that is assigned to the fixed distances.
#     Defaults to -1,000,000.
#
# @return A distance object containing the fixed distances. For items
#     that are part of the same cluster.
#
#
edit_distances <- function(distances, clustering, value = -1000000) {
  n_groups <- length(unique(clustering))
  for (i in 1:n_groups) {
    items <- which(clustering == i) ## which items are in the i'th cluster
    ## two for-loops to tap all pairwise distances that may require
    ## fixing (a lot of unnecessary iterations, probably)
    for (j in 1:(length(items) - 1)) {
      for (t in 2:length(items)) {
        distances[items[j], items[t]] <- value
        distances[items[t], items[j]] <- value
      }
    }
  }
  return(distances)
}

#' Compute k-plus variables
#'
#'
#' @param x A vector, matrix or data.frame of data points. Rows
#'     correspond to elements and columns correspond to features. A
#'     vector represents a single feature.
#' @param T The number of distribution moments for which variables are generated.
#' @param standardize Logical, should all columns of the output be standardized
#'  (defaults to TRUE).
#'
#' @return A matrix containing all columns of \code{x} and all additional
#' columns of k-plus variables. If \code{x} has M columns, the output matrix
#' has M * T columns.
#'
#' @details
#' 
#' The k-plus criterion is an extension of the k-means criterion
#' (i.e., the "variance", see \code{\link{variance_objective}}).
#' In \code{\link{kplus_anticlustering}}, equalizing means and variances
#' simultaneously (and possibly additional distribution moments) is
#' accomplished by internally appending new variables to the data
#' input \code{x}. When using only the  variance as additional criterion, the
#' new variables represent the squared difference of each data point to
#' the mean of the respective column. All columns are then included---in
#' addition to the original data---in standard k-means
#' anticlustering. The logic is readily extended towards higher order moments,
#' see Papenberg (2024). This function gives users the possibility to generate
#' k-plus variables themselves, which offers some additional flexibility when
#' conducting k-plus anticlustering.
#'
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' @references
#' 
#' Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for 
#' Maximizing Between-Group Similarity. British Journal of Mathematical and 
#' Statistical Psychology, 77(1), 80--102. https://doi.org/10.1111/bmsp.12315
#' 
#' @export
#'
#' @examples
#' 
#' # Use Schaper data set for example
#' data(schaper2019)
#' features <- schaper2019[, 3:6]
#' K <- 3
#' N <- nrow(features)
#'
#' # Some equivalent ways of doing k-plus anticlustering:
#' 
#' init_groups <- sample(rep_len(1:3, N))
#' table(init_groups)
#' 
#' kplus_groups1 <- anticlustering(
#'   features,
#'   K = init_groups,
#'   objective = "kplus",
#'   standardize = TRUE,
#'   method = "local-maximum"
#' )
#'
#' kplus_groups2 <- anticlustering(
#'   kplus_moment_variables(features, T = 2), # standardization included by default
#'   K = init_groups,
#'   objective = "variance", # (!)
#'   method = "local-maximum"
#' )
#' 
#' # this function uses standardization by default unlike anticlustering():
#' kplus_groups3 <- kplus_anticlustering(
#'   features, 
#'   K = init_groups,
#'   method = "local-maximum"
#' )
#'
#' all(kplus_groups1 == kplus_groups2)
#' all(kplus_groups1 == kplus_groups3)
#' all(kplus_groups2 == kplus_groups3)
#' 

kplus_moment_variables <- function(x, T, standardize = TRUE) {
  validate_data_matrix(x)
  validate_input(T, "T", greater_than = 1, must_be_integer = TRUE, len = 1,
                 not_na = TRUE, not_function = TRUE)
  validate_input(standardize, "standardize", objmode = "logical", len = 1,
                 input_set = c(TRUE, FALSE), not_na = TRUE, not_function = TRUE)
  x <- as.matrix(x)
  M <- ncol(x)
  for (i in 2:T) {
    x <- cbind(x, moment_features(x, T))
  }
  colnames(x) <- paste0(colnames(x), "^", rep(1:T, each = M))
  if (!standardize) {
    return(x)
  }
  scale(x)
}

#' Exact weighted cluster editing
#'
#' Optimally solves weighted cluster editing (also known as »correlation clustering« or
#' »clique partitioning problem«). 
#'
#' @param x A N x N similarity matrix. Larger values indicate stronger
#'     agreement / similarity between a pair of data points
#' @param solver Optional argument; if passed, has to be either "glpk" or
#'   "symphony". See details.
#'   
#' @return An integer vector representing the cluster affiliation of each data point
#' 
#'
#' @examples
#' \donttest{
#' features <- swiss
#' distances <- dist(scale(swiss))
#' hist(distances)
#' # Define agreement as being close enough to each other.
#' # By defining low agreement as -1 and high agreement as +1, we
#' # solve *unweighted* cluster editing
#' agreements <- ifelse(as.matrix(distances) < 3, 1, -1)
#' clusters <- wce(agreements)
#' plot(swiss, col = clusters, pch = 19)
#' }
#' 
#' @export
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#' 
#' @details 
#' 
#' Finds the clustering that maximizes the sum of pairwise similarities within clusters. 
#' In the input some similarities should be negative (indicating dissimilarity) because 
#' otherwise the maximum sum of similarities is obtained by simply joining all elements 
#' within a single big cluster. The function uses a "solver" to optimize
#' the clustering objective. See \code{\link{optimal_anticlustering}}
#' for an overview of the solvers that are available.
#' 
#' 
#' @references
#'
#' 
#' Bansal, N., Blum, A., & Chawla, S. (2004). Correlation clustering. 
#' Machine Learning, 56, 89–113. 
#' 
#' Böcker, S., & Baumbach, J. (2013). Cluster editing. In Conference on 
#' Computability in Europe (pp. 33–44).
#' 
#' Grötschel, M., & Wakabayashi, Y. (1989). A cutting plane algorithm
#' for a clustering problem. Mathematical Programming, 45, 59-96.
#' 
#' Wittkop, T., Emig, D., Lange, S., Rahmann, S., Albrecht, M., Morris, J. H., ..., Baumbach,
#' J. (2010). Partitioning biological data with transitivity clustering. Nature Methods, 7,
#' 419–420. 
#'

wce <- function(x, solver = NULL) {

  if (argument_exists(solver)) {
    validate_input(solver, "solver", objmode = "character", len = 1,
                   input_set = c("glpk", "symphony", "lpSolve", "gurobi"), not_na = TRUE, not_function = TRUE)
  } else {
    solver <- find_ilp_solver()
  }

  validate_data_matrix(x)
  if (!is_distance_matrix(x)) {
    stop("The input via argument `weights` is not a similarity matrix, ",
         "the upper and lower triangulars of your matrix differ.")
  }
  
  x <- as.matrix(x)
  ilp <- anticlustering_ilp(x, K = 0, FALSE) # k is irrelevant
  solution <- solve_ilp(ilp, "max", solver)
  ilp_to_groups(solution, nrow(x))
}

# Use blocking for anticlustering (anticluster within a block, which a level of a cagegorical variable, 
# or a stratum). When working on subsequent blocks, it "remembers" the assignment of prior blocks
# when optimizing an anticlustering objective.

# including constraints (must-link /cannot-link) would require re-indexing, I will at first not allow this
# because it is additional work for me.

blocked_anticlustering <- function(
    x, K, objective = "diversity", method = "exchange", 
    categories = NULL, cannot_link = NULL, blocks
) {
  stopifnot(method %in% c("exchange", "local-maximum"))

  validate_input(K, "K", len = 1) # here we can only generate equal-sized groups
  
  N <- nrow(x)

  if (!argument_exists(blocks)) stop("Can only use blocked anticlustering when argument 'blocks' is given.")  
  blocks <- merge_into_one_variable(blocks) # use "strata"
  
  n_blocks <- length(unique(blocks))
  blocksizes <- table(blocks)

  # Anticlustering with blocking: 
  condition_blocked <- rep(NA, N)
  # different initialization needed if there are cannot-link constraints
  if (argument_exists(cannot_link)) {
    condition_blocked <- optimal_cannot_link_reduced(
      N = N, K = K, 
      target_groups = table(initialize_clusters(N = N, K = K, NULL)),
      cannot_link = cannot_link
    )
    x <- convert_to_distances(x, squared = objective == "variance")
    x[cleanup_cannot_link_indices(cannot_link)] <- -(sum(x) + 1) # edit distances to maintain cannot-link restrictions
  }
  for (i in 1:n_blocks) {
    previous_groups <- condition_blocked # just for asserting at the end that no previous assignments were changed
    select <- blocks <= i
    if (is_distance_matrix(x)) {
      input <- x[select, select, drop = FALSE] 
    } else {
      input <- x[select, , drop = FALSE] 
    }
    target_groups <- table(initialize_clusters(N = sum(select), K = K, NULL))
    initial_groups <- add_unassigned_elements(target_groups, condition_blocked[select], N = N, K = K)
    # ensure that previous conditions are still as before:
    stopifnot(all(condition_blocked[select] == initial_groups, na.rm = TRUE))
    # Now do anticlustering with restriction that previously assigned conditions must not change.
    # Define who can be exchanged during anticlustering optimization (goes into the `categories` argument)
    exchange_partners <- get_blocking_exchange_parters(condition_blocked, select)
    condition_blocked[select] <- anticlustering(
      input, 
      K = initial_groups, 
      objective = objective,
      method = method,
      categories = cbind(exchange_partners, categories[select])
    )
    # ensure that previous conditions are still as before:
    stopifnot(all(condition_blocked[blocks < i] == previous_groups[blocks < i], na.rm = TRUE)) 
  }
  
  # verify that output has correct structure: 
  tab <- table(blocks, condition_blocked)
  stopifnot(all(abs(tab[ ,1] - tab[, 2]) <= 1)) # this test only works if equal group were requested, which are currently allowed only
  condition_blocked
}

# Create input for categories argument in anticlustering(), that ensures that previously blocked subjects
# are not changed (only used to compute the anticlustering objective)
get_blocking_exchange_parters <- function(condition_blocked, select) {
  exchange_partners <- rep(1, sum(select))
  is_already_assigned <- !is.na(condition_blocked[select])
  groups_already_assigned <- condition_blocked[select][is_already_assigned]
  exchange_partners[is_already_assigned] <- sample(sum(is_already_assigned), size = sum(is_already_assigned)) + 1
  exchange_partners
}

#' Three phase search with dynamic population size heuristic
#'
#' This function implements the three phase search algorithm TPSPD for
#' anticlustering by Yang et al. (2022; <doi.org/10.1016/j.ejor.2022.02.003>).
#' The description of their algorithm is
#' given in Section 2 of their paper (in particular, see the
#' Pseudocode in Algorithm 1).
#' 
#' 
#' @param x The data input, as in \code{\link{anticlustering}}.
#' @param K Number of anticlusters to be formed.
#' @param N Number of elements.
#' @param objective The anticlustering objective, can be "diversity" or "dispersion".
#' @param number_iterations A number that defines how many times the steps in the search algorithm are repeated.
#' @param clusters A vector of length K that specifies the number of elements each cluster can contain. 
#' If this vector is not NULL, the lower and upper bounds will be disregarded.
#' @param beta_max The algorithm begins with a pool of random initial solutions of size beta_max. 
#'  Over time, the size of the solution pool decreases linearly until it reaches beta_min.
#' @param beta_min The minimum solution pool size the algorithm should reach before making a determination.
#' @param lower_bound Minimum number of elements in each anticluster. By default, anticlusters are of equal size,
#'  calculated as the total number of items divided by the number of clusters.
#' @param upper_bound Maximum number of elements in each anticluster. By default, anticlusters are of equal size,
#'  calculated as the total number of items divided by the number of clusters.
#' @param theta_max Parameter for the strength of undirected perturbation, 
#' which decreases linearly over time from theta_max to theta_min.
#' @param theta_min Parameter for the strength of undirected perturbation, 
#' which decreases linearly over time from theta_max to theta_min.
#' @param eta_max Parameter that specifies how many times the steps in the direct perturbation are executed.
#' @param alpha Parameter for weighing the discrimination of a slightly worse local optimal child solution.
#'     
#' @details Details of the implementation of the algorithm can be found 
#'  in the pseudocode of the paper Yang et al. (2022). However, we performed one change
#'  as compared to the original description of the algorithm: Instead of 
#'  setting a time limit, we define the number of iterations the algorithm 
#'  performs before terminating (via argument \code{number_iterations}).
#' 
#' @return A vector of length N that assigns a group (i.e, a number
#'     between 1 and \code{K}) to each input element
#' 
#' @author Hannah Hengelbrock \email{Hannah.Hengelbrock@@hhu.de}, 
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#' 
#' @export
#' 
#' @examples 
#' 
#' # Generate some random data
#' N <- 120
#' M <- 5
#' K <- 4
#' dat <- matrix(rnorm(N * M), ncol = M)
#' distances <- dist(dat)
#'
#' # Perform three hase serach algorithm
#' result1 <- three_phase_search_anticlustering(dat, K, N)
#'
#' # Compute objectives function
#' diversity_objective(distances, result1)
#' 
#' # Standard algorithm:
#' result2 <- anticlustering(distances, K=K, method="local-maximum", repetitions = 10)
#' diversity_objective(distances, result2)
#' 
#' 
#' @references
#' 
#' Xiao Yang et al. “A three-phase search approach with dynamic population size for solving 
#' the maximally diverse grouping problem”. In: European Journal of Operational Research
#' 302.3 (2022) <doi:10.1016/j.ejor.2022.02.003>
#' 
three_phase_search_anticlustering <- function(x, K, N, objective = "diversity",
    number_iterations=50, clusters = NULL, upper_bound  = NULL, lower_bound  = NULL, 
    beta_max = 15,  theta_max = NULL, theta_min = NULL, beta_min = NULL, eta_max=3, alpha=0.05) {

    input_validation_threephase_search(x, K, N, objective, clusters, number_iterations, upper_bound, 
    theta_max, theta_min, lower_bound, beta_max, beta_min, eta_max, alpha)
    
    distances <- convert_to_distances(x) 

    if (is.null(lower_bound)) {
       lower_bound <- floor(N/K)
    } 
    if (is.null(upper_bound)) {
       upper_bound <- ceiling(N/K)
    }   
    
    if (N <= 400  & is.null(theta_max) & is.null(theta_min) & is.null(beta_min)) {
    	theta_max  <- 1.2
    	theta_min  <- 0.1
    	beta_min  <- 2
    } else if ( is.null(theta_max) & is.null(theta_min) & is.null(beta_min)) {
    	theta_max  <- 2.0
    	theta_min  <- 1.0
    	beta_min  <- 1
    }

     # create result vector for results to use in C
    result_vector <- numeric(N)
    
    # If clusters are not predefined, set the value to -1 so the C implementation knows 
    # to evenly distribute the number of elements based on K and the boundaries.
    if (is.null(clusters)) {
      clusters <- rep(-1, K)
    }
     
     if (objective == "diversity") {
     results <- .C("three_phase_search_dynamic_population_size",
                  distances = as.double(distances),
                  N_in = as.integer(N),
                  K_in = as.integer(K),
                  number_of_iterations = as.integer(number_iterations),
                  clusters = as.integer(clusters),
                  upper_bound = as.integer(upper_bound),
                  lower_bound = as.integer(lower_bound),
                  Beta_max = as.integer(beta_max),
                  elapsed_time = as.integer(0),
                  Theta_max = as.double(theta_max),
                  Theta_min = as.double(theta_min),
                  Beta_min = as.integer(beta_min),
                  Eta_max = as.integer(eta_max),
                  Alpha = as.double(alpha),
                  result = as.integer(result_vector),
                  score = as.double(0.0),
                  mem_error = as.integer(0),
                  PACKAGE = "anticlust"
                  )
     } else if (objective == "dispersion") {
      results <- .C("three_phase_search_dispersion",
                  distances = as.double(distances),
                  N_in = as.integer(N),
                  K_in = as.integer(K),
                  number_of_iterations = as.integer(number_iterations),
                  clusters = as.integer(clusters),
                  upper_bound = as.integer(upper_bound),
                  lower_bound = as.integer(lower_bound),
                  Beta_max = as.integer(beta_max),
                  elapsed_time = as.integer(0),
                  Theta_max = as.double(theta_max),
                  Theta_min = as.double(theta_min),
                  Beta_min = as.integer(beta_min),
                  Eta_max = as.integer(eta_max),
                  Alpha = as.double(alpha),
                  result = as.integer(result_vector),
                  score = as.double(0.0),
                  mem_error = as.integer(0),
                  PACKAGE = "anticlust"
                  )
     } else {
      stop("Objective funtion for TPSDP is not defined.")
     }

     results[["mem_error"]]
     if (results[["mem_error"]] == 1) {
       stop("Could not allocate enough memory.")
     }

    return(results$result + 1)
}

input_validation_threephase_search <- function(x, K, N, objective, clusters, number_iterations, upper_bound, lower_bound, 
theta_max, theta_min, beta_max, beta_min, eta_max, alpha) {

    # cluster vector
    if (!is.null(clusters)) {
    validate_input(clusters, "clusters", len = K)
    }
      # Objective
    validate_input(
    objective, "objective", objmode = "character", len = 1,
    input_set = c("diversity", "dispersion"), not_na = TRUE,  not_function = TRUE
    )
    validate_input(K, "K",  must_be_integer = TRUE, not_na = TRUE)
    validate_input(lower_bound, "lower_bound", greater_than = 0, must_be_integer = TRUE)
    validate_input(upper_bound, "upper_bound", greater_than = 0, must_be_integer = TRUE)
    validate_input(beta_max, "beta_max", greater_than = 0, must_be_integer = TRUE)
    validate_input(beta_min, "beta_min", greater_than = 0, must_be_integer = TRUE)
    validate_input(eta_max, "eta_max", greater_than = 0, must_be_integer = TRUE)
    validate_input(theta_max, "theta_max", greater_than = 0.0)
    validate_input(theta_min, "theta_min", greater_than = 0.0)
    validate_input(eta_max, "eta_max", greater_than = 0, must_be_integer = TRUE)
    validate_input(alpha, "alpha", greater_than = 0.0)
}

#' Random sampling employing a categorical constraint
#' 
#' This function can be used to obtain a stratified split of a data set.
#'
#' @param categories A matrix or vector of one or more categorical variables.
#' @param K The number of groups that are returned. 
#'
#' @return A vector representing the sample each element was assigned to.
#' 
#' @details 
#' 
#' This function can be used to obtain a stratified split of a data set. 
#' Using this function is like calling \code{\link{anticlustering}} with 
#' argument `categories`, but without optimizing a clustering objective. The
#' categories are just evenly split between samples. Apart from the restriction 
#' that categories are balanced between samples, the split is random.
#'
#' @importFrom stats rbinom
#'
#' @export
#' 
#' @examples 
#' 
#' data(schaper2019)
#' categories <- schaper2019$room
#' groups <- categorical_sampling(categories, K = 6)
#' table(groups, categories)
#' 
#' # Unequal sized groups
#' groups <- categorical_sampling(categories, K = c(24, 24, 48))
#' table(groups, categories)
#' 
#' # Heavily unequal sized groups, is harder to balance the groups
#' groups <- categorical_sampling(categories, K = c(51, 19, 26))
#' table(groups, categories)
#' 

categorical_sampling <- function(categories, K) {
  categories <- merge_into_one_variable(categories)
  N <- length(categories)
  validate_input(K, "K", objmode = "numeric", 
                 must_be_integer = TRUE, not_na = TRUE)

  if (length(K) == N) {
    K <- table(K)
  }
  
  cats <- data.frame(
    categories = categories,
    order = 1:N
  )
  cats <- sort_by_col(cats, "categories")
  if (sum(K) == N && length(K) > 1) { # Unequal group sizes were requested
    gdc <- gcd_set(K)
    if (gdc > 1) {
      # if the GDC of all categories is > 1, we can use this procedure:
      K <- K / gdc
      init <- rep_len(rep(1:length(K), K), N)
      samples <- unlist(by(init, cats$categories, sample_))
    } else {
      samples <- generate_groups(K)
    }
  } else {
    validate_input(K, "K", objmode = "numeric", len = 1, 
                   greater_than = 1, smaller_than = N)
    init <- rep_len(1:K, N)
    samples <- unlist(by(init, cats$categories, sample_))
  } 
  cats$samples <- samples
  sort_by_col(cats, "order")$samples
}

# sample function must be redefined for categorical sampling just in case
# there is a category with only 1 member
sample_ <- function(x, ...) {
  if (length(x) == 1) {
    return(x)
  }
  sample(x, ...)
}



# Generate groups in case of heaviliy uneven group size requirements

# Outer function repeatedly calls generate_groups_() until constraints
# on group sizes are met exactly

generate_groups <- function(tab) {
  loop <- TRUE
  while (loop) {
    samples <- generate_groups_(tab)
    # loop until the number of categories fits the requirements
    loop <- any(sort(table(samples)) != sort(tab))
  }
  samples
}

# Generate group affiliations for all N cases, may not exactly fit 
# the requirements
generate_groups_ <- function(tab) {
  N <- sum(tab) 
  k <- length(tab)
  samples <- unlist(
    replicate(
      N / sum(floor(tab / min(tab))), 
      generate_groups_one_set(tab)
    )
  )
  samples[1:sum(tab)]
}

# Generate one set of group affiliations
generate_groups_one_set <- function(tab) {
  k <- length(tab)
  proportions <- tab / min(tab)
  has_to_be_in <- floor(proportions)
  samples <- rep(1:k, has_to_be_in)
  add <- which(as.logical(rbinom(k, 1, proportions - has_to_be_in)))
  sample(c(samples, add))
}

# Functions to find greatest common denominator among a set of values
gcd <- function(x, y) ifelse(y, Recall(y, x %% y), x)
gcd_ <- function(x) gcd(x[1], x[2])

gcd_all <- function(x) {
  apply(combn(x, 2), 2, gcd_)
}

gcd_set <- function(x) {
  x <- unique(x)
  gdc_found <- FALSE 
  while (!gdc_found) {
    x <- unique(gcd_all(x))
    if (all(x == x[1])) {
      gdc_found <- TRUE
    }
    if (any(x == 1)) { # cannot find a greatest common denominator!
      return(1)
    }
  }
  x[1]
}

library("anticlust")

# test that function for greatest common denominator works correctly
x <- c(20, 20, 40, 40, 60, 60, 80, 80)
expect_equal(anticlust:::gcd_set(x), 20)

x <- 1:5
expect_equal(anticlust:::gcd_set(x), 1)

x <- c(9, 9, 9, 5)
expect_equal(anticlust:::gcd_set(x), 1)

x <- c(2, 8, 10, 10, 10, 10, 20, 20, 30, 30, 30, 1000, 1000, 10000000000)
expect_equal(anticlust:::gcd_set(x), 2)

x <- seq(3, 999, by = 3)
expect_equal(anticlust:::gcd_set(x), 3)

x <- c(1, seq(3, 999, by = 3))
expect_equal(anticlust:::gcd_set(x), 1)

#' K-plus anticlustering
#' 
#' Perform anticlustering using the k-plus objective to maximize between-group 
#' similarity. This function implements the k-plus anticlustering method described 
#' in Papenberg (2024; <doi:10.1111/bmsp.12315>).
#'
#' @param x A feature matrix where rows correspond to elements and columns
#'     correspond to variables (a single numeric variable can be
#'     passed as a vector).
#' @param K How many anticlusters should be created. Alternatively:
#'     (a) A vector describing the size of each group, or (b) a vector
#'     of length \code{nrow(x)} describing how elements are assigned
#'     to anticlusters before the optimization starts.
#' @param variance Boolean: Should the k-plus objective include a term to 
#'   maximize between-group similarity with regard to the variance? 
#'   (Default = TRUE)
#' @param skew Boolean: Should the k-plus objective include a term to 
#'   maximize between-group similarity with regard to skewness? 
#'   (Default = FALSE)
#' @param kurtosis Boolean: Should the k-plus objective include a term to 
#'   maximize between-group similarity with regard to kurtosis? 
#'   (Default = FALSE)
#' @param covariances Boolean: Should the k-plus objective include a term to 
#'   maximize between-group similarity with regard to covariance structure? 
#'   (Default = FALSE)
#' @param T Optional argument: An integer specifying how many
#'   distribution moments should be equalized between groups. 
#' @param standardize Boolean. If \code{TRUE}, the data is standardized through 
#'     a call to \code{\link{scale}} before the optimization starts. 
#'     Defaults to TRUE. See details.
#' @param ... Arguments passed down to \code{\link{anticlustering}}. All of the 
#'   arguments are supported except for \code{objective}. 
#' 
#' @details 
#'     
#'     This function implements the unweighted sum approach for k-plus 
#'     anticlustering. Details are given in Papenberg (2024). 
#'     
#'     The optional argument \code{T} denotes the number of distribution 
#'     moments that are considered in the anticlustering process. For example,
#'     \code{T = 4} will lead to similar means, variances, skew and kurtosis. 
#'     For the first four moments, it is also possible to use the boolean
#'     convenience arguments \code{variance}, \code{skew} and \code{kurtosis}; the
#'     mean (the first moment) is always included and cannot be "turned off".
#'     If the argument \code{T} is used, it overrides the arguments
#'     \code{variance}, \code{skew} and \code{kurtosis} (corresponding to
#'     the second, third and fourth moment), ignoring their values.
#'     
#'     The \code{standardization} is applied to all original features and the 
#'     additional k-plus features that are appended to the data set in order 
#'     to optimize the k-plus criterion. When using standardization, 
#'     all criteria such as means, variances and skewness receive a comparable
#'     weight during the optimization. It is usually recommended not
#'     to change the default setting \code{standardization = TRUE}.
#'    
#'     This function can use any arguments that are also possible in 
#'     \code{\link{anticlustering}}
#'     (except for `objective` because the objective optimized here
#'     is the k-plus objective; to use a different objective, 
#'     call \code{\link{anticlustering}} directly). Any arguments that are
#'     not explicitly changed here (i.e., \code{standardize = TRUE}) receive the 
#'     default given in \code{\link{anticlustering}} 
#'     (e.g., \code{method = "exchange"}.)
#'     
#' @examples 
#' 
#' # Generate some data
#' N <- 180
#' M <- 4
#' features <- matrix(rnorm(N * M), ncol = M)
#' # standard k-plus anticlustering: optimize similarity with regard to mean and variance:
#' cl <- kplus_anticlustering(features, K = 3, method = "local-maximum")
#' mean_sd_tab(features, cl)
#' # Visualize an anticlustering solution:
#' plot(features, col = palette()[2:4][cl], pch = c(16:18)[cl])
#' 
#' # Also optimize with regard to skewness and kurtosis
#' cl2 <- kplus_anticlustering(
#'   features, 
#'   K = 3, 
#'   method = "local-maximum", 
#'   skew = TRUE, 
#'   kurtosis = TRUE
#' )
#' 
#' # The following two calls are equivalent: 
#' init_clusters <- sample(rep_len(1:3, nrow(features)))
#' # 1.
#' x1 <- kplus_anticlustering(
#'   features, 
#'   K = init_clusters, 
#'   variance = TRUE,
#'   skew = TRUE
#' )
#' # 2. 
#' x2 <- kplus_anticlustering(
#'   features, 
#'   K = init_clusters, 
#'   T = 3
#' )
#' # Verify: 
#' all(x1 == x2)
#' 
#' @export
#' 
#' @references
#' 
#' Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for 
#' Maximizing Between-Group Similarity. British Journal of Mathematical and 
#' Statistical Psychology, 77(1), 80--102. https://doi.org/10.1111/bmsp.12315
#' 


kplus_anticlustering <- function(
    x, K, 
    variance = TRUE,
    skew = FALSE,
    kurtosis = FALSE,
    covariances = FALSE,
    T = NULL,
    standardize = TRUE,
    ...) {
  
  validate_input_kplus(x, K, variance, skew, kurtosis, covariances, T, ...)
  
  x <- as.matrix(x)
  M <- ncol(x) 
  feature_bools <- c(variance, skew, kurtosis)
  
  # Unfortunately, I have to extract all ellipsis arguments by hand because
  # preclustering / categories must be used here before calling anticlustering()...
  # (I know this is really ugly...)
  preclustering <- get_argument_from_ellipsis("preclustering", ...)
  preclustering <- ifelse(is.null(preclustering), FALSE, preclustering) # default = FALSE
  categories <- get_argument_from_ellipsis("categories", ...) # default = NULL
  categories <- get_categorical_constraints(x, K, preclustering, categories)
  method <- get_argument_from_ellipsis("method", ...) 
  method <- ifelse(is.null(method), "exchange", method) # default = "exchange"
  repetitions <- get_argument_from_ellipsis("repetitions", ...) # default = NULL

  if (is.null(T)) {
    moments <- (2:4)[feature_bools]
  } else {
    moments <- 2:T
  }
  # determine the number of features after augmentation, for validation below
  M_augmented <- M + M * length(moments) + ifelse(covariances, choose(M, 2), 0)

  # Add additional k-plus features: 
  #   1. Moments
  augmented_data <- x
  for (i in seq_along(moments)) {
    augmented_data <- cbind(augmented_data, moment_features(x, moments[i]))
  }
  #   2. Covariance    
  if (covariances == TRUE) {
    augmented_data <- cbind(augmented_data, covariance_features(x))
  }
  
  # Validation of number of features
  stopifnot(ncol(augmented_data) == M_augmented)
  
  # Call anticlustering
  anticlustering(
    augmented_data, 
    K = K, 
    standardize = standardize, 
    objective = "variance", 
    categories = categories,
    method = method,
    repetitions = repetitions
  )
}

# function to compute features for variance
# moment = 2 -> variance; 3 = skew; 4 = kurtosis
moment_features <- function(data, moment = 2) {
  apply(data, 2, function(x) (x - mean(x, na.rm = TRUE))^moment)
}

# Compute k-covariances features for an original number of M features.
# That is choose(M, 2) additional features, i.e., O(M^2) additional features.

covariance_features <- function(data) {
  M <- ncol(data)
  if (M < 2) {
    stop("k-covariances can only be used when at least two features are present.")
  }
  feature_combinations <- t(combn(M, 2))
  # store new features in matrix
  cov_feature_matrix <- matrix(ncol = nrow(feature_combinations), nrow = nrow(data))
  for (i in 1:nrow(feature_combinations)) {
    f1 <- feature_combinations[i, 1]
    f2 <- feature_combinations[i, 2]
    cov_feature_matrix[, i] <- (data[, f1] - mean(data[, f1], na.rm = TRUE)) * (data[, f2] - mean(data[, f2], na.rm = TRUE))
  }
  cov_feature_matrix
}

#' Get argument from ellipsis
#' @param argument Name of the argument
#' @param ... The list of additional arguments
#' @examples
#' 
#' get_argument_from_ellipsis(argument = "moep", objective = 1:4)
#' get_argument_from_ellipsis(argument = "objective", objective = 1:4)
#' 
#' @noRd
#' 

get_argument_from_ellipsis <- function(argument, ...) {
  args <- as.list(substitute(list(...)))[-1L]
  if (!argument %in% names(args)) {
    return(NULL)
  }
  eval(args[[argument]])
}


validate_input_kplus <- function(x, K, variance, skew, kurtosis, covariances, T, standardize, ...) {
  validate_data_matrix(x)
  # K is validated in anticlustering()
  validate_input(variance, "variance", len = 1,
                 input_set = c(TRUE, FALSE), not_na = TRUE, not_function = TRUE)
  validate_input(skew, "skew", len = 1,
                 input_set = c(TRUE, FALSE), not_na = TRUE, not_function = TRUE)
  validate_input(kurtosis, "kurtosis", len = 1,
                 input_set = c(TRUE, FALSE), not_na = TRUE, not_function = TRUE)
  validate_input(covariances, "covariances", len = 1,
                 input_set = c(TRUE, FALSE), not_na = TRUE, not_function = TRUE)
  if (!is.null(T)) {
    validate_input(T, "T", greater_than = 1, must_be_integer = TRUE, len = 1,
                   not_na = TRUE, not_function = TRUE)
  }
  elipsis_arg_names <- names(lapply(substitute(list(...))[-1], deparse))
  if ("objective" %in% elipsis_arg_names) {
    stop("You cannot set the `objective` argument - this is k-plus anticlustering!",
    " \n(-> k-plus is the objective, an extension to k-means anticlustering, see the documenation).")
  }
}

#' Sort a matrix or data frame by one column
#'
#' @param x A matrix or data frame
#' @param col An numeric index or character vector of length 1 to select
#'     the column by which the table is sorted.
#' @param decreasing logical. Should the sort order be increasing or
#'     decreasing?
#'
#' @return The matrix or data frame sorted by `col`
#'
#' @noRd
sort_by_col <- function(x, col, decreasing = FALSE) {
  if (length(col) != 1)
    stop("False input, can only select by one column")
  x[order(x[, col], decreasing = decreasing), ]
}

library("anticlust")

# data table sorting works
mat <- cbind(10:1, 1:10)
colnames(mat) <- c("one", "two")
expect_equal(mat[, 1], 10:1)
expect_equal(mat[, 2], 1:10)
expect_equal(anticlust:::sort_by_col(mat, 1)[, 1], 1:10)
expect_equal(anticlust:::sort_by_col(mat, 1)[, "one"], 1:10)
expect_equal(anticlust:::sort_by_col(mat, 1)[, 2], 10:1)
expect_equal(anticlust:::sort_by_col(mat, 1)[, "two"], 10:1)
expect_error(anticlust:::sort_by_col(mat, 1:2))
expect_error(anticlust:::sort_by_col(mat, c(TRUE, FALSE)))
expect_error(anticlust:::sort_by_col(mat, c("TRUE", "FALSE")))
expect_error(anticlust:::sort_by_col(1, 1))

#' Means and standard deviations by group variable formatted in table
#' 
#' @param features A data frame of features
#' @param groups A grouping vector
#' @param decimals The number of decimals
#' @param na.rm Should NAs be removed prior to computing stats 
#'     (Default = FALSE)
#' @param return_diff Boolean. Should an additional row be printed that 
#'     contains the difference between minimum and maximum
#' 
#' @return A table that illustrates means and standard deviations (in brackets)
#' 
#' @examples 
#'
#' data(iris)
#' mean_sd_tab(iris[, -5], iris[, 5])
#' 
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#' 
#' @export

mean_sd_tab <- function(features, groups, decimals = 2, na.rm  = FALSE, 
                        return_diff = FALSE) {
  K <- length(unique(groups))
  # Means
  means <- by(features, groups, colMeans, na.rm = na.rm)
  rows <- names(means)
  means <- t(matrix(unlist(means), ncol = K))
  mode(means) <- "numeric"
  ncols <- K
  if (return_diff) {
    rows <- c(rows, "Diff")
    ncols <- K + 1
    means <- rbind(means, apply(means, 2, range_diff, na.rm = na.rm))
  }
  means <- apply(means, 2, force_decimals, decimals = decimals)
  # SDS
  sds <- by(features, groups, function(x) apply(x, 2, sd, na.rm = na.rm))
  sds <- t(matrix(unlist(sds), ncol = K))
  mode(sds) <- "numeric"
  if (return_diff) {
    sds <- rbind(sds, apply(sds, 2, range_diff, na.rm = na.rm))
  }
  sds <- apply(sds, 2, force_decimals, decimals = decimals)
  # Merge Means and SDS
  values <- paste0(paste(means, sds, sep = " ("), ")")
  dim(values) <- c(ncols, length(means) / (ncols))
  colnames(values) <- colnames(features)
  rownames(values) <- rows
  values
}

# Determine the maximum range in a vector
range_diff <- function(x, na.rm) {
  x <- round(x, 2) # so that it is consistent in the table
  diff(range(x, na.rm = na.rm))
}

#' Force printing a specified number of decimals for a number
#'
#' @param x the numeric values to be printed
#' @param decimals how many decimals are to be printed. Defaults to 2.
#'
#' @return The number in the required format
#'
#' @examples
#' 
#' force_decimals(c(1.23456, 0.873, 2.3456))
#' force_decimals(c(1.23456, 0.873, 2.3456), 3)
#' 
#' @author Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' @noRd
force_decimals <- function(x, decimals = 2) {
  format(round(x, decimals), nsmall = decimals, scientific = FALSE)
}

#' Create balanced clusters of equal size
#'
#' @param x The data input. Can be one of two structures: (1) A feature
#'     matrix where rows correspond to elements and columns correspond
#'     to variables (a single numeric variable can be passed as a
#'     vector). (2) An N x N matrix dissimilarity matrix; can be an
#'     object of class \code{dist} (e.g., returned by
#'     \code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
#'     where the entries of the upper and lower triangular matrix
#'     represent pairwise dissimilarities.
#' @param K How many clusters should be created.
#' @param method One of "centroid" or "ilp". See Details.
#' @param solver Optional. The solver used to obtain the optimal method 
#'     if \code{method  = "ilp"}. Currently supports "glpk" and "symphony". 
#'     Is ignored for \code{method = "centroid"}.
#'
#' @return An integer vector representing the cluster affiliation of 
#'     each data point
#'
#' @details
#'
#' This function partitions a set of elements into \code{K}
#' equal-sized clusters. The function offers two methods: a heuristic
#' and an exact method. The heuristic (\code{method = "centroid"})
#' first computes the centroid of all data points. If the input is a
#' feature matrix, the centroid is defined as the mean vector of all
#' columns. If the input is a dissimilarity matrix, the most central
#' element acts as the centroid; the most central element is defined
#' as the element having the minimum maximal distance to all other
#' elements. After identifying the centroid, the algorithm proceeds as
#' follows: The element having the highest distance from the centroid
#' is clustered with its \code{(N/K) - 1} nearest neighbours
#' (neighbourhood is defined according to the Euclidean distance if
#' the data input is a feature matrix). From the remaining elements,
#' again the element farthest to the centroid is selected and
#' clustered with its \code{(N/K) - 1} neighbours; the procedure is
#' repeated until all elements are part of a cluster.
#'
#' An exact method (\code{method = "ilp"}) can be used to solve
#' equal-sized weighted cluster editing optimally (implements the
#' integer linear program described in Papenberg and Klau, 2020; 
#' (8) - (10), (12) - (13)). The cluster editing objective is the 
#' sum of pairwise distances
#' within clusters; clustering is accomplished by minimizing this
#' objective. If the argument \code{x} is a features matrix, the
#' Euclidean distance is computed as the basic unit of the cluster
#' editing objective. If another distance measure is preferred, users
#' may pass a self-computed dissimiliarity matrix via the argument
#' \code{x}. 
#' 
#' The optimal \code{method = "ilp"} uses a "solver" to optimize
#' the clustering objective. See \code{\link{optimal_anticlustering}}
#' for an overview of the solvers that are available.
#'
#' @source
#'
#' The centroid method was originally developed and contributed by
#' Meik Michalke. It was later rewritten by Martin Papenberg, who
#' also implemented the integer linear programming method.
#'
#' @export
#'
#' @author
#' Martin Papenberg \email{martin.papenberg@@hhu.de}
#'
#' Meik Michalke \email{meik.michalke@@hhu.de}
#'
#' @examples
#'
#' # Cluster a data set and visualize results
#' N <- 1000
#' lds <- data.frame(f1 = rnorm(N), f2 = rnorm(N))
#' cl <- balanced_clustering(lds, K = 10)
#' plot_clusters(lds, clusters = cl)
#' 
#' # Repeat using a distance matrix as input
#' cl2 <- balanced_clustering(dist(lds), K = 10)
#' plot_clusters(lds, clusters = cl2)
#'
#' @references
#'
#' Grötschel, M., & Wakabayashi, Y. (1989). A cutting plane algorithm
#' for a clustering problem. Mathematical Programming, 45, 59–96.
#'
#' Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
#' data sets into equivalent parts. Psychological Methods, 26(2), 
#' 161–174. https://doi.org/10.1037/met0000301.
#'

balanced_clustering <- function(x, K, method = "centroid", solver = NULL) {

  validate_data_matrix(x, FALSE)
  validate_input_optimal_anticlustering(x, K, "diversity", solver, NULL)
  validate_input(
    method, 
    "method", 
    len = 1, 
    input_set = c("centroid", "ilp"), 
    not_na = TRUE
  )
  data <- to_matrix(x)
  
  if (method == "ilp") {
    solver <- ifelse(argument_exists(solver), solver, find_ilp_solver())
    return(balanced_cluster_editing(data, K, solver))
  }
  nn_centroid_clustering(data, NROW(data) / K)
}

library(tinytest)
library(anticlust)

N <- 12
M <- 5
K <- 2
dat <- matrix(rnorm(N * M), ncol = M)

# just make sure that all runs okay
balanced_clustering(dat, K, method = "ilp", solver = "glpk")

expect_error(
  balanced_clustering(dat, K = 5, method = "ilp", solver = "glpk")
)

expect_error(
  balanced_clustering(dat, K = c(2, 2, 2, 6), method = "ilp", solver = "glpk")
)

expect_error(
  balanced_clustering(dat, K = 5, method = "ilp", solver = "FOO")
)

---
title: "Anticlustering in 2025"
output: rmarkdown::html_vignette
author: Martin Papenberg
vignette: >
  %\VignetteIndexEntry{Anticlustering_in_2025}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
set.seed(1)

```

```{r setup}
library(anticlust)
```

This vignette is (in October 2025) the most up-to-date reference for doing anticlustering. Here, I summarize some of the things I learned about---or implemented for---anticlustering, which may not be documented in earlier work (e.g., papers, other vignettes). For example, if you have used `anticlust` before, you may be interested to know about recent changes, available from version 0.8.12 onward: 

- Categorical variables can now be directly included as the data input when defined as `factor`s (which is R's data structure for categorical variables)
- Missing values are now handled directly by `anticlustering()`

I will use these changes in the running example below. Of course, in both cases (categorical variables and `NA`s), a more specialized handling could be preferred in certain settings; defaults can never be optimal for all applications. 

Moreover, we have implemented the three phase search algorithm by [Yang et al.](https://doi.org/10.1016/j.ejor.2022.02.003), which seems to be the most powerful algorithm for anticlustering that we have included so far. I will showcase it below. 

To reproduce the running example, install the "MASS" package that contains the `survey` data set that is used.

```{r}
library(MASS)
data(survey)       # load data set
nrow(survey)       # number of students
head(survey, n=10) # look at the data
sapply(survey, anyNA) # most variables contain missing values
``` 

The `survey` data set contains some demographic information on a student sample. Imagine you wish to assign students to three different dormitories and want to create a similar groups of students in each house. In this example I will use the following features (brackets contain the description contained in the R documentation, `?MASS::survey`):

**Categorical features:**

- `Sex`: The sex of the student (Factor with levels "Male" and "Female".)
- `W.Hnd`: writing hand of student. (Factor, with levels "Left" and "Right".)
- `Exer`: how often the student exercises. (Factor, with levels "Freq" (frequently), "Some", "None".)
- `Smoke`: how much the student smokes. (Factor, levels "Heavy", "Regul" (regularly), "Occas" (occasionally), "Never".)

All of the categorical features are already defined as factor variables, which I will make use of. 

**Numeric features:**

- `Pulse`: pulse rate of student (beats per minute).
- `Height`: height of the student in centimetres.
- `Age`: age of the student in years.

I will select these features as a new data frame and apply the default anticlustering method to create 3 equal-sized groups.

```{r}

features <- c("Sex", "W.Hnd", "Exer", "Smoke", "Pulse", "Height", "Age")

survey$House <- anticlustering(
  survey[, features], 
  K = 3
)

```

I use the `tableone` package to illustrate the descriptive statistics by group. For numeric variables, we get means and standard deviations; for categorical variables, we get counts and percentages. It even gives a *p*-value as a (sort of) measure of imbalance among houses for each variable. For categorical variables, the *p*-value is the result of a chi squared test. Higher values indicate that the observed distribution of categories among houses is more uniform. For numeric variables, the `oneway.test` is used, which compares 2 or several means (for 2 means, it is equivalent to the *t*-test). Higher *p*-values indicate less discrepancy between group means (however, other characteristics such as the standard deviation are not considered).

```{r}
library(tableone)

CreateTableOne(features, strata = "House", data = survey)

```

By default, the `anticlustering()` method maximizes the `diversity` objective, which strives for a similar overall data distribution among groups. It is difficult to assess the adequateness of the solution without a comparison. So let us first add an random assignment as baseline comparison: 

```{r}
survey$Rnd_House <- sample(survey$House)

CreateTableOne(features, strata = "Rnd_House", data = survey)

```

The *p*-values tend to be lower, and the *p*-value for height would even be considered "significant" in an inference testing setting (however, note that such an interpretation is not appropriate in this setting). So, apparently, anticlustering helped to form a more even observed distribution of the input variables among houses. 

## Improving the results

I think that the result from the "default" anticlustering method are okay'ish. However, let's assume we want to ensure an even more optimized assignment where the variables are more even among houses. Here I list three options:

- Use standardization (**always recommended**)
- Use better algorithm (**usually recommended**)
- Change the objective function (**recommended depending on requirements of application**)

### Standardization

A first step to improve results is using standardization. It always should be used (especially when using categorical and numeric variables), but as discussed elsewhere, it is not the default. Let's repeat our previous code but now use standardization with the `standardize` argument. 

```{r}
survey$House2 <- anticlustering(
  survey[, features], 
  K = 3,
  standardize = TRUE
)

CreateTableOne(features, strata = "House2", data = survey)

```

Standardization has a remarkable effect, in particular on the categorical variables. So you definitely should use it: It ensures that all variables are equally important when conducting anticlustering. 

### Using a better algorithm

The `method` argument determines the algorithm, which conducts the anticlustering optimization. The default is `exchange`, which is a compromise between speed and solution quality. In many applications, it is possible to use a better algorithm because speed is not an issue. Using method = `local-maximum` basically maintains the same algorithm, but runs longer until a so-called local maximum is reached. It would be a good next step. However, here I directly use the tree phase search algorithm, which is new in `anticlust` and outperforms the `local-maximum` method. Sometimes, it can even be faster. 

```{r}
survey$House3 <- anticlustering(
  survey[, features], 
  K = 3,
  standardize = TRUE,
  method = "3phase"
)
CreateTableOne(features, strata = "House3", data = survey)
```

It seems that using the improved algorithm also translated into a more even distribution of variables among groups. 

### Changing the objective function

In some settings we may be interests in optimizing similarity with regard to certain distribution characteristics among groups. One possibility is to focus on the average of each variable. Though I strongly discourage focusing on mean values alone (e.g., see Papenberg, 2024), let's discuss here how it could be done. To just focus on mean values, we can replace the default diversity objective by the *k*-means criterion (also called the *variance*). The objective function is set via the `objective` argument. It defines how the similarity between groups is computed and it is the so called optimization criterion; the anticlustering algorithm proceeds in such a way that the grouping maximizes the objective function. 

```{r}
survey$House4 <- anticlustering(
  survey[, features], 
  K = 3,
  standardize = TRUE,
  method = "3phase",
  objective = "variance"
)
CreateTableOne(features, strata = "House4", data = survey)

```

Overall, using the variance objective does not really seem to improve results. Maybe, the mean values for `Pulse` are more evenly distributed. Additionally, we can see that the standard deviations for some of the numeric variables (in particular, `Pulse`) are highly dissimilar. Thus, the distribution of this variable is not similar between groups even though the means are similar: Group 1 has little spread around the mean and thus consists of "average" students regarding pulse; group 3 has a large spread around the mean and therefore tends to consist of students with either very high or very low pulse. This even can be illustrated visually, when using house affiliation to color the Pulse values: 

```{r}

colors <- c("#a9a9a9", "#df536b", "#61d04f")

ord <- order(survey$Pulse)
# Plot the data while visualizing the different clusters
plot(
  survey$Pulse[ord], 
  col = colors[survey$House4[ord]], 
  pch = 19, 
  ylab = "Pulse", 
  xlab = "Students (ordered by pulse)"
)
legend("bottomright", legend = paste("Group", 1:3), col = colors, pch = 19)

```

This plot is an astonishing illustration why the *k*-means objective can be problematic (and why you should not only rely on *p* values to assess whether the distribution of numeric variables differs between groups). Admittedly, this problem also occurred with the previous solution that used the default diversity objective. However, it is more likely to occur with the *k*-means approach.

Let's use the *k*-plus objective that tries to correct for the mean/SD problem by trying to match both means and standard deviations of the input variables:

```{r}
survey$House5 <- anticlustering(
  survey[, features], 
  K = 3,
  method = "3phase",
  objective = "kplus",
  standardize = TRUE
)
CreateTableOne(features, strata = "House5", data = survey)

```

This solution has the most even distribution for the pulse variable of all our solutions. However, it is slightly worse with regard to balance in the handedness variable. This shows that there is oftentimes no unique best solution, and there are tradeoffs to consider. For example, using more variables makes the problem harder on an algorithmic level. We might ask if it is necessary to balance handedness between houses? If not, we may leave out the variable? Or is some loss in balance acceptable for some variables? In this application, I would argue that the imbalance in handedness is not severe (right handedness differs between 93.7% and 89.7% between houses). However, the imbalance observed in the standard deviation of the pulse variable---that we obtained when using the diversity and *k*-means objectives---could be considered severe. So we might want to stick with the *k*-plus solution. In an actual application this decision is up to the user, it cannot be done automatically or optimally by an algorithm. You should try out different approaches and use the one which is best for your case. 

### Unequal group sizes

Let's assume we have three different sized houses. We can incorporate this using the `K` argument by describing the size of each group. Some things are to consider with this change in requirements. First, let's just repeat the anticlustering assignment using the default diversity objective, but using different group sizes. 

```{r}
survey$House6 <- anticlustering(
  survey[, features], 
  K = c(137, 50, 50),
  standardize = TRUE,
  method = "3phase"
)
CreateTableOne(features, strata = "House6", data = survey)
```

Wow, this is really bad! We certainly would not want such an assignment. For the categorical variables, we now have strongly different distributions. For the numeric variables, mean values are similar between groups, but the standard deviations are highly different. Why did this happen? **The diversity is actually not a good measure of between-group similarity when group sizes are unequal.** We could use the "average diversity", which takes into consideration group sizes. Currently, the three phase algorithm only supports the diversity objective for unequal-sized groups, so we must choose a "classical" algorithm:

```{r}
survey$House7 <- anticlustering(
  survey[, features], 
  K = c(137, 50, 50),
  standardize = TRUE,
  method = "local-maximum",
  repetitions = 10, # increasing repetitions may be helpful with method = "local-maximum"
  objective = "average-diversity"
)
CreateTableOne(features, strata = "House7", data = survey)
```

This looks much better! Still, the pulse variable seems to produce some weird results with a diversity-based objective, so we may again prefer the *k*-plus objective, which is also adequate for unequal-group sizes: 

```{r}
survey$House8 <- anticlustering(
  survey[, features], 
  K = c(137, 50, 50),
  standardize = TRUE,
  method = "local-maximum",
  repetitions = 10, 
  objective = "kplus"
)
CreateTableOne(features, strata = "House8", data = survey)
```

It seems that now the age variable suffers from differences in standard deviations, albeit to a lesser degree than the pulse variable when using the average diversity objective. Solving this issue is not trivial. When inspecting the variable, we find out that it is highly skewed with 2 outliers of age > 70:

```{r}
hist(survey$Age)
sort(survey$Age, decreasing = TRUE)[1:10]
```

We can try to enforce passing these two students to different houses. This can be done by creating a new categorical variable encoding if a student is an outlier regarding age. This variable can be passed to the `categories` argument. The `categories` argument implements hard constraints on the assignment of categorical variables. 

```{r}
survey$is_age_outlier <- factor(survey$Age > 70)
survey$House9 <- anticlustering(
  survey[, features], 
  K = c(137, 50, 50),
  standardize = TRUE,
  method = "local-maximum",
  repetitions = 10, 
  objective = "kplus",
  categories = survey$is_age_outlier
)
CreateTableOne(features, strata = "House9", data = survey)
```

The age outliers are now assigned to different houses, which was not the case with our previous assignment: 

```{r}
table(survey$is_age_outlier, survey$House9) # new assignment using `categories` argument
table(survey$is_age_outlier, survey$House8) # old assignment not using `categories` argument
```

In total, I think the the results are now slightly improved due to enforcing to balance the age outliers. Still, with an uneven distribution as observed here for age, creating similar (unequal-sized!) groups with regard to many variables is not a trivial issue. It really is quite a difficult problem; anticlustering is our best shot at solving it satisfactorily nevertheless. However, anticlustering cannot solve issues with the data themselves.

## Further reading

If you require some theoretical background on the stuff discussed here, here you go: 

### Categorical variables

In the original introduction of the `anticlust` package (Papenberg and Klau, 2021), we introduced the handling of categorical variables as hard constraints via the `categories` argument. In Papenberg, Wang., et al. (2025), we discussed using binary ("one hot") encoding for categorical variables and using them as part of the optimization criterion (i.e., the first argument in `anticlustering()`). 

### Objectives

The diversity objective has long been known in the context of anticlustering and cluster analysis. The *k*-plus criterion was presented in Papenberg (2024), and the average diversity was discussed in Papenberg, Breuer, et al. (2025), where the term average diversity was first employed. The objective was also used by Mohebi et al. (2022). I am not aware of other earlier usages in the context of anticlustering applications.

### Algorithms

The default anticlustering algorithm method = "exchange" was described in Papenberg and Klau (2021). However, this algorithm is actually just a shorter version of the local maximum method (i.e., method = "local-maximum"), which corresponds to the algorithm "LCW" by Weitz and Lakshminarayanan (1996, 1998). The three phase search algorithm for anticlustering was presented by Yang et al. (2022). For `anticlust`, we implemented some changes of their default procedure, in particular they used a maximum "time" that the algorithm runs before terminating. For consistency with our other methods in `anticlust`, we are however using a maximum number of "repetitions". Our default is 50 repetitions, which may be increased (but not decreased!) with the `repetitions` argument in `anticlustering()`. This adaptation was described and evaluated by Hengelbrock (2024). 

## References

Hengelbrock, H. (2024). *Evaluating the three-phase search approach with dynamic population size for anticlustering*. Unpublished bachelor thesis.  

Mohebi, S., Zanella, A., & Zorzi, M. (2022). *Pilot reuse in cell-free massive MIMO systems: A diverse clustering approach*. arXiv. https://doi.org/10.48550/arXiv.2212.08872

Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for Maximizing Between-Group Similarity. *British Journal of Mathematical and Statistical Psychology, 77* (1), 80--102. https://doi.org/10.1111/bmsp.12315

Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition data sets into equivalent parts. *Psychological Methods, 26*(2), 161--174. https://doi.org/10.1037/met0000301

Papenberg, M., Wang, C., Diop, M., Bukhari, S. H., Oskotsky, B., Davidson, B. R., ... & Oskotsky, T. T. (2025). Anticlustering for sample allocation to minimize batch effects. *Cell Reports Methods, 5*(8). https://doi.org/10.1016/j.crmeth.2025.101137

Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. (2025). Extending the Bicriterion Approach for Anticlustering: Exact and Hybrid Approaches. Psychometrika. Advance online publication. https://doi.org/10.1017/psy.2025.10052

Weitz, R. R., & Lakshminarayanan, S. (1996). On a heuristic for the final exam scheduling problem. *Journal of the Operational Research Society, 47*(4), 599--600.

Weitz, R. R., & Lakshminarayanan, S. (1998). An empirical comparison of heuristic methods for creating maximally diverse groups. *Journal of the Operational Research Society, 49*(6), 635--646. https://doi.org/10.1057/palgrave.jors.2600510

Yang, X., Cai, Z., Jin, T., Tang, Z., & Gao, S. (2022). A three-phase search approach with dynamic population size for solving the maximally diverse grouping problem. *European Journal of Operational Research, 302*(3), 925--953. https://doi.org/10.1016/j.ejor.2022.02.003

---
title: "Using the R package anticlust for stimulus selection in experiments"
output: 
  rmarkdown::html_vignette:
    number_sections: false
    df_print: kable
author: Martin Papenberg
vignette: >
  %\VignetteIndexEntry{Using the R package anticlust for stimulus selection in experiments}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
library(knitr)
opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
# define a method for objects of the class data.frame
knit_print.matrix = function(x, ...) {
    res = paste(c("", "", kable(x, row.names = TRUE)), collapse = "\n")
    asis_output(res)
}
# register the method
registerS3method("knit_print", "matrix", knit_print.matrix)

```

This tutorial teaches you how to use the `R` package `anticlust` for 
stimulus selection in psychological experiments. All code can easily be 
reproduced via Copy & Paste. The tutorial discusses the following 
functionalities:

- Match similar stimuli based on covariates of interest
- Minimize differences between stimulus sets with regard to some variables
- Maximize differences between stimulus sets with regard to some variables
- Balance the occurrence of a categorical variable between stimulus sets

To follow the code in this tutorial, load the `anticlust` 
package first:

```{r}
library(anticlust)
```

For the examples in this document, we use norming data for a stimulus 
set provided by Schaper, Kuhlmann and Bayen (2019a, 2019b). It is 
available when the package `anticlust` is loaded: 

```{r}
data("schaper2019")
# look at the data
head(schaper2019)
```

```{r, echo = FALSE}
cols <- toString(paste0("\`", names(schaper2019)[3:6], "\`"))
```

The item pool consists of 96 German words, given in the column `item`. 
Each word represents an object that is either typically found in a 
bathroom or in a kitchen. For their experiments, Schaper et al. 
partitioned the pool into 3 word lists that should be as similar as 
possible with regard to four numeric criteria (that is: `r cols`). 
Typically, stimulus sets may contain more than 96 elements (and the 
selection usually becomes more effective when the pool is larger), but 
the logic for stimulus selection that is applied in this tutorial can be 
transfered to arbitrarily large stimulus sets.

## Example 1a: Maximize differences in frequency

In the first example, I select two word lists that differ on frequency 
but are as similar as possible with regard to consistency ratings and 
the number of syllables. First, I need to define the boundaries that 
define "high" and "low" frequency. Note that `frequency` is 
reverse-coded such that low values actually indicate high frequency. 
Here, I arbitrarily define values below 18 as "high" frequency and above 
19 as "low" frequency, but any user-defined splits are possible.

```{r}
schaper2019 <- within(schaper2019, {
  freq <- ifelse(frequency < 18, "high", NA)
  freq <- ifelse(frequency > 19, "low", freq)
})
```

This code defined a new column `freq` for the data set `schaper2019` 
that is either "low", "high" or missing (`NA`). Let's check it out:

```{r}
schaper2019$freq
```

Before matching high and low frequency words, it is necessary to remove 
cases that are not selected, i.e. where the entry in `freq` is `NA`:

```{r}
selected <- subset(schaper2019, !is.na(freq))
# see how many cases were selected:
table(selected$freq)
```

Now, I use the `anticlust` function `matching()` to find paired words of 
high and low frequency that are as similar as possible on consistency 
and the number of syllables:

```{r}
# Match the conditions based on covariates
covariates <- scale(selected[, 3:5])
selected$matches <- matching(
  covariates, 
  match_between = selected$freq,
  match_within = selected$room,
  match_extreme_first = FALSE
)
```

The first argument defines the covariates that should be similar in both 
sets. I used the function `scale()` before passing the covariates to the 
matching function to standardize the variables. This way, each variable 
has the same weight in the matching process, which is usually desirable. 
The argument `match_between` is the grouping variable based on `frequency` 
that I just defined. The argument `match_within` is used to ensure that 
matches are selected between words belonging to the same room. The 
argument `match_extreme_first` is used to guide the behaviour of the 
matching algorithm, but its precise meaning is not important for now 
(for more information, see `?matching`). Generally, if we plan to only 
keep a subset of all matches -- as we do in the current application --, 
we set `match_extreme_first` to `FALSE`. If we want to keep all 
elements, `match_extreme_first = TRUE` is usually better. 

Next, let us check out some of the matches we created:

```{r}
subset(selected, matches == 1)
subset(selected, matches == 2)
```

The matches are numbered by similarity, meaning that matches with 
grouping number 1 are most similar. Therefore, we can easily select the 
8 best-matched groups of words that we would like include in our 
experiment:

```{r}
# Select the 8 best matches:
final_selection <- subset(selected, matches <= 8)
```

Last, we check the quality of the results by investigating the 
descriptive statistics -- means and standard deviations (in brackets) --
by condition:

```{r}
# Check quality of the selection:
mean_sd_tab(
  final_selection[, 3:6], 
  final_selection$freq
)
```

The descriptive statistics are similar across the sets for the 
consistency ratings and the number of syllables, and dissimilar for 
frequency, which is good. In general, the results can even be improved 
if we can select from a larger item pool (then, more matches are 
possible).

If we are not sure how many items should be part of our experiment, the 
function `plot_similarity()` may help. It plots an index of similarity 
(see `?plot_similarity`) for each match:

```{r, similarity-plot}
plot_similarity(
  covariates, 
  groups = selected$matches
)
```

The figure depicts the sum of pairwise dissimilarities (based on the 
Euclidean distance) variance per match; the larger the value, the less 
homogenous the match. It seems indeed that the first eight matches are 
most similar; after the 8th match, there is already a notable decline in 
similarity.

## Example 1b: Two-factorial design

```{r}
# Reload the data for next example
data("schaper2019")
```

In another example, we construct a two-factorial design: we create 
groups that differ on `rating_consistent` and `rating_inconsistent` 
(these properties are orthogonally crossed) but are similar on frequency 
and the number of syllables. First, we categorize the variables 
frequency and syllables into two levels, respectively. This time, I just 
use median splits, which means that I do not exclude any data in this 
first step:

```{r}
schaper2019 <- within(schaper2019, {
  incon <- ifelse(rating_inconsistent < median(rating_inconsistent), "low incon", NA)
  incon <- ifelse(rating_inconsistent >= median(rating_inconsistent), "high incon", incon)
  con <- ifelse(rating_consistent <= median(rating_consistent), "low con", NA)
  con <- ifelse(rating_consistent > median(rating_consistent), "high con", con)
})
```

Let us check how many words are left per condition: 

```{r}
table(schaper2019$con, schaper2019$incon)
```

Next, we conduct a matching between the four groups that resulted from 
crossing frequency and syllables:

```{r}
# Match the conditions based on covariates
covariates <- scale(schaper2019[, c("frequency", "syllables")])
schaper2019$matches <- matching(
  covariates, 
  match_between = schaper2019[, c("con", "incon")],
  match_extreme_first = FALSE
)
```

In this application, each match consists of 4 words because we have 4 
conditions, e.g.:

```{r}
subset(schaper2019, matches == 1)
```

To decide how many matched groups we would like to keep, let's check out 
the similarity plot:

```{r}
# Plot covariate similarity by match:
plot_similarity(covariates, schaper2019$matches)
```

Based on the plot, we select the 10 best matches:

```{r}
# Select the 5 best matches:
final_selection <- subset(schaper2019, matches <= 10)
```

Last, we check quality of the selection by printing the descriptive 
statistics by condition:

```{r}
mean_sd_tab(
  subset(final_selection, select = 3:6), 
  paste(final_selection$con, final_selection$incon)
)
```

Again, the covariates are quite similar between sets, but with a larger 
item pool even better results could be achieved.

## Example 2: Anticlustering

```{r}
# Reload the data for next example
data("schaper2019")
```

In the next example, we wish to partition the entire pool of 96 items 
into 3 sets that are as similar as possible on all variables. That means 
that there is no independent variable that varies between conditions; 
the experimental manipulation is independent of the intrinsic stimulus 
properties. Creating stimulus sets that are overall similar to each 
other can be done using anticlustering (Papenberg & Klau, 2020):

```{r}
## Conduct anticlustering (assign all items to three similar groups)
schaper2019$anticluster <- anticlustering(
  schaper2019[, 3:6], 
  K = 3,
  objective = "variance"
)

## check out quality of the solution
mean_sd_tab(
  subset(schaper2019, select = 3:6), 
  schaper2019$anticluster
)
```

## Example 2b: Anticlustering on subset selection

```{r}
# Reload the data for next example
data("schaper2019")
```

If we do not want to include all 96 words in our experiment, we can 
again use the `matching()` function to select a subset of similar 
stimuli that we employ. Again, we wish the select three groups that 
are as similar as possible on all variables, but we only use a subset
of all 96 items. To achieve this goal, we create triplets of similar
stimuli using `matching()`; afterwards, the items belonging to the 
same triplet will be assigned to different experimental sets. 
We use the argument `match_within` to ensure that all triplets
consist of items belonging to the same room:

```{r}
# First, identify triplets of similar word, within room
covariates <- scale(schaper2019[, 3:6])
schaper2019$triplet <- matching(
  covariates,
  p = 3,
  match_within = schaper2019$room
)

# check out the two most similar triplets:
subset(schaper2019, triplet == 1)
subset(schaper2019, triplet == 2)

# Select the 10 best triplets
best <- subset(schaper2019, triplet <= 10)
```

Now, we use `anticlustering()` to assign the matched words to different
sets:


```{r}
best$anticluster <- anticlustering(
  best[, 3:6], 
  K = 3,
  categories = best$triplet,
  objective = "variance"
)
```

In this function call, we pass the `triplets` to the argument `categories`, 
thus ensuring that the matched items are assigned to different sets. We can confirm
this worked by looking at the cross table of `triplet` and `anticluster`:

```{r}
table(best$triplet, best$anticluster)
```

Anticlustering strives to assign the matched triplets to the different sets in 
such a way that all sets are as similar as possible. Note that by ensuring
that the triplets only consisted of items from the same room, the room
was also balanced across anticlusters: 

```{r}
table(best$room, best$anticluster)
```

Last, we check out the descriptive statistics by anticluster, confirming 
that the sets are indeed quite similar on all numeric attributes:

```{r}
## check out quality of the solution
mean_sd_tab(
  subset(best, select = 3:6), 
  best$anticluster
)
```

## References

Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition data sets into equivalent parts. *Psychological Methods, 26*(2), 161–174. https://doi.org/10.1037/met0000301

Schaper, M. L., Kuhlmann, B. G., & Bayen, U. J. (2019a). Metacognitive expectancy effects in source monitoring: Beliefs, in-the-moment experiences, or both? Journal of Memory and Language, 107, 95–110. https://doi.org/10.1016/j.jml.2019.03.009

Schaper, M. L., Kuhlmann, B. G., & Bayen, U. J. (2019b). Metamemory expectancy illusion and schema-consistent guessing in source monitoring. Journal of Experimental Psychology: Learning, Memory, and Cognition, 45, 470. https://doi.org/10.1037/xlm0000602

---
title: "Using categorical variables with anticlustering"
output: rmarkdown::html_vignette
author: Martin Papenberg
vignette: >
  %\VignetteIndexEntry{Using categorical variables with anticlustering}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
set.seed(123)
```

```{r setup}
library(anticlust)
```

In this vignette I explore two ways to incorporate categorical variables with anticlustering. The main function of `anticlust` is `anticlustering()`, and it has an argument `categories`. It can be used easily enough: We just pass the numeric variables as first argument (`x`) and our categorical variable(s) to `categories`. I will use the penguin data set from the `palmerpenguins` package to illustrate the usage: 

```{r}
library(palmerpenguins)
# First exclude cases with missing values
df <- na.omit(penguins)
head(df)
nrow(df)
```

In the data set, each row represents a penguin, and the data set has four numeric variables (bill_length_mm, bill_depth_mm, flipper_length_mm, body_mass_g) and several categorical variables (species, island, sex) as descriptions of the penguins.

Let's call `anticlustering()` to divide the `r nrow(df)` penguins into 3 groups. We use the four the numeric variables as first argument (i.e., the anticlustering objective is computed on the basis of the numeric variables), and the penguins' sex as categorical variable:

```{r}
numeric_vars <- df[, c("bill_length_mm", "bill_depth_mm", "flipper_length_mm", "body_mass_g")]
groups <- anticlustering(
  numeric_vars, 
  K = 3,
  categories = df$sex
)
```

Let's check out how well our categorical variables are balanced:

```{r}
table(groups, df$sex)
```

A perfect split! Similarly, we could use the species as categorical variable: 

```{r}
groups <- anticlustering(
  numeric_vars, 
  K = 3,
  categories = df$species
)

table(groups, df$species)
```

As good as it could be! Now, let's use both categorical variables at the same time:

```{r}
groups <- anticlustering(
  numeric_vars, 
  K = 3,
  categories = df[, c("species", "sex")]
)

table(groups, df$sex)
table(groups, df$species) 
```

The results for the sex variable are worse than previously when we only considered one variable at a time. This is because when using multiple variables with the `categories` argument, all columns are "merged" into a single column, and each combination of sex / species is treated as a separate category. Some information on the original variables is lost, and the results may become less optimal---while being still pretty okay here. Alas, using only the `categories` argument, we cannot improve this balancing even if a better split with regard to both categorical variables would be possible.

## Categorical variables as numeric variables

A second possibility to incorporate categorical variables is to treat them as numeric variables and use them as part of the first argument `x`, which is used to compute the anticlustering objective (e.g., the diversity or variance). This approach can lead to better results when multiple categorical variables are available, and / or if the group sizes are unequal. I discuss the approach by the example of k-means anticlustering, but using the diversity objective is also possible (in principle, any reasonable way to transform categorical variables to pairwise dissimilarities would work).

To use categorical variables as part of the anticlustering objective, we first generate a matrix of the categorical variables in binary representation using the `anticlust` convenience function `categories_to_binary()`.[^modelmatrix] Because k-means anticlustering optimizes similarity with regard to means, k-means anticlustering applied to this binary matrix will even out the proportion of each category in each group (this is because the mean of a binary variable is the proportion of `1`s in that variable).

[^modelmatrix]: Internally, `categories_to_binary()` is wrapper around the base `R` function `model.matrix()`.

```{r}
binary_categories <- categories_to_binary(df[, c("species", "sex")])
# see ?categories_to_binary
head(binary_categories)
```

```{r}
groups <- anticlustering(
  binary_categories,
  K = 3,
  method = "local-maximum", 
  objective = "variance",
  repetitions = 10,
  standardize = TRUE
)
table(groups, df$sex)
table(groups, df$species)
```

The results are quite convincing. In particular, the penguins' sex is better balanced than previously when we used the argument `categories`. If we have multiple categorical variables and / or unequal-sized groups, it may be useful to try out the k-means optimization version of including categorical variables, instead of (only) using the `categories` argument. If we also wish to ensure that the categorical variables *in their combination* are balanced between groups (i.e., the proportion of the penguins' sex is roughly the same for each species in each group), we could set the optional argument `use_combinations` of `categories_to_binary()` to `TRUE`:


```{r}
binary_categories <- categories_to_binary(df[, c("species", "sex")], use_combinations = TRUE)
groups <- anticlustering(
  binary_categories,
  K = 3,
  method = "local-maximum", 
  objective = "variance",
  repetitions = 10,
  standardize = TRUE
)
table(groups, df$sex)
table(groups, df$species) 
table(groups, df$sex, df$species)
```

Note that we only evenly distributed the categorical variable between groups and did not consider any numeric variables. Fortunately, also considering the numeric variables is possible, and can we accomplish that in two different ways: 

(a) we first optimize similarity with regard to the categorical variable(s) via k-means anticlustering, and then insert the resulting group assignment as a "hard constraint" into `anticlustering()`
(b) we simultaneous optimize similarity with regard to numeric and categorical variables

We discuss both approaches in the following.

### a. Sequential optimization

We use the output vector `groups` of the previous call to `anticlustering()`---which convincingly balanced our categorical variables---as input to the `K` argument in an additional call to `anticlustering()`. The `groups` vector is used as the initial group assignment before the anticlustering optimization starts. In this group assignment, the categories are already well balanced. We additionally pass the two categorical variables to `categories`, thus ensuring that the balancing of the categorical variable is never changed throughout the optimization process:[^exchange]

[^exchange]: Only elements that have the same value in `categories` are exchanged between clusters throughout the optimization algorithm, so the initial balancing of the categories is never changed when the algorithm runs.

```{r}
final_groups <- anticlustering(
  numeric_vars,
  K = groups,
  standardize = TRUE,
  method = "local-maximum",
  categories = df[, c("species", "sex")]
)

table(groups, df$sex)
table(groups, df$species)
mean_sd_tab(numeric_vars, final_groups)
```

The results are convincing, both with regard to the numeric variables and the categorical variables. 

### b. Simultaneous optimization

We can simultaneously consider the numeric and categorical variables in the optimization process. Note that this approach only works with the k-means and k-plus objectives, because only k-means adequately deals with the categorical variables (at least when using the approach described here). 
Using the simultaneous approach, we just pass all variables (representing binary categories and numeric variables) as a single matrix to the first argument of `anticlustering()`. Do not use the `categories` argument here!

```{r}
final_groups <- anticlustering(
  cbind(numeric_vars, binary_categories),
  K = 3,
  standardize = TRUE,
  method = "local-maximum", 
  objective = "variance",
  repetitions = 10
)

table(groups, df$sex)
table(groups, df$species)
mean_sd_tab(numeric_vars, final_groups)
```

The following code extends the simultaneous optimization approach towards k-plus anticlustering, which ensures that standard deviations as well as means are similar between groups (and not only the means, which is achieved via standard k-means anticlustering):

```{r}
final_groups <- anticlustering(
  cbind(kplus_moment_variables(numeric_vars, T = 2), binary_categories),
  K = 3,
  method = "local-maximum", 
  objective = "variance", 
  repetitions = 10
)

table(groups, df$sex)
table(groups, df$species)
mean_sd_tab(numeric_vars, final_groups)
``` 

While we use `objective = "variance"`---indicating that the k-means objective is used---this code actually performs k-plus anticlustering because the first argument takes as input the augmented k-plus variable matrix[^kplus]. We see that the standard deviations are now also quite evenly matched between groups (which is unlike when using standard k-means anticlustering). 

[^kplus]: This is how k-plus anticlustering actually works: It reuses the k-means criterion but uses additional "k-plus" variables as input. More information on the k-plus approach is given in the documentation: `?kplus_moment_variables` and `?kplus_anticlustering`.

In the end: You should try out the different approaches for dealing with categorical variables and see which one works best for you!

---
title: "Speeding up anticlustering"
output: rmarkdown::html_vignette
author: Martin Papenberg
vignette: >
  %\VignetteIndexEntry{Speeding up anticlustering}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

set.seed(1234)
```

```{r setup}
library(anticlust)
```

This vignette documents various ways by which the speed of the anticlustering method implemented in the `anticlust` package can be adjusted. Speedup is particularly useful for large data sets when the default anticlustering algorithm becomes too slow. A fast method may also be desirable for testing purposes, even if the final anticlustering is based on a slower method.

## The exchange algorithm

The default anticlustering algorithm works by exchanging data points between clusters in such a way that exchanges improve the anticlustering objective as much as possible. Details on the exchange method may also be found in Papenberg and Klau (2021; https://doi.org/10.1037/met0000301), Papenberg (2024; https://doi.org/10.1111/bmsp.12315), or the `anticlust` documentation (`?anticlustering`). Basically, running more exchanges tends to improve the results, but the improvements are diminishing with many repetitions---especially for large data sets. So, to speed up anticlustering, we can reduce the number of exchanges. Here we will learn how to do that. However, first we learn how to slow down anticlustering. Slowing down can lead to better results and is recommended if you have the time.

## Slowing down

The default exchange algorithm (`anticlustering(..., method = "exchange")`) iterates through all input elements and attempts to improve the anticlustering by swapping each input element with a element that is currently assigned to a different cluster. No swap is conducted if an element cannot be swapped in such a way that the anticlustering objective is improved. The process stops after all possible exchanges have been evaluated for each element. When the number of input elements is $N$, this process leads to approximately $N^2$---or $O(N^2)$---attempted exchanges, because each element is swapped with all elements that are currently assigned to a different cluster. To give a concrete example, when having $N = 100$ data points and $K = 4$ equal-sized groups, 75 swaps are evaluated for each element and the best swap is realized. This leads to 100 * 75 = `r 100 * 75` exchanges that have to be conducted during the entire exchange algorithm, and for each exchange the objective function has to be re-evaluated. This is less exchanges than $N^2 = 100^2 = 10000$ because we skip exchanges with the 25 elements that are currently in the same cluster (including itself). However, according to the [Big O notation](https://en.wikipedia.org/wiki/Big_O_notation), we would still classify the number of exchanges as $O(N^2)$, independent of the number of groups. Thus, the total theoretical run time of the exchange method is $O(N^2)$ multiplied with the effort needed to compute an anticlustering objective. 

The results of the exchange method can be improved by not stopping after a single iteration through the data set; instead we may repeat the process until no single exchange is able to further improve the anticlustering objective, i.e., until a local maximum is found. This happens if we use `anticlustering(..., method = "local-maximum")`. This method corresponds to the algorithm "LCW" in  Weitz and Lakshminarayanan (1998). Using the local maximum method leads to more exchanges and thus to longer running time, but also better results than the default exchange method.  

Let's compare the two exchange methods with regard to their running time, using the iris data set, which contains `r nrow(iris)` elements.

```{r}

K <- 3
system.time(anticlustering(iris[, -5], K = K, method = "exchange"))
system.time(anticlustering(iris[, -5], K = K, method = "local-maximum"))

```

Depending on how many iterations are needed, the default exchange method can be much faster than the local maximum method. Generally I would recommend to use `method = "local-maximum"` for better results, but if speed is an issue, stick with the default.

To slow down even more: The exchange process may be restarted several times, each time using a different initial grouping of the elements. This is accomplished when specifying the `repetitions` argument, which defaults to 1 repetition of the exchange / local maximum algorithm. Thus, for better results, we may increase the number of repetitions:

```{r}

K <- 3
system.time(anticlustering(iris[, -5], K = K, method = "exchange"))
system.time(anticlustering(iris[, -5], K = K, method = "local-maximum"))
system.time(anticlustering(iris[, -5], K = K, method = "local-maximum", repetitions = 10))

```

In this case, sticking with the default leads to run times that are much much faster. Still, if your data set is not too large, using several repetitions may be useful (but you can judge yourself via the results). The good news is that fewer exchanges may be enough in large data sets: anticlustering generally becomes easier with more data. 

## Getting fast: Using fewer exchange partners

If the default exchange method is not fast enough for your taste, it is possible to use fewer exchange partners during the anticlustering process. By default, the exchange method evaluates each possible exchange with all elements that are currently assigned to a different cluster, leading to $O(N^2)$ exchanges. If we only use a fixed number of exchange partners per element, we can reduce the number of exchanges to $O(N)$, corresponding to a gain of an order of magnitude in terms of run time. Using fewer exchange partners for each element may decrease the quality of the results, and is generally only recommended if speed is an issue, e.g. for large data sets. But the run time will be considerably faster. We will consider three possibilities of using fewer exchange partners  in `anticlust`: `fast_anticlustering()`, preclustering, and an additional secret hack. 

### fast_anticlustering()

In Papenberg and Klau (2021), we described the function `fast_anticlustering()` as a method to speed up the anticlustering process for large data sets. It optimizes the k-means objective because computing all pairwise distances as is done when optimizing the "diversity" (i.e., the default in `anticlustering()`) is not feasible for very large data sets (for about N > 20000 on my personal computer). Moreover, `fast_anticlustering()` directly takes as argument the number of exchange partners for each element, via the argument `k_neighbours`. In this case, the application is straight forward: 

```{r}
N <- 5000
M <- 2
data <- matrix(rnorm(N * M), ncol = M)
start <- Sys.time()
groups1 <- fast_anticlustering(data, K = 2)  # default uses all exchange partners
Sys.time() - start 
```

The default behaviour in `fast_anticlustering()` is to use all exchange partners. Using fewer exchange partners can lead to much faster run time: 

```{r}
start <- Sys.time()
groups2 <- fast_anticlustering(data, K = 2, k_neighbours = 20)
Sys.time() - start 
```

Was there a cost to this speedup? 

```{r}
variance_objective(data, groups1)
variance_objective(data, groups2)
```

Frankly, there is no observable difference with regard to the objective that was obtained, but the call to `fast_anticlustering()`, which used fewer exchange partners, was much faster. In general, for large data sets, using fewer exchange partners may not impair the results and instead lead to heavily reduced run times. *Sometimes, there is free lunch*.

By default, `fast_anticlustering()` selects exchange partners though a nearest neighbour search when specifying `k_neighbours`: similar elements serve as exchange partners. The nearest neighbour search, which is done once in the beginning, only has O(N log(N)) run time and therefore does not strongly contribute to the overall run time. It is possible to suppress the nearest neighbour search by passing custom exchange partners using the `exchange_partners` argument of `fast_anticlustering()`. Exchange partners can for example be generated by `generate_exchange_partners()`, but a custom list may also be used. See the documentation (`?fast_anticlustering` and `?generate_exchange_partners()`) for more information.

### Preclustering

A second way of doing using fewer exchange partners is by including "preclustering" restrictions with the standard anticlustering function `anticlustering()`. Unlike `fast_anticlustering()`, preclustering works with all anticlustering objectives that are available in `anticlustering()` (e.g., the diversity). When setting `preclustering = TRUE`, the optimization restricts the number of exchange partners to `K - 1` (very similar) elements. While the logic is similar to the nearest neighbour approach in `fast_anticlustering()`, the difference is that the exchange partners consist of mutually exclusive groups when using preclustering. For example, the first, third and tenth element may only serve as exchange partners for each other, and none of them is exchanged with an "outsider" of this particular group. With `fast_anticlustering()`, each element has its own separate list of exchange partners, which can even be user generated when using the argument `exchange_partners`. In graph terms, with preclustering, the exchange partners form a clique, whereas with `fast_anticlustering()`, any (number of) connections are possible.

Note that the preclustering algorithm, which has to be performed prior to the anticlustering algorithm, has $O(N^2)$ run time, which is slower than the nearest neighbour search in `fast_anticlustering()`. It nevertheless leads to strongly improved run times for larger data sets:

```{r}
N <- 1000
M <- 5
K <- 3
data <- matrix(rnorm(N*M), ncol = M)
system.time(anticlustering(data, K = K))
system.time(anticlustering(data, K = K, preclustering = TRUE))
```

Still, the preclustering approach is probably not necessarily recommended to speed up anticlustering for very large data sets; it is useful if you are interested in enforcing the preclustering restrictions in the groupings.

### Secret hack

There is also an additional "hidden" method to make the `anticlustering()` function run faster. This method also relies on using fewer exchange partners during the exchange process, but does not use preclustering or the approach used in `fast_anticlustering()`. This approach is documented here and mostly relies on a dirty "hack" involving the `anticlustering()` argument `categories`. Since it works with `anticlustering()`, it can be used for all anticlustering objectives (unlike `fast_anticlustering()`). Let's see how it works:

The first step that I am using here is not strictly necessary---in the next section, we will learn more about what this accomplishes---but let's create the initial clusters before calling `anticlustering()`. This grouping is the basis on which the exchange procedure starts to improve the anticlustering:

```{r}
N <- nrow(iris)
K <- 3
initial_clusters <- sample(rep_len(1:K, N))
initial_clusters
table(initial_clusters)
```

Now, the argument `categories` can be used to define which elements serve as exchange partners for each other. Lets create random groups of 10 elements that serve as exchange elements for each other: 

```{r}
exchange_partners <- sample(rep_len(1:(N/10), N)) #somewhat ugly but works
exchange_partners
table(exchange_partners)
```

The variable `exchange_partners` now defines groups of elements that are exchanged with each other: When passed to `categories`, only elements having the same value in `exchange_partners` serve as exchange partners for each other -- this is how the `categories` argument operates. Thus, each element is only swapped with 9 other elements instead of all `r N` elements. 

Now let's call anticlustering using the exchange partners we just defined:

```{r}
system.time(anticlustering(iris[, -5], K = initial_clusters))
system.time(anticlustering(iris[, -5], K = initial_clusters, categories = exchange_partners))
```

Well, there is not a lot going on here with this very small data set (N = 150), so let's do this for a larger data set with 1000 data points.

```{r}
N <- 1000
M <- 2
K <- 5
data <- matrix(rnorm(M*N), ncol = M)

initial_clusters <- sample(rep_len(1:K, N))
exchange_partners <- sample(rep_len(1:(N/10), N))

system.time(anticlustering(data, K = initial_clusters))
system.time(anticlustering(data, K = initial_clusters, categories = exchange_partners))
```

The speedup is enormous! 

### Including categorical variables

The previous "hacky" approach to speed up anticlustering used the `categories` argument. We should now reflect how this was accomplished, and first note that the `categories` argument usually has a different purpose: It is used to evenly distribute a categorical variable across groups---we did not care for that in the previous example. 

For example, coming back to the iris data set, we may require to evenly distribute the species of the iris plants across 5 groups of plants: 

```{r}
groups <- anticlustering(iris[, -5], K = 5, categories = iris$Species)
table(groups, iris$Species)
```

How does the `categories` argument accomplish the even spread of the species? First, the initial grouping of the elements is not random, but instead a "stratified split", which ensures that a categorical variable occurs an equal number of times in each split. `anticlust` has the function `categorical_sampling()` for this purpose, which is called by `anticlustering()` internally before the exchange algorithm starts. After conducting the initial stratified split, only plants belonging to the same species serve as exchange partners for each other. This second purpose of the `categories` argument is the one that we used above to restrict the number of exchange partners to speed up anticlustering: Only elements that have the same value in `categories` serve as exchange partners for each other.

In the example above, we prevented `anticlustering()` from conducting a stratified split on the basis of the `categories` argument because we passed the initial grouping of the variables ourselves. The insight that the `categories` argument has a twofold purpose---one of which can be shut down by using the `K` argument as the initial grouping vector---leads to the following approach, where I combine the speedup aspect of `categories` with the aspect of conducting a stratified split.

First, we conduct a manual stratified split as the initial grouping vector for the `K` argument in `anticlustering()`:

```{r}
initial_groups <- categorical_sampling(iris$Species, K = 5)
table(initial_groups, iris$Species) # even!
```

Next, as in the previous section, we generate a vector that defines groups of pairwise exchange partners.

```{r}
N <- nrow(iris)
exchange_partners <- sample(rep_len(1:(N/10), N))
```

Now, and this is the crucial part, we pass to the argument `categories` a matrix that contains both the species as well as the `exchange_partners` vector, and to the argument `K` the vector that encodes the stratified split. This way we ensure that: 

(a) the species is split evenly between groups at the start of the algorithm (argument `K`) 
(b) the number of exchange partners is restricted to 9 elements (one column of the argument `categories`)
(c) the exchange partners are from the same species, thereby ensuring that the species remains evenly distributed between groups (the other column of `categories`)

```{r}
groups <- anticlustering(
  iris[, -5],
  K = initial_groups, 
  categories = cbind(iris$Species, exchange_partners)
)
```

The groups are still balanced after `anticlustering()` was called:

```{r}
table(groups, iris$Species)
```

## Objective function

The package `anticlust` primarily implements two objective functions for anticlustering: k-means and the diversity.[^kplusdispersion] The k-means criterion is well-known in cluster analysis and is computed as the sum of the squared Euclidean distances between all data points and the centroid of their respective cluster: the variance. The diversity is the overall sum of all pairwise distances between elements that are grouped in the same cluster (for details, see Papenberg & Klau, 2021). By default, `anticlustering()` optimizes the "Euclidean diversity": the diversity objective using the Euclidean distance as measure of pairwise dissimilarity.

[^kplusdispersion]: Actually, four objectives are natively supported for the `anticlustering()` argument `objective`: `"diversity"`, `"variance"` (i.e, k-means), `"kplus"` and `"dispersion"`. However, the k-plus objective as implemented in `anticlust` effectively re-uses the original k-means criterion and just extends the input data internally. So, k-plus anticlustering will be somewhat slower than k-means anticlustering because the number of variables $M$ does contribute to the overall run time; however, the run time is usually dominated by $N$, the number of elements. The dispersion objective has a different goal than the other objectives as it does not strive for between-group similarity, so it cannot be used as an alternative to the other objectives.

As explained in the previous section, the anticlustering algorithms recompute the objective function for each attempted exchange. Computing the objective is therefore the major contributor to overall run time. In `anticlust`, I exploit the fact that anticlustering objectives can be recomputed faster when only two items have swapped between clusters and the objective value prior to the exchange is known. For example, computing the diversity objective "from scratch" is in $O(N^2)$. When the cluster affiliation of only two items differs between swaps, it is however not necessary to spend the entire $O(N^2)$ time during each exchange. Instead, by "cleverly" updating the objective before and after the swap, the computation reduces to $O(N)$, leading to about $O(N^3)$ for the entire exchange method (instead of $O(N^4)$---this is a huge difference). When using a fixed number of exchange partners as described in the previous section, we are left with a run time of $O(N^2)$ *in total*. However, note that for very large data sets, using the diversity objective may not be feasible at all. The reason for this is that a quadratic matrix of between-item distances has to be computed and stored in memory. It is my experience that on a personal computer this becomes difficult for about > 20000 elements (where the distance matrix has 20000^2 = `r format(20000^2, scientific = FALSE)` elements).

Computing the standard k-means objective (which is done when using `anticlustering(..., objective = "variance"`) is in $O(M \cdot N)$, where $M$ is the number of variables. The function `anticlustering()` uses some optimizations during the exchange process to prevent the entire re-computation of the cluster centroids for each exchange, which otherwise consumes most of the run time. However, this does not change the theoretical $O(M \cdot N)$ run time of the computation. The function `fast_anticlustering()` uses a different -- but equivalent -- formulation of the k-means objective where the re-computation of the objective only depends on $M$, but no longer on $N$.[^minimization] This reduces the run time by an additional order of magnitude and makes k-means anticlustering applicable to very large data sets (even in the millions). 

[^minimization]: `fast_anticlustering()` minimizes the weighted sum of squared distances between cluster centroids and the overall data centroid; the distances between all individual data points and their cluster center are no longer computed. 

Because the k-means objective has some disadvantages for anticlustering (see Papenberg, 2024),[^kmeansbad] you may consider using the k-plus objective, which extends---and effectively re-uses---the k-means criterion. It can also be applied to very large data sets using `fast_anticlustering()`. For example, the following code---using $N = 100000$, $K = 5$, 3 features and 10 exchange partners per element---runs quite quickly:

[^kmeansbad]: The primary disadvantage is that k-means anticlustering only leads to similarity in means, but not in standard deviations or any other distribution aspects; the k-plus criterion can be used to equalize any distribution moments between groups.

```{r}
N <- 100000
M <- 3
K <- 5
data <- matrix(rnorm(M*N), ncol = M)

start <- Sys.time()
groups <- fast_anticlustering(
  kplus_moment_variables(data, T = 2), 
  K = K, 
  exchange_partners = generate_exchange_partners(10, N = N)
)
Sys.time() - start
mean_sd_tab(data, groups) # means and standard deviations are similar
```

When k-plus anticlustering is conducted with `fast_anticlustering()`, we have to augment the numeric input with so called k-plus variables to ensure that k-plus anticlustering is actually performed (because otherwise `fast_anticlustering()` just performs k-means anticlustering). We should also use the argument `exchange_partners` instead of relying on the default nearest neighbour search, because searching nearest neighbours based on the k-plus variables does not really make sense (even though it probably wouldn't hurt too much).

## References

Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition data sets into equivalent parts. *Psychological Methods, 26*(2), 161--174. https://doi.org/10.1037/met0000301.

Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for Maximizing Between-Group Similarity. *British Journal of Mathematical and Statistical Psychology, 77* (1), 80--102. https://doi.org/10.1111/bmsp.12315

Weitz, R., & Lakshminarayanan, S. (1998). An empirical comparison of heuristic methods
for creating maximally diverse groups. *Journal of the Operational Research Society, 49*(6), 635--646. https://doi.org/10.1057/palgrave.jors.2600510

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper-clustering.R
\name{balanced_clustering}
\alias{balanced_clustering}
\title{Create balanced clusters of equal size}
\source{
The centroid method was originally developed and contributed by
Meik Michalke. It was later rewritten by Martin Papenberg, who
also implemented the integer linear programming method.
}
\usage{
balanced_clustering(x, K, method = "centroid", solver = NULL)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A feature
matrix where rows correspond to elements and columns correspond
to variables (a single numeric variable can be passed as a
vector). (2) An N x N matrix dissimilarity matrix; can be an
object of class \code{dist} (e.g., returned by
\code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
where the entries of the upper and lower triangular matrix
represent pairwise dissimilarities.}

\item{K}{How many clusters should be created.}

\item{method}{One of "centroid" or "ilp". See Details.}

\item{solver}{Optional. The solver used to obtain the optimal method 
if \code{method  = "ilp"}. Currently supports "glpk" and "symphony". 
Is ignored for \code{method = "centroid"}.}
}
\value{
An integer vector representing the cluster affiliation of 
    each data point
}
\description{
Create balanced clusters of equal size
}
\details{
This function partitions a set of elements into \code{K}
equal-sized clusters. The function offers two methods: a heuristic
and an exact method. The heuristic (\code{method = "centroid"})
first computes the centroid of all data points. If the input is a
feature matrix, the centroid is defined as the mean vector of all
columns. If the input is a dissimilarity matrix, the most central
element acts as the centroid; the most central element is defined
as the element having the minimum maximal distance to all other
elements. After identifying the centroid, the algorithm proceeds as
follows: The element having the highest distance from the centroid
is clustered with its \code{(N/K) - 1} nearest neighbours
(neighbourhood is defined according to the Euclidean distance if
the data input is a feature matrix). From the remaining elements,
again the element farthest to the centroid is selected and
clustered with its \code{(N/K) - 1} neighbours; the procedure is
repeated until all elements are part of a cluster.

An exact method (\code{method = "ilp"}) can be used to solve
equal-sized weighted cluster editing optimally (implements the
integer linear program described in Papenberg and Klau, 2020; 
(8) - (10), (12) - (13)). The cluster editing objective is the 
sum of pairwise distances
within clusters; clustering is accomplished by minimizing this
objective. If the argument \code{x} is a features matrix, the
Euclidean distance is computed as the basic unit of the cluster
editing objective. If another distance measure is preferred, users
may pass a self-computed dissimiliarity matrix via the argument
\code{x}. 

The optimal \code{method = "ilp"} uses a "solver" to optimize
the clustering objective. See \code{\link{optimal_anticlustering}}
for an overview of the solvers that are available.
}
\examples{

# Cluster a data set and visualize results
N <- 1000
lds <- data.frame(f1 = rnorm(N), f2 = rnorm(N))
cl <- balanced_clustering(lds, K = 10)
plot_clusters(lds, clusters = cl)

# Repeat using a distance matrix as input
cl2 <- balanced_clustering(dist(lds), K = 10)
plot_clusters(lds, clusters = cl2)

}
\references{
Grötschel, M., & Wakabayashi, Y. (1989). A cutting plane algorithm
for a clustering problem. Mathematical Programming, 45, 59–96.

Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
data sets into equivalent parts. Psychological Methods, 26(2), 
161–174. https://doi.org/10.1037/met0000301.
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}

Meik Michalke \email{meik.michalke@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Optimal_Dispersion.R
\name{optimal_dispersion}
\alias{optimal_dispersion}
\title{Maximize dispersion for K groups}
\usage{
optimal_dispersion(
  x,
  K,
  solver = NULL,
  max_dispersion_considered = NULL,
  min_dispersion_considered = NULL,
  npartitions = 1,
  time_limit = NULL
)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A
feature matrix where rows correspond to elements and columns
correspond to variables (a single numeric variable can be
passed as a vector). (2) An N x N matrix dissimilarity matrix;
can be an object of class \code{dist} (e.g., returned by
\code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
where the entries of the upper and lower triangular matrix
represent pairwise dissimilarities.}

\item{K}{The number of groups or a vector describing the size of
each group.}

\item{solver}{Optional argument; currently supports "lpSolve", 
"glpk", "symphony", and "gurobi". See \code{\link{optimal_anticlustering}}.}

\item{max_dispersion_considered}{Optional argument used for early
stopping. If the dispersion found is equal to or exceeds this
value, a solution having the previous best dispersion is
returned.}

\item{min_dispersion_considered}{Optional argument used for
speeding up the algorithm computation.  If passed, the
dispersion is optimized starting from this value instead the
global minimum distance.}

\item{npartitions}{The number of groupings that are returned, each
having an optimal dispersion value (defaults to 1).}

\item{time_limit}{Time limit in seconds, given to the solver.
Default is there is no time limit.}
}
\value{
A list with four elements:  
   \item{dispersion}{The optimal dispersion}
   \item{groups}{An assignment of elements to groups (vector)}
   \item{edges}{A matrix of 2 columns. Each row contains the indices of 
   elements that had to be investigated to find the dispersion (i.e., each pair
   of elements cannot be part of the same group in order to achieve maximum 
   dispersion).}
   \item{dispersions_considered}{All distances that were tested until the dispersion was found.}
}
\description{
Maximize dispersion for K groups
}
\details{
The dispersion is the minimum distance between two elements
  within the same group. This function implements an optimal method
  to maximize the dispersion. If the data input \code{x} is a
  feature matrix and not a dissimilarity matrix, the pairwise
  Euclidean distance is used. It uses the algorithm OptDispF presented 
  Papenberg, Breuer, et al. (2025). It was originally developed as part of 
  Max Diekhoff's Bachelor thesis at the Computer Science Department
  at Heinrich Heine University Düsseldorf.

  To find out which items are not allowed to be grouped in the same
  cluster for maximum dispersion, the algorithm sequentially builds
  instances of a graph coloring problem, using an integer linear
  programming (ILP) representation. It is possible to specify the ILP solver via the argument
  \code{solver} (See \code{\link{optimal_anticlustering}} for more
  information on this argument). Optimally solving the maximum
  dispersion problem is NP-hard for K > 2 and therefore
  computationally infeasible for larger data sets. For K = 3 and K
  = 4, it seems that this approach scales up to several 100
  elements, or even > 1000 for K = 3 (at least when using the
  Symphony or gurobi solver).  For larger data sets, use the heuristic
  approaches in \code{\link{anticlustering}} or
  \code{\link{bicriterion_anticlustering}}. However, note that for
  K = 2, the optimal approach is usually much faster than the
  heuristics.
  
  In the output, the element \code{edges} defines which elements
  must be in separate clusters in order to achieve maximum
  dispersion. All elements not listed here can be changed
  arbitrarily between clusters without reducing the dispersion.  If
  the maximum possible dispersion corresponds to the minimum
  dispersion in the data set, the output elements \code{edges} and
  \code{groups} are set to \code{NULL} because all possible
  groupings have the same value of dispersion.  In this case the
  output element \code{dispersions_considered} has length 1.
  
  If a \code{time_limit} is set and the function cannot find in the optimal
  dispersion in the given time, it will throw an error.
}
\note{
If the SYMPHONY solver is used, an unfortunate "message" is
    printed to the console when this function terminates:

sym_get_col_solution(): No solution has been stored!

This message is no reason to worry and instead is a direct result
of the algorithm finding the optimal value for the dispersion.
Unfortunately, this message is generated in the C code underlying
the SYMPHONY library (via the printing function \code{printf}),
which cannot be prevented in R.
}
\examples{

N <- 30
M <- 5
K <- 3
data <- matrix(rnorm(N*M), ncol = M)
distances <- dist(data)

opt <- optimal_dispersion(distances, K = K)
opt

# Compare to bicriterion heuristic:
groups_heuristic <- anticlustering(
  distances, 
  K = K,
  method = "brusco", 
  objective = "dispersion", 
  repetitions = 100
)
c(
  OPT = dispersion_objective(distances, opt$groups),
  HEURISTIC = dispersion_objective(distances, groups_heuristic)
)

# Different group sizes are possible:
table(optimal_dispersion(distances, K = c(15, 10, 5))$groups)

# Induce cannot-link constraints by maximizing the dispersion:
solvable <- matrix(1, ncol = 6, nrow = 6)
solvable[2, 1] <- -1
solvable[3, 1] <- -1
solvable[4, 1] <- -1
solvable <- as.dist(solvable)
solvable

# An optimal solution has to put item 1 in a different group than 
# items 2, 3 and 4 -> this is possible for K = 2
optimal_dispersion(solvable, K = 2)$groups

# It no longer works when item 1 can also not be linked with item 5:
# (check out output!)
unsolvable <- as.matrix(solvable)
unsolvable[5, 1] <- -1
unsolvable <- as.dist(unsolvable)
unsolvable
optimal_dispersion(unsolvable, K = 2)
# But:
optimal_dispersion(unsolvable, K = c(2, 4)) # group sizes, not number of groups

}
\references{
Diekhoff (2023). Maximizing dispersion for anticlustering. Retrieved from 
https://www.cs.hhu.de/fileadmin/redaktion/Fakultaeten/Mathematisch-Naturwissenschaftliche_Fakultaet/Informatik/Algorithmische_Bioinformatik/Bachelor-_Masterarbeiten/2831963_ba_ifo_AbschlArbeit_klau_mapap102_madie120_20230203_1815.pdf  

Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. (2025). 
Extending the Bicriterion Approach for Anticlustering: Exact and Hybrid 
Approaches. Psychometrika. Advance online publication. https://doi.org/10.1017/psy.2025.10052
}
\seealso{
\code{\link{dispersion_objective}} \code{\link{anticlustering}}
}
\author{
Max Diekhoff

Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimal_anticlustering.R
\name{optimal_anticlustering}
\alias{optimal_anticlustering}
\title{Optimal ("exact") algorithms for anticlustering}
\usage{
optimal_anticlustering(x, K, objective, solver = NULL, time_limit = NULL)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A feature
matrix where rows correspond to elements and columns correspond
to variables (a single numeric variable can be passed as a
vector). (2) An N x N matrix dissimilarity matrix; can be an
object of class \code{dist} (e.g., returned by
\code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
where the entries of the upper and lower triangular matrix
represent pairwise dissimilarities.}

\item{K}{How many anticlusters should be created or alternatively:
(a) A vector describing the size of each group (the latter
currently only works for \code{objective = "dispersion")}.}

\item{objective}{The anticlustering objective, can be "diversity",
"variance", "kplus" or "dispersion".}

\item{solver}{Optional. The solver used to obtain the optimal
method.  Currently supports "glpk", "symphony", "lpSolve" and "gurobi". 
See details.}

\item{time_limit}{Time limit in seconds, given to the solver.
Default is there is no time limit.}
}
\value{
A vector of length N that assigns a group (i.e, a number
    between 1 and \code{K}) to each input element.
}
\description{
Wrapper function that gives access to all optimal algorithms for anticlustering 
that are available in anticlust.
}
\details{
This is a wrapper for all optimal methods supported in anticlust
(currently and in the future).  As compared to
\code{\link{anticlustering}}, it allows to specify the solver to
obtain an optimal solution, a time limit, and it can be used to obtain optimal
solutions for all supported anticlustering objectives (variance,
diversity, k-plus, dispersion). For the objectives "variance",
"diversity" and "kplus", the optimal ILP method in Papenberg and
Klau (2021) is used, which maximizes the sum of all pairwise
intra-cluster distances (given user specified number of clusters,
for equal-sized clusters).  To employ k-means anticlustering
(i.e. set \code{objective = "variance"}), the squared Euclidean
distance is used. For k-plus anticlustering, the squared Euclidean
distance based on the extended k-plus data matrix is used (see
\code{\link{kplus_moment_variables}}).  For the diversity (and the
dispersion), the Euclidean distance is used by default, but any
user-defined dissimilarity matrix is possible. The equivalence
of the k-means/k-plus objectives and the diversity, which enables 
using the ILP approach by Papenberg and Klau for the k-means/k-plus 
objectives, was discussed in Papenberg, Breuer, et al. (2025). 
Also see the examples. 


The dispersion is solved optimal using the algorithm OptDispF presented in 
Papenberg, Breuer, et al. (2025). Also see \code{\link{optimal_dispersion}}.

The optimal methods make use of "solvers" that actually implement
the algorithm for finding optimal solutions. The package anticlust
supports four solvers:

\itemize{
  \item{The default solver lpSolve (<https://sourceforge.net/projects/lpsolve/>).}
  \item{GNU linear programming kit (<http://www.gnu.org/software/glpk/>), 
  available from the package Rglpk and requested using \code{solver = "glpk"}.
  The R package Rglpk has to be installed manually if this solver should be used.}
  \item{The Symphony solver (<https://github.com/coin-or/SYMPHONY>),
  available from the package Rsymphony and requested using \code{solver = "symphony"}.
  (The package Rsymphony has to be installed manually if this solver should be used).}
  \item{The commercial gurobi solver, see https://www.gurobi.com/downloads/.}
}

In general, the commercial gurobi solver is best at solving exact anticlustering
problems. If you have a license to use it, use it!
Among the open source alternatives, for the maximum dispersion problem, 
it seems that the Symphony
solver is fastest, while the lpSolve solver seems to be good for
maximum diversity. However, note that in general the dispersion can
be solved optimally for much larger data sets than the diversity.

If a \code{time_limit} is set and the function cannot find in the optimal
objective in the given time, it will throw an error.
}
\examples{

data <- matrix(rnorm(24), ncol = 2)

# These calls are equivalent for k-means anticlustering:
optimal_anticlustering(data, K = 2, objective = "variance")
optimal_anticlustering(dist(data)^2, K = 2, objective = "diversity")

# These calls are equivalent for k-plus anticlustering:
optimal_anticlustering(data, K = 2, objective = "kplus")
optimal_anticlustering(dist(kplus_moment_variables(data, 2))^2, K = 2, objective = "diversity")

}
\references{
Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
data sets into equivalent parts. Psychological Methods, 26(2), 
161–174. https://doi.org/10.1037/met0000301.

Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. (2025). 
Extending the Bicriterion Approach for Anticlustering: Exact and Hybrid 
Approaches. Psychometrika. Advance online publication. https://doi.org/10.1017/psy.2025.10052
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-similarity.R
\name{plot_similarity}
\alias{plot_similarity}
\title{Plot similarity objective by cluster}
\usage{
plot_similarity(x, groups)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A data matrix
where rows correspond to elements and columns correspond to
features (a single numeric feature can be passed as a vector). (2)
An N x N matrix dissimilarity matrix; can be an object of class
\code{dist} (e.g., returned by \code{\link{dist}} or
\code{\link{as.dist}}) or a \code{matrix} where the entries of
the upper and lower triangular matrix represent the pairwise
dissimilarities.}

\item{groups}{A grouping vector of length N, usually the output
of \code{\link{matching}}.}
}
\value{
The diversity (sum of distances) by group.
}
\description{
Plot similarity objective by cluster
}
\details{
Plots the sum of pairwise distances by group.
}
\examples{

# Match elements and plot similarity by match
N <- 100
lds <- data.frame(f1 = rnorm(N), f2 = rnorm(N))
pairs <- matching(lds, p = 2)
plot_similarity(lds, pairs)

}
\seealso{
\code{\link{diversity_objective}}
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/objective-kplus.R
\name{kplus_moment_variables}
\alias{kplus_moment_variables}
\title{Compute k-plus variables}
\usage{
kplus_moment_variables(x, T, standardize = TRUE)
}
\arguments{
\item{x}{A vector, matrix or data.frame of data points. Rows
correspond to elements and columns correspond to features. A
vector represents a single feature.}

\item{T}{The number of distribution moments for which variables are generated.}

\item{standardize}{Logical, should all columns of the output be standardized
(defaults to TRUE).}
}
\value{
A matrix containing all columns of \code{x} and all additional
columns of k-plus variables. If \code{x} has M columns, the output matrix
has M * T columns.
}
\description{
Compute k-plus variables
}
\details{
The k-plus criterion is an extension of the k-means criterion
(i.e., the "variance", see \code{\link{variance_objective}}).
In \code{\link{kplus_anticlustering}}, equalizing means and variances
simultaneously (and possibly additional distribution moments) is
accomplished by internally appending new variables to the data
input \code{x}. When using only the  variance as additional criterion, the
new variables represent the squared difference of each data point to
the mean of the respective column. All columns are then included---in
addition to the original data---in standard k-means
anticlustering. The logic is readily extended towards higher order moments,
see Papenberg (2024). This function gives users the possibility to generate
k-plus variables themselves, which offers some additional flexibility when
conducting k-plus anticlustering.
}
\examples{

# Use Schaper data set for example
data(schaper2019)
features <- schaper2019[, 3:6]
K <- 3
N <- nrow(features)

# Some equivalent ways of doing k-plus anticlustering:

init_groups <- sample(rep_len(1:3, N))
table(init_groups)

kplus_groups1 <- anticlustering(
  features,
  K = init_groups,
  objective = "kplus",
  standardize = TRUE,
  method = "local-maximum"
)

kplus_groups2 <- anticlustering(
  kplus_moment_variables(features, T = 2), # standardization included by default
  K = init_groups,
  objective = "variance", # (!)
  method = "local-maximum"
)

# this function uses standardization by default unlike anticlustering():
kplus_groups3 <- kplus_anticlustering(
  features, 
  K = init_groups,
  method = "local-maximum"
)

all(kplus_groups1 == kplus_groups2)
all(kplus_groups1 == kplus_groups3)
all(kplus_groups2 == kplus_groups3)

}
\references{
Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for 
Maximizing Between-Group Similarity. British Journal of Mathematical and 
Statistical Psychology, 77(1), 80--102. https://doi.org/10.1111/bmsp.12315
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/objective-dispersion.R
\name{dispersion_objective}
\alias{dispersion_objective}
\title{Cluster dispersion}
\usage{
dispersion_objective(x, clusters)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A
feature matrix where rows correspond to elements and columns
correspond to variables (a single numeric variable can be
passed as a vector). (2) An N x N matrix dissimilarity matrix;
can be an object of class \code{dist} (e.g., returned by
\code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
where the entries of the upper and lower triangular matrix
represent pairwise dissimilarities.}

\item{clusters}{A vector representing (anti)clusters (e.g.,
returned by \code{\link{anticlustering}}).}
}
\description{
Compute the dispersion objective for a given clustering (i.e., the
minimum distance between two elements within the same cluster).
}
\details{
The dispersion is the minimum distance between two elements within
the same cluster. When the input \code{x} is a feature matrix, the
Euclidean distance is used as the distance unit. Maximizing the
dispersion maximizes the minimum heterogeneity within clusters and
is an anticlustering task.
}
\examples{

N <- 50 # number of elements
M <- 2  # number of variables per element
K <- 5  # number of clusters
random_data <- matrix(rnorm(N * M), ncol = M)
random_clusters <- sample(rep_len(1:K, N))
dispersion_objective(random_data, random_clusters)

# Maximize the dispersion 
optimized_clusters <- anticlustering(
  random_data,
  K = random_clusters, 
  objective = dispersion_objective
)
dispersion_objective(random_data, optimized_clusters)

}
\references{
Brusco, M. J., Cradit, J. D., & Steinley, D. (2020). Combining
diversity and dispersion criteria for anticlustering: A bicriterion
approach. British Journal of Mathematical and Statistical
Psychology, 73, 275-396. https://doi.org/10.1111/bmsp.12186
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weighted-cluster-editing.R
\name{wce}
\alias{wce}
\title{Exact weighted cluster editing}
\usage{
wce(x, solver = NULL)
}
\arguments{
\item{x}{A N x N similarity matrix. Larger values indicate stronger
agreement / similarity between a pair of data points}

\item{solver}{Optional argument; if passed, has to be either "glpk" or
"symphony". See details.}
}
\value{
An integer vector representing the cluster affiliation of each data point
}
\description{
Optimally solves weighted cluster editing (also known as »correlation clustering« or
»clique partitioning problem«).
}
\details{
Finds the clustering that maximizes the sum of pairwise similarities within clusters. 
In the input some similarities should be negative (indicating dissimilarity) because 
otherwise the maximum sum of similarities is obtained by simply joining all elements 
within a single big cluster. The function uses a "solver" to optimize
the clustering objective. See \code{\link{optimal_anticlustering}}
for an overview of the solvers that are available.
}
\examples{
\donttest{
features <- swiss
distances <- dist(scale(swiss))
hist(distances)
# Define agreement as being close enough to each other.
# By defining low agreement as -1 and high agreement as +1, we
# solve *unweighted* cluster editing
agreements <- ifelse(as.matrix(distances) < 3, 1, -1)
clusters <- wce(agreements)
plot(swiss, col = clusters, pch = 19)
}

}
\references{
Bansal, N., Blum, A., & Chawla, S. (2004). Correlation clustering. 
Machine Learning, 56, 89–113. 

Böcker, S., & Baumbach, J. (2013). Cluster editing. In Conference on 
Computability in Europe (pp. 33–44).

Grötschel, M., & Wakabayashi, Y. (1989). A cutting plane algorithm
for a clustering problem. Mathematical Programming, 45, 59-96.

Wittkop, T., Emig, D., Lange, S., Rahmann, S., Albrecht, M., Morris, J. H., ..., Baumbach,
J. (2010). Partitioning biological data with transitivity clustering. Nature Methods, 7,
419–420.
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bicriterion_iterated_local_search_call.R
\name{bicriterion_anticlustering}
\alias{bicriterion_anticlustering}
\title{Bicriterion iterated local search heuristic}
\usage{
bicriterion_anticlustering(
  x,
  K,
  R = NULL,
  W = c(1e-06, 1e-05, 1e-04, 0.001, 0.01, 0.1, 0.5, 0.99, 0.999, 0.999999),
  Xi = c(0.05, 0.1),
  dispersion_distances = NULL,
  average_diversity = FALSE,
  init_partitions = NULL,
  return = "paretoset"
)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A
feature matrix where rows correspond to elements and columns
correspond to variables (a single numeric variable can be
passed as a vector). (2) An N x N matrix dissimilarity matrix;
can be an object of class \code{dist} (e.g., returned by
\code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
where the entries of the upper and lower triangular matrix
represent pairwise dissimilarities.}

\item{K}{How many anticlusters should be created. Alternatively:
(a) A vector describing the size of each group, or (b) a vector
of length \code{nrow(x)} describing how elements are assigned
to anticlusters before the optimization starts.}

\item{R}{The desired number of restarts for the algorithm. By
default, both phases (MBPI + ILS) of the algorithm are
performed once.  See details.}

\item{W}{Optional argument, a vector of weights defining the
relative importance of dispersion and diversity (0 <= W <=
1). See details.}

\item{Xi}{Optional argument, specifies probability of swapping
elements during the iterated local search. See examples.}

\item{dispersion_distances}{A distance matrix used to compute the
dispersion if the dispersion should not be computed on the
basis of argument \code{x}.}

\item{average_diversity}{Boolean. Compute the diversity not as a
global sum across all pairwise within-group distances, but as
the sum of the average of within-group distances.}

\item{init_partitions}{A matrix of initial partitions (rows =
partitions; columns = elements) that serve as starting
partitions during the iterations of the first phase of the BILS
(i.e., the MBPI).  If not passed, a new random partition is
generated at the start of each iteration (which is the default
behaviour).}

\item{return}{Either "paretoset" (default), "best-diversity", 
"best-average-diversity", "best-dispersion". See below.}
}
\value{
By default, a \code{matrix} of anticlustering partitions
    (i.e., the approximated pareto set). Each row corresponds to a
    partition, each column corresponds to an input element. If the
    argument \code{return} is set to either "best-diversity", 
    "best-average-diversity", or "best-dispersion", it only returns one 
    partition (as a vector), that maximizes the respective objective.
}
\description{
This function implements the bicriterion algorithm BILS for
anticlustering by Brusco et al. (2020;
<doi:10.1111/bmsp.12186>). The description of their algorithm is
given in Section 3 of their paper (in particular, see the
Pseudocode in their Figure 2). It also implements some extensions
to the bicriterion approach for anticlustering described in 
Papenberg, Breuer, et al. (2025).
}
\details{
The bicriterion algorithm by Brusco et al. (2020) aims to
simultaneously optimize two anticlustering criteria: the
\code{\link{diversity_objective}} and the
\code{\link{dispersion_objective}}. It returns a list of partitions
that approximate the pareto set of efficient solutions across both
criteria. By considering both the diversity and dispersion, this
algorithm is well-suited for maximizing overall within-group
heterogeneity. To select a partition among the approximated pareto
set, it is reasonable to plot the objectives for each partition
(see Examples).

The arguments \code{R}, \code{W} and \code{Xi} are named for
consistency with Brusco et al. (2020). The argument \code{K} is
used for consistency with other functions in anticlust; Brusco et
al. used `G` to denote the number of groups. However, note that
\code{K} can not only be used to denote the number of equal-sized
groups, but also to specify group sizes, as in
\code{\link{anticlustering}}.

This function implements the combined bicriterion algorithm BILS,
which consists of two phases: The multistart bicriterion pairwise
interchange heuristic (MBPI, which is a local maximum search
heuristic similar to \code{method = "local-maximum"} in
\code{\link{anticlustering}}), and the iterated local search (ILS),
which is an improvement procedure that tries to overcome local
optima.  The argument \code{R} denotes the number of restarts of
the two phases of the algorithm. If \code{R} has length 1, half of
the repetitions perform the first phase MBPI), the other half
perform the ILS.  If \code{R} has length 2, the first entry
indicates the number of restarts of MBPI the second entry indicates
the number of restarts of ILS.  The argument \code{W} denotes the
relative weight given to the diversity and dispersion criterion in
a given run of the search heuristic. In each run, the a weight is
randomly selected from the vector \code{W}. As default values, we
use the weights that Brusco et al. used in their analyses. All
values in \code{W} have to be in [0, 1]; larger values indicate
that diversity is more important, whereas smaller values indicate
that dispersion is more important; \code{w = .5} implies the same
weight for both criteria. The argument \code{Xi} is the probability
that an element is swapped during the iterated local search
(specifically, Xi has to be a vector of length 2, denoting the
range of a uniform distribution from which the probability of
swapping is selected). For \code{Xi}, the default is selected
consistent with the analyses by Brusco et al.

If the data input \code{x} is a feature matrix (that is: each row
is a "case" and each column is a "variable"), a matrix of the
Euclidean distances is computed as input to the algorithm. If a
different measure of dissimilarity is preferred, you may pass a
self-generated dissimilarity matrix via the argument \code{x}.  The
argument \code{dispersion_distances} can additionally be used if
the dispersion should be computed on the basis of a different
distance matrix.

As of anticlust version 0.8.6, this function includes some extensions 
to the original BILS by Brusco et al. (2020). These extensions are are 
implemented through the optional arguments \code{dispersion_distances}, 
\code{average_diversity}, \code{init_partitions}, and \code{return}, and
were described in Papenberg, Breuer, et al. (2025). 
If these arguments are not changed, the function performs the "vanilla" BILS 
as described in Brusco et al. If multiple \code{init_partitions} are given, 
ensure that each partition (i.e., each row of\code{init_partitions}) has the exact
same output of \code{\link{table}}.
}
\note{
For technical reasons, the pareto set returned by this function has
a limit of 500 partitions. Usually however, the
algorithm usually finds much fewer partitions. There is one following exception:
We do not recommend to use this method when the input data is
one-dimensional where the algorithm may identify too many
equivalent partitions causing it to run very slowly (see section 5.6 in 
Breuer, 2020).
}
\examples{
# Generate some random data
M <- 3
N <- 80
K <- 4
data <- matrix(rnorm(N * M), ncol = M)

# Perform bicriterion algorithm, use 200 repetitions
pareto_set <- bicriterion_anticlustering(data, K = K, R = 200)

# Compute objectives for all solutions
diversities_pareto <- apply(pareto_set, 1, diversity_objective, x = data)
dispersions_pareto <- apply(pareto_set, 1, dispersion_objective, x = data)

# Plot the pareto set
plot(
  diversities_pareto,
  dispersions_pareto,
  col = "blue",
  cex = 2,
  pch = as.character(1:NROW(pareto_set))
)

# Get some random solutions for comparison
rnd_solutions <- t(replicate(n = 200, sample(pareto_set[1, ])))

# Compute objectives for all random solutions
diversities_rnd <- apply(rnd_solutions, 1, diversity_objective, x = data)
dispersions_rnd <- apply(rnd_solutions, 1, dispersion_objective, x = data)

# Plot random solutions and pareto set. Random solutions are far away 
# from the good solutions in pareto set
plot(
  diversities_rnd, dispersions_rnd, 
  col = "red",
  xlab = "Diversity",
  ylab = "Dispersion",
  ylim = c(
    min(dispersions_rnd, dispersions_pareto), 
    max(dispersions_rnd, dispersions_pareto)
  ),
  xlim = c(
    min(diversities_rnd, diversities_pareto), 
    max(diversities_rnd, diversities_pareto)
  )
)

# Add approximated pareto set from bicriterion algorithm:
points(diversities_pareto, dispersions_pareto, col = "blue", cex = 2, pch = 19)

}
\references{
Brusco, M. J., Cradit, J. D., & Steinley, D. (2020). Combining
diversity and dispersion criteria for anticlustering: A bicriterion
approach. British Journal of Mathematical and Statistical
Psychology, 73, 275-396. https://doi.org/10.1111/bmsp.12186

Breuer (2020). Using anticlustering to maximize diversity and dispersion:
Comparing exact and heuristic approaches. Bachelor thesis.

Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. (2025). 
Extending the Bicriterion Approach for Anticlustering: Exact and Hybrid 
Approaches. Psychometrika. Advance online publication. https://doi.org/10.1017/psy.2025.10052
}
\author{
Martin Breuer \email{M.Breuer@hhu.de}, Martin Papenberg
    \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schaper2019.R
\docType{data}
\name{schaper2019}
\alias{schaper2019}
\title{Ratings for 96 words}
\format{
A data frame with 96 rows and 7 variables
\describe{
  \item{item}{The name of an object (in German)}
  \item{room}{The room in which the item is typically found; can be
      'kitchen' or 'bathroom'}
  \item{rating_consistent}{How expected would it
      be to find the \code{item} in the typical \code{room}}
  \item{rating_inconsistent}{How expected would it
      be to find the \code{item} in the atypical \code{room}}
  \item{syllables}{The number of syllables in the object name}
  \item{frequency}{A value indicating the relative frequency of the
      object name in German language (lower values indicate higher
      frequency)}
  \item{list}{Represents the set affiliation of the \code{item} as
      realized in experiments by Schaper et al.}
}
}
\source{
Courteously provided by Marie Lusia Schaper and Ute Bayen.
}
\usage{
schaper2019
}
\description{
A stimulus set that was used in experiments by Schaper, Kuhlmann and
Bayen (2019a; 2019b). The item pool consists of 96 German words. 
Each word represents an object that is either 
typically found in a bathroom or in a kitchen.
}
\examples{

head(schaper2019)
features <- schaper2019[, 3:6]

# Optimize the variance criterion
# (tends to maximize similarity in feature means)
anticlusters <- anticlustering(
  features,
  K = 3,
  objective = "variance",
  categories = schaper2019$room,
  method = "exchange"
)

# Means are quite similar across sets:
by(features, anticlusters, function(x) round(colMeans(x), 2))
# Check differences in standard deviations:
by(features, anticlusters, function(x) round(apply(x, 2, sd), 2))
# Room is balanced between the three sets:
table(Room = schaper2019$room, Set = anticlusters)

# Maximize the diversity criterion
ac_dist <- anticlustering(
  features,
  K = 3,
  objective = "diversity",
  categories = schaper2019$room,
  method = "exchange"
)
# With the distance criterion, means tend to be less similar,
# but standard deviations tend to be more similar:
by(features, ac_dist, function(x) round(colMeans(x), 2))
by(features, ac_dist, function(x) round(apply(x, 2, sd), 2))


}
\references{
Schaper, M. L., Kuhlmann, B. G., & Bayen, U. J. (2019a). Metacognitive expectancy
effects in source monitoring: Beliefs, in-the-moment experiences, or both? Journal
of Memory and Language, 107, 95–110. https://doi.org/10.1016/j.jml.2019.03.009

Schaper, M. L., Kuhlmann, B. G., & Bayen, U. J. (2019b). Metamemory expectancy illusion
and schema-consistent guessing in source monitoring. Journal of Experimental
Psychology: Learning, Memory, and Cognition, 45, 470.
https://doi.org/10.1037/xlm0000602
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/descriptives_by_cluster.R
\name{mean_sd_tab}
\alias{mean_sd_tab}
\title{Means and standard deviations by group variable formatted in table}
\usage{
mean_sd_tab(features, groups, decimals = 2, na.rm = FALSE, return_diff = FALSE)
}
\arguments{
\item{features}{A data frame of features}

\item{groups}{A grouping vector}

\item{decimals}{The number of decimals}

\item{na.rm}{Should NAs be removed prior to computing stats 
(Default = FALSE)}

\item{return_diff}{Boolean. Should an additional row be printed that 
contains the difference between minimum and maximum}
}
\value{
A table that illustrates means and standard deviations (in brackets)
}
\description{
Means and standard deviations by group variable formatted in table
}
\examples{

data(iris)
mean_sd_tab(iris[, -5], iris[, 5])

}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brunel2025.R
\docType{data}
\name{brunel2025}
\alias{brunel2025}
\title{Completion norms for 403 French sentences}
\format{
A data frame with 403 rows and 7 variables
\describe{
  \item{target_word}{The word (in French)}
  \item{sentence}{The sentence (in French), which is completed by \code{target_word}}
  \item{target_word_emotionality}{Categorical: Is the target word classified as "positive", "negative" or "neutral"}
  \item{sentence_emotionality}{Categorical: Is the sentence classified as 
    "mixt", "negative", "neutral", "non specified", or "positive"}
  \item{percentage_target_word}{Numeric between 0 and 1; percentage of a test sample that identified the target word}
  \item{valence_target_word}{Numeric, average valence rating for the target word}
  \item{arousal_target_word}{Numeric, average arousal rating for the target word}
}
}
\source{
This data set was distributed as supplementary online materials to the paper by Brunel et al. (2025), 
and is available from the Open Science Repository via https://osf.io/7pc46. Retrieved on October 2, 2025.
The data was published under a CC-By Attribution 4.0 International license, which permits redistribution.
The anticlust packages includes the raw and unchanged data file as provided by Jeremy Brunel, and a 
script that converts the raw data into the package data.
}
\usage{
brunel2025
}
\description{
Completion norms for 403 French sentences
}
\examples{

groups <- anticlustering(
  brunel2025[, 3:7],
  K = 4, 
  objective = "kplus", 
  standardize = TRUE,
  method = "3phase"
)
# check out descriptive statistics for each variable for each of 4 groups:
# Numeric variables
mean_sd_tab(brunel2025[, 5:7], groups)
# Categorical variables:
table(groups, brunel2025$target_word_emotionality)
table(groups, brunel2025$sentence_emotionality)

}
\references{
Brunel, J., Dujardin, E., Delord, S. (2025) Emotional valence, cloze 
probability, and entropy: Completion norms for 403 French sentences. 
Behavior Research Methods 57(81). https://doi.org/10.3758/s13428-025-02604-7
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in
%   R/wrapper-three-phase-search-dynamic-population-size.R
\name{three_phase_search_anticlustering}
\alias{three_phase_search_anticlustering}
\title{Three phase search with dynamic population size heuristic}
\usage{
three_phase_search_anticlustering(
  x,
  K,
  N,
  objective = "diversity",
  number_iterations = 50,
  clusters = NULL,
  upper_bound = NULL,
  lower_bound = NULL,
  beta_max = 15,
  theta_max = NULL,
  theta_min = NULL,
  beta_min = NULL,
  eta_max = 3,
  alpha = 0.05
)
}
\arguments{
\item{x}{The data input, as in \code{\link{anticlustering}}.}

\item{K}{Number of anticlusters to be formed.}

\item{N}{Number of elements.}

\item{objective}{The anticlustering objective, can be "diversity" or "dispersion".}

\item{number_iterations}{A number that defines how many times the steps in the search algorithm are repeated.}

\item{clusters}{A vector of length K that specifies the number of elements each cluster can contain. 
If this vector is not NULL, the lower and upper bounds will be disregarded.}

\item{upper_bound}{Maximum number of elements in each anticluster. By default, anticlusters are of equal size,
calculated as the total number of items divided by the number of clusters.}

\item{lower_bound}{Minimum number of elements in each anticluster. By default, anticlusters are of equal size,
calculated as the total number of items divided by the number of clusters.}

\item{beta_max}{The algorithm begins with a pool of random initial solutions of size beta_max. 
Over time, the size of the solution pool decreases linearly until it reaches beta_min.}

\item{theta_max}{Parameter for the strength of undirected perturbation, 
which decreases linearly over time from theta_max to theta_min.}

\item{theta_min}{Parameter for the strength of undirected perturbation, 
which decreases linearly over time from theta_max to theta_min.}

\item{beta_min}{The minimum solution pool size the algorithm should reach before making a determination.}

\item{eta_max}{Parameter that specifies how many times the steps in the direct perturbation are executed.}

\item{alpha}{Parameter for weighing the discrimination of a slightly worse local optimal child solution.}
}
\value{
A vector of length N that assigns a group (i.e, a number
    between 1 and \code{K}) to each input element
}
\description{
This function implements the three phase search algorithm TPSPD for
anticlustering by Yang et al. (2022; <doi.org/10.1016/j.ejor.2022.02.003>).
The description of their algorithm is
given in Section 2 of their paper (in particular, see the
Pseudocode in Algorithm 1).
}
\details{
Details of the implementation of the algorithm can be found 
 in the pseudocode of the paper Yang et al. (2022). However, we performed one change
 as compared to the original description of the algorithm: Instead of 
 setting a time limit, we define the number of iterations the algorithm 
 performs before terminating (via argument \code{number_iterations}).
}
\examples{

# Generate some random data
N <- 120
M <- 5
K <- 4
dat <- matrix(rnorm(N * M), ncol = M)
distances <- dist(dat)

# Perform three hase serach algorithm
result1 <- three_phase_search_anticlustering(dat, K, N)

# Compute objectives function
diversity_objective(distances, result1)

# Standard algorithm:
result2 <- anticlustering(distances, K=K, method="local-maximum", repetitions = 10)
diversity_objective(distances, result2)


}
\references{
Xiao Yang et al. “A three-phase search approach with dynamic population size for solving 
the maximally diverse grouping problem”. In: European Journal of Operational Research
302.3 (2022) <doi:10.1016/j.ejor.2022.02.003>
}
\author{
Hannah Hengelbrock \email{Hannah.Hengelbrock@hhu.de}, 
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample-by-category.R
\name{categorical_sampling}
\alias{categorical_sampling}
\title{Random sampling employing a categorical constraint}
\usage{
categorical_sampling(categories, K)
}
\arguments{
\item{categories}{A matrix or vector of one or more categorical variables.}

\item{K}{The number of groups that are returned.}
}
\value{
A vector representing the sample each element was assigned to.
}
\description{
This function can be used to obtain a stratified split of a data set.
}
\details{
This function can be used to obtain a stratified split of a data set. 
Using this function is like calling \code{\link{anticlustering}} with 
argument `categories`, but without optimizing a clustering objective. The
categories are just evenly split between samples. Apart from the restriction 
that categories are balanced between samples, the split is random.
}
\examples{

data(schaper2019)
categories <- schaper2019$room
groups <- categorical_sampling(categories, K = 6)
table(groups, categories)

# Unequal sized groups
groups <- categorical_sampling(categories, K = c(24, 24, 48))
table(groups, categories)

# Heavily unequal sized groups, is harder to balance the groups
groups <- categorical_sampling(categories, K = c(51, 19, 26))
table(groups, categories)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/objective-diversity.R
\name{diversity_objective}
\alias{diversity_objective}
\title{(Anti)cluster editing "diversity" objective}
\usage{
diversity_objective(x, clusters)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A data matrix
where rows correspond to elements and columns correspond to
features (a single numeric feature can be passed as a vector). (2)
An N x N matrix dissimilarity matrix; can be an object of class
\code{dist} (e.g., returned by \code{\link{dist}} or
\code{\link{as.dist}}) or a \code{matrix} where the entries of
the upper and lower triangular matrix represent the pairwise
dissimilarities.}

\item{clusters}{A vector representing (anti)clusters (e.g.,
returned by \code{\link{anticlustering}}).}
}
\value{
The cluster editing objective
}
\description{
Compute the diversity for a given clustering.
}
\details{
The objective function used in (anti)cluster editing is the
diversity, i.e., the sum of the pairwise distances between elements
within the same groups. When the input \code{x} is a feature
matrix, the Euclidean distance is computed as the basic distance
unit of this objective.
}
\examples{

data(iris)
distances <- dist(iris[1:60, -5])
## Clustering
clusters <- balanced_clustering(distances, K = 3)
# This is low:
diversity_objective(distances, clusters)
## Anticlustering
anticlusters <- anticlustering(distances, K = 3)
# This is higher:
diversity_objective(distances, anticlusters)

}
\references{
Brusco, M. J., Cradit, J. D., & Steinley, D. (2020). Combining
diversity and dispersion criteria for anticlustering: A bicriterion
approach. British Journal of Mathematical and Statistical
Psychology, 73, 275-396. https://doi.org/10.1111/bmsp.12186

Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
data sets into equivalent parts. Psychological Methods, 26(2), 
161–174. https://doi.org/10.1037/met0000301.
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate-partitions.R
\name{generate_partitions}
\alias{generate_partitions}
\title{Generate all partitions of same cardinality}
\usage{
generate_partitions(N, K, generate_permutations = FALSE)
}
\arguments{
\item{N}{The total N. \code{K} has to be dividble
by \code{N}.}

\item{K}{How many partitions}

\item{generate_permutations}{If TRUE, all permutations are returned,
resulting in duplicate partitions.}
}
\value{
A list of all partitions (or permutations if
\code{generate_permutations} is \code{TRUE}).
}
\description{
Generate all partitions of same cardinality
}
\details{
In principle, anticlustering can be solved to optimality by
generating all possible partitions of N items into K groups.
The example code below illustrates how to do this.
However, this approach only works for small N because the
number of partitions grows exponentially with N.

The partition c(1, 2, 2, 1)
is the same as the partition c(2, 1, 1, 2) but they correspond
to different permutations of the elements [1, 1, 2, 2]. If the argument
\code{generate_permutations} is \code{TRUE}, all permutations are
returned. To solve balanced anticlustering exactly, it is sufficient
to inspect all partitions while ignoring duplicated permutations.
}
\examples{

## Generate all partitions to solve k-means anticlustering
## to optimality.

N <- 14
K <- 2
features <- matrix(sample(N * 2, replace = TRUE), ncol = 2)
partitions <- generate_partitions(N, K)
length(partitions) # number of possible partitions

## Create an objective function that takes the partition
## as first argument (then, we can use sapply to compute
## the objective for each partition)
var_obj <- function(clusters, features) {
  variance_objective(features, clusters)
}

all_objectives <- sapply(
  partitions,
  FUN = var_obj,
  features = features
)

## Check out distribution of the objective over all partitions:
hist(all_objectives) # many large, few low objectives
## Get best k-means anticlustering objective:
best_obj <- max(all_objectives)
## It is possible that there are multiple best solutions:
sum(all_objectives == best_obj)
## Select one best partition:
best_anticlustering <- partitions[all_objectives == best_obj][[1]]
## Look at mean for each partition:
by(features, best_anticlustering, function(x) round(colMeans(x), 2))


## Get best k-means clustering objective:
min_obj <- min(all_objectives)
sum(all_objectives == min_obj)
## Select one best partition:
best_clustering <- partitions[all_objectives == min_obj][[1]]

## Plot minimum and maximum objectives:
user_par <- par("mfrow")
par(mfrow = c(1, 2))
plot_clusters(
  features,
  best_anticlustering,
  illustrate_variance = TRUE,
  main = "Maximum variance"
)
plot_clusters(
  features,
  best_clustering,
  illustrate_variance = TRUE,
  main = "Minimum variance"
)
par(mfrow = user_par)

}
\references{
Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
data sets into equivalent parts. Psychological Methods, 26(2), 
161–174. https://doi.org/10.1037/met0000301.
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper-matching.R
\name{matching}
\alias{matching}
\title{Matching}
\usage{
matching(
  x,
  p = 2,
  match_between = NULL,
  match_within = NULL,
  match_extreme_first = TRUE,
  target_group = NULL,
  sort_output = TRUE
)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A feature
matrix where rows correspond to elements and columns correspond
to variables (a single numeric variable can be passed as a
vector). (2) An N x N matrix dissimilarity matrix; can be an
object of class \code{dist} (e.g., returned by
\code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
where the entries of the upper and lower triangular matrix
represent pairwise dissimilarities.}

\item{p}{The size of the groups; the default is 2, in which case
the function returns pairs.}

\item{match_between}{An optional vector, \code{data.frame} or
matrix representing one or several categorical constraints. If
passed, the argument \code{p} is ignored and matches are sought
between elements of different categories.}

\item{match_within}{An optional vector, \code{data.frame} or matrix
representing one or several categorical constraints. If passed,
matches are sought between elements of the same category.}

\item{match_extreme_first}{Logical: Determines if matches are first
sought for extreme elements first or for central
elements. Defaults to \code{TRUE}.}

\item{target_group}{Currently, the options "none",
smallest" and "diverse" are supported. See Details.}

\item{sort_output}{Boolean. If \code{TRUE} (default), the output clusters 
are sorted by similarity. See Details.}
}
\value{
An integer vector encoding the matches. See Details for
    more information.
}
\description{
Conduct K-partite or unrestricted (minimum distance) matching to
find pairs or groups of similar elements. By default, finding
matches is based on the Euclidean distance between data points, but
a custom dissimilarity measure can also be employed.
}
\details{
If the data input \code{x} is a feature matrix, matching is based
on the Euclidean distance between data points. If the argument
\code{x} is a dissimilarity matrix, matching is based on the
user-specified dissimilarities. To find matches, the algorithm
proceeds by selecting a target element and then searching its
nearest neighbours. Critical to the behaviour or the algorithm is
the order in which target elements are selected. By default, the
most extreme elements are selected first, i.e., elements with the
highest distance to the centroid of the data set (see
\code{\link{balanced_clustering}} that relies on the same
algorithm). Set the argument \code{match_extreme_first} to
\code{FALSE}, to enforce that elements close to the centroid are
first selected as targets.

If the argument \code{match_between} is passed and the groups
specified via this argument are of different size, target elements
are selected from the smallest group by default (because in this
group, all elements can be matched). However, it is also possible
to specify how matches are selected through the option
\code{target_group}. When specifying \code{"none"}, matches are
always selected from extreme elements, irregardless of the group
sizes (or from central elements first if \code{match_extreme_first
= FALSE}). With option \code{"smallest"}, matches are selected from
the smallest group. With option \code{"diverse"}, matches are
selected from the most heterogenous group according to the sum of
pairwise distances within groups.

The output is an integer vector encoding which elements have been
matched. The grouping numbers are sorted by similarity. That is,
elements with the grouping number »1« have the highest intra-group
similarity, followed by 2 etc (groups having the same similarity
index are still assigned a different grouping number,
though). Similarity is measured as the sum of pairwise (Euclidean)
distances within groups (see \code{\link{diversity_objective}}). To 
prevent sorting by similarity (this is some extra computational burden),
set \code{sort_output = FALSE}. Some unmatched elements may be \code{NA}. 
This happens if it is not
possible to evenly split the item pool evenly into groups of size
\code{p} or if the categories described by the argument
\code{match_between} are of different size.
}
\note{
It is possible to specify grouping restrictions via
    \code{match_between} and \code{match_within} at the same time.
}
\examples{

# Find triplets
N <- 120
lds <- data.frame(f1 = rnorm(N), f2 = rnorm(N))
triplets <- matching(lds, p = 3)
plot_clusters(
  lds,
  clusters = triplets,
  within_connection = TRUE
)

# Bipartite matching with unequal-sized groups:
# Only selects matches for some elements
N <- 100
data <- matrix(rnorm(N), ncol = 1)
groups <- sample(1:2, size = N, replace = TRUE, prob = c(0.8, 0.2))
matched <- matching(data[, 1], match_between = groups)
plot_clusters(
  cbind(groups, data), 
  clusters = matched, 
  within_connection = TRUE
)

# Match objects from the same category only
matched <- matching(
  schaper2019[, 3:6], 
  p = 3, 
  match_within = schaper2019$room
)
head(table(matched, schaper2019$room))

# Match between different plant species in the »iris« data set
species <- iris$Species != "versicolor"
matched <- matching(
  iris[species, 1], 
  match_between = iris[species, 5]
)
# Adjust `match_extreme_first` argument
matched2 <- matching(
  iris[species, 1], 
  match_between = iris[species, 5],
  match_extreme_first = FALSE
)
# Plot the matching results
user_par <- par("mfrow")
par(mfrow = c(1, 2))
data <- data.frame(
  Species = as.numeric(iris[species, 5]),
  Sepal.Length = iris[species, 1]
)
plot_clusters(
  data,
  clusters = matched,
  within_connection = TRUE,
  main = "Extreme elements matched first"
)
plot_clusters(
  data,
  clusters = matched2,
  within_connection = TRUE,
  main = "Central elements matched first"
)
par(mfrow = user_par)


}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/categories_to_binary.R
\name{categories_to_binary}
\alias{categories_to_binary}
\title{Get binary representation of categorical variables}
\usage{
categories_to_binary(categories, use_combinations = FALSE)
}
\arguments{
\item{categories}{A vector, data.frame or matrix representing one
or several categorical variables}

\item{use_combinations}{Logical, should the output also include columns representing
the combination / interaction of the categories (defaults to \code{FALSE}).}
}
\value{
A matrix encoding the categorical variable(s) in binary form.
}
\description{
Get binary representation of categorical variables
}
\details{
The conversion of categorical variables to binary variables is done via
\code{\link[stats]{model.matrix}}. Since version 0.8.9, each category
of a categorical variable is coded by a separate variable ('one hot' encoding).
So we do not use 'dummy' coding, which is often used to encode predictors 
in statistical analysis. Dummy coding uses a reference category that is not
explicitly modelled via a separate variable. This implies that there is
a different distance to the reference category than among the other 
categories, which is unwarranted in anticlustering (thanks to Gunnar Klau
for noting this). See examples. 

This function can be used to include categorical variables as part of the 
optimization criterion in anticlustering, rather than including them as hard constraints as done when using the 
argument \code{categories} in \code{\link{anticlustering}} (or \code{\link{fast_anticlustering}}). 
This way, categorical variables are treated as numeric variables, 
which can be useful when there are several
categorical variables or when the group sizes are unequal (or both).
See examples. Please see the vignette 'Using categorical variables with anticlustering'
for more information on this approach.

Since version 0.8.12, this function treats NA in the categorical input variables
as a distinct category. This implies that different observations having an NA are treated
as having the same value on this variable. If this is not desired, you need to do some other
preprocessing of the NAs before calling this function.
}
\examples{

# Illustrate why dummy encoding is not appropriate for anticlustering.
# Use 3 elements of the iris data set, with different levels of 'Species'
input <- iris[c(1, 51, 101), "Species", drop = FALSE]
input
# Default dummy encoding: 
(dummy <- model.matrix(~ . , input))
dist(dummy) 
# Distance between versicolor and virginica is larger than among setosa and 
# versicolor, and setosa and virginica. This would bias the anticlustering 
# computation.
(one_hot <- categories_to_binary(input))
dist(one_hot) # all the same distances 
dist(one_hot)^2 # or use squared Euclidean distance (corresponds to Manhattan distance in this case)
dist(one_hot, method = "manhattan")

# We could also get one-hot encoding directly via model.matrix(), which is 
# used in categories_to_binary():
model.matrix(~ . -1, input) # suppresses the 'intercept'

# Use Schaper data set for anticlustering example
data(schaper2019)
features <- schaper2019[, 3:6]
K <- 3
N <- nrow(features) 

# - Generate data input for k-means anticlustering -
# We conduct k-plus anticlustering by first generating k-plus variables, 
# and also include the categorical variable as "numeric" input for the 
# k-means optimization (rather than as input for the argument categories)

input_data <- cbind(
  kplus_moment_variables(features, T = 2), 
  categories_to_binary(schaper2019$room) 
)

kplus_groups <- anticlustering(
  input_data, 
  K = K,
  objective = "variance",
  method = "local-maximum", 
  repetitions = 10
)
mean_sd_tab(features, kplus_groups)
table(kplus_groups, schaper2019$room) # argument categories was not used!

 
}
\references{
Papenberg, M., Wang, C., Diop, M., Bukhari, S. H., Oskotsky, B., Davidson, B. R., 
Vo, K. C., Liu, B., Irwin, J. C., Combes, A., Gaudilliere, B., Li, J., Stevenson, D. K., 
Klau, G. W., Giudice, L. C., Sirota, M., & Oskotsky, T. T. (2025). Anticlustering
for sample allocation to minimize batch effects. Cell Reports Methods, 5(8), 
101137. https://doi.org/10.1016/j.crmeth.2025.101137
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper-anticlustering.R
\name{anticlustering}
\alias{anticlustering}
\title{Anticlustering}
\usage{
anticlustering(
  x,
  K,
  objective = "diversity",
  method = "exchange",
  preclustering = FALSE,
  categories = NULL,
  repetitions = NULL,
  standardize = FALSE,
  cannot_link = NULL,
  must_link = NULL,
  blocks = NULL
)
}
\arguments{
\item{x}{The data input. Can be one of two structures: (1) A
feature matrix where rows correspond to elements and columns
correspond to variables (a single numeric variable can be
passed as a vector). (2) An N x N matrix dissimilarity matrix;
can be an object of class \code{dist} (e.g., returned by
\code{\link{dist}} or \code{\link{as.dist}}) or a \code{matrix}
where the entries of the upper and lower triangular matrix
represent pairwise dissimilarities.}

\item{K}{How many anticlusters should be created. Alternatively:
(a) A vector describing the size of each group, or (b) a vector
of length \code{nrow(x)} describing how elements are assigned
to anticlusters before the optimization starts.}

\item{objective}{The objective to be maximized. The options
"diversity" (default; previously called "distance", which is
still supported), "average-diversity", "variance", "kplus" and "dispersion" are
natively supported. May also be a user-defined function. See
Details.}

\item{method}{One of "exchange" (default) , "local-maximum",
"brusco", "ilp", "2PML", or "3phase".  See Details.}

\item{preclustering}{Boolean. Should a preclustering be conducted
before anticlusters are created? Defaults to \code{FALSE}. See
Details.}

\item{categories}{A vector, data.frame or matrix representing one
or several categorical variables whose distribution should be similar 
between groups. See Details.}

\item{repetitions}{The number of times a search heuristic is
initiated when using \code{method = "exchange"}, \code{method =
"local-maximum"}, \code{method = "brusco"}, \code{method = "3phase"}, 
or \code{method = "2PML"}. In the end, the best objective found across 
the repetitions is returned.}

\item{standardize}{Boolean. If \code{TRUE} and \code{x} is a
feature matrix, the data is standardized through a call to
\code{\link{scale}} before the optimization starts. This
argument is silently ignored if \code{x} is a distance matrix.}

\item{cannot_link}{A 2 column matrix where each row has the indices 
of two elements that must not be assigned to the same anticluster. 
Alternatively a vector of length \code{nrow(x)}; elements having 
the same value in this vector cannot be assigned to the same anticluster.}

\item{must_link}{A numeric vector of length \code{nrow(x)}. Elements having 
the same value in this vector are assigned to the same anticluster.}

\item{blocks}{A vector, data.frame or matrix representing one
or several categorical variables. Anticlustering is done sequentially 
within blocks. See details.}
}
\value{
A vector of length N that assigns a group (i.e, a number
    between 1 and \code{K}) to each input element.
}
\description{
Partition a pool of elements into groups (i.e., anticlusters) with
the aim of creating high within-group heterogeneity and high
between-group similarity.  Anticlustering is accomplished by
maximizing instead of minimizing a clustering objective function.
Implements anticlustering methods as described in Papenberg and
Klau (2021; <doi:10.1037/met0000301>), Brusco et al. 
(2020; <doi:10.1111/bmsp.12186>), Papenberg (2024; 
<doi:10.1111/bmsp.12315>), Papenberg, Wang, et al. (2025; 
<doi:10.1016/j.crmeth.2025.101137>), Papenberg, Breuer, et al. (2025; 
<doi:10.1017/psy.2025.10052>), and Yang et al. (2022; <doi:10.1016/j.ejor.2022.02.003>).
}
\details{
This function is used to solve anticlustering. That is, the data
input is divided into \code{K} groups in such a way that elements
within groups are heterogeneous and the different groups are
similar. Anticlustering is accomplished by maximizing instead of
minimizing a clustering objective function. The maximization of
five objectives is natively supported (other
functions can also defined by the user as described below):

\itemize{
  \item{the diversity, setting 
        \code{objective = "diversity"} (this is the default objective)}
  \item{the average diversity, which normalizes the diversity by cluster size, 
       setting \code{objective = "average-diversity"}}
  \item{the k-means (or "variance") objective, setting \code{objective = "variance"}}
  \item{the k-plus objective, an extension of the k-means objective,
        setting \code{objective = "kplus"}}
  \item{the dispersion, which is the minimum distance between 
        any two elements within the same cluster (setting 
        \code{objective = "dispersion"})}
}

The k-means objective is the within-group variance---that is, the
sum of the squared distances between each element and its cluster
center (see \code{\link{variance_objective}}). K-means
anticlustering focuses on minimizing differences with regard to the
means of the input variables (that is, the columns in \code{x}), but it ignores any other distribution
characterstics such as the variance / standard deviation. K-plus anticlustering
(using \code{objective = "kplus"}) is an extension of the k-means criterion that also
minimizes differences with regard to the standard
deviations between groups (for details see \code{\link{kplus_anticlustering}}). K-plus
anticlustering can also be extended towards higher order moments such as skew and kurtosis; 
to consider these additional distribution characteristics, use the function
\code{\link{kplus_anticlustering}}. Setting \code{objective = "kplus"} in 
\code{anticlustering} function will only consider means 
and standard deviations (in my experience, this is what users usually want). 
It is strongly recommended to set the argument \code{standardize = TRUE} 
when using the k-plus objective.

The "diversity" objective is the sum of pairwise
distances of elements within the same groups (see
\code{\link{diversity_objective}}). Hence, anticlustering using the diversity 
criterion maximizes between-group similarity
by maximizing within-group heterogeneity (represented as the sum of all pairwise distances). 
In the publication that introduced
the \code{anticlust} package (Papenberg & Klau, 2021), we used the term "anticluster 
editing" to refer to the maximization of the diversity, because the reversed 
procedure - minimizing the diversity - is also known as "cluster editing".
If the diversity is computed on the basis of the Euclidean distance (which is the default
behaviour if \code{x} is a feature matrix), the diversity is an all rounder objective that 
tends to equate all distribution characteristics between groups (such as means, variances, ...). 
Note that the equivalence of within-group heterogeneity and between-group similarity only
holds for equal-sized groups. For unequal-sized groups, it is recommended to
use a different objective when striving for overall between-group similarity,
e.g., the k-plus objective or the \code{"average-diversity"}. The average diversity
was introduced in version 0.8.6, and it is more useful if groups are not 
equal-sized. The average diversity normalizes the sum of intra-cluster distances 
by group size. If all groups are equal-sized, it is equivalent to the 
regular diversity. 

The "dispersion" is the minimum distance between any two elements
that are part of the same cluster; maximization of this objective
ensures that any two elements within the same group are as
dissimilar from each other as possible. Applications that require
high within-group heterogeneity often require to maximize the
dispersion. Oftentimes, it is useful to also consider the diversity
and not only the dispersion; to optimize both objectives at the
same time, see the function
\code{\link{bicriterion_anticlustering}}.

In the standard case, groups of equal size are generated. Adjust
the argument \code{K} to create groups of different size (see
Examples).


\strong{Data input}

The data input \code{x} is usually a feature matrix where each each row
is a "case" and each column is a "variable". The default
\code{objective = "diversity"} or the \code{objective = "dispersion"} are computed
using a dissimilarity matrix as input, however. The default behaviour for 
feature matrices is to compute the Euclidean distance as the 
basic unit of these objectives, using a call to \code{\link{dist}}. 
The combination \code{objective = "diversity"} + 
Euclidean distance was evaluated by us (e.g., Papenberg and Klau, 2021; 
Papenberg, 2024; Papenberg, Wang, et al., 2025) and is generally considered 
to be a good default for equal-sized groups. For unequal-sized groups, prefer 
\code{objective = "average-diversity"}. 

If a different type of dissimilarity measure is preferred (e.g., the Gower distance), 
you may pass a self-generated dissimilarity matrix via the argument \code{x}. This dissimilarity
matrix can be an object of class \code{dist} (e.g., returned by \code{\link{dist}} 
or \code{\link{as.dist}}) or a \code{matrix} where the entries of the upper 
and lower triangular matrix represent pairwise dissimilarities. If the input
is a matrix rather than of class \code{dist}, the function internally tests if
the input is a dissimilarity matrix (because a \code{matrix} could also be a 
feature matrix). It assumes the input is a dissimilarity matrix if the 
upper and lower triangular consists of the same values, and if the diagonal 
only consists of the same values. If the input is a \code{data.frame},
this function always assumes that it is a feature matrix. In a \code{data.frame},
each column must be either \code{numeric} or \code{factor} (the latter for 
categorical variables).

As of version 0.8.12, this function supports handling of missing values (\code{NA}). 
In this case the function \code{\link[stats]{dist}} handles the \code{NA}s when converting
features to pairwise distances. Note that \code{\link[stats]{dist}} always handles
missing values even for objectives that are usually computed using a feature 
matrix as input (i.e., \code{objective = "variance"} and \code{objective = "kplus"}).
In this case, we exploit an equivalence between the \code{objective = "variance"} and
the \code{objective = "average-diversity"}: When the squared Euclidean distance
\code{objective = "variance"} and the \code{objective = "average-diversity"}
are actually equivalent, so we can optimize the \code{objective = "average-diversity"}
in order to perform k-means (and thus, k-plus) anticlustering. See Papenberg, 
Breuer, et al. (2025) for details.


\strong{Algorithms for anticlustering}

By default, a heuristic method is employed for anticlustering: the
exchange method (\code{method = "exchange"}). First, elements are
randomly assigned to anticlusters (It is also possible to
explicitly specify the initial assignment using the argument
\code{K}; in this case, \code{K} has length \code{nrow(x)}.) Based
on the initial assignment, elements are systematically swapped
between anticlusters in such a way that each swap improves the
objective value. For an element, each possible swap with elements
in other clusters is simulated; then, the one swap is performed
that improves the objective the most, but a swap is only conducted
if there is an improvement at all. This swapping procedure is
repeated for each element. When using \code{method =
"local-maximum"}, the exchange method does not terminate after the
first iteration over all elements; instead, the swapping continues
until a local maximum is reached. This method corresponds to the algorithm 
"LCW" by Weitz and Lakshminarayanan (1998). This means that after the
exchange process has been conducted once for each data point, the
algorithm restarts with the first element and proceeds to conduct
exchanges until the objective cannot be improved.

When setting \code{preclustering = TRUE}, only the \code{K - 1}
most similar elements serve as exchange partners for each element,
which can speed up the optimization (more information
on the preclustering heuristic follows below). If the \code{categories} argument
is used, only elements having the same value in \code{categories} serve as exchange
partners.

Using \code{method = "brusco"} implements the local bicriterion
iterated local search (BILS) heuristic by Brusco et al. (2020) and
returns the partition that best optimized either the diversity or
the dispersion during the optimization process. The function
\code{\link{bicriterion_anticlustering}} can also be used to run
the algorithm by Brusco et al., but it returns multiple partitions
that approximate the optimal pareto efficient set according to both
objectives (diversity and dispersion). Thus, to fully utilize the
BILS algorithm, use the function
\code{\link{bicriterion_anticlustering}}.

Since version 0.8.12, \code{method = "3phase"} calls the three phase search algorithm 
for anticlustering by Yang et al. (2022). For anticlust, we implemented some
changes of their default procedure, in particular they used a maximum "time"
that the algorithm runs before terminating. For consistency with our other 
methods in anticlust, we are however using a maximum number of repetitions. 
Our default is 50 repetitions, which may be increased (but not decreased!) 
with the \code{repetitions} argument. 

\strong{Optimal anticlustering}

Usually, heuristics are employed to tackle anticlustering problems,
and their performance is generally very satisfying.  However,
heuristics do not investigate all possible group assignments and
therefore do not (necessarily) find the
"globally optimal solution", i.e., a partitioning that has the best
possible value with regard to the objective that is optimized.  Enumerating
all possible partitions in order to find the best solution,
however, quickly becomes impossible with increasing N, and
therefore it is not possible to find a global optimum this
way. Because all anticlustering problems considered here are also
NP-hard, there is also no (known) clever algorithm that might
identify the best solution without considering all possibilities -
at least in the worst case. Integer linear programming (ILP) is an
approach for tackling NP hard problems that nevertheless tries to
be clever when finding optimal solutions: It does not necessarily
enumerate all possibilities but is still guaranteed to return the
optimal solution. Still, for NP hard problems such as
anticlustering, ILP methods will also fail at some point (i.e.,
when N increases).

\code{anticlust} implements optimal solution algorithms via integer
linear programming. In order to use the ILP methods, set
\code{method = "ilp"}. The integer linear program optimizing the
diversity was described in Papenberg & Klau, (2021; (8) -
(12)). It can also be used to optimize the k-means and k-plus objectives,
but you actually have to use the function \code{\link{optimal_anticlustering}}
for these objectives. The documentation of the function
\code{\link{optimal_dispersion}} and \code{\link{optimal_anticlustering}}
contain more information on the optimal anticlustering algorithms.

\strong{Categorical variables}

There are several ways to balance categorical variables among anticlusters (also 
see the package vignette "Using categorical variables with anticlustering").
The first way is to treat them as "hard constraints" via the argument 
\code{categories} (see Papenberg & Klau, 2021). If done so, balancing the 
categorical variable is accomplished via \code{categorical_sampling} through
a stratified split before the anticlustering optimization. After that, the 
balance is never changed when the algorithm runs (hence, it is a "hard constraint"). 
When \code{categories} has multiple columns (i.e., there are multiple 
categorical variables), each combination of categories is treated as a
 distinct category by the exchange method (i.e., the multiple columns
are "merged" into a single column). This behaviour may lead
to less than optimal results on the level of each single categorical variable.
In this case, it may be useful to treat the categorical variables as part of 
the numeric data, i.e., the first argument \code{x} via binary coding 
(e.g. using \code{\link{categories_to_binary}}). The examples show how to do this 
when using the bicriterion algorithm by Bruso et al. Using the argument 
\code{categories} is only available for the classical exchange procedures, 
that is, for \code{method = "exchange"} and \code{method = "local-maximum"}. 

As of version 0.8.12, the data input \code{x} can contain variables of type \code{factor}
to represent categorical variables. In this case, they are internally converted to 
binary via \code{\link{categories_to_binary}}, which no longer needs to be done 
manually by users (and which can be tricky in particular when combining categorical 
and numeric variables, and when using the k-plus objective). However note that the internal
conversion always uses \code{use_combinations = FALSE}, which may not always 
what users need. In this case, I still recommend using \code{\link{categories_to_binary}} 
manually.

The argument \code{blocks} (available as of version 0.8.13) is similar 
to \code{categories}. Anticlustering is repeated within each stratum
(i.e., each levels or combination of levels of \code{blocks} -- as in \code{categories}
mulitple variables are merged into a single variable to form strata). Anticlustering
in later blocks "remembers" the assignment of previous blocks. This way, 
overall balance as well as balance within blocks is maximized. This is unlike 
when using \code{categories}, which only maximizes overall balance, but not 
specifically within strata. 

\strong{Anticlustering with constraints}

Versions 0.8.6 and 0.8.7 of anticlust introduced the possibility to induce
cannot-link and must-link constraints with anticlustering with 
the arguments \code{cannot_link} and \code{must_link}, respectively.
Cannot-link constraints ensure that pairs of items are assigned to different
clusters. They are given as a 2-column matrix, where each row has the indices
of two elements, which must not be assigned to the same cluster. It is possible
that a set of cannot-link constraints cannot be fulfilled. To verify whether
the constraints cannot be fulfilled (and to actually assign elements 
while respecting the constraints), a graph coloring algorithm algorithm is used.
This algorithm is is actually the same method as used in \code{\link{optimal_dispersion}}. 
The graph coloring algorithm uses an ILP solver and it greatly profits (that is,
it may be much faster) from the Rsymphony package, which is not installed as 
a necessary dependency with anticlust. It is therefore recommended to 
manually install the Rsymphony package, which is then automatically 
selected as solver when using the \code{must_link} argument. If you have
access to the gurobi solver and have the gurobi R package installed, it will
be selected as solver (which is even faster than Symphony). As of version 0.8.11, 
it is also possible to use \code{cannot_link} as a vector. In this case it 
is ensured that elements having the same value in \code{cannot_link} are not
linked in the same cluster. The methods for handling \code{cannot_link} constraints
have been described in Papenberg, Breuer, et al. (2025). 

Must-link constraints are passed as a single vector of length \code{nrow(x)}.
Positions that have the same numeric index are assigned to the same anticluster 
(if the constraints can be fulfilled). When including must-link constraints, 
\code{method = "2PML"} performs a specialized search heuristic that potentially
yields better results than \code{method = "local-maximum"}. The must-link 
functionality and the 2PML algorithm was introduced in Papenberg, Wang, et al. (2025).

The examples illustrate the usage of the \code{must_link} and \code{cannot_link}
arguments. Currently, the different kinds of constraints (arguments \code{must_link}, 
\code{cannot_link}, and \code{categories}) cannot be used together, but this 
may change in future versions. 

\strong{Preclustering}

A useful heuristic for anticlustering is to form small groups of
very similar elements and assign these to different groups. This
logic is used as a preprocessing when setting \code{preclustering =
TRUE}. That is, before the anticlustering objective is optimized, a
cluster analysis identifies small groups of similar elements (pairs
if \code{K = 2}, triplets if \code{K = 3}, and so forth). The
optimization of the anticlustering objective is then conducted
under the constraint that these matched elements cannot be assigned
to the same group. When using the exchange algorithm, preclustering
is conducted using a call to \code{\link{matching}}. When using
\code{method = "ilp"}, the preclustering optimally finds groups of
minimum pairwise distance by solving the integer linear program
described in Papenberg and Klau (2021; (8) - (10), (12) - (13)).
Note that when combining preclustering restrictions with \code{method = "ilp"},
the anticlustering result is no longer guaranteed to be globally optimal, but
only optimal given the preclustering restrictions.


\strong{Optimize a custom objective function}

It is possible to pass a \code{function} to the argument
\code{objective}. See \code{\link{dispersion_objective}} for an
example. If \code{objective} is a function, the exchange method
assigns elements to anticlusters in such a way that the return
value of the custom function is maximized (hence, the function
should return larger values when the between-group similarity is
higher). The custom function has to take two arguments: the first
is the data argument, the second is the clustering assignment. That
is, the argument \code{x} will be passed down to the user-defined
function as first argument. \strong{However, only after}
\code{\link{as.matrix}} has been called on \code{x}. This implies
that in the function body, columns of the data set cannot be
accessed using \code{data.frame} operations such as
\code{$}. Objects of class \code{dist} will be converted to matrix
as well.
}
\examples{

# Use default method ("exchange") and the default diversity criterion, also include
# a categorical variable via argument `categories`:
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  K = 3,
  categories = schaper2019$room
)
# Compare feature means and standard deviations by anticluster
mean_sd_tab(schaper2019[, 3:6], anticlusters)
# Verify that the "room" is balanced across anticlusters:
table(anticlusters, schaper2019$room)

# Use multiple starts of the algorithm to improve the objective and
# optimize the k-means criterion ("variance")
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  objective = "variance",
  K = 3,
  categories = schaper2019$room,
  method = "local-maximum", # better search algorithm
  repetitions = 20 # multiple restarts of the algorithm 
)
# Compare means and standard deviations by anticluster
mean_sd_tab(schaper2019[, 3:6], anticlusters)

# Use different group sizes and optimize the extended k-means
# criterion ("kplus")
anticlusters <- anticlustering(
  schaper2019[, 3:6],
  objective = "kplus",
  K = c(24, 24, 48),
  categories = schaper2019$room,
  repetitions = 20,
  method = "local-maximum",
  standardize = TRUE # ususally recommended
)

# Use cannot_link constraints: Element 1 must not be linked with elements 2 to 10:
cl_matrix <- matrix(c(rep(1, 9), 2:10), ncol = 2)
cl <- anticlustering(
  schaper2019[, 3:6],
  K = 10,
  cannot_link = cl_matrix
)
all(cl[1] != cl[2:10])

# Use cannot_link constraints: Element 1 must be linked with elements 2 to 10.
# Element 11 must be linked with elements 12-20.
must_link <- rep(NA, nrow(schaper2019))
must_link[1:10] <- 1
must_link[11:20] <- 2
cl <- anticlustering(
  schaper2019[, 3:6],
  K = 3,
  must_link = must_link
)
cl[1:10]
cl[11:20]

# Use the heuristic by Brusco et al. (2020) for k-plus anticlustering
# Include categorical variable as part of the optimization criterion rather 
# than the argument categories!
anticlusters <- anticlustering(
  cbind(
    kplus_moment_variables(schaper2019[, 3:6], 2), 
    categories_to_binary(schaper2019$room)
  ),
  objective = "variance", # k-plus anticlustering because of the input above!
  K = 3,
  repetitions = 20,
  method = "brusco"
)

mean_sd_tab(schaper2019[, 3:6], anticlusters)
table(anticlusters, schaper2019$room)

}
\references{
Brusco, M. J., Cradit, J. D., & Steinley, D. (2020). Combining
diversity and dispersion criteria for anticlustering: A bicriterion
approach. British Journal of Mathematical and Statistical
Psychology, 73, 275-396. https://doi.org/10.1111/bmsp.12186

Papenberg, M., Breuer, M., Diekhoff, M., Tran, N. K., & Klau, G. W. 
(2025). Extending the Bicriterion Approach for Anticlustering: 
Exact and Hybrid Approaches. Advance online publication. Psychometrika.

Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
data sets into equivalent parts. Psychological Methods, 26(2), 
161–174. https://doi.org/10.1037/met0000301.

Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for 
Maximizing Between-Group Similarity. British Journal of Mathematical and 
Statistical Psychology, 77(1), 80-102. https://doi.org/10.1111/bmsp.12315

Papenberg, M., Wang, C., Diop, M., Bukhari, S. H., Oskotsky, B., Davidson, B. R., 
Vo, K. C., Liu, B., Irwin, J. C., Combes, A., Gaudilliere, B., Li, J., Stevenson, D. K., 
Klau, G. W., Giudice, L. C., Sirota, M., & Oskotsky, T. T. (2025). Anticlustering
for sample allocation to minimize batch effects. Cell Reports Methods, 5(8), 
101137. https://doi.org/10.1016/j.crmeth.2025.101137

Späth, H. (1986). Anticlustering: Maximizing the variance criterion.
Control and Cybernetics, 15, 213-218.

Weitz, R. R., & Lakshminarayanan, S. (1998). An empirical comparison of 
heuristic methods for creating maximally diverse groups. Journal of the 
Operational Research Society, 49(6), 635-646. https://doi.org/10.1057/palgrave.jors.2600510

Yang, X., Cai, Z., Jin, T., Tang, Z., & Gao, S. (2022). A three-phase search 
approach with dynamic population size for solving the maximally diverse grouping 
problem. European Journal of Operational Research, 302(3), 925-953. 
https://doi.org/10.1016/j.ejor.2022.02.003
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fast_anticlustering.R
\name{fast_anticlustering}
\alias{fast_anticlustering}
\title{Fast anticlustering}
\usage{
fast_anticlustering(
  x,
  K,
  k_neighbours = Inf,
  categories = NULL,
  exchange_partners = NULL
)
}
\arguments{
\item{x}{A numeric vector, matrix or data.frame of data points.
Rows correspond to elements and columns correspond to
features. A vector represents a single numeric feature.}

\item{K}{How many anticlusters should be created. Alternatively:
(a) A vector describing the size of each group, or (b) a vector
of length \code{nrow(x)} describing how elements are assigned
to anticlusters before the optimization starts.}

\item{k_neighbours}{The number of nearest neighbours that serve as
exchange partner for each element. See details.}

\item{categories}{A vector, data.frame or matrix representing one
or several categorical constraints.}

\item{exchange_partners}{Optional argument. A list of length
\code{NROW(x)} specifying for each element the indices of the
elements that serve as exchange partners. If used, this
argument overrides the \code{k_neighbours} argument. See
examples.}
}
\description{
Increasing the speed of (k-means / k-plus) anticlustering by (1) 
conducting fewer exchanges during the optimization and (2) using an alternative
formulation of the k-means objective. Makes anticlustering applicable to 
quite large data sets.
}
\details{
This function was created to make anticlustering applicable to
large data sets (e.g., several 100,000 elements). It optimizes the
k-means objective because computing all pairwise distances as is
done when optimizing the "diversity" (i.e., the default in
\code{\link{anticlustering}}) is not feasible for very large data
sets (for about N > 20000 on my personal computer). Using
\code{fast_anticlustering} for k-plus anticlustering is also
possible by applying \code{\link{kplus_moment_variables}} on the
input (and possibly by using the argument \code{exchange_partners},
see Examples).

The function \code{fast_anticlustering} employs a speed-optimized
exchange method, which is basically equivalent to \code{method =
"exchange"} in \code{\link{anticlustering}}, but may reduce the number
of exchanges that are investigated for each input element. The number of 
exchange partners per element has to be set using the argument \code{k_neighbours}. By
default, it is set to \code{Inf}, meaning that all possible swaps are
tested. If \code{k_neighbours} is set differently (which is usually recommended when running 
this function), the default behaviour is to generate exchange partners using a
nearest neighbour search (using the function \code{\link[RANN]{nn2}}
from the \code{RANN} package). Using more exchange partners can improve the quality of
the results, but also increase run time. Note that for very large
data sets, anticlustering generally becomes "easier" (even a random
split may yield satisfactory results), so using few exchange
partners is usually not a problem. 

It is possible to specify custom exchange partners using the
argument \code{exchange_partners} instead of relying on the default
nearest neighbour search.  When using \code{exchange_partners}, it
is not necessary that each element has the same number of exchange
partners; this is why the argument \code{exchange_partners} has to
be a \code{list} instead of a \code{matrix} or
\code{data.frame}. Exchange partners can for example be generated
by \code{\link{generate_exchange_partners}} (see Examples), but a
custom list may also be used. Note that categorical constraints
induced via \code{categories} may not be respected during the
optimization if the \code{exchange_partners} argument allows
exchanges between members of different categories, so care must be
taken when combining the arguments \code{exchange_partners} and
\code{categories}.

In \code{anticlustering(..., objective = "variance")}, the run time
of computing the k-means objective is in O(M N), where N is the
total number of elements and M is the number of variables. This is
because the variance is computed as the sum of squared distances
between all data points and their cluster centers.  The function
\code{fast_anticlustering} uses a different - but equivalent -
formulation of the k-means objective, where the re-computation of
the objective only depends and M but no longer on N. In
particular, this variant of k-means anticlustering minimizes 
the weighted sum of squared distances between
cluster centroids and the overall data centroid; the distances
between all individual data points and their cluster center are not
computed (Späth, 1986). Using the different objective formulation 
reduces the run time by an
order of magnitude and makes k-means anticlustering applicable to
very large data sets (even in the millions). For a fixed number of
exchange partners (specified using the argument
\code{k_neighbours}), the approximate run time of
\code{fast_anticlustering} is in O(M N). The algorithm
\code{method = "exchange"} in \code{\link{anticlustering}} with
\code{objective = "variance"} has a run time of O(M N^3). 
Thus, \code{fast_anticlustering} can improve the run time
by two orders of magnitude as compared to the standard exchange
algorithm. The nearest neighbour search, which is done in the
beginning usually does not strongly contribute to the overall
run time. It is nevertheless possible to suppress the nearest
neighbour search by using the \code{exchange_partners} argument.

When setting the \code{categories} argument, exchange partners
(i.e., nearest neighbours) will be generated from the same
category. Note that when \code{categories} has multiple columns,
each combination of these categories is treated as a distinct
category by the exchange method. You can also use
\code{\link{categories_to_binary}} to potentially improve results
for several categorical variables, instead of using the argument
\code{categories}.
}
\examples{

## Use fewer or more exchange partners to adjust speed (vs. quality tradeoff)
features <- iris[, - 5]
N <- nrow(features)
init <- sample(rep_len(1:3, N)) # same starting point for all calls:
groups1 <- fast_anticlustering(features, K = init) # default: all exchanges
groups2 <- fast_anticlustering(features, K = init, k_neighbours = 20) 
groups3 <- fast_anticlustering(features, K = init, k_neighbours = 2)

variance_objective(features, groups1)
variance_objective(features, groups2)
variance_objective(features, groups3)

# K-plus anticlustering is straight forward when sticking with the default
# for k_neighbours
kplus_anticlusters <- fast_anticlustering(
  kplus_moment_variables(features, T = 2), 
  K = 3
)
mean_sd_tab(features, kplus_anticlusters)

# Some care is needed when applying k-plus using with this function 
# while using a reduced number of exchange partners generated in the 
# nearest neighbour search. Then we:
# 1) Use kplus_moment_variables() on the numeric input
# 2) Generate custom exchange_partners because otherwise nearest 
#    neighbours are internally selected based on the extended k-plus 
#    variables returned by kplus_moment_variables() 
#    (which does not really make sense)
kplus_anticlusters <- fast_anticlustering(
  kplus_moment_variables(features, T = 2), 
  K = 3,
  exchange_partners = generate_exchange_partners(120, features = features, method = "RANN")
 )
mean_sd_tab(features, kplus_anticlusters)
# Or we use random exchange partners: 
kplus_anticlusters <- fast_anticlustering(
  kplus_moment_variables(features, T = 2), 
  K = 3,
  exchange_partners = generate_exchange_partners(120, N = nrow(features), method = "random")
)
mean_sd_tab(features, kplus_anticlusters)


# Working on several 1000 elements is very fast (Here n = 10000, m = 2)
data <- matrix(rnorm(10000 * 2), ncol = 2)
start <- Sys.time()
groups <- fast_anticlustering(data, K = 5, k_neighbours = 5)
Sys.time() - start 

}
\references{
Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
data sets into equivalent parts. Psychological Methods, 26(2), 
161–174. https://doi.org/10.1037/met0000301.

Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for 
Maximizing Between-Group Similarity. British Journal of Mathematical and 
Statistical Psychology, 77(1), 80--102. https://doi.org/10.1111/bmsp.12315

Späth, H. (1986). Anticlustering: Maximizing the variance criterion.
Control and Cybernetics, 15, 213-218.
}
\seealso{
\code{\link{anticlustering}}

\code{\link{kplus_moment_variables}}

\code{\link{categories_to_binary}}

\code{\link{variance_objective}}

\code{\link{generate_exchange_partners}}
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/number-of-partitions.R
\name{n_partitions}
\alias{n_partitions}
\title{Number of equal sized partitions}
\usage{
n_partitions(N, K)
}
\arguments{
\item{N}{How many elements}

\item{K}{How many partitions}
}
\value{
The number of partitions
}
\description{
Number of equal sized partitions
}
\examples{
n_partitions(20, 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper-k-plus-anticlustering.R
\name{kplus_anticlustering}
\alias{kplus_anticlustering}
\title{K-plus anticlustering}
\usage{
kplus_anticlustering(
  x,
  K,
  variance = TRUE,
  skew = FALSE,
  kurtosis = FALSE,
  covariances = FALSE,
  T = NULL,
  standardize = TRUE,
  ...
)
}
\arguments{
\item{x}{A feature matrix where rows correspond to elements and columns
correspond to variables (a single numeric variable can be
passed as a vector).}

\item{K}{How many anticlusters should be created. Alternatively:
(a) A vector describing the size of each group, or (b) a vector
of length \code{nrow(x)} describing how elements are assigned
to anticlusters before the optimization starts.}

\item{variance}{Boolean: Should the k-plus objective include a term to 
maximize between-group similarity with regard to the variance? 
(Default = TRUE)}

\item{skew}{Boolean: Should the k-plus objective include a term to 
maximize between-group similarity with regard to skewness? 
(Default = FALSE)}

\item{kurtosis}{Boolean: Should the k-plus objective include a term to 
maximize between-group similarity with regard to kurtosis? 
(Default = FALSE)}

\item{covariances}{Boolean: Should the k-plus objective include a term to 
maximize between-group similarity with regard to covariance structure? 
(Default = FALSE)}

\item{T}{Optional argument: An integer specifying how many
distribution moments should be equalized between groups.}

\item{standardize}{Boolean. If \code{TRUE}, the data is standardized through 
a call to \code{\link{scale}} before the optimization starts. 
Defaults to TRUE. See details.}

\item{...}{Arguments passed down to \code{\link{anticlustering}}. All of the 
arguments are supported except for \code{objective}.}
}
\description{
Perform anticlustering using the k-plus objective to maximize between-group 
similarity. This function implements the k-plus anticlustering method described 
in Papenberg (2024; <doi:10.1111/bmsp.12315>).
}
\details{
This function implements the unweighted sum approach for k-plus 
    anticlustering. Details are given in Papenberg (2024). 
    
    The optional argument \code{T} denotes the number of distribution 
    moments that are considered in the anticlustering process. For example,
    \code{T = 4} will lead to similar means, variances, skew and kurtosis. 
    For the first four moments, it is also possible to use the boolean
    convenience arguments \code{variance}, \code{skew} and \code{kurtosis}; the
    mean (the first moment) is always included and cannot be "turned off".
    If the argument \code{T} is used, it overrides the arguments
    \code{variance}, \code{skew} and \code{kurtosis} (corresponding to
    the second, third and fourth moment), ignoring their values.
    
    The \code{standardization} is applied to all original features and the 
    additional k-plus features that are appended to the data set in order 
    to optimize the k-plus criterion. When using standardization, 
    all criteria such as means, variances and skewness receive a comparable
    weight during the optimization. It is usually recommended not
    to change the default setting \code{standardization = TRUE}.
   
    This function can use any arguments that are also possible in 
    \code{\link{anticlustering}}
    (except for `objective` because the objective optimized here
    is the k-plus objective; to use a different objective, 
    call \code{\link{anticlustering}} directly). Any arguments that are
    not explicitly changed here (i.e., \code{standardize = TRUE}) receive the 
    default given in \code{\link{anticlustering}} 
    (e.g., \code{method = "exchange"}.)
}
\examples{

# Generate some data
N <- 180
M <- 4
features <- matrix(rnorm(N * M), ncol = M)
# standard k-plus anticlustering: optimize similarity with regard to mean and variance:
cl <- kplus_anticlustering(features, K = 3, method = "local-maximum")
mean_sd_tab(features, cl)
# Visualize an anticlustering solution:
plot(features, col = palette()[2:4][cl], pch = c(16:18)[cl])

# Also optimize with regard to skewness and kurtosis
cl2 <- kplus_anticlustering(
  features, 
  K = 3, 
  method = "local-maximum", 
  skew = TRUE, 
  kurtosis = TRUE
)

# The following two calls are equivalent: 
init_clusters <- sample(rep_len(1:3, nrow(features)))
# 1.
x1 <- kplus_anticlustering(
  features, 
  K = init_clusters, 
  variance = TRUE,
  skew = TRUE
)
# 2. 
x2 <- kplus_anticlustering(
  features, 
  K = init_clusters, 
  T = 3
)
# Verify: 
all(x1 == x2)

}
\references{
Papenberg, M. (2024). K-plus Anticlustering: An Improved k-means Criterion for 
Maximizing Between-Group Similarity. British Journal of Mathematical and 
Statistical Psychology, 77(1), 80--102. https://doi.org/10.1111/bmsp.12315
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fast_anticlustering.R
\name{generate_exchange_partners}
\alias{generate_exchange_partners}
\title{Get exchange partners for fast_anticlustering()}
\usage{
generate_exchange_partners(
  n_exchange_partners,
  N = NULL,
  features = NULL,
  method = "random",
  categories = NULL
)
}
\arguments{
\item{n_exchange_partners}{The number of exchange partners per
element}

\item{N}{The number of elements for which exchange partners; can be
\code{NULL} if \code{features} is passed (it is ignored if
\code{features} is passed).}

\item{features}{The features for which nearest neighbours are
sought if \code{method = "RANN"}.  May be NULL if random
exchange partners are generated.}

\item{method}{Currently supports "random" (default), "RANN" and
"restricted_random". See details.}

\item{categories}{A vector, data.frame or matrix representing one
or several categorical constraints.}
}
\value{
A list of length \code{N}. Is usually used as input to the
    argument \code{exchange_partners} in
    \code{\link{fast_anticlustering}}.  Then, the i'th element of
    the list contains the indices of the exchange partners that are
    used for the i'th element.
}
\description{
Get exchange partners for fast_anticlustering()
}
\details{
The \code{method = "RANN"} generates exchange partners using a
nearest neighbour search via \code{\link[RANN]{nn2}} from the
\code{RANN} package; \code{methode = "restricted_random"} generates
random exchange partners but ensures that for each element, no
duplicates are generated and that the element itself does not occur
as exchange partner (this is the slowest method, and I would not
recommend it for large N); \code{method = "random"} (default) does
not impose these restrictions and generates unrescricted random
partners (it may therefore generate duplicates and the element
itself as exchange partner).

When setting the \code{categories} argument and using \code{method
= "RANN"}, exchange partners (i.e., nearest neighbours) will be
generated from the same category; \code{methode =
"restricted_random"} will also adhere to categorical constraints
induced via \code{categories} (i.e. each element only receives
exchange partners from the same category as itself); \code{methode
= "random"} cannot incoorporate categorical restrictions.
}
\examples{

# Restricted random method generates no duplicates per element and cannot return 
# the element itself as exchange partner
generate_exchange_partners(5, N = 10, method = "restricted_random")
# "random" simply randomizes with replacement and without restrictions
# (categorical restrictions are also not possible; is much faster for large data sets)
generate_exchange_partners(5, N = 10, method = "random")
# May return less than 5 exchange partners if there are not enough members 
# of the same category: 
generate_exchange_partners(
  5, N = 10, 
  method = "restricted_random", 
  categories = cbind(schaper2019$room, schaper2019$frequency)
)
# using nearest neighbour search (unlike RANN::nn2, this does not 
# return the ID of the element itself as neighbour)
generate_exchange_partners(5, features = schaper2019[, 3:5], method = "RANN")[1:3]
# compare with RANN directly:
RANN::nn2(schaper2019[, 3:5], k = 6)$nn.idx[1:3, ] # note k = 6

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/objective-variance.R
\name{variance_objective}
\alias{variance_objective}
\title{Objective value for the variance criterion}
\usage{
variance_objective(x, clusters)
}
\arguments{
\item{x}{A vector, matrix or data.frame of data points. Rows
correspond to elements and columns correspond to features. A
vector represents a single feature.}

\item{clusters}{A vector representing (anti)clusters (e.g., returned
by \code{\link{anticlustering}} or
\code{\link{balanced_clustering}})}
}
\value{
The total within-cluster variance
}
\description{
Compute the k-means variance objective for a given clustering.
}
\details{
The variance objective is given by the sum of the squared
errors between cluster centers and individual data points. It is the
objective function used in k-means clustering, see
\code{\link{kmeans}}.
}
\examples{

data(iris)
## Clustering
clusters <- balanced_clustering(
  iris[, -5],
  K = 3
)
# This is low:
variance_objective(
  iris[, -5],
  clusters
)
## Anticlustering
anticlusters <- anticlustering(
  iris[, -5],
  K = 3,
  objective = "variance"
)
# This is higher:
variance_objective(
  iris[, -5],
  anticlusters
)

# Illustrate variance objective
N <- 18
data <- matrix(rnorm(N * 2), ncol = 2)
cl <- balanced_clustering(data, K = 3)
plot_clusters(data, cl, illustrate_variance = TRUE)
}
\references{
Jain, A. K. (2010). Data clustering: 50 years beyond k-means.
Pattern Recognition Letters, 31, 651–666.

Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition 
data sets into equivalent parts. Psychological Methods, 26(2), 
161–174. https://doi.org/10.1037/met0000301.

Späth, H. (1986). Anticlustering: Maximizing the variance criterion.
Control and Cybernetics, 15, 213–218.
}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plot_clusters}
\alias{plot_clusters}
\title{Visualize a cluster analysis}
\usage{
plot_clusters(
  features,
  clusters,
  within_connection = FALSE,
  between_connection = FALSE,
  illustrate_variance = FALSE,
  show_axes = FALSE,
  xlab = NULL,
  ylab = NULL,
  xlim = NULL,
  ylim = NULL,
  main = "",
  cex = 1.2,
  cex.axis = 1.2,
  cex.lab = 1.2,
  lwd = 1.5,
  lty = 2,
  frame.plot = FALSE,
  cex_centroid = 2
)
}
\arguments{
\item{features}{A data.frame or matrix representing the features that
are plotted. Must have two columns.}

\item{clusters}{A vector representing the clustering}

\item{within_connection}{Boolean. Connect the elements within each
clusters through lines? Useful to illustrate a graph structure.}

\item{between_connection}{Boolean. Connect the elements between each
clusters through lines? Useful to illustrate a graph structure.
(This argument only works for two clusters).}

\item{illustrate_variance}{Boolean. Illustrate the variance criterion
in the plot?}

\item{show_axes}{Boolean, display values on the x and y-axis? Defaults
to `FALSE`.}

\item{xlab}{The label for the x-axis}

\item{ylab}{The label for the y-axis}

\item{xlim}{The limits for the x-axis}

\item{ylim}{The limits for the y-axis}

\item{main}{The title of the plot}

\item{cex}{The size of the plotting symbols, see \code{\link{par}}}

\item{cex.axis}{The size of the values on the axes}

\item{cex.lab}{The size of the labels of the axes}

\item{lwd}{The width of the lines connecting elements.}

\item{lty}{The line type of the lines connecting elements
(see \code{\link{par}}).}

\item{frame.plot}{a logical indicating whether a box should be drawn
around the plot.}

\item{cex_centroid}{The size of the cluster center symbol (has an
effect only if \code{illustrate_variance} is \code{TRUE})}
}
\description{
Visualize a cluster analysis
}
\details{
In most cases, the argument \code{clusters} is a vector
  returned by one of the functions \code{\link{anticlustering}},
  \code{\link{balanced_clustering}} or \code{\link{matching}}. 
  However, the plotting function can also be used to plot the results 
  of other cluster functions such as \code{\link{kmeans}}. This function
  is usually just used to get a fast impression of the results of an 
  (anti)clustering assignment, but limited in its functionality. 
  It is useful for depicting the intra-cluster connections using 
  argument \code{within_connection}.
}
\examples{

N <- 15
features <- matrix(runif(N * 2), ncol = 2)
K <- 3
clusters <- balanced_clustering(features, K = K)
anticlusters <- anticlustering(features, K = K)
user_par <- par("mfrow")
par(mfrow = c(1, 2))
plot_clusters(features, clusters, main = "Cluster editing", within_connection = TRUE)
plot_clusters(features, anticlusters, main = "Anticluster editing", within_connection = TRUE)
par(mfrow = user_par)

}
\author{
Martin Papenberg \email{martin.papenberg@hhu.de}
}
